@@ -1,32 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1997-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: bt_recno.c,v 12.28 2006/08/24 14:44:44 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_recno.c,v 11.86 2001/10/04 18:07:19 bostic Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <limits.h>
-#include <stdio.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
-#include "db_page.h"
-#include "btree.h"
-#include "db_ext.h"
-#include "db_shash.h"
-#include "lock.h"
-#include "qam.h"
-#include "txn.h"
+#include "dbinc/db_page.h"
+#include "dbinc/btree.h"
+#include "dbinc/lock.h"
+#include "dbinc/mp.h"
 
 static int  __ram_add __P((DBC *, db_recno_t *, DBT *, u_int32_t, u_int32_t));
 static int  __ram_source __P((DB *));
@@ -61,7 +48,7 @@ static int  __ram_update __P((DBC *, db_recno_t, int));
 	}								\
 }
 #define	CD_ISSET(cp)							\
-	(F_ISSET(cp, C_RENUMBER) && F_ISSET(cp, C_DELETED))
+	(F_ISSET(cp, C_RENUMBER) && F_ISSET(cp, C_DELETED) ? 1 : 0)
 
 /*
  * Macros for comparing the ordering of two cursors.
@@ -88,17 +75,34 @@ static int  __ram_update __P((DBC *, db_recno_t, int));
  * Do we need to log the current cursor adjustment?
  */
 #define	CURADJ_LOG(dbc)							\
-	(DB_LOGGING((dbc)) && (dbc)->txn != NULL && (dbc)->txn->parent != NULL)
+	(DBC_LOGGING((dbc)) && (dbc)->txn != NULL && (dbc)->txn->parent != NULL)
+
+/*
+ * After a search, copy the found page into the cursor, discarding any
+ * currently held lock.
+ */
+#define	STACK_TO_CURSOR(cp, ret) {					\
+	int __t_ret;							\
+	(cp)->page = (cp)->csp->page;					\
+	(cp)->pgno = (cp)->csp->page->pgno;				\
+	(cp)->indx = (cp)->csp->indx;					\
+	if ((__t_ret = __TLPUT(dbc, (cp)->lock)) != 0 && (ret) == 0)	\
+		ret = __t_ret;						\
+	(cp)->lock = (cp)->csp->lock;					\
+	(cp)->lock_mode = (cp)->csp->lock_mode;				\
+}
 
 /*
  * __ram_open --
  *	Recno open function.
  *
- * PUBLIC: int __ram_open __P((DB *, const char *, db_pgno_t, u_int32_t));
+ * PUBLIC: int __ram_open __P((DB *,
+ * PUBLIC:      DB_TXN *, const char *, db_pgno_t, u_int32_t));
  */
 int
-__ram_open(dbp, name, base_pgno, flags)
+__ram_open(dbp, txn, name, base_pgno, flags)
 	DB *dbp;
+	DB_TXN *txn;
 	const char *name;
 	db_pgno_t base_pgno;
 	u_int32_t flags;
@@ -107,13 +111,11 @@ __ram_open(dbp, name, base_pgno, flags)
 	DBC *dbc;
 	int ret, t_ret;
 
+	COMPQUIET(name, NULL);
 	t = dbp->bt_internal;
 
-	/* Initialize the remaining fields/methods of the DB. */
-	dbp->stat = __bam_stat;
-
 	/* Start up the tree. */
-	if ((ret = __bam_read_root(dbp, name, base_pgno, flags)) != 0)
+	if ((ret = __bam_read_root(dbp, txn, base_pgno, flags)) != 0)
 		return (ret);
 
 	/*
@@ -128,9 +130,9 @@ __ram_open(dbp, name, base_pgno, flags)
 		return (ret);
 
 	/* If we're snapshotting an underlying source file, do it now. */
-	if (F_ISSET(dbp, DB_RE_SNAPSHOT)) {
+	if (F_ISSET(dbp, DB_AM_SNAPSHOT)) {
 		/* Allocate a cursor. */
-		if ((ret = dbp->cursor(dbp, NULL, &dbc, 0)) != 0)
+		if ((ret = __db_cursor(dbp, NULL, &dbc, 0)) != 0)
 			return (ret);
 
 		/* Do the snapshot. */
@@ -139,11 +141,11 @@ __ram_open(dbp, name, base_pgno, flags)
 			ret = 0;
 
 		/* Discard the cursor. */
-		if ((t_ret = dbc->c_close(dbc)) != 0 && ret == 0)
+		if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
 			ret = t_ret;
 	}
 
-	return (0);
+	return (ret);
 }
 
 /*
@@ -173,8 +175,8 @@ __ram_append(dbc, key, data)
 
 	/* Return the record number. */
 	if (ret == 0)
-		ret = __db_retcopy(dbc->dbp, key, &cp->recno, sizeof(cp->recno),
-		    &dbc->rkey->data, &dbc->rkey->ulen);
+		ret = __db_retcopy(dbc->dbp->dbenv, key, &cp->recno,
+		    sizeof(cp->recno), &dbc->rkey->data, &dbc->rkey->ulen);
 
 	return (ret);
 }
@@ -195,8 +197,7 @@ __ram_c_del(dbc)
 	DB *dbp;
 	DB_LSN lsn;
 	DBT hdr, data;
-	EPG *epg;
-	int exact, ret, stack;
+	int exact, nc, ret, stack, t_ret;
 
 	dbp = dbc->dbp;
 	cp = (BTREE_CURSOR *)dbc->internal;
@@ -218,7 +219,7 @@ __ram_c_del(dbc)
 		return (DB_KEYEMPTY);
 
 	/* Search the tree for the key; delete only deletes exact matches. */
-	if ((ret = __bam_rsearch(dbc, &cp->recno, S_DELETE, 1, &exact)) != 0)
+	if ((ret = __bam_rsearch(dbc, &cp->recno, SR_DELETE, 1, &exact)) != 0)
 		goto err;
 	if (!exact) {
 		ret = DB_NOTFOUND;
@@ -226,13 +227,12 @@ __ram_c_del(dbc)
 	}
 	stack = 1;
 
-	/* Copy the page into the cursor, discarding any currently held lock. */
-	cp->page = cp->csp->page;
-	cp->pgno = cp->csp->page->pgno;
-	cp->indx = cp->csp->indx;
-	(void)__TLPUT(dbc, cp->lock);
-	cp->lock = cp->csp->lock;
-	cp->lock_mode = cp->csp->lock_mode;
+	/* Copy the page into the cursor. */
+	if ((ret = __memp_dirty(dbp->mpf, &cp->csp->page, dbc->txn, 0)) != 0)
+		goto err;
+	STACK_TO_CURSOR(cp, ret);
+	if (ret != 0)
+		goto err;
 
 	/*
 	 * If re-numbering records, the on-page deleted flag can only mean
@@ -245,7 +245,7 @@ __ram_c_del(dbc)
 	 * delete records they never created, the latter is an error because
 	 * if the record was "deleted", we could never have found it.
 	 */
-	if (B_DISSET(GET_BKEYDATA(cp->page, cp->indx)->type)) {
+	if (B_DISSET(GET_BKEYDATA(dbp, cp->page, cp->indx)->type)) {
 		ret = DB_KEYEMPTY;
 		goto err;
 	}
@@ -254,11 +254,13 @@ __ram_c_del(dbc)
 		/* Delete the item, adjust the counts, adjust the cursors. */
 		if ((ret = __bam_ditem(dbc, cp->page, cp->indx)) != 0)
 			goto err;
-		__bam_adjust(dbc, -1);
-		if (__ram_ca(dbc, CA_DELETE) > 0 &&
-		    CURADJ_LOG(dbc) && (ret = __bam_rcuradj_log(dbp->dbenv,
-		    dbc->txn, &lsn, 0, dbp->log_fileid, CA_DELETE,
-		    cp->root, cp->recno, cp->order)) != 0)
+		if ((ret = __bam_adjust(dbc, -1)) != 0)
+			goto err;
+		if ((ret = __ram_ca(dbc, CA_DELETE, &nc)) != 0)
+			goto err;
+		if (nc > 0 &&
+		    CURADJ_LOG(dbc) && (ret = __bam_rcuradj_log(dbp, dbc->txn,
+		    &lsn, 0, CA_DELETE, cp->root, cp->recno, cp->order)) != 0)
 			goto err;
 
 		/*
@@ -275,21 +277,11 @@ __ram_c_del(dbc)
 		 * are closed, and then clean it up.
 		 */
 		if (NUM_ENT(cp->page) == 0 && PGNO(cp->page) != cp->root) {
-			/*
-			 * We already have a locked stack of pages.  However,
-			 * there are likely entries in the stack that aren't
-			 * going to be emptied by removing the single reference
-			 * to the emptied page (or one of its parents).
-			 */
-			for (epg = cp->sp; epg <= cp->csp; ++epg)
-				if (NUM_ENT(epg->page) <= 1)
-					break;
-
 			/*
 			 * We want to delete a single item out of the last page
-			 * that we're not deleting, back up to that page.
+			 * that we're not deleting.
 			 */
-			ret = __bam_dpages(dbc, --epg);
+			ret = __bam_dpages(dbc, 0, 0);
 
 			/*
 			 * Regardless of the return from __bam_dpages, it will
@@ -303,14 +295,10 @@ __ram_c_del(dbc)
 		if ((ret = __bam_ditem(dbc, cp->page, cp->indx)) != 0)
 			goto err;
 
-		B_TSET(bk.type, B_KEYDATA, 1);
+		B_TSET_DELETED(bk.type, B_KEYDATA);
 		bk.len = 0;
-		memset(&hdr, 0, sizeof(hdr));
-		hdr.data = &bk;
-		hdr.size = SSZA(BKEYDATA, data);
-		memset(&data, 0, sizeof(data));
-		data.data = (void *)"";
-		data.size = 0;
+		DB_INIT_DBT(hdr, &bk, SSZA(BKEYDATA, data));
+		DB_INIT_DBT(data, "", 0);
 		if ((ret = __db_pitem(dbc,
 		    cp->page, cp->indx, BKEYDATA_SIZE(0), &hdr, &data)) != 0)
 			goto err;
@@ -318,8 +306,8 @@ __ram_c_del(dbc)
 
 	t->re_modified = 1;
 
-err:	if (stack)
-		__bam_stkrel(dbc, STK_CLRDBC);
+err:	if (stack && (t_ret = __bam_stkrel(dbc, STK_CLRDBC)) != 0 && ret == 0)
+		ret = t_ret;
 
 	return (ret);
 }
@@ -381,8 +369,13 @@ retry:	switch (flags) {
 		 * we have to avoid incrementing the record number so that we
 		 * return the right record by virtue of renumbering the tree.
 		 */
-		if (CD_ISSET(cp))
+		if (CD_ISSET(cp)) {
+			/*
+			 * Clear the flag, we've moved off the deleted record.
+			 */
+			CD_CLR(cp);
 			break;
+		}
 
 		if (cp->recno != RECNO_OOB) {
 			++cp->recno;
@@ -478,7 +471,7 @@ retry:	switch (flags) {
 	for (;; ++cp->recno) {
 		/* Search the tree for the record. */
 		if ((ret = __bam_rsearch(dbc, &cp->recno,
-		    F_ISSET(dbc, DBC_RMW) ? S_FIND_WR : S_FIND,
+		    F_ISSET(dbc, DBC_RMW) ? SR_FIND_WR : SR_FIND,
 		    1, &exact)) != 0)
 			goto err;
 		if (!exact) {
@@ -486,16 +479,10 @@ retry:	switch (flags) {
 			goto err;
 		}
 
-		/*
-		 * Copy the page into the cursor, discarding any currently
-		 * held lock.
-		 */
-		cp->page = cp->csp->page;
-		cp->pgno = cp->csp->page->pgno;
-		cp->indx = cp->csp->indx;
-		(void)__TLPUT(dbc, cp->lock);
-		cp->lock = cp->csp->lock;
-		cp->lock_mode = cp->csp->lock_mode;
+		/* Copy the page into the cursor. */
+		STACK_TO_CURSOR(cp, ret);
+		if (ret != 0)
+			goto err;
 
 		/*
 		 * If re-numbering records, the on-page deleted flag means this
@@ -506,7 +493,7 @@ retry:	switch (flags) {
 		 * walking through off-page duplicates, and fail if they were
 		 * requested explicitly by the application.
 		 */
-		if (B_DISSET(GET_BKEYDATA(cp->page, cp->indx)->type))
+		if (B_DISSET(GET_BKEYDATA(dbp, cp->page, cp->indx)->type))
 			switch (flags) {
 			case DB_NEXT:
 			case DB_PREV:
@@ -534,7 +521,7 @@ retry:	switch (flags) {
 
 		if (flags == DB_GET_BOTH ||
 		    flags == DB_GET_BOTHC || flags == DB_GET_BOTH_RANGE) {
-			if ((ret = __bam_cmp(dbp, data,
+			if ((ret = __bam_cmp(dbp, dbc->txn, data,
 			    cp->page, cp->indx, __bam_defcmp, &cmp)) != 0)
 				return (ret);
 			if (cmp == 0)
@@ -549,11 +536,10 @@ retry:	switch (flags) {
 	}
 
 	/* Return the key if the user didn't give us one. */
-	if (!F_ISSET(dbc, DBC_OPD)) {
-		if (flags != DB_SET && flags != DB_SET_RANGE)
-			ret = __db_retcopy(dbp,
-			     key, &cp->recno, sizeof(cp->recno),
-			     &dbc->rkey->data, &dbc->rkey->ulen);
+	if (!F_ISSET(dbc, DBC_OPD) && !F_ISSET(key, DB_DBT_ISSET)) {
+		ret = __db_retcopy(dbp->dbenv,
+		    key, &cp->recno, sizeof(cp->recno),
+		    &dbc->rkey->data, &dbc->rkey->ulen);
 		F_SET(key, DB_DBT_ISSET);
 	}
 
@@ -578,14 +564,16 @@ __ram_c_put(dbc, key, data, flags, pgnop)
 {
 	BTREE_CURSOR *cp;
 	DB *dbp;
+	DB_ENV *dbenv;
 	DB_LSN lsn;
-	int exact, nc, ret, t_ret;
 	u_int32_t iiflags;
+	int exact, nc, ret, t_ret;
 	void *arg;
 
 	COMPQUIET(pgnop, NULL);
 
 	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
 	cp = (BTREE_CURSOR *)dbc->internal;
 
 	/*
@@ -606,26 +594,24 @@ __ram_c_put(dbc, key, data, flags, pgnop)
 			    &cp->recno, data, DB_APPEND, 0)) != 0)
 				return (ret);
 			if (CURADJ_LOG(dbc) &&
-			    (ret = __bam_rcuradj_log(dbp->dbenv,
-			    dbc->txn, &lsn, 0, dbp->log_fileid, CA_ICURRENT,
-			    cp->root, cp->recno, cp->order)))
+			    (ret = __bam_rcuradj_log(dbp, dbc->txn, &lsn, 0,
+			    CA_ICURRENT, cp->root, cp->recno, cp->order)) != 0)
 				return (ret);
 			return (0);
+		default:
+			break;
 		}
 
 	/*
 	 * Handle normal DB_KEYFIRST/DB_KEYLAST;  for a recno, which has
 	 * no duplicates, these are identical and mean "put the given
 	 * datum at the given recno".
-	 *
-	 * Note that the code here used to be in __ram_put;  now, we
-	 * go through the access-method-common __db_put function, which
-	 * handles DB_NOOVERWRITE, so we and __ram_add don't have to.
 	 */
-	if (flags == DB_KEYFIRST || flags == DB_KEYLAST) {
+	if (flags == DB_KEYFIRST ||
+	    flags == DB_KEYLAST || flags == DB_NOOVERWRITE) {
 		ret = __ram_getno(dbc, key, &cp->recno, 1);
 		if (ret == 0 || ret == DB_NOTFOUND)
-			ret = __ram_add(dbc, &cp->recno, data, 0, 0);
+			ret = __ram_add(dbc, &cp->recno, data, flags, 0);
 		return (ret);
 	}
 
@@ -642,17 +628,18 @@ __ram_c_put(dbc, key, data, flags, pgnop)
 	else
 		iiflags = flags;
 
-split:	if ((ret = __bam_rsearch(dbc, &cp->recno, S_INSERT, 1, &exact)) != 0)
+split:	if ((ret = __bam_rsearch(dbc, &cp->recno, SR_INSERT, 1, &exact)) != 0)
 		goto err;
 	/*
 	 * An inexact match is okay;  it just means we're one record past the
 	 * end, which is reasonable if we're marked deleted.
 	 */
-	DB_ASSERT(exact || CD_ISSET(cp));
+	DB_ASSERT(dbenv, exact || CD_ISSET(cp));
 
-	cp->page = cp->csp->page;
-	cp->pgno = cp->csp->page->pgno;
-	cp->indx = cp->csp->indx;
+	/* Copy the page into the cursor. */
+	STACK_TO_CURSOR(cp, ret);
+	if (ret != 0)
+		goto err;
 
 	ret = __bam_iitem(dbc, key, data, iiflags, 0);
 	t_ret = __bam_stkrel(dbc, STK_CLRDBC);
@@ -670,7 +657,8 @@ split:	if ((ret = __bam_rsearch(dbc, &cp->recno, S_INSERT, 1, &exact)) != 0)
 
 	switch (flags) {			/* Adjust the cursors. */
 	case DB_AFTER:
-		nc = __ram_ca(dbc, CA_IAFTER);
+		if ((ret = __ram_ca(dbc, CA_IAFTER, &nc)) != 0)
+			goto err;
 
 		/*
 		 * We only need to adjust this cursor forward if we truly added
@@ -682,19 +670,18 @@ split:	if ((ret = __bam_rsearch(dbc, &cp->recno, S_INSERT, 1, &exact)) != 0)
 
 		/* Only log if __ram_ca found any relevant cursors. */
 		if (nc > 0 && CURADJ_LOG(dbc) &&
-		    (ret = __bam_rcuradj_log(dbp->dbenv,
-		    dbc->txn, &lsn, 0, dbp->log_fileid, CA_IAFTER,
+		    (ret = __bam_rcuradj_log(dbp, dbc->txn, &lsn, 0, CA_IAFTER,
 		    cp->root, cp->recno, cp->order)) != 0)
 			goto err;
 		break;
 	case DB_BEFORE:
-		nc = __ram_ca(dbc, CA_IBEFORE);
+		if ((ret = __ram_ca(dbc, CA_IBEFORE, &nc)) != 0)
+			goto err;
 		--cp->recno;
 
 		/* Only log if __ram_ca found any relevant cursors. */
 		if (nc > 0 && CURADJ_LOG(dbc) &&
-		    (ret = __bam_rcuradj_log(dbp->dbenv,
-		    dbc->txn, &lsn, 0, dbp->log_fileid, CA_IBEFORE,
+		    (ret = __bam_rcuradj_log(dbp, dbc->txn, &lsn, 0, CA_IBEFORE,
 		    cp->root, cp->recno, cp->order)) != 0)
 			goto err;
 		break;
@@ -703,20 +690,25 @@ split:	if ((ret = __bam_rsearch(dbc, &cp->recno, S_INSERT, 1, &exact)) != 0)
 		 * We only need to do an adjustment if we actually
 		 * added an item, which we only would have done if the
 		 * cursor was marked deleted.
-		 *
-		 * Only log if __ram_ca found any relevant cursors.
 		 */
-		if (CD_ISSET(cp) && __ram_ca(dbc, CA_ICURRENT) > 0 &&
-		    CURADJ_LOG(dbc) && (ret = __bam_rcuradj_log(
-		    dbp->dbenv, dbc->txn, &lsn, 0, dbp->log_fileid,
+		if (!CD_ISSET(cp))
+			break;
+
+		/* Only log if __ram_ca found any relevant cursors. */
+		if ((ret = __ram_ca(dbc, CA_ICURRENT, &nc)) != 0)
+			goto err;
+		if (nc > 0 && CURADJ_LOG(dbc) &&
+		    (ret = __bam_rcuradj_log(dbp, dbc->txn, &lsn, 0,
 		    CA_ICURRENT, cp->root, cp->recno, cp->order)) != 0)
 			goto err;
 		break;
+	default:
+		break;
 	}
 
 	/* Return the key if we've created a new record. */
 	if (!F_ISSET(dbc, DBC_OPD) && (flags == DB_AFTER || flags == DB_BEFORE))
-		ret = __db_retcopy(dbp, key, &cp->recno,
+		ret = __db_retcopy(dbenv, key, &cp->recno,
 		    sizeof(cp->recno), &dbc->rkey->data, &dbc->rkey->ulen);
 
 	/* The cursor was reset, no further delete adjustment is necessary. */
@@ -729,39 +721,37 @@ err:	CD_CLR(cp);
  * __ram_ca --
  *	Adjust cursors.  Returns the number of relevant cursors.
  *
- * PUBLIC: int __ram_ca __P((DBC *, ca_recno_arg));
+ * PUBLIC: int __ram_ca __P((DBC *, ca_recno_arg, int *));
  */
 int
-__ram_ca(dbc_arg, op)
+__ram_ca(dbc_arg, op, foundp)
 	DBC *dbc_arg;
 	ca_recno_arg op;
+	int *foundp;
 {
 	BTREE_CURSOR *cp, *cp_arg;
 	DB *dbp, *ldbp;
 	DB_ENV *dbenv;
 	DBC *dbc;
 	db_recno_t recno;
-	int adjusted, found;
 	u_int32_t order;
+	int adjusted, found;
 
 	dbp = dbc_arg->dbp;
 	dbenv = dbp->dbenv;
 	cp_arg = (BTREE_CURSOR *)dbc_arg->internal;
 	recno = cp_arg->recno;
 
-	found = 0;
-
 	/*
 	 * It only makes sense to adjust cursors if we're a renumbering
 	 * recno;  we should only be called if this is one.
 	 */
-	DB_ASSERT(F_ISSET(cp_arg, C_RENUMBER));
+	DB_ASSERT(dbenv, F_ISSET(cp_arg, C_RENUMBER));
 
-	MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_LOCK(dbenv, dbenv->mtx_dblist);
 	/*
 	 * Adjust the cursors.  See the comment in __bam_ca_delete().
-	 */
-	/*
+	 *
 	 * If we're doing a delete, we need to find the highest
 	 * order of any cursor currently pointing at this item,
 	 * so we can assign a higher order to the newly deleted
@@ -769,33 +759,34 @@ __ram_ca(dbc_arg, op)
 	 * the cursor list.
 	 */
 	if (op == CA_DELETE) {
-		order = 1;
-		for (ldbp = __dblist_get(dbenv, dbp->adj_fileid);
+		FIND_FIRST_DB_MATCH(dbenv, dbp, ldbp);
+		for (order = 1;
 		    ldbp != NULL && ldbp->adj_fileid == dbp->adj_fileid;
-		    ldbp = LIST_NEXT(ldbp, dblistlinks)) {
-			MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
-			for (dbc = TAILQ_FIRST(&ldbp->active_queue);
-			    dbc != NULL; dbc = TAILQ_NEXT(dbc, links)) {
+		    ldbp = TAILQ_NEXT(ldbp, dblistlinks)) {
+			MUTEX_LOCK(dbenv, dbp->mutex);
+			TAILQ_FOREACH(dbc, &ldbp->active_queue, links) {
 				cp = (BTREE_CURSOR *)dbc->internal;
 				if (cp_arg->root == cp->root &&
 				    recno == cp->recno && CD_ISSET(cp) &&
-				    order <= cp->order)
+				    order <= cp->order &&
+				    !MVCC_SKIP_CURADJ(dbc, cp->root))
 					order = cp->order + 1;
 			}
-			MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+			MUTEX_UNLOCK(dbenv, dbp->mutex);
 		}
 	} else
 		order = INVALID_ORDER;
 
 	/* Now go through and do the actual adjustments. */
-	for (ldbp = __dblist_get(dbenv, dbp->adj_fileid);
+	FIND_FIRST_DB_MATCH(dbenv, dbp, ldbp);
+	for (found = 0;
 	    ldbp != NULL && ldbp->adj_fileid == dbp->adj_fileid;
-	    ldbp = LIST_NEXT(ldbp, dblistlinks)) {
-		MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
-		for (dbc = TAILQ_FIRST(&ldbp->active_queue);
-		    dbc != NULL; dbc = TAILQ_NEXT(dbc, links)) {
+	    ldbp = TAILQ_NEXT(ldbp, dblistlinks)) {
+		MUTEX_LOCK(dbenv, dbp->mutex);
+		TAILQ_FOREACH(dbc, &ldbp->active_queue, links) {
 			cp = (BTREE_CURSOR *)dbc->internal;
-			if (cp_arg->root != cp->root)
+			if (cp_arg->root != cp->root ||
+			    MVCC_SKIP_CURADJ(dbc, cp->root))
 				continue;
 			++found;
 			adjusted = 0;
@@ -837,7 +828,7 @@ __ram_ca(dbc_arg, op)
 				 * marked undeleted and point to the new
 				 * item.
 				 */
-				DB_ASSERT(CD_ISSET(cp_arg));
+				DB_ASSERT(dbenv, CD_ISSET(cp_arg));
 				if (C_EQUAL(cp_arg, cp)) {
 					CD_CLR(cp);
 					break;
@@ -861,11 +852,13 @@ iafter:				if (!adjusted && C_LESSTHAN(cp_arg, cp)) {
 				break;
 			}
 		}
-		MUTEX_THREAD_UNLOCK(dbp->dbenv, dbp->mutexp);
+		MUTEX_UNLOCK(dbp->dbenv, dbp->mutex);
 	}
-	MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_UNLOCK(dbenv, dbenv->mtx_dblist);
 
-	return (found);
+	if (foundp != NULL)
+		*foundp = found;
+	return (0);
 }
 
 /*
@@ -888,7 +881,7 @@ __ram_getno(dbc, key, rep, can_create)
 
 	/* Check the user's record number. */
 	if ((recno = *(db_recno_t *)key->data) == 0) {
-		__db_err(dbp->dbenv, "illegal record number of 0");
+		__db_errx(dbp->dbenv, "illegal record number of 0");
 		return (EINVAL);
 	}
 	if (rep != NULL)
@@ -967,17 +960,19 @@ static int
 __ram_source(dbp)
 	DB *dbp;
 {
+	DB_ENV *dbenv;
 	BTREE *t;
 	char *source;
 	int ret;
 
+	dbenv = dbp->dbenv;
 	t = dbp->bt_internal;
 
 	/* Find the real name, and swap out the one we had before. */
-	if ((ret = __db_appname(dbp->dbenv,
-	    DB_APP_DATA, NULL, t->re_source, 0, NULL, &source)) != 0)
+	if ((ret = __db_appname(dbenv,
+	    DB_APP_DATA, t->re_source, 0, NULL, &source)) != 0)
 		return (ret);
-	__os_freestr(dbp->dbenv, t->re_source);
+	__os_free(dbenv, t->re_source);
 	t->re_source = source;
 
 	/*
@@ -986,9 +981,9 @@ __ram_source(dbp)
 	 * much care other than we'll complain if there are any modifications
 	 * when it comes time to write the database back to the source.
 	 */
-	if ((t->re_fp = fopen(t->re_source, "r")) == NULL) {
-		ret = errno;
-		__db_err(dbp->dbenv, "%s: %s", t->re_source, db_strerror(ret));
+	if ((t->re_fp = fopen(t->re_source, "rb")) == NULL) {
+		ret = __os_get_errno();
+		__db_err(dbenv, ret, "%s", t->re_source);
 		return (ret);
 	}
 
@@ -1018,6 +1013,7 @@ __ram_writeback(dbp)
 	t = dbp->bt_internal;
 	dbenv = dbp->dbenv;
 	fp = NULL;
+	pad = NULL;
 
 	/* If the file wasn't modified, we're done. */
 	if (!t->re_modified)
@@ -1029,8 +1025,20 @@ __ram_writeback(dbp)
 		return (0);
 	}
 
+	/*
+	 * We step through the records, writing each one out.  Use the record
+	 * number and the dbp->get() function, instead of a cursor, so we find
+	 * and write out "deleted" or non-existent records.  The DB handle may
+	 * be threaded, so allocate memory as we go.
+	 */
+	memset(&key, 0, sizeof(key));
+	key.size = sizeof(db_recno_t);
+	key.data = &keyno;
+	memset(&data, 0, sizeof(data));
+	F_SET(&data, DB_DBT_REALLOC);
+
 	/* Allocate a cursor. */
-	if ((ret = dbp->cursor(dbp, NULL, &dbc, 0)) != 0)
+	if ((ret = __db_cursor(dbp, NULL, &dbc, 0)) != 0)
 		return (ret);
 
 	/*
@@ -1056,59 +1064,48 @@ __ram_writeback(dbp)
 	 */
 	if ((ret =
 	    __ram_update(dbc, DB_MAX_RECORDS, 0)) != 0 && ret != DB_NOTFOUND)
-		return (ret);
+		goto err;
 
 	/*
 	 * Close any existing file handle and re-open the file, truncating it.
 	 */
 	if (t->re_fp != NULL) {
 		if (fclose(t->re_fp) != 0) {
-			ret = errno;
+			ret = __os_get_errno();
+			__db_err(dbenv, ret, "%s", t->re_source);
 			goto err;
 		}
 		t->re_fp = NULL;
 	}
-	if ((fp = fopen(t->re_source, "w")) == NULL) {
-		ret = errno;
-		__db_err(dbenv, "%s: %s", t->re_source, db_strerror(ret));
+	if ((fp = fopen(t->re_source, "wb")) == NULL) {
+		ret = __os_get_errno();
+		__db_err(dbenv, ret, "%s", t->re_source);
 		goto err;
 	}
 
-	/*
-	 * We step through the records, writing each one out.  Use the record
-	 * number and the dbp->get() function, instead of a cursor, so we find
-	 * and write out "deleted" or non-existent records.  The DB handle may
-	 * be threaded, so allocate memory as we go.
-	 */
-	memset(&key, 0, sizeof(key));
-	key.size = sizeof(db_recno_t);
-	key.data = &keyno;
-	memset(&data, 0, sizeof(data));
-	F_SET(&data, DB_DBT_MALLOC);
-
 	/*
 	 * We'll need the delimiter if we're doing variable-length records,
 	 * and the pad character if we're doing fixed-length records.
 	 */
 	delim = t->re_delim;
-	if (F_ISSET(dbp, DB_RE_FIXEDLEN)) {
-		if ((ret = __os_malloc(dbenv, t->re_len, &pad)) != 0)
-			goto err;
-		memset(pad, t->re_pad, t->re_len);
-	} else
-		COMPQUIET(pad, NULL);
 	for (keyno = 1;; ++keyno) {
-		switch (ret = dbp->get(dbp, NULL, &key, &data, 0)) {
+		switch (ret = __db_get(dbp, NULL, &key, &data, 0)) {
 		case 0:
-			if (data.size != 0 && (u_int32_t)fwrite(
-			    data.data, 1, data.size, fp) != data.size)
+			if (data.size != 0 &&
+			    fwrite(data.data, 1, data.size, fp) != data.size)
 				goto write_err;
 			break;
 		case DB_KEYEMPTY:
-			if (F_ISSET(dbp, DB_RE_FIXEDLEN) &&
-			    (u_int32_t)fwrite(pad, 1, t->re_len, fp) !=
-			    t->re_len)
-				goto write_err;
+			if (F_ISSET(dbp, DB_AM_FIXEDLEN)) {
+				if (pad == NULL) {
+					if ((ret = __os_malloc(
+					    dbenv, t->re_len, &pad)) != 0)
+						goto err;
+					memset(pad, t->re_pad, t->re_len);
+				}
+				if (fwrite(pad, 1, t->re_len, fp) != t->re_len)
+					goto write_err;
+			}
 			break;
 		case DB_NOTFOUND:
 			ret = 0;
@@ -1116,12 +1113,11 @@ __ram_writeback(dbp)
 		default:
 			goto err;
 		}
-		if (!F_ISSET(dbp, DB_RE_FIXEDLEN) &&
+		if (!F_ISSET(dbp, DB_AM_FIXEDLEN) &&
 		    fwrite(&delim, 1, 1, fp) != 1) {
-write_err:		ret = errno;
-			__db_err(dbp->dbenv,
-			    "%s: write failed to backing file: %s",
-			    t->re_source, strerror(ret));
+write_err:		ret = __os_get_errno();
+			__db_err(dbenv, ret,
+			    "%s: write failed to backing file", t->re_source);
 			goto err;
 		}
 	}
@@ -1129,18 +1125,21 @@ write_err:		ret = errno;
 err:
 done:	/* Close the file descriptor. */
 	if (fp != NULL && fclose(fp) != 0) {
+		t_ret = __os_get_errno();
+		__db_err(dbenv, t_ret, "%s", t->re_source);
 		if (ret == 0)
-			ret = errno;
-		__db_err(dbenv, "%s: %s", t->re_source, db_strerror(errno));
+			ret = t_ret;
 	}
 
 	/* Discard the cursor. */
-	if ((t_ret = dbc->c_close(dbc)) != 0 && ret == 0)
+	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
 		ret = t_ret;
 
 	/* Discard memory allocated to hold the data items. */
 	if (data.data != NULL)
-		__os_free(dbenv, data.data, data.ulen);
+		__os_ufree(dbenv, data.data);
+	if (pad != NULL)
+		__os_free(dbenv, pad);
 
 	if (ret == 0)
 		t->re_modified = 0;
@@ -1176,7 +1175,7 @@ __ram_sread(dbc, top)
 	 * The record data return memory is used by __bam_iitem, which
 	 * we'll indirectly call, so use the key so as not to collide.
 	 */
-	len = F_ISSET(dbp, DB_RE_FIXEDLEN) ? t->re_len : 256;
+	len = F_ISSET(dbp, DB_AM_FIXEDLEN) ? t->re_len : 256;
 	rdata = &dbc->my_rkey;
 	if (rdata->ulen < len) {
 		if ((ret = __os_realloc(
@@ -1185,16 +1184,16 @@ __ram_sread(dbc, top)
 			rdata->data = NULL;
 			return (ret);
 		}
-		rdata->ulen = len;
+		rdata->ulen = (u_int32_t)len;
 	}
 
 	memset(&data, 0, sizeof(data));
 	while (recno < top) {
 		data.data = rdata->data;
 		data.size = 0;
-		if (F_ISSET(dbp, DB_RE_FIXEDLEN))
+		if (F_ISSET(dbp, DB_AM_FIXEDLEN))
 			for (len = t->re_len; len > 0; --len) {
-				if ((ch = getc(t->re_fp)) == EOF) {
+				if ((ch = fgetc(t->re_fp)) == EOF) {
 					if (data.size == 0)
 						goto eof;
 					break;
@@ -1203,7 +1202,7 @@ __ram_sread(dbc, top)
 			}
 		else
 			for (;;) {
-				if ((ch = getc(t->re_fp)) == EOF) {
+				if ((ch = fgetc(t->re_fp)) == EOF) {
 					if (data.size == 0)
 						goto eof;
 					break;
@@ -1261,18 +1260,26 @@ __ram_add(dbc, recnop, data, flags, bi_flags)
 	u_int32_t flags, bi_flags;
 {
 	BTREE_CURSOR *cp;
-	int exact, ret, stack;
+	int exact, ret, stack, t_ret;
 
 	cp = (BTREE_CURSOR *)dbc->internal;
 
 retry:	/* Find the slot for insertion. */
 	if ((ret = __bam_rsearch(dbc, recnop,
-	    S_INSERT | (flags == DB_APPEND ? S_APPEND : 0), 1, &exact)) != 0)
+	    SR_INSERT | (flags == DB_APPEND ? SR_APPEND : 0), 1, &exact)) != 0)
 		return (ret);
 	stack = 1;
-	cp->page = cp->csp->page;
-	cp->pgno = cp->csp->page->pgno;
-	cp->indx = cp->csp->indx;
+
+	/* Copy the page into the cursor. */
+	STACK_TO_CURSOR(cp, ret);
+	if (ret != 0)
+		goto err;
+
+	if (exact && flags == DB_NOOVERWRITE && !CD_ISSET(cp) &&
+	    !B_DISSET(GET_BKEYDATA(dbc->dbp, cp->page, cp->indx)->type)) {
+		ret = DB_KEYEXIST;
+		goto err;
+	}
 
 	/*
 	 * The application may modify the data based on the selected record
@@ -1322,8 +1329,8 @@ retry:	/* Find the slot for insertion. */
 		goto err;
 	}
 
-err:	if (stack)
-		__bam_stkrel(dbc, STK_CLRDBC);
+err:	if (stack && (t_ret = __bam_stkrel(dbc, STK_CLRDBC)) != 0 && ret == 0)
+		ret = t_ret;
 
 	return (ret);
 }
