@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1990, 1993, 1994, 1995, 1996
@@ -35,23 +35,17 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: bt_rsearch.c,v 12.11 2006/08/24 14:44:44 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: bt_rsearch.c,v 11.27 2001/07/24 18:31:00 bostic Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-#endif
-
 #include "db_int.h"
-#include "db_page.h"
-#include "btree.h"
-#include "db_shash.h"
-#include "lock.h"
+#include "dbinc/db_page.h"
+#include "dbinc/btree.h"
+#include "dbinc/lock.h"
+#include "dbinc/mp.h"
 
 /*
  * __bam_rsearch --
@@ -77,11 +71,12 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 	db_lockmode_t lock_mode;
 	db_pgno_t pg;
 	db_recno_t recno, t_recno, total;
-	int ret, stack;
+	int ret, stack, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
 	cp = (BTREE_CURSOR *)dbc->internal;
+	h = NULL;
 
 	BT_STK_CLR(cp);
 
@@ -100,41 +95,14 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 	 *
 	 * Retrieve the root page.
 	 */
-	pg = cp->root;
-	stack = LF_ISSET(S_STACK) ? 1 : 0;
-	lock_mode = stack ? DB_LOCK_WRITE : DB_LOCK_READ;
-	if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
-		return (ret);
-	if ((ret = mpf->get(mpf, &pg, 0, &h)) != 0) {
-		/* Did not read it, so we can release the lock */
-		(void)__LPUT(dbc, lock);
-		return (ret);
-	}
 
-	/*
-	 * Decide if we need to save this page; if we do, write lock it.
-	 * We deliberately don't lock-couple on this call.  If the tree
-	 * is tiny, i.e., one page, and two threads are busily updating
-	 * the root page, we're almost guaranteed deadlocks galore, as
-	 * each one gets a read lock and then blocks the other's attempt
-	 * for a write lock.
-	 */
-	if (!stack &&
-	    ((LF_ISSET(S_PARENT) && (u_int8_t)(stop + 1) >= h->level) ||
-	    (LF_ISSET(S_WRITE) && h->level == LEAFLEVEL))) {
-		(void)mpf->put(mpf, h, 0);
-		(void)__LPUT(dbc, lock);
-		lock_mode = DB_LOCK_WRITE;
-		if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
-			return (ret);
-		if ((ret = mpf->get(mpf, &pg, 0, &h)) != 0) {
-			/* Did not read it, so we can release the lock */
-			(void)__LPUT(dbc, lock);
-			return (ret);
-		}
-		stack = 1;
-	}
+	if ((ret = __bam_get_root(dbc, cp->root, stop, flags, &stack)) != 0)
+		return (ret);
+	lock_mode = cp->csp->lock_mode;
+	lock = cp->csp->lock;
+	h = cp->csp->page;
 
+	BT_STK_CLR(cp);
 	/*
 	 * If appending to the tree, set the record number now -- we have the
 	 * root page locked.
@@ -148,7 +116,7 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 	 * a fast check now.
 	 */
 	total = RE_NREC(h);
-	if (LF_ISSET(S_APPEND)) {
+	if (LF_ISSET(SR_APPEND)) {
 		*exactp = 0;
 		*recnop = recno = total + 1;
 	} else {
@@ -157,7 +125,7 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 			*exactp = 1;
 		else {
 			*exactp = 0;
-			if (!LF_ISSET(S_PAST_EOF) || recno > total + 1) {
+			if (!LF_ISSET(SR_PAST_EOF) || recno > total + 1) {
 				/*
 				 * Keep the page locked for serializability.
 				 *
@@ -166,9 +134,11 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 				 * eliminate any concurrency.  A possible fix
 				 * would be to lock the last leaf page instead.
 				 */
-				(void)mpf->put(mpf, h, 0);
-				(void)__TLPUT(dbc, lock);
-				return (DB_NOTFOUND);
+				ret = __memp_fput(mpf, h, 0);
+				if ((t_ret =
+				    __TLPUT(dbc, lock)) != 0 && ret == 0)
+					ret = t_ret;
+				return (ret == 0 ? DB_NOTFOUND : ret);
 			}
 		}
 	}
@@ -198,14 +168,20 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 			for (t_recno = 0, indx = 0;; indx += adjust) {
 				if (indx >= NUM_ENT(h)) {
 					*exactp = 0;
-					if (!LF_ISSET(S_PAST_EOF) ||
+					if (!LF_ISSET(SR_PAST_EOF) ||
 					    recno > t_recno + 1) {
-						ret = DB_NOTFOUND;
+						ret = __memp_fput(mpf, h, 0);
+						h = NULL;
+						if ((t_ret = __TLPUT(dbc,
+						    lock)) != 0 && ret == 0)
+							ret = t_ret;
+						if (ret == 0)
+							ret = DB_NOTFOUND;
 						goto err;
 					}
 				}
-				if (!B_DISSET(
-				    GET_BKEYDATA(h, indx + deloffset)->type) &&
+				if (!B_DISSET(GET_BKEYDATA(dbp, h,
+				    indx + deloffset)->type) &&
 				    ++t_recno == recno)
 					break;
 			}
@@ -218,7 +194,7 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 			return (0);
 		case P_IBTREE:
 			for (indx = 0, top = NUM_ENT(h);;) {
-				bi = GET_BINTERNAL(h, indx);
+				bi = GET_BINTERNAL(dbp, h, indx);
 				if (++indx == top || total + bi->nrecs >= recno)
 					break;
 				total += bi->nrecs;
@@ -237,7 +213,7 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 			return (0);
 		case P_IRECNO:
 			for (indx = 0, top = NUM_ENT(h);;) {
-				ri = GET_RINTERNAL(h, indx);
+				ri = GET_RINTERNAL(dbp, h, indx);
 				if (++indx == top || total + ri->nrecs >= recno)
 					break;
 				total += ri->nrecs;
@@ -249,19 +225,20 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 		}
 		--indx;
 
+		/* Return if this is the lowest page wanted. */
+		if (stop == LEVEL(h)) {
+			BT_STK_ENTER(dbp->dbenv,
+			    cp, h, indx, lock, lock_mode, ret);
+			if (ret != 0)
+				goto err;
+			return (0);
+		}
 		if (stack) {
-			/* Return if this is the lowest page wanted. */
-			if (LF_ISSET(S_PARENT) && stop == h->level) {
-				BT_STK_ENTER(dbp->dbenv,
-				    cp, h, indx, lock, lock_mode, ret);
-				if (ret != 0)
-					goto err;
-				return (0);
-			}
 			BT_STK_PUSH(dbp->dbenv,
 			    cp, h, indx, lock, lock_mode, ret);
 			if (ret != 0)
 				goto err;
+			h = NULL;
 
 			lock_mode = DB_LOCK_WRITE;
 			if ((ret =
@@ -273,34 +250,40 @@ __bam_rsearch(dbc, recnop, flags, stop, exactp)
 			 * page in the stack.  If we do, write lock it and
 			 * never unlock it.
 			 */
-			if ((LF_ISSET(S_PARENT) &&
-			    (u_int8_t)(stop + 1) >= (u_int8_t)(h->level - 1)) ||
-			    (h->level - 1) == LEAFLEVEL)
+			if ((LF_ISSET(SR_PARENT) &&
+			    (u_int8_t)(stop + 1) >= (u_int8_t)(LEVEL(h) - 1)) ||
+			    (LEVEL(h) - 1) == LEAFLEVEL)
 				stack = 1;
 
-			(void)mpf->put(mpf, h, 0);
+			if ((ret = __memp_fput(mpf, h, 0)) != 0)
+				goto err;
+			h = NULL;
 
 			lock_mode = stack &&
-			    LF_ISSET(S_WRITE) ? DB_LOCK_WRITE : DB_LOCK_READ;
+			    LF_ISSET(SR_WRITE) ? DB_LOCK_WRITE : DB_LOCK_READ;
 			if ((ret = __db_lget(dbc,
-			    LCK_COUPLE, pg, lock_mode, 0, &lock)) != 0) {
+			    LCK_COUPLE_ALWAYS, pg, lock_mode, 0, &lock)) != 0) {
 				/*
 				 * If we fail, discard the lock we held.  This
 				 * is OK because this only happens when we are
 				 * descending the tree holding read-locks.
 				 */
-				__LPUT(dbc, lock);
+				(void)__LPUT(dbc, lock);
 				goto err;
 			}
 		}
 
-		if ((ret = mpf->get(mpf, &pg, 0, &h)) != 0)
+		if ((ret = __memp_fget(mpf, &pg, dbc->txn, 0, &h)) != 0)
 			goto err;
 	}
 	/* NOTREACHED */
 
-err:	BT_STK_POP(cp);
+err:	if (h != NULL && (t_ret = __memp_fput(mpf, h, 0)) != 0 && ret == 0)
+		ret = t_ret;
+
+	BT_STK_POP(cp);
 	__bam_stkrel(dbc, 0);
+
 	return (ret);
 }
 
@@ -332,26 +315,28 @@ __bam_adjust(dbc, adjust)
 	for (epg = cp->sp; epg <= cp->csp; ++epg) {
 		h = epg->page;
 		if (TYPE(h) == P_IBTREE || TYPE(h) == P_IRECNO) {
-			if (DB_LOGGING(dbc)) {
-				if ((ret = __bam_cadjust_log(dbp->dbenv,
-				    dbc->txn, &LSN(h), 0, dbp->log_fileid,
-				    PGNO(h), &LSN(h), (u_int32_t)epg->indx,
-				    adjust, PGNO(h) == root_pgno ?
+			if ((ret = __memp_dirty(mpf, &h, dbc->txn, 0)) != 0)
+				return (ret);
+			epg->page = h;
+			if (DBC_LOGGING(dbc)) {
+				if ((ret = __bam_cadjust_log(dbp, dbc->txn,
+				    &LSN(h), 0, PGNO(h), &LSN(h),
+				    (u_int32_t)epg->indx, adjust,
+				    PGNO(h) == root_pgno ?
 				    CAD_UPDATEROOT : 0)) != 0)
 					return (ret);
 			} else
 				LSN_NOT_LOGGED(LSN(h));
 
 			if (TYPE(h) == P_IBTREE)
-				GET_BINTERNAL(h, epg->indx)->nrecs += adjust;
+				GET_BINTERNAL(dbp, h, epg->indx)->nrecs +=
+				    adjust;
 			else
-				GET_RINTERNAL(h, epg->indx)->nrecs += adjust;
+				GET_RINTERNAL(dbp, h, epg->indx)->nrecs +=
+				    adjust;
 
 			if (PGNO(h) == root_pgno)
 				RE_NREC_ADJ(h, adjust);
-
-			if ((ret = mpf->set(mpf, h, DB_MPOOL_DIRTY)) != 0)
-				return (ret);
 		}
 	}
 	return (0);
@@ -373,7 +358,7 @@ __bam_nrecs(dbc, rep)
 	DB_MPOOLFILE *mpf;
 	PAGE *h;
 	db_pgno_t pgno;
-	int ret;
+	int ret, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
@@ -381,25 +366,27 @@ __bam_nrecs(dbc, rep)
 	pgno = dbc->internal->root;
 	if ((ret = __db_lget(dbc, 0, pgno, DB_LOCK_READ, 0, &lock)) != 0)
 		return (ret);
-	if ((ret = mpf->get(mpf, &pgno, 0, &h)) != 0)
+	if ((ret = __memp_fget(mpf, &pgno, dbc->txn, 0, &h)) != 0)
 		return (ret);
 
 	*rep = RE_NREC(h);
 
-	(void)mpf->put(mpf, h, 0);
-	(void)__TLPUT(dbc, lock);
+	ret = __memp_fput(mpf, h, 0);
+	if ((t_ret = __TLPUT(dbc, lock)) != 0 && ret == 0)
+		ret = t_ret;
 
-	return (0);
+	return (ret);
 }
 
 /*
  * __bam_total --
  *	Return the number of records below a page.
  *
- * PUBLIC: db_recno_t __bam_total __P((PAGE *));
+ * PUBLIC: db_recno_t __bam_total __P((DB *, PAGE *));
  */
 db_recno_t
-__bam_total(h)
+__bam_total(dbp, h)
+	DB *dbp;
 	PAGE *h;
 {
 	db_recno_t nrecs;
@@ -412,25 +399,26 @@ __bam_total(h)
 	case P_LBTREE:
 		/* Check for logically deleted records. */
 		for (indx = 0; indx < top; indx += P_INDX)
-			if (!B_DISSET(GET_BKEYDATA(h, indx + O_INDX)->type))
+			if (!B_DISSET(
+			    GET_BKEYDATA(dbp, h, indx + O_INDX)->type))
 				++nrecs;
 		break;
 	case P_LDUP:
 		/* Check for logically deleted records. */
 		for (indx = 0; indx < top; indx += O_INDX)
-			if (!B_DISSET(GET_BKEYDATA(h, indx)->type))
+			if (!B_DISSET(GET_BKEYDATA(dbp, h, indx)->type))
 				++nrecs;
 		break;
 	case P_IBTREE:
 		for (indx = 0; indx < top; indx += O_INDX)
-			nrecs += GET_BINTERNAL(h, indx)->nrecs;
+			nrecs += GET_BINTERNAL(dbp, h, indx)->nrecs;
 		break;
 	case P_LRECNO:
 		nrecs = NUM_ENT(h);
 		break;
 	case P_IRECNO:
 		for (indx = 0; indx < top; indx += O_INDX)
-			nrecs += GET_RINTERNAL(h, indx)->nrecs;
+			nrecs += GET_RINTERNAL(dbp, h, indx)->nrecs;
 		break;
 	}
 
