@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1990, 1993, 1994, 1995, 1996
@@ -39,24 +39,113 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: bt_search.c,v 11.50 2004/07/23 17:21:09 bostic Exp $
+ * $Id: bt_search.c,v 12.24 2006/08/24 14:44:45 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/db_page.h"
-#include "dbinc/db_shash.h"
 #include "dbinc/btree.h"
 #include "dbinc/lock.h"
 #include "dbinc/mp.h"
 
+/*
+ * __bam_get_root --
+ *	Fetch the root of a tree and see if we want to keep
+ * it in the stack.
+ *
+ * PUBLIC: int __bam_get_root __P((DBC *, db_pgno_t, int, u_int32_t, int *));
+ */
+int
+__bam_get_root(dbc, pg, slevel, flags, stack)
+	DBC *dbc;
+	db_pgno_t pg;
+	int slevel;
+	u_int32_t flags;
+	int *stack;
+{
+	BTREE_CURSOR *cp;
+	DB *dbp;
+	DB_LOCK lock;
+	DB_MPOOLFILE *mpf;
+	PAGE *h;
+	db_lockmode_t lock_mode;
+	int ret, t_ret;
+
+	dbp = dbc->dbp;
+	mpf = dbp->mpf;
+	cp = (BTREE_CURSOR *)dbc->internal;
+	/*
+	 * If write-locking pages, we need to know whether or not to acquire a
+	 * write lock on a page before getting it.  This depends on how deep it
+	 * is in tree, which we don't know until we acquire the root page.  So,
+	 * if we need to lock the root page we may have to upgrade it later,
+	 * because we won't get the correct lock initially.
+	 *
+	 * Retrieve the root page.
+	 */
+try_again:
+	*stack = LF_ISSET(SR_STACK) &&
+	      (dbc->dbtype == DB_RECNO || F_ISSET(cp, C_RECNUM));
+	lock_mode = DB_LOCK_READ;
+	if (*stack ||
+	    LF_ISSET(SR_DEL) || (LF_ISSET(SR_NEXT) && LF_ISSET(SR_WRITE)))
+		lock_mode = DB_LOCK_WRITE;
+	if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
+		return (ret);
+	if ((ret = __memp_fget(mpf, &pg, dbc->txn, 0, &h)) != 0) {
+		/* Did not read it, so we can release the lock */
+		(void)__LPUT(dbc, lock);
+		return (ret);
+	}
+
+	/*
+	 * Decide if we need to save this page; if we do, write lock it.
+	 * We deliberately don't lock-couple on this call.  If the tree
+	 * is tiny, i.e., one page, and two threads are busily updating
+	 * the root page, we're almost guaranteed deadlocks galore, as
+	 * each one gets a read lock and then blocks the other's attempt
+	 * for a write lock.
+	 */
+	if (!*stack &&
+	    ((LF_ISSET(SR_PARENT) && (u_int8_t)(slevel + 1) >= LEVEL(h)) ||
+	    (LF_ISSET(SR_WRITE) && LEVEL(h) == LEAFLEVEL) ||
+	    (LF_ISSET(SR_START) && slevel == LEVEL(h)))) {
+		if (!STD_LOCKING(dbc))
+			goto no_relock;
+		ret = __memp_fput(mpf, h, 0);
+		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
+			return (ret);
+		lock_mode = DB_LOCK_WRITE;
+		if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
+			return (ret);
+		if ((ret = __memp_fget(mpf, &pg, dbc->txn, 0, &h)) != 0) {
+			/* Did not read it, so we can release the lock */
+			(void)__LPUT(dbc, lock);
+			return (ret);
+		}
+		if (!((LF_ISSET(SR_PARENT) &&
+		    (u_int8_t)(slevel + 1) >= LEVEL(h)) ||
+		    (LF_ISSET(SR_WRITE) && LEVEL(h) == LEAFLEVEL) ||
+		    (LF_ISSET(SR_START) && slevel == LEVEL(h)))) {
+			/* Someone else split the root, start over. */
+			ret = __memp_fput(mpf, h, 0);
+			if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+				ret = t_ret;
+			if (ret != 0)
+				return (ret);
+			goto try_again;
+		}
+no_relock:	*stack = 1;
+	}
+	BT_STK_ENTER(dbp->dbenv, cp, h, 0, lock, lock_mode, ret);
+
+	return (ret);
+}
+
 /*
  * __bam_search --
  *	Search a btree for a key.
@@ -65,17 +154,18 @@
  * PUBLIC:     const DBT *, u_int32_t, int, db_recno_t *, int *));
  */
 int
-__bam_search(dbc, root_pgno, key, flags, stop, recnop, exactp)
+__bam_search(dbc, root_pgno, key, flags, slevel, recnop, exactp)
 	DBC *dbc;
 	db_pgno_t root_pgno;
 	const DBT *key;
 	u_int32_t flags;
-	int stop, *exactp;
+	int slevel, *exactp;
 	db_recno_t *recnop;
 {
 	BTREE *t;
 	BTREE_CURSOR *cp;
 	DB *dbp;
+	DB_ENV *dbenv;
 	DB_LOCK lock;
 	DB_MPOOLFILE *mpf;
 	PAGE *h;
@@ -87,6 +177,7 @@ __bam_search(dbc, root_pgno, key, flags, stop, recnop, exactp)
 	int (*func) __P((DB *, const DBT *, const DBT *));
 
 	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
 	mpf = dbp->mpf;
 	cp = (BTREE_CURSOR *)dbc->internal;
 	h = NULL;
@@ -102,64 +193,17 @@ __bam_search(dbc, root_pgno, key, flags, stop, recnop, exactp)
 	 * deleted items, and if we are locking pairs of pages.  In addition,
 	 * if we're modifying record numbers, we have to lock the entire tree
 	 * regardless.  See btree.h for more details.
-	 *
-	 * If write-locking pages, we need to know whether or not to acquire a
-	 * write lock on a page before getting it.  This depends on how deep it
-	 * is in tree, which we don't know until we acquire the root page.  So,
-	 * if we need to lock the root page we may have to upgrade it later,
-	 * because we won't get the correct lock initially.
-	 *
-	 * Retrieve the root page.
 	 */
-try_again:
-	pg = root_pgno == PGNO_INVALID ? cp->root : root_pgno;
-	stack = LF_ISSET(S_STACK) && F_ISSET(cp, C_RECNUM);
-	lock_mode = stack ? DB_LOCK_WRITE : DB_LOCK_READ;
-	if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
-		return (ret);
-	if ((ret = __memp_fget(mpf, &pg, 0, &h)) != 0) {
-		/* Did not read it, so we can release the lock */
-		(void)__LPUT(dbc, lock);
-		return (ret);
-	}
 
-	/*
-	 * Decide if we need to save this page; if we do, write lock it.
-	 * We deliberately don't lock-couple on this call.  If the tree
-	 * is tiny, i.e., one page, and two threads are busily updating
-	 * the root page, we're almost guaranteed deadlocks galore, as
-	 * each one gets a read lock and then blocks the other's attempt
-	 * for a write lock.
-	 */
-	if (!stack &&
-	    ((LF_ISSET(S_PARENT) && (u_int8_t)(stop + 1) >= h->level) ||
-	    (LF_ISSET(S_WRITE) && h->level == LEAFLEVEL))) {
-		ret = __memp_fput(mpf, h, 0);
-		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
-			ret = t_ret;
-		if (ret != 0)
-			return (ret);
-		lock_mode = DB_LOCK_WRITE;
-		if ((ret = __db_lget(dbc, 0, pg, lock_mode, 0, &lock)) != 0)
-			return (ret);
-		if ((ret = __memp_fget(mpf, &pg, 0, &h)) != 0) {
-			/* Did not read it, so we can release the lock */
-			(void)__LPUT(dbc, lock);
-			return (ret);
-		}
-		if (!((LF_ISSET(S_PARENT) &&
-		    (u_int8_t)(stop + 1) >= h->level) ||
-		    (LF_ISSET(S_WRITE) && h->level == LEAFLEVEL))) {
-			/* Someone else split the root, start over. */
-			ret = __memp_fput(mpf, h, 0);
-			if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
-				ret = t_ret;
-			if (ret != 0)
-				return (ret);
-			goto try_again;
-		}
-		stack = 1;
-	}
+	if (root_pgno == PGNO_INVALID)
+		root_pgno = cp->root;
+	if ((ret = __bam_get_root(dbc, root_pgno, slevel, flags, &stack)) != 0)
+		return (ret);
+	lock_mode = cp->csp->lock_mode;
+	lock = cp->csp->lock;
+	h = cp->csp->page;
+
+	BT_STK_CLR(cp);
 
 	/* Choose a comparison function. */
 	func = F_ISSET(dbc, DBC_OPD) ?
@@ -168,6 +212,23 @@ try_again:
 
 	for (;;) {
 		inp = P_INP(dbp, h);
+		adjust = TYPE(h) == P_LBTREE ? P_INDX : O_INDX;
+		if (LF_ISSET(SR_MIN | SR_MAX)) {
+			if (LF_ISSET(SR_MIN) || NUM_ENT(h) == 0)
+				indx = 0;
+			else if (TYPE(h) == P_LBTREE)
+				indx = NUM_ENT(h) - 2;
+			else
+				indx = NUM_ENT(h) - 1;
+
+			if (LEVEL(h) == LEAFLEVEL ||
+			     (!LF_ISSET(SR_START) && LEVEL(h) == slevel)) {
+				if (LF_ISSET(SR_NEXT))
+					goto get_next;
+				goto found;
+			}
+			goto next;
+		}
 		/*
 		 * Do a binary search on the current page.  If we're searching
 		 * a Btree leaf page, we have to walk the indices in groups of
@@ -175,16 +236,20 @@ try_again:
 		 * page, they're an index per page item.  If we find an exact
 		 * match on a leaf page, we're done.
 		 */
-		adjust = TYPE(h) == P_LBTREE ? P_INDX : O_INDX;
 		for (base = 0,
 		    lim = NUM_ENT(h) / (db_indx_t)adjust; lim != 0; lim >>= 1) {
 			indx = base + ((lim >> 1) * adjust);
-			if ((ret =
-			    __bam_cmp(dbp, key, h, indx, func, &cmp)) != 0)
+			if ((ret = __bam_cmp(dbp, dbc->txn, key,
+			     h, indx, func, &cmp)) != 0)
 				goto err;
 			if (cmp == 0) {
-				if (TYPE(h) == P_LBTREE || TYPE(h) == P_LDUP)
+				if (LEVEL(h) == LEAFLEVEL ||
+				    (!LF_ISSET(SR_START) &&
+				    LEVEL(h) == slevel)) {
+					if (LF_ISSET(SR_NEXT))
+						goto get_next;
 					goto found;
+				}
 				goto next;
 			}
 			if (cmp > 0) {
@@ -197,19 +262,21 @@ try_again:
 		 * No match found.  Base is the smallest index greater than
 		 * key and may be zero or a last + O_INDX index.
 		 *
-		 * If it's a leaf page, return base as the "found" value.
+		 * If it's a leaf page or the stopping point,
+		 * return base as the "found" value.
 		 * Delete only deletes exact matches.
 		 */
-		if (TYPE(h) == P_LBTREE || TYPE(h) == P_LDUP) {
+		if (LEVEL(h) == LEAFLEVEL ||
+		    (!LF_ISSET(SR_START) && LEVEL(h) == slevel)) {
 			*exactp = 0;
 
-			if (LF_ISSET(S_EXACT)) {
+			if (LF_ISSET(SR_EXACT)) {
 				ret = DB_NOTFOUND;
 				goto err;
 			}
 
-			if (LF_ISSET(S_STK_ONLY)) {
-				BT_STK_NUM(dbp->dbenv, cp, h, base, ret);
+			if (LF_ISSET(SR_STK_ONLY)) {
+				BT_STK_NUM(dbenv, cp, h, base, ret);
 				if ((t_ret =
 				    __LPUT(dbc, lock)) != 0 && ret == 0)
 					ret = t_ret;
@@ -218,6 +285,43 @@ try_again:
 					ret = t_ret;
 				return (ret);
 			}
+			if (LF_ISSET(SR_NEXT)) {
+get_next:			/*
+				 * The caller could have asked for a NEXT
+				 * at the root if the tree recently collapsed.
+				 */
+				if (PGNO(h) == root_pgno) {
+					ret = DB_NOTFOUND;
+					goto err;
+				}
+				/*
+				 * Save the root of the subtree
+				 * and drop the rest of the subtree
+				 * and search down again starting at
+				 * the next child.
+				 */
+				if ((ret = __LPUT(dbc, lock)) != 0)
+					goto err;
+				if ((ret = __memp_fput(mpf, h, 0)) != 0)
+					goto err;
+				h = NULL;
+				LF_SET(SR_MIN);
+				LF_CLR(SR_NEXT);
+				indx = cp->sp->indx + 1;
+				if (indx == NUM_ENT(cp->sp->page)) {
+					ret = DB_NOTFOUND;
+					cp->csp++;
+					goto err;
+				}
+				h = cp->sp->page;
+				cp->sp->page = NULL;
+				lock = cp->sp->lock;
+				LOCK_INIT(cp->sp->lock);
+				if ((ret = __bam_stkrel(dbc, STK_NOLOCK)) != 0)
+					goto err;
+				stack = 1;
+				goto next;
+			}
 
 			/*
 			 * !!!
@@ -227,8 +331,9 @@ try_again:
 			 * to find an undeleted record.  This is handled by the
 			 * calling routine.
 			 */
-			BT_STK_ENTER(dbp->dbenv,
-			    cp, h, base, lock, lock_mode, ret);
+			if (LF_ISSET(SR_DEL) && cp->csp == cp->sp)
+				cp->csp++;
+			BT_STK_ENTER(dbenv, cp, h, base, lock, lock_mode, ret);
 			if (ret != 0)
 				goto err;
 			return (0);
@@ -252,9 +357,13 @@ next:		if (recnop != NULL)
 
 		pg = GET_BINTERNAL(dbp, h, indx)->pgno;
 
-		if (LF_ISSET(S_STK_ONLY)) {
-			if (stop == h->level) {
-				BT_STK_NUM(dbp->dbenv, cp, h, indx, ret);
+		/* See if we are at the level to start stacking. */
+		if (LF_ISSET(SR_START) && slevel == LEVEL(h))
+			stack = 1;
+
+		if (LF_ISSET(SR_STK_ONLY)) {
+			if (slevel == LEVEL(h)) {
+				BT_STK_NUM(dbenv, cp, h, indx, ret);
 				if ((t_ret =
 				    __LPUT(dbc, lock)) != 0 && ret == 0)
 					ret = t_ret;
@@ -263,7 +372,7 @@ next:		if (recnop != NULL)
 					ret = t_ret;
 				return (ret);
 			}
-			BT_STK_NUMPUSH(dbp->dbenv, cp, h, indx, ret);
+			BT_STK_NUMPUSH(dbenv, cp, h, indx, ret);
 			(void)__memp_fput(mpf, h, 0);
 			h = NULL;
 			if ((ret = __db_lget(dbc,
@@ -278,14 +387,25 @@ next:		if (recnop != NULL)
 			}
 		} else if (stack) {
 			/* Return if this is the lowest page wanted. */
-			if (LF_ISSET(S_PARENT) && stop == h->level) {
-				BT_STK_ENTER(dbp->dbenv,
+			if (LF_ISSET(SR_PARENT) && slevel == LEVEL(h)) {
+				BT_STK_ENTER(dbenv,
 				    cp, h, indx, lock, lock_mode, ret);
 				if (ret != 0)
 					goto err;
 				return (0);
 			}
-			BT_STK_PUSH(dbp->dbenv,
+			if (LF_ISSET(SR_DEL) && NUM_ENT(h) > 1) {
+				/*
+				 * There was a page with a singleton pointer
+				 * to a non-empty subtree.
+				 */
+				cp->csp--;
+				if ((ret = __bam_stkrel(dbc, STK_NOLOCK)) != 0)
+					goto err;
+				stack = 0;
+				goto do_del;
+			}
+			BT_STK_PUSH(dbenv,
 			    cp, h, indx, lock, lock_mode, ret);
 			if (ret != 0)
 				goto err;
@@ -301,17 +421,70 @@ next:		if (recnop != NULL)
 			 * page in the return stack.  If so, lock it and never
 			 * unlock it.
 			 */
-			if ((LF_ISSET(S_PARENT) &&
-			    (u_int8_t)(stop + 1) >= (u_int8_t)(h->level - 1)) ||
-			    (h->level - 1) == LEAFLEVEL)
+			if ((LF_ISSET(SR_PARENT) &&
+			    (u_int8_t)(slevel + 1) >= (LEVEL(h) - 1)) ||
+			    (LEVEL(h) - 1) == LEAFLEVEL)
 				stack = 1;
 
-			if ((ret = __memp_fput(mpf, h, 0)) != 0)
-				goto err;
-			h = NULL;
+			/*
+			 * Returning a subtree.  See if we have hit the start
+			 * point if so save the parent and set stack.
+			 * Otherwise free the parent and temporarily
+			 * save this one.
+			 * For SR_DEL we need to find a page with 1 entry.
+			 * For SR_NEXT we want find the minimal subtree
+			 * that contains the key and the next page.
+			 * We save pages as long as we are at the right
+			 * edge of the subtree.  When we leave the right
+			 * edge, then drop the subtree.
+			 */
+			if (!LF_ISSET(SR_DEL | SR_NEXT)) {
+				if ((ret = __memp_fput(mpf, h, 0)) != 0)
+					goto err;
+				goto lock_next;
+			}
 
-			lock_mode = stack &&
-			    LF_ISSET(S_WRITE) ? DB_LOCK_WRITE : DB_LOCK_READ;
+			if ((LF_ISSET(SR_DEL) && NUM_ENT(h) == 1)) {
+				stack = 1;
+				LF_SET(SR_WRITE);
+				/* Push the parent. */
+				cp->csp++;
+				/* Push this node. */
+				BT_STK_PUSH(dbenv, cp, h,
+				     indx, lock, lock_mode, ret);
+				if (ret != 0)
+					goto err;
+				LOCK_INIT(lock);
+			} else {
+			/*
+			 * See if we want to save the tree so far.
+			 * If we are looking for the next key,
+			 * then we must save this node if we are
+			 * at the end of the page.  If not then
+			 * discard anything we have saved so far.
+			 * For delete only keep one node until
+			 * we find a singleton.
+			 */
+do_del:				if (cp->csp->page != NULL) {
+					if (LF_ISSET(SR_NEXT) &&
+					     indx == NUM_ENT(h) - 1)
+						cp->csp++;
+					else if ((ret =
+					    __bam_stkrel(dbc, STK_NOLOCK)) != 0)
+						goto err;
+				}
+				/* Save this node. */
+				BT_STK_ENTER(dbenv, cp,
+				    h, indx, lock, lock_mode, ret);
+				if (ret != 0)
+					goto err;
+				LOCK_INIT(lock);
+			}
+
+lock_next:		h = NULL;
+
+			if (stack && LF_ISSET(SR_WRITE))
+				lock_mode = DB_LOCK_WRITE;
 			if ((ret = __db_lget(dbc,
 			    LCK_COUPLE_ALWAYS, pg, lock_mode, 0, &lock)) != 0) {
 				/*
@@ -320,10 +493,12 @@ next:		if (recnop != NULL)
 				 * descending the tree holding read-locks.
 				 */
 				(void)__LPUT(dbc, lock);
+				if (LF_ISSET(SR_DEL | SR_NEXT))
+					cp->csp++;
 				goto err;
 			}
 		}
-		if ((ret = __memp_fget(mpf, &pg, 0, &h)) != 0)
+		if ((ret = __memp_fget(mpf, &pg, dbc->txn, 0, &h)) != 0)
 			goto err;
 	}
 	/* NOTREACHED */
@@ -340,12 +515,12 @@ found:	*exactp = 1;
 	 * all duplicate sets that are not on overflow pages exist on a
 	 * single leaf page.
 	 */
-	if (TYPE(h) == P_LBTREE) {
-		if (LF_ISSET(S_DUPLAST))
+	if (TYPE(h) == P_LBTREE && NUM_ENT(h) > P_INDX) {
+		if (LF_ISSET(SR_DUPLAST))
 			while (indx < (db_indx_t)(NUM_ENT(h) - P_INDX) &&
 			    inp[indx] == inp[indx + P_INDX])
 				indx += P_INDX;
-		else
+		else if (LF_ISSET(SR_DUPFIRST))
 			while (indx > 0 &&
 			    inp[indx] == inp[indx - P_INDX])
 				indx -= P_INDX;
@@ -354,13 +529,13 @@ found:	*exactp = 1;
 	/*
 	 * Now check if we are allowed to return deleted items; if not, then
 	 * find the next (or previous) non-deleted duplicate entry.  (We do
-	 * not move from the original found key on the basis of the S_DELNO
+	 * not move from the original found key on the basis of the SR_DELNO
 	 * flag.)
 	 */
-	DB_ASSERT(recnop == NULL || LF_ISSET(S_DELNO));
-	if (LF_ISSET(S_DELNO)) {
+	DB_ASSERT(dbenv, recnop == NULL || LF_ISSET(SR_DELNO));
+	if (LF_ISSET(SR_DELNO)) {
 		deloffset = TYPE(h) == P_LBTREE ? O_INDX : 0;
-		if (LF_ISSET(S_DUPLAST))
+		if (LF_ISSET(SR_DUPLAST))
 			while (B_DISSET(GET_BKEYDATA(dbp,
 			    h, indx + deloffset)->type) && indx > 0 &&
 			    inp[indx] == inp[indx - adjust])
@@ -388,7 +563,7 @@ found:	*exactp = 1;
 		 * duplicates and record numbers in the same tree.
 		 */
 		if (recnop != NULL) {
-			DB_ASSERT(TYPE(h) == P_LBTREE);
+			DB_ASSERT(dbenv, TYPE(h) == P_LBTREE);
 
 			for (i = 0; i < indx; i += P_INDX)
 				if (!B_DISSET(
@@ -400,14 +575,17 @@ found:	*exactp = 1;
 		}
 	}
 
-	if (LF_ISSET(S_STK_ONLY)) {
-		BT_STK_NUM(dbp->dbenv, cp, h, indx, ret);
+	if (LF_ISSET(SR_STK_ONLY)) {
+		BT_STK_NUM(dbenv, cp, h, indx, ret);
 		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
 			ret = t_ret;
 		if ((t_ret = __memp_fput(mpf, h, 0)) != 0 && ret == 0)
 			ret = t_ret;
-	} else
-		BT_STK_ENTER(dbp->dbenv, cp, h, indx, lock, lock_mode, ret);
+	} else {
+		if (LF_ISSET(SR_DEL) && cp->csp == cp->sp)
+			cp->csp++;
+		BT_STK_ENTER(dbenv, cp, h, indx, lock, lock_mode, ret);
+	}
 	if (ret != 0)
 		goto err;
 
@@ -471,6 +649,13 @@ __bam_stkrel(dbc, flags)
 			 */
 			epg->page = NULL;
 		}
+		/*
+		 * We set this if we need to release our pins,
+		 * but are not logically ready to have the pages
+		 * visible.
+		 */
+		if (LF_ISSET(STK_PGONLY))
+			continue;
 		if (LF_ISSET(STK_NOLOCK)) {
 			if ((t_ret = __LPUT(dbc, epg->lock)) != 0 && ret == 0)
 				ret = t_ret;
@@ -480,7 +665,8 @@ __bam_stkrel(dbc, flags)
 	}
 
 	/* Clear the stack, all pages have been released. */
-	BT_STK_CLR(cp);
+	if (!LF_ISSET(STK_PGONLY))
+		BT_STK_CLR(cp);
 
 	return (ret);
 }
