@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996, 1997, 1998, 1999, 2000
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1989, 1991, 1993
@@ -31,18 +31,15 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: getcwd.c,v 12.5 2006/08/24 14:45:09 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: getcwd.c,v 11.6 2000/03/28 21:50:05 ubell Exp $";
-#endif /* not lint */
+#include "db_int.h"
 
 #ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-#include <sys/stat.h>
-
 #if HAVE_DIRENT_H
 # include <dirent.h>
 # define NAMLEN(dirent) strlen((dirent)->d_name)
@@ -59,16 +56,8 @@ static const char revid[] = "$Id: getcwd.c,v 11.6 2000/03/28 21:50:05 ubell Exp
 #  include <ndir.h>
 # endif
 #endif
-
-#include <errno.h>
-#include <stdio.h>
-#include <stdlib.h>
-#include <string.h>
-#include <unistd.h>
 #endif
 
-#include "db_int.h"
-
 #define	ISDOT(dp) \
 	(dp->d_name[0] == '.' && (dp->d_name[1] == '\0' || \
 	    (dp->d_name[1] == '.' && dp->d_name[2] == '\0')))
@@ -121,7 +110,7 @@ getcwd(pt, size)
 		ept = pt + size;
 	} else {
 		if ((ret =
-		    __os_malloc(NULL, ptsize = 1024 - 4, NULL, &pt)) != 0) {
+		    __os_malloc(NULL, ptsize = 1024 - 4, &pt)) != 0) {
 			__os_set_errno(ret);
 			return (NULL);
 		}
@@ -135,7 +124,7 @@ getcwd(pt, size)
 	 * Should always be enough (it's 340 levels).  If it's not, allocate
 	 * as necessary.  Special case the first stat, it's ".", not "..".
 	 */
-	if ((ret = __os_malloc(NULL, upsize = 1024 - 4, NULL, &up)) != 0)
+	if ((ret = __os_malloc(NULL, upsize = 1024 - 4, &up)) != 0)
 		goto err;
 	eup = up + 1024;
 	bup = up;
@@ -168,7 +157,7 @@ getcwd(pt, size)
 			 * been that way and stuff would probably break.
 			 */
 			bcopy(bpt, pt, ept - bpt);
-			__os_free(up, upsize);
+			__os_free(NULL, up);
 			return (pt);
 		}
 
@@ -178,7 +167,7 @@ getcwd(pt, size)
 		 * possible component name, plus a trailing NULL.
 		 */
 		if (bup + 3  + MAXNAMLEN + 1 >= eup) {
-			if (__os_realloc(NULL, upsize *= 2, NULL, &up) != 0)
+			if (__os_realloc(NULL, upsize *= 2, &up) != 0)
 				goto err;
 			bup = up;
 			eup = up + upsize;
@@ -239,7 +228,7 @@ getcwd(pt, size)
 			}
 			off = bpt - pt;
 			len = ept - bpt;
-			if (__os_realloc(NULL, ptsize *= 2, NULL, &pt) != 0)
+			if (__os_realloc(NULL, ptsize *= 2, &pt) != 0)
 				goto err;
 			bpt = pt + off;
 			ept = pt + ptsize;
@@ -262,12 +251,12 @@ notfound:
 	 * didn't find the current directory in its parent directory, set
 	 * errno to ENOENT.
 	 */
-	if (__os_get_errno() == 0)
+	if (__os_get_errno_ret_zero() == 0)
 		__os_set_errno(save_errno == 0 ? ENOENT : save_errno);
 	/* FALLTHROUGH */
 err:
 	if (ptsize)
-		__os_free(pt, ptsize);
-	__os_free(up, upsize);
+		__os_free(NULL, pt);
+	__os_free(NULL, up);
 	return (NULL);
 }
