@@ -1,23 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2003
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1997-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: cxx_logc.cpp,v 12.4 2006/08/24 14:45:13 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: cxx_logc.cpp,v 11.11 2003/03/27 20:05:04 merrells Exp $";
-#endif /* not lint */
-
-#include <errno.h>
-#include <string.h>
+#include "db_int.h"
 
 #include "db_cxx.h"
 #include "dbinc/cxx_int.h"
 
-#include "db_int.h"
 #include "dbinc/db_page.h"
 #include "dbinc_auto/db_auto.h"
 #include "dbinc_auto/crdel_auto.h"
@@ -56,7 +52,7 @@ int DbLogc::get(DbLsn *lsn, Dbt *data, u_int32_t _flags)
 	ret = logc->get(logc, lsn, data, _flags);
 
 	if (!DB_RETOK_LGGET(ret)) {
-		if (ret == ENOMEM && DB_OVERFLOWED_DBT(data))
+		if (ret == DB_BUFFER_SMALL)
 			DB_ERROR_DBT(DbEnv::get_DbEnv(logc->dbenv),
 				"DbLogc::get", data, ON_ERROR_UNKNOWN);
 		else
