@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1995, 1996
@@ -35,30 +35,22 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: db_dispatch.c,v 11.169 2004/10/27 16:44:26 ubell Exp $
+ * $Id: db_dispatch.c,v 12.26 2006/08/24 14:45:15 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/db_page.h"
-#include "dbinc/db_shash.h"
 #include "dbinc/hash.h"
+#include "dbinc/fop.h"
+#include "dbinc/lock.h"
 #include "dbinc/log.h"
 #include "dbinc/mp.h"
-#include "dbinc/fop.h"
 #include "dbinc/txn.h"
 
-#ifndef HAVE_FTRUNCATE
-static int __db_limbo_fix __P((DB *,
-    DB_TXN *, DB_TXNLIST *, db_pgno_t *, DBMETA *, db_limbo_state));
+static int __db_limbo_fix __P((DB *, DB_TXN *,
+		DB_TXNLIST *, db_pgno_t *, DBMETA *, db_limbo_state));
 static int __db_limbo_bucket __P((DB_ENV *,
 	     DB_TXN *, DB_TXNLIST *, db_limbo_state));
 static int __db_limbo_move __P((DB_ENV *, DB_TXN *, DB_TXN *, DB_TXNLIST *));
@@ -66,11 +58,10 @@ static int __db_limbo_prepare __P(( DB *, DB_TXN *, DB_TXNLIST *));
 static int __db_lock_move __P((DB_ENV *,
 		u_int8_t *, db_pgno_t, db_lockmode_t, DB_TXN *, DB_TXN *));
 static int __db_txnlist_pgnoadd __P((DB_ENV *, DB_TXNHEAD *,
-		int32_t, u_int8_t [DB_FILE_ID_LEN], char *, db_pgno_t));
-#endif
-static int __db_txnlist_find_internal __P((DB_ENV *,
-    void *, db_txnlist_type, u_int32_t, u_int8_t[DB_FILE_ID_LEN],
-    DB_TXNLIST **, int, u_int32_t *));
+		int32_t, u_int8_t *, char *, db_pgno_t));
+static int __db_txnlist_find_internal __P((DB_ENV *, DB_TXNHEAD *,
+		db_txnlist_type, u_int32_t, u_int8_t *, DB_TXNLIST **,
+		int, u_int32_t *));
 
 /*
  * __db_dispatch --
@@ -83,7 +74,7 @@ static int __db_txnlist_find_internal __P((DB_ENV *,
  *
  * PUBLIC: int __db_dispatch __P((DB_ENV *,
  * PUBLIC:     int (**)__P((DB_ENV *, DBT *, DB_LSN *, db_recops, void *)),
- * PUBLIC:     size_t, DBT *, DB_LSN *, db_recops, void *));
+ * PUBLIC:     size_t, DBT *, DB_LSN *, db_recops, DB_TXNHEAD *));
  */
 int
 __db_dispatch(dbenv, dtab, dtabsize, db, lsnp, redo, info)
@@ -93,7 +84,7 @@ __db_dispatch(dbenv, dtab, dtabsize, db, lsnp, redo, info)
 	DBT *db;		/* The log record upon which to dispatch. */
 	DB_LSN *lsnp;		/* The lsn of the record being dispatched. */
 	db_recops redo;		/* Redo this op (or undo it). */
-	void *info;
+	DB_TXNHEAD *info;	/* Transaction list. */
 {
 	DB_LSN prev_lsn;
 	u_int32_t rectype, status, txnid;
@@ -104,7 +95,7 @@ __db_dispatch(dbenv, dtab, dtabsize, db, lsnp, redo, info)
 	make_call = ret = 0;
 
 	/* If we don't have a dispatch table, it's hard to dispatch. */
-	DB_ASSERT(dtab != NULL);
+	DB_ASSERT(dbenv, dtab != NULL);
 
 	/*
 	 * If we find a record that is in the user's number space and they
@@ -235,7 +226,9 @@ __db_dispatch(dbenv, dtab, dtabsize, db, lsnp, redo, info)
 			break;
 
 		default:
-			if (txnid != 0) {
+			if (txnid == 0)
+				status = 0;
+			else {
 				ret = __db_txnlist_find(dbenv,
 				    info, txnid, &status);
 
@@ -314,7 +307,7 @@ __db_dispatch(dbenv, dtab, dtabsize, db, lsnp, redo, info)
 			 * as our sanity check.
 			 */
 			if (rectype > dtabsize || dtab[rectype] == NULL) {
-				__db_err(dbenv,
+				__db_errx(dbenv,
 				    "Illegal record type %lu in log",
 				    (u_long)rectype);
 				return (EINVAL);
@@ -364,14 +357,14 @@ __db_add_recovery(dbenv, dtab, dtabsize, func, ndx)
  *	Initialize transaction linked list.
  *
  * PUBLIC: int __db_txnlist_init __P((DB_ENV *,
- * PUBLIC:     u_int32_t, u_int32_t, DB_LSN *, void *));
+ * PUBLIC:     u_int32_t, u_int32_t, DB_LSN *, DB_TXNHEAD **));
  */
 int
 __db_txnlist_init(dbenv, low_txn, hi_txn, trunc_lsn, retp)
 	DB_ENV *dbenv;
 	u_int32_t low_txn, hi_txn;
 	DB_LSN *trunc_lsn;
-	void *retp;
+	DB_TXNHEAD **retp;
 {
 	DB_TXNHEAD *headp;
 	u_int32_t size, tmp;
@@ -427,7 +420,7 @@ __db_txnlist_init(dbenv, low_txn, hi_txn, trunc_lsn, retp)
 	}
 	ZERO_LSN(headp->ckplsn);
 
-	*(void **)retp = headp;
+	*retp = headp;
 	return (0);
 }
 
@@ -436,23 +429,21 @@ __db_txnlist_init(dbenv, low_txn, hi_txn, trunc_lsn, retp)
  *	Add an element to our transaction linked list.
  *
  * PUBLIC: int __db_txnlist_add __P((DB_ENV *,
- * PUBLIC:     void *, u_int32_t, u_int32_t, DB_LSN *));
+ * PUBLIC:     DB_TXNHEAD *, u_int32_t, u_int32_t, DB_LSN *));
  */
 int
-__db_txnlist_add(dbenv, listp, txnid, status, lsn)
+__db_txnlist_add(dbenv, hp, txnid, status, lsn)
 	DB_ENV *dbenv;
-	void *listp;
+	DB_TXNHEAD *hp;
 	u_int32_t txnid, status;
 	DB_LSN *lsn;
 {
-	DB_TXNHEAD *hp;
 	DB_TXNLIST *elp;
 	int ret;
 
 	if ((ret = __os_malloc(dbenv, sizeof(DB_TXNLIST), &elp)) != 0)
 		return (ret);
 
-	hp = (DB_TXNHEAD *)listp;
 	LIST_INSERT_HEAD(&hp->head[DB_TXNLIST_MASK(hp, txnid)], elp, links);
 
 	elp->type = TXNLIST_TXNID;
@@ -464,8 +455,8 @@ __db_txnlist_add(dbenv, listp, txnid, status, lsn)
 	if (lsn != NULL && IS_ZERO_LSN(hp->maxlsn) && status == TXN_COMMIT)
 		hp->maxlsn = *lsn;
 
-	DB_ASSERT(lsn == NULL ||
-	    status != TXN_COMMIT || log_compare(&hp->maxlsn, lsn) >= 0);
+	DB_ASSERT(dbenv, lsn == NULL ||
+	    status != TXN_COMMIT || LOG_COMPARE(&hp->maxlsn, lsn) >= 0);
 
 	return (0);
 }
@@ -474,19 +465,19 @@ __db_txnlist_add(dbenv, listp, txnid, status, lsn)
  * __db_txnlist_remove --
  *	Remove an element from our transaction linked list.
  *
- * PUBLIC: int __db_txnlist_remove __P((DB_ENV *, void *, u_int32_t));
+ * PUBLIC: int __db_txnlist_remove __P((DB_ENV *, DB_TXNHEAD *, u_int32_t));
  */
 int
-__db_txnlist_remove(dbenv, listp, txnid)
+__db_txnlist_remove(dbenv, hp, txnid)
 	DB_ENV *dbenv;
-	void *listp;
+	DB_TXNHEAD *hp;
 	u_int32_t txnid;
 {
 	DB_TXNLIST *entry;
 	u_int32_t status;
 
 	return (__db_txnlist_find_internal(dbenv,
-	    listp, TXNLIST_TXNID, txnid, NULL, &entry, 1, &status));
+	    hp, TXNLIST_TXNID, txnid, NULL, &entry, 1, &status));
 }
 
 /*
@@ -497,22 +488,19 @@ __db_txnlist_remove(dbenv, listp, txnid)
  * recovery, we are going to virtually truncate the log and we need
  * to retain the last checkpoint before the truncation point.
  *
- * PUBLIC: void __db_txnlist_ckp __P((DB_ENV *, void *, DB_LSN *));
+ * PUBLIC: void __db_txnlist_ckp __P((DB_ENV *, DB_TXNHEAD *, DB_LSN *));
  */
 void
-__db_txnlist_ckp(dbenv, listp, ckp_lsn)
+__db_txnlist_ckp(dbenv, hp, ckp_lsn)
 	DB_ENV *dbenv;
-	void *listp;
+	DB_TXNHEAD *hp;
 	DB_LSN *ckp_lsn;
 {
-	DB_TXNHEAD *hp;
 
 	COMPQUIET(dbenv, NULL);
 
-	hp = (DB_TXNHEAD *)listp;
-
 	if (IS_ZERO_LSN(hp->ckplsn) && !IS_ZERO_LSN(hp->maxlsn) &&
-	    log_compare(&hp->maxlsn, ckp_lsn) >= 0)
+	    LOG_COMPARE(&hp->maxlsn, ckp_lsn) >= 0)
 		hp->ckplsn = *ckp_lsn;
 }
 
@@ -520,26 +508,24 @@ __db_txnlist_ckp(dbenv, listp, ckp_lsn)
  * __db_txnlist_end --
  *	Discard transaction linked list.
  *
- * PUBLIC: void __db_txnlist_end __P((DB_ENV *, void *));
+ * PUBLIC: void __db_txnlist_end __P((DB_ENV *, DB_TXNHEAD *));
  */
 void
-__db_txnlist_end(dbenv, listp)
+__db_txnlist_end(dbenv, hp)
 	DB_ENV *dbenv;
-	void *listp;
-{
 	DB_TXNHEAD *hp;
-	DB_TXNLIST *p;
+{
 	u_int32_t i;
+	DB_TXNLIST *p;
 
-	if ((hp = (DB_TXNHEAD *)listp) == NULL)
+	if (hp == NULL)
 		return;
 
 	for (i = 0; i < hp->nslots; i++)
 		while (hp != NULL && (p = LIST_FIRST(&hp->head[i])) != NULL) {
-			LIST_REMOVE(p, links);
 			switch (p->type) {
 			case TXNLIST_LSN:
-				__os_free(dbenv, p->u.l.lsn_array);
+				__os_free(dbenv, p->u.l.lsn_stack);
 				break;
 			case TXNLIST_DELETE:
 			case TXNLIST_PGNO:
@@ -551,12 +537,13 @@ __db_txnlist_end(dbenv, listp)
 				 */
 				break;
 			}
+			LIST_REMOVE(p, links);
 			__os_free(dbenv, p);
 		}
 
 	if (hp->gen_array != NULL)
 		__os_free(dbenv, hp->gen_array);
-	__os_free(dbenv, listp);
+	__os_free(dbenv, hp);
 }
 
 /*
@@ -568,12 +555,12 @@ __db_txnlist_end(dbenv, listp)
  *	was generated while not in a transaction.
  *
  * PUBLIC: int __db_txnlist_find __P((DB_ENV *,
- * PUBLIC:     void *, u_int32_t, u_int32_t *));
+ * PUBLIC:     DB_TXNHEAD *, u_int32_t, u_int32_t *));
  */
 int
-__db_txnlist_find(dbenv, listp, txnid, statusp)
+__db_txnlist_find(dbenv, hp, txnid, statusp)
 	DB_ENV *dbenv;
-	void *listp;
+	DB_TXNHEAD *hp;
 	u_int32_t txnid, *statusp;
 {
 	DB_TXNLIST *entry;
@@ -581,7 +568,7 @@ __db_txnlist_find(dbenv, listp, txnid, statusp)
 	if (txnid == 0)
 		return (DB_NOTFOUND);
 
-	return (__db_txnlist_find_internal(dbenv, listp,
+	return (__db_txnlist_find_internal(dbenv, hp,
 	    TXNLIST_TXNID, txnid, NULL, &entry, 0, statusp));
 }
 
@@ -590,32 +577,30 @@ __db_txnlist_find(dbenv, listp, txnid, statusp)
  *	Change the status of an existing transaction entry.
  *	Returns DB_NOTFOUND if no such entry exists.
  *
- * PUBLIC: int __db_txnlist_update __P((DB_ENV *,
- * PUBLIC:     void *, u_int32_t, u_int32_t, DB_LSN *, u_int32_t *, int));
+ * PUBLIC: int __db_txnlist_update __P((DB_ENV *, DB_TXNHEAD *,
+ * PUBLIC:     u_int32_t, u_int32_t, DB_LSN *, u_int32_t *, int));
  */
 int
-__db_txnlist_update(dbenv, listp, txnid, status, lsn, ret_status, add_ok)
+__db_txnlist_update(dbenv, hp, txnid, status, lsn, ret_status, add_ok)
 	DB_ENV *dbenv;
-	void *listp;
+	DB_TXNHEAD *hp;
 	u_int32_t txnid, status;
 	DB_LSN *lsn;
 	u_int32_t *ret_status;
 	int add_ok;
 {
-	DB_TXNHEAD *hp;
 	DB_TXNLIST *elp;
 	int ret;
 
 	if (txnid == 0)
 		return (DB_NOTFOUND);
 
-	hp = (DB_TXNHEAD *)listp;
 	ret = __db_txnlist_find_internal(dbenv,
-	    listp, TXNLIST_TXNID, txnid, NULL, &elp, 0, ret_status);
+	    hp, TXNLIST_TXNID, txnid, NULL, &elp, 0, ret_status);
 
 	if (ret == DB_NOTFOUND && add_ok) {
 		*ret_status = status;
-		return (__db_txnlist_add(dbenv, listp, txnid, status, lsn));
+		return (__db_txnlist_add(dbenv, hp, txnid, status, lsn));
 	}
 	if (ret != 0)
 		return (ret);
@@ -640,9 +625,9 @@ __db_txnlist_update(dbenv, listp, txnid, status, lsn, ret_status, add_ok)
  */
 static int
 __db_txnlist_find_internal(dbenv,
-    listp, type, txnid, uid, txnlistp, delete, statusp)
+    hp, type, txnid, uid, txnlistp, delete, statusp)
 	DB_ENV *dbenv;
-	void *listp;
+	DB_TXNHEAD *hp;
 	db_txnlist_type type;
 	u_int32_t  txnid;
 	u_int8_t uid[DB_FILE_ID_LEN];
@@ -651,14 +636,13 @@ __db_txnlist_find_internal(dbenv,
 	u_int32_t *statusp;
 {
 	struct __db_headlink *head;
-	DB_TXNHEAD *hp;
 	DB_TXNLIST *p;
 	u_int32_t generation, hash, i;
 	int ret;
 
 	ret = 0;
 
-	if ((hp = (DB_TXNHEAD *)listp) == NULL)
+	if (hp == NULL)
 		return (DB_NOTFOUND);
 
 	switch (type) {
@@ -674,7 +658,7 @@ __db_txnlist_find_internal(dbenv,
 			    (txnid >= hp->gen_array[i].txn_min ||
 			    txnid <= hp->gen_array[i].txn_max))
 				break;
-		DB_ASSERT(i <= hp->generation);
+		DB_ASSERT(dbenv, i <= hp->generation);
 		generation = hp->gen_array[i].generation;
 		break;
 	case TXNLIST_PGNO:
@@ -688,8 +672,7 @@ __db_txnlist_find_internal(dbenv,
 	}
 
 	head = &hp->head[DB_TXNLIST_MASK(hp, hash)];
-
-	for (p = LIST_FIRST(head); p != NULL; p = LIST_NEXT(p, links)) {
+	LIST_FOREACH(p, head, links) {
 		if (p->type != type)
 			continue;
 		switch (type) {
@@ -703,6 +686,7 @@ __db_txnlist_find_internal(dbenv,
 		case TXNLIST_PGNO:
 			if (memcmp(uid, p->u.p.uid, DB_FILE_ID_LEN) != 0)
 				continue;
+			*statusp = TXN_OK;
 			break;
 		case TXNLIST_DELETE:
 		case TXNLIST_LSN:
@@ -712,12 +696,14 @@ __db_txnlist_find_internal(dbenv,
 		if (delete == 1) {
 			LIST_REMOVE(p, links);
 			__os_free(dbenv, p);
+			*txnlistp = NULL;
 		} else if (p != LIST_FIRST(head)) {
 			/* Move it to head of list. */
 			LIST_REMOVE(p, links);
 			LIST_INSERT_HEAD(head, p, links);
-		}
-		*txnlistp = p;
+			*txnlistp = p;
+		} else
+			*txnlistp = p;
 		return (ret);
 	}
 
@@ -729,16 +715,15 @@ __db_txnlist_find_internal(dbenv,
  *	Change the current generation number.
  *
  * PUBLIC: int __db_txnlist_gen __P((DB_ENV *,
- * PUBLIC:       void *, int, u_int32_t, u_int32_t));
+ * PUBLIC:       DB_TXNHEAD *, int, u_int32_t, u_int32_t));
  */
 int
-__db_txnlist_gen(dbenv, listp, incr, min, max)
+__db_txnlist_gen(dbenv, hp, incr, min, max)
 	DB_ENV *dbenv;
-	void *listp;
+	DB_TXNHEAD *hp;
 	int incr;
 	u_int32_t min, max;
 {
-	DB_TXNHEAD *hp;
 	int ret;
 
 	/*
@@ -753,7 +738,6 @@ __db_txnlist_gen(dbenv, listp, incr, min, max)
 	 * is given the generation number of the first range it falls into
 	 * in the stack.
 	 */
-	hp = (DB_TXNHEAD *)listp;
 	if (incr < 0) {
 		--hp->generation;
 		memmove(hp->gen_array, &hp->gen_array[1],
@@ -775,71 +759,76 @@ __db_txnlist_gen(dbenv, listp, incr, min, max)
 	return (0);
 }
 
-#define	TXN_BUBBLE(AP, MAX) {						\
-	DB_LSN __tmp;							\
-	u_int32_t __j;							\
-									\
-	for (__j = 0; __j < MAX - 1; __j++)				\
-		if (log_compare(&AP[__j], &AP[__j + 1]) < 0) {		\
-			__tmp = AP[__j];				\
-			AP[__j] = AP[__j + 1];				\
-			AP[__j + 1] = __tmp;				\
-		}							\
-}
-
 /*
  * __db_txnlist_lsnadd --
- *	Add to or re-sort the transaction list lsn entry.  Note that since this
- *	is used during an abort, the __txn_undo code calls into the "recovery"
- *	subsystem explicitly, and there is only a single TXNLIST_LSN entry on
- *	the list.
+ *	Save the prev_lsn from a txn_child record.
  *
- * PUBLIC: int __db_txnlist_lsnadd __P((DB_ENV *, void *, DB_LSN *, u_int32_t));
+ * PUBLIC: int __db_txnlist_lsnadd __P((DB_ENV *, DB_TXNHEAD *, DB_LSN *));
  */
 int
-__db_txnlist_lsnadd(dbenv, listp, lsnp, flags)
+__db_txnlist_lsnadd(dbenv, hp, lsnp)
 	DB_ENV *dbenv;
-	void *listp;
-	DB_LSN *lsnp;
-	u_int32_t flags;
-{
 	DB_TXNHEAD *hp;
+	DB_LSN *lsnp;
+{
 	DB_TXNLIST *elp;
-	u_int32_t i;
 	int ret;
 
-	hp = (DB_TXNHEAD *)listp;
+	if (IS_ZERO_LSN(*lsnp))
+		return (0);
 
-	for (elp = LIST_FIRST(&hp->head[0]);
-	    elp != NULL; elp = LIST_NEXT(elp, links))
+	LIST_FOREACH(elp, &hp->head[0], links)
 		if (elp->type == TXNLIST_LSN)
 			break;
 
-	if (elp == NULL)
+	if (elp == NULL) {
+		if ((ret = __db_txnlist_lsninit(dbenv, hp, lsnp)) != 0)
+			return (ret);
 		return (DB_SURPRISE_KID);
+	}
 
-	if (LF_ISSET(TXNLIST_NEW)) {
-		if (elp->u.l.ntxns >= elp->u.l.maxn) {
-			if ((ret = __os_realloc(dbenv,
-			    2 * elp->u.l.maxn * sizeof(DB_LSN),
-			    &elp->u.l.lsn_array)) != 0)
-				return (ret);
-			elp->u.l.maxn *= 2;
+	if (elp->u.l.stack_indx == elp->u.l.stack_size) {
+		elp->u.l.stack_size <<= 1;
+		if ((ret = __os_realloc(dbenv, sizeof(DB_LSN) *
+		     elp->u.l.stack_size, &elp->u.l.lsn_stack)) != 0) {
+			__db_txnlist_end(dbenv, hp);
+			return (ret);
 		}
-		elp->u.l.lsn_array[elp->u.l.ntxns++] = *lsnp;
-	} else
-		/* Simply replace the 0th element. */
-		elp->u.l.lsn_array[0] = *lsnp;
+	}
+	elp->u.l.lsn_stack[elp->u.l.stack_indx++] = *lsnp;
 
-	/*
-	 * If we just added a new entry and there may be NULL entries, so we
-	 * have to do a complete bubble sort, not just trickle a changed entry
-	 * around.
-	 */
-	for (i = 0; i < (!LF_ISSET(TXNLIST_NEW) ? 1 : elp->u.l.ntxns); i++)
-		TXN_BUBBLE(elp->u.l.lsn_array, elp->u.l.ntxns);
+	return (0);
+}
 
-	*lsnp = elp->u.l.lsn_array[0];
+/*
+ * __db_txnlist_lsnget --
+ *
+ * PUBLIC: int __db_txnlist_lsnget __P((DB_ENV *,
+ * PUBLIC:     DB_TXNHEAD *, DB_LSN *, u_int32_t));
+ *	Get the lsn saved from a txn_child record.
+ */
+int
+__db_txnlist_lsnget(dbenv, hp, lsnp, flags)
+	DB_ENV *dbenv;
+	DB_TXNHEAD *hp;
+	DB_LSN *lsnp;
+	u_int32_t flags;
+{
+	DB_TXNLIST *elp;
+
+	COMPQUIET(dbenv, NULL);
+	COMPQUIET(flags, 0);
+
+	LIST_FOREACH(elp, &hp->head[0], links)
+		if (elp->type == TXNLIST_LSN)
+			break;
+
+	if (elp == NULL || elp->u.l.stack_indx == 0) {
+		ZERO_LSN(*lsnp);
+		return (0);
+	}
+
+	*lsnp = elp->u.l.lsn_stack[--elp->u.l.stack_indx];
 
 	return (0);
 }
@@ -865,13 +854,13 @@ __db_txnlist_lsninit(dbenv, hp, lsnp)
 		goto err;
 	LIST_INSERT_HEAD(&hp->head[0], elp, links);
 
-	if ((ret = __os_malloc(dbenv,
-	    12 * sizeof(DB_LSN), &elp->u.l.lsn_array)) != 0)
-		goto err;
 	elp->type = TXNLIST_LSN;
-	elp->u.l.maxn = 12;
-	elp->u.l.ntxns = 1;
-	elp->u.l.lsn_array[0] = *lsnp;
+	if ((ret = __os_malloc(dbenv,
+	    sizeof(DB_LSN) * DB_LSN_STACK_SIZE, &elp->u.l.lsn_stack)) != 0)
+		goto err;
+	elp->u.l.stack_indx = 1;
+	elp->u.l.stack_size = DB_LSN_STACK_SIZE;
+	elp->u.l.lsn_stack[0] = *lsnp;
 
 	return (0);
 
@@ -879,23 +868,20 @@ err:	__db_txnlist_end(dbenv, hp);
 	return (ret);
 }
 
-#ifndef HAVE_FTRUNCATE
 /*
  * __db_add_limbo -- add pages to the limbo list.
  *	Get the file information and call pgnoadd for each page.
  *
- * PUBLIC: #ifndef HAVE_FTRUNCATE
  * PUBLIC: int __db_add_limbo __P((DB_ENV *,
- * PUBLIC:      void *, int32_t, db_pgno_t, int32_t));
- * PUBLIC: #endif
+ * PUBLIC:      DB_TXNHEAD *, int32_t, db_pgno_t, u_int32_t));
  */
 int
-__db_add_limbo(dbenv, info, fileid, pgno, count)
+__db_add_limbo(dbenv, hp, fileid, pgno, count)
 	DB_ENV *dbenv;
-	void *info;
+	DB_TXNHEAD *hp;
 	int32_t fileid;
 	db_pgno_t pgno;
-	int32_t count;
+	u_int32_t count;
 {
 	DB_LOG *dblp;
 	FNAME *fnp;
@@ -907,7 +893,7 @@ __db_add_limbo(dbenv, info, fileid, pgno, count)
 
 	do {
 		if ((ret =
-		    __db_txnlist_pgnoadd(dbenv, info, fileid, fnp->ufid,
+		    __db_txnlist_pgnoadd(dbenv, hp, fileid, fnp->ufid,
 		    R_ADDR(&dblp->reginfo, fnp->name_off), pgno)) != 0)
 			return (ret);
 		pgno++;
@@ -948,10 +934,8 @@ __db_add_limbo(dbenv, info, fileid, pgno, count)
  * "create list and write meta-data page" algorithm.  Otherwise, we're in
  * an abort and doing the "use compensating transaction" algorithm.
  *
- * PUBLIC: #ifndef HAVE_FTRUNCATE
  * PUBLIC: int __db_do_the_limbo __P((DB_ENV *,
  * PUBLIC:     DB_TXN *, DB_TXN *, DB_TXNHEAD *, db_limbo_state));
- * PUBLIC: #endif
  */
 int
 __db_do_the_limbo(dbenv, ptxn, txn, hp, state)
@@ -983,7 +967,7 @@ __db_do_the_limbo(dbenv, ptxn, txn, hp, state)
 	}
 
 err:	if (ret != 0) {
-		__db_err(dbenv, "Fatal error in abort of an allocation");
+		__db_errx(dbenv, "Fatal error in abort of an allocation");
 		ret = __db_panic(dbenv, ret);
 	}
 
@@ -1136,9 +1120,9 @@ retry:		dbp_created = 0;
 			dbp_created = 1;
 
 			/* It is ok if the file is nolonger there. */
-			ret = __db_open(dbp,
-			    t, elp->u.p.fname, NULL, DB_UNKNOWN,
-			    DB_ODDFILESIZE, __db_omode("rw----"), PGNO_BASE_MD);
+			ret = __db_open(dbp, t, elp->u.p.fname, NULL,
+			    DB_UNKNOWN, DB_ODDFILESIZE, __db_omode(OWNER_RW),
+			    PGNO_BASE_MD);
 			if (ret == ENOENT)
 				goto next;
 		}
@@ -1153,10 +1137,12 @@ retry:		dbp_created = 0;
 		mpf = dbp->mpf;
 		last_pgno = PGNO_INVALID;
 
-		if (meta == NULL && 
+		if (meta == NULL &&
 		    (ctxn == NULL || state == LIMBO_COMPENSATE)) {
 			pgno = PGNO_BASE_MD;
-			if ((ret = __memp_fget(mpf, &pgno, 0, &meta)) != 0)
+			if ((ret = __memp_fget(mpf, &pgno, txn,
+			    (state != LIMBO_PREPARE) ? DB_MPOOL_DIRTY : 0,
+			    &meta)) != 0)
 				goto err;
 			last_pgno = meta->free;
 		}
@@ -1194,13 +1180,11 @@ retry:		dbp_created = 0;
 			 * off the recovery flag briefly.
 			 */
 			if (state == LIMBO_COMPENSATE)
-				F_CLR(
-				    (DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER);
+				F_CLR(dbenv->lg_handle, DBLOG_RECOVER);
 			ret = __txn_commit(ctxn, DB_TXN_NOSYNC);
 			ctxn = NULL;
 			if (state == LIMBO_COMPENSATE)
-				F_SET(
-				    (DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER);
+				F_SET(dbenv->lg_handle, DBLOG_RECOVER);
 			if (ret != 0)
 				goto retry;
 		}
@@ -1225,7 +1209,7 @@ retry:		dbp_created = 0;
 			 * first so that it isn't pinned when we try to sync.
 			 */
 			if (!IS_RECOVERING(dbenv) && !T_RESTORED(txn))
-				__db_err(dbenv, "Flushing free list to disk");
+				__db_errx(dbenv, "Flushing free list to disk");
 			if ((ret = __memp_fput(mpf, meta, 0)) != 0)
 				goto err;
 			meta = NULL;
@@ -1240,18 +1224,16 @@ retry:		dbp_created = 0;
 			 */
 			if ((ret = __db_sync(dbp)) == 0) {
 				pgno = PGNO_BASE_MD;
-				if ((ret =
-				    __memp_fget(mpf, &pgno, 0, &meta)) != 0)
+				if ((ret = __memp_fget(mpf, &pgno, txn,
+				    DB_MPOOL_DIRTY, &meta)) != 0)
 					goto err;
 				meta->free = last_pgno;
-				if ((ret = __memp_fput(mpf,
-				     meta, DB_MPOOL_DIRTY)) != 0)
+				if ((ret = __memp_fput(mpf, meta, 0)) != 0)
 					goto err;
 				meta = NULL;
 			} else {
-				__db_err(dbenv,
-				    "%s: %s", dbp->fname, db_strerror(ret));
-				__db_err(dbenv, "%s: %s %s", dbp->fname,
+				__db_err(dbenv, ret, "%s", dbp->fname);
+				__db_errx(dbenv, "%s: %s %s", dbp->fname,
 				    "allocation flush failed, some free pages",
 				    "may not appear in the free list");
 				ret = 0;
@@ -1324,8 +1306,8 @@ __db_limbo_fix(dbp, ctxn, elp, lastp, meta, state)
 		if (pgno == PGNO_INVALID)
 			continue;
 
-		if ((ret =
-		    __memp_fget(mpf, &pgno, DB_MPOOL_CREATE, &pagep)) != 0) {
+		if ((ret = __memp_fget(mpf, &pgno, ctxn,
+		    DB_MPOOL_CREATE | DB_MPOOL_EDIT, &pagep)) != 0) {
 			if (ret != ENOSPC)
 				goto err;
 			continue;
@@ -1342,7 +1324,7 @@ __db_limbo_fix(dbp, ctxn, elp, lastp, meta, state)
 					if (next == pgno)
 						break;
 					if ((ret = __memp_fget(mpf,
-					    &next, 0, &freep)) != 0)
+					    &next, ctxn, 0, &freep)) != 0)
 						goto err;
 					next = NEXT_PGNO(freep);
 					if ((ret =
@@ -1382,7 +1364,16 @@ __db_limbo_fix(dbp, ctxn, elp, lastp, meta, state)
 				 * marked compensating, but in case we didn't
 				 * do the open, we have to mark it explicitly.
 				 */
-				F_SET(dbc, DBC_COMPENSATE);
+				F_SET(dbc, DBC_DONTLOCK);
+
+				/*
+				 * If aborting a txn for a different process
+				 * via XA or failchk, DB_AM_RECOVER will be
+				 * set but we need to log the compensating
+				 * transactions.
+				 */
+				F_CLR(dbc, DBC_RECOVER);
+
 				ret = __db_free(dbc, pagep);
 				pagep = NULL;
 
@@ -1407,15 +1398,15 @@ __db_limbo_fix(dbp, ctxn, elp, lastp, meta, state)
 			elp->u.p.pgno_array[i] = PGNO_INVALID;
 
 		if (pagep != NULL) {
-			ret = __memp_fput(mpf, pagep, DB_MPOOL_DIRTY);
+			ret = __memp_fput(mpf, pagep, 0);
 			pagep = NULL;
 		}
 		if (ret != 0)
 			goto err;
 	}
 
-err:	if (pagep != NULL &&
-	    (t_ret = __memp_fput(mpf, pagep, DB_MPOOL_DIRTY)) != 0 && ret == 0)
+err:	if (pagep != NULL && (t_ret = __memp_fput(mpf, pagep, 0)) != 0 &&
+	    ret == 0)
 		ret = t_ret;
 	if (dbc != NULL && (t_ret = __db_c_close(dbc)) != 0 && ret == 0)
 		ret = t_ret;
@@ -1446,8 +1437,8 @@ __db_limbo_prepare(dbp, txn, elp)
 	for (i = 0; i < elp->u.p.nentries; i++) {
 		pgno = elp->u.p.pgno_array[i];
 
-		if ((ret =
-		    __memp_fget(mpf, &pgno, DB_MPOOL_CREATE, &pagep)) != 0) {
+		if ((ret = __memp_fget(mpf,
+		    &pgno, txn, DB_MPOOL_CREATE, &pagep)) != 0) {
 			if (ret != ENOSPC)
 				return (ret);
 			continue;
@@ -1530,31 +1521,26 @@ __db_txnlist_pgnoadd(dbenv, hp, fileid, uid, fname, pgno)
 
 err:	return (ret);
 }
-#endif
 
 #ifdef DEBUG
 /*
  * __db_txnlist_print --
  *	Print out the transaction list.
  *
- * PUBLIC: void __db_txnlist_print __P((void *));
+ * PUBLIC: void __db_txnlist_print __P((DB_TXNHEAD *));
  */
 void
-__db_txnlist_print(listp)
-	void *listp;
-{
+__db_txnlist_print(hp)
 	DB_TXNHEAD *hp;
+{
 	DB_TXNLIST *p;
 	u_int32_t i;
 	char *txntype;
 
-	hp = (DB_TXNHEAD *)listp;
-
 	printf("Maxid: %lu Generation: %lu\n",
 	    (u_long)hp->maxid, (u_long)hp->generation);
 	for (i = 0; i < hp->nslots; i++)
-		for (p = LIST_FIRST(&hp->head[i]);
-		    p != NULL; p = LIST_NEXT(p, links)) {
+		LIST_FOREACH(p, &hp->head[i], links) {
 			if (p->type != TXNLIST_TXNID) {
 				printf("Unrecognized type: %d\n", p->type);
 				continue;
