@@ -1,34 +1,27 @@
 /*
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1998-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1998-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: db_join.c,v 12.14 2006/08/24 14:45:15 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: db_join.c,v 11.46 2001/08/15 13:30:27 bostic Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
-#include "db_page.h"
-#include "db_join.h"
-#include "db_am.h"
-#include "btree.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_join.h"
+#include "dbinc/btree.h"
 
-static int __db_join_close __P((DBC *));
+static int __db_join_close_pp __P((DBC *));
 static int __db_join_cmp __P((const void *, const void *));
 static int __db_join_del __P((DBC *, u_int32_t));
 static int __db_join_get __P((DBC *, DBT *, DBT *, u_int32_t));
+static int __db_join_get_pp __P((DBC *, DBT *, DBT *, u_int32_t));
 static int __db_join_getnext __P((DBC *, DBT *, DBT *, u_int32_t, u_int32_t));
+static int __db_join_primget __P((DB *,
+    DB_TXN *, u_int32_t, DBT *, DBT *, u_int32_t));
 static int __db_join_put __P((DBC *, DBT *, DBT *, u_int32_t));
 
 /*
@@ -66,7 +59,7 @@ static int __db_join_put __P((DBC *, DBT *, DBT *, u_int32_t));
  * cursor method of a DB, join cursors are created through an explicit
  * call to DB->join.
  *
- * The curslist is an array of existing, intialized cursors and primary
+ * The curslist is an array of existing, initialized cursors and primary
  * is the DB of the primary file.  The data item that joins all the
  * cursors in the curslist is used as the key into the primary and that
  * key and data are returned.  When no more items are left in the join
@@ -83,25 +76,18 @@ __db_join(primary, curslist, dbcp, flags)
 	DB_ENV *dbenv;
 	DBC *dbc;
 	JOIN_CURSOR *jc;
+	size_t ncurs, nslots;
+	u_int32_t i;
 	int ret;
-	u_int32_t i, ncurs, nslots;
-
-	COMPQUIET(nslots, 0);
-
-	PANIC_CHECK(primary->dbenv);
-
-	if ((ret = __db_joinchk(primary, curslist, flags)) != 0)
-		return (ret);
 
+	dbenv = primary->dbenv;
 	dbc = NULL;
 	jc = NULL;
-	dbenv = primary->dbenv;
 
 	if ((ret = __os_calloc(dbenv, 1, sizeof(DBC), &dbc)) != 0)
 		goto err;
 
-	if ((ret = __os_calloc(dbenv,
-	    1, sizeof(JOIN_CURSOR), &jc)) != 0)
+	if ((ret = __os_calloc(dbenv, 1, sizeof(JOIN_CURSOR), &jc)) != 0)
 		goto err;
 
 	if ((ret = __os_malloc(dbenv, 256, &jc->j_key.data)) != 0)
@@ -120,7 +106,7 @@ __db_join(primary, curslist, dbcp, flags)
 	 * the number of cursors involved in the join, because the
 	 * list is NULL-terminated.
 	 */
-	ncurs = jc->j_curslist - curslist;
+	ncurs = (size_t)(jc->j_curslist - curslist);
 	nslots = ncurs + 1;
 
 	/*
@@ -186,7 +172,7 @@ __db_join(primary, curslist, dbcp, flags)
 		jc->j_fdupcurs[i] = NULL;
 		jc->j_exhausted[i] = 0;
 	}
-	jc->j_ncurs = ncurs;
+	jc->j_ncurs = (u_int32_t)ncurs;
 
 	/*
 	 * If DB_JOIN_NOSORT is not set, optimize secondary cursors by
@@ -206,47 +192,81 @@ __db_join(primary, curslist, dbcp, flags)
 	 * because this is the last thing that can fail.  Modifier of this
 	 * function beware!
 	 */
-	if ((ret = jc->j_curslist[0]->c_dup(jc->j_curslist[0], jc->j_workcurs,
-	    DB_POSITIONI)) != 0)
+	if ((ret =
+	    __db_c_dup(jc->j_curslist[0], jc->j_workcurs, DB_POSITION)) != 0)
 		goto err;
 
-	dbc->c_close = __db_join_close;
+	dbc->c_close = __db_join_close_pp;
 	dbc->c_del = __db_join_del;
-	dbc->c_get = __db_join_get;
+	dbc->c_get = __db_join_get_pp;
 	dbc->c_put = __db_join_put;
-	dbc->internal = (DBC_INTERNAL *) jc;
+	dbc->internal = (DBC_INTERNAL *)jc;
 	dbc->dbp = primary;
 	jc->j_primary = primary;
 
+	/* Stash the first cursor's transaction here for easy access. */
+	dbc->txn = curslist[0]->txn;
+
 	*dbcp = dbc;
 
-	MUTEX_THREAD_LOCK(dbenv, primary->mutexp);
+	MUTEX_LOCK(dbenv, primary->mutex);
 	TAILQ_INSERT_TAIL(&primary->join_queue, dbc, links);
-	MUTEX_THREAD_UNLOCK(dbenv, primary->mutexp);
+	MUTEX_UNLOCK(dbenv, primary->mutex);
 
 	return (0);
 
 err:	if (jc != NULL) {
 		if (jc->j_curslist != NULL)
-			__os_free(dbenv,
-			    jc->j_curslist, nslots * sizeof(DBC *));
+			__os_free(dbenv, jc->j_curslist);
 		if (jc->j_workcurs != NULL) {
 			if (jc->j_workcurs[0] != NULL)
-				__os_free(dbenv,
-				    jc->j_workcurs[0], sizeof(DBC));
-			__os_free(dbenv,
-			    jc->j_workcurs, nslots * sizeof(DBC *));
+				(void)__db_c_close(jc->j_workcurs[0]);
+			__os_free(dbenv, jc->j_workcurs);
 		}
 		if (jc->j_fdupcurs != NULL)
-			__os_free(dbenv,
-			    jc->j_fdupcurs, nslots * sizeof(DBC *));
+			__os_free(dbenv, jc->j_fdupcurs);
 		if (jc->j_exhausted != NULL)
-			__os_free(dbenv,
-			    jc->j_exhausted, nslots * sizeof(u_int8_t));
-		__os_free(dbenv, jc, sizeof(JOIN_CURSOR));
+			__os_free(dbenv, jc->j_exhausted);
+		__os_free(dbenv, jc);
 	}
 	if (dbc != NULL)
-		__os_free(dbenv, dbc, sizeof(DBC));
+		__os_free(dbenv, dbc);
+	return (ret);
+}
+
+/*
+ * __db_join_close_pp --
+ *	DBC->c_close pre/post processing for join cursors.
+ */
+static int
+__db_join_close_pp(dbc)
+	DBC *dbc;
+{
+	DB_ENV *dbenv;
+	DB_THREAD_INFO *ip;
+	DB *dbp;
+	int handle_check, ret, t_ret;
+
+	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
+
+	PANIC_CHECK(dbenv);
+
+	ENV_ENTER(dbenv, ip);
+
+	handle_check = IS_ENV_REPLICATED(dbenv);
+	if (handle_check &&
+	    (ret = __db_rep_enter(dbp, 1, 0, dbc->txn != NULL)) != 0) {
+		handle_check = 0;
+		goto err;
+	}
+
+	ret = __db_join_close(dbc);
+
+	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+		ret = t_ret;
+
+err:	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -276,29 +296,108 @@ __db_join_del(dbc, flags)
 	return (EINVAL);
 }
 
+/*
+ * __db_join_get_pp --
+ *	DBjoin->get pre/post processing.
+ */
+static int
+__db_join_get_pp(dbc, key, data, flags)
+	DBC *dbc;
+	DBT *key, *data;
+	u_int32_t flags;
+{
+	DB *dbp;
+	DB_ENV *dbenv;
+	DB_THREAD_INFO *ip;
+	u_int32_t handle_check, save_flags;
+	int ret, t_ret;
+
+	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
+
+	/* Save the original flags value. */
+	save_flags = flags;
+
+	PANIC_CHECK(dbenv);
+
+	if (LF_ISSET(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW)) {
+		if (!LOCKING_ON(dbenv))
+			return (__db_fnl(dbenv, "DBcursor->c_get"));
+		LF_CLR(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW);
+	}
+
+	switch (flags) {
+	case 0:
+	case DB_JOIN_ITEM:
+		break;
+	default:
+		return (__db_ferr(dbenv, "DBcursor->c_get", 0));
+	}
+
+	/*
+	 * A partial get of the key of a join cursor don't make much sense;
+	 * the entire key is necessary to query the primary database
+	 * and find the datum, and so regardless of the size of the key
+	 * it would not be a performance improvement.  Since it would require
+	 * special handling, we simply disallow it.
+	 *
+	 * A partial get of the data, however, potentially makes sense (if
+	 * all possible data are a predictable large structure, for instance)
+	 * and causes us no headaches, so we permit it.
+	 */
+	if (F_ISSET(key, DB_DBT_PARTIAL)) {
+		__db_errx(dbenv,
+		    "DB_DBT_PARTIAL may not be set on key during join_get");
+		return (EINVAL);
+	}
+
+	ENV_ENTER(dbenv, ip);
+
+	handle_check = IS_ENV_REPLICATED(dbenv);
+	if (handle_check &&
+	    (ret = __db_rep_enter(dbp, 1, 0, dbc->txn != NULL)) != 0) {
+		handle_check = 0;
+		goto err;
+	}
+
+	/* Restore the original flags value. */
+	flags = save_flags;
+
+	ret = __db_join_get(dbc, key, data, flags);
+
+	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+		ret = t_ret;
+
+err:	ENV_LEAVE(dbenv, ip);
+	__dbt_userfree(dbenv, key, NULL, NULL);
+	return (ret);
+}
+
 static int
 __db_join_get(dbc, key_arg, data_arg, flags)
 	DBC *dbc;
 	DBT *key_arg, *data_arg;
 	u_int32_t flags;
 {
-	DBT *key_n, key_n_mem;
 	DB *dbp;
 	DBC *cp;
+	DBT *key_n, key_n_mem;
+	DB_ENV *dbenv;
 	JOIN_CURSOR *jc;
 	int db_manage_data, ret;
 	u_int32_t i, j, operation, opmods;
 
 	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
 	jc = (JOIN_CURSOR *)dbc->internal;
 
-	PANIC_CHECK(dbp->dbenv);
-
 	operation = LF_ISSET(DB_OPFLAGS_MASK);
-	opmods = LF_ISSET(DB_RMW | DB_DIRTY_READ);
 
-	if ((ret = __db_joingetchk(dbp, key_arg, flags)) != 0)
-		return (ret);
+	/* !!!
+	 * If the set of flags here changes, check that __db_join_primget
+	 * is updated to handle them properly.
+	 */
+	opmods = LF_ISSET(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW);
 
 	/*
 	 * Since we are fetching the key as a datum in the secondary indices,
@@ -306,8 +405,8 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 	 * management flags.  If necessary, use a stack-allocated DBT;
 	 * we'll appropriately copy and/or allocate the data later.
 	 */
-	if (F_ISSET(key_arg, DB_DBT_USERMEM) ||
-	    F_ISSET(key_arg, DB_DBT_MALLOC)) {
+	if (F_ISSET(key_arg,
+	    DB_DBT_MALLOC | DB_DBT_USERCOPY | DB_DBT_USERMEM)) {
 		/* We just use the default buffer;  no need to go malloc. */
 		key_n = &key_n_mem;
 		memset(key_n, 0, sizeof(DBT));
@@ -318,6 +417,8 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 		 */
 		key_n = key_arg;
 	}
+	if (F_ISSET(key_arg, DB_DBT_USERCOPY))
+		key_arg->data = NULL;
 
 	/*
 	 * If our last attempt to do a get on the primary key failed,
@@ -327,13 +428,12 @@ __db_join_get(dbc, key_arg, data_arg, flags)
 		goto samekey;
 	F_CLR(jc, JOIN_RETRY);
 
-retry:	ret = jc->j_workcurs[0]->c_real_get(jc->j_workcurs[0],
-	    &jc->j_key, key_n,
+retry:	ret = __db_c_get(jc->j_workcurs[0], &jc->j_key, key_n,
 	    opmods | (jc->j_exhausted[0] ? DB_NEXT_DUP : DB_CURRENT));
 
-	if (ret == ENOMEM) {
+	if (ret == DB_BUFFER_SMALL) {
 		jc->j_key.ulen <<= 1;
-		if ((ret = __os_realloc(dbp->dbenv,
+		if ((ret = __os_realloc(dbenv,
 		    jc->j_key.ulen, &jc->j_key.data)) != 0)
 			goto mem_err;
 		goto retry;
@@ -356,7 +456,7 @@ retry:	ret = jc->j_workcurs[0]->c_real_get(jc->j_workcurs[0],
 	 */
 	for (i = 1; i < jc->j_ncurs; i++) {
 		if (jc->j_fdupcurs[i] != NULL &&
-		    (ret = jc->j_fdupcurs[i]->c_close(jc->j_fdupcurs[i])) != 0)
+		    (ret = __db_c_close(jc->j_fdupcurs[i])) != 0)
 			goto err;
 		jc->j_fdupcurs[i] = NULL;
 	}
@@ -377,12 +477,11 @@ retry:	ret = jc->j_workcurs[0]->c_real_get(jc->j_workcurs[0],
 
 	/* We have the first element; now look for it in the other cursors. */
 	for (i = 1; i < jc->j_ncurs; i++) {
-		DB_ASSERT(jc->j_curslist[i] != NULL);
+		DB_ASSERT(dbenv, jc->j_curslist[i] != NULL);
 		if (jc->j_workcurs[i] == NULL)
 			/* If this is NULL, we need to dup curslist into it. */
-			if ((ret = jc->j_curslist[i]->c_dup(
-			    jc->j_curslist[i], jc->j_workcurs + i,
-			    DB_POSITIONI)) != 0)
+			if ((ret = __db_c_dup(jc->j_curslist[i],
+			    &jc->j_workcurs[i], DB_POSITION)) != 0)
 				goto err;
 
 retry2:		cp = jc->j_workcurs[i];
@@ -433,7 +532,7 @@ retry2:		cp = jc->j_workcurs[i];
 					 * and let strange things happen--we
 					 * can't make rope childproof.
 					 */
-					if ((ret = jc->j_workcurs[j]->c_close(
+					if ((ret = __db_c_close(
 					    jc->j_workcurs[j])) != 0)
 						goto err;
 					if (!SORTED_SET(jc, 0) ||
@@ -446,10 +545,10 @@ retry2:		cp = jc->j_workcurs[i];
 						jc->j_workcurs[j] = NULL;
 					else
 						/* Partial reset suffices. */
-						if ((jc->j_fdupcurs[j]->c_dup(
+						if ((__db_c_dup(
 						    jc->j_fdupcurs[j],
 						    &jc->j_workcurs[j],
-						    DB_POSITIONI)) != 0)
+						    DB_POSITION)) != 0)
 							goto err;
 					jc->j_exhausted[j] = 0;
 				}
@@ -465,27 +564,25 @@ retry2:		cp = jc->j_workcurs[i];
 			for (j = i + 1;
 			    jc->j_workcurs[j] != NULL;
 			    j++) {
-				if ((ret = jc->j_workcurs[j]->c_close(
-				    jc->j_workcurs[j])) != 0)
+				if ((ret =
+				    __db_c_close(jc->j_workcurs[j])) != 0)
 					goto err;
 				jc->j_exhausted[j] = 0;
-				if (jc->j_fdupcurs[j] != NULL &&
-				    (ret = jc->j_fdupcurs[j]->c_dup(
-				    jc->j_fdupcurs[j], &jc->j_workcurs[j],
-				    DB_POSITIONI)) != 0)
-					goto err;
-				else
+				if (jc->j_fdupcurs[j] == NULL)
 					jc->j_workcurs[j] = NULL;
+				else if ((ret = __db_c_dup(jc->j_fdupcurs[j],
+				    &jc->j_workcurs[j], DB_POSITION)) != 0)
+					goto err;
 			}
 			goto retry2;
 			/* NOTREACHED */
 		}
 
-		if (ret == ENOMEM) {
+		if (ret == DB_BUFFER_SMALL) {
 			jc->j_key.ulen <<= 1;
-			if ((ret = __os_realloc(dbp->dbenv, jc->j_key.ulen,
+			if ((ret = __os_realloc(dbenv, jc->j_key.ulen,
 			    &jc->j_key.data)) != 0) {
-mem_err:			__db_err(dbp->dbenv,
+mem_err:			__db_errx(dbenv,
 				    "Allocation failed for join key, len = %lu",
 				    (u_long)jc->j_key.ulen);
 				goto err;
@@ -519,9 +616,8 @@ mem_err:			__db_err(dbp->dbenv,
 		 * duplicate duplicates;  store this into jc->j_fdupcurs[i].
 		 */
 		if (SORTED_SET(jc, i) && jc->j_fdupcurs[i] == NULL && (ret =
-		    cp->c_dup(cp, &jc->j_fdupcurs[i], DB_POSITIONI)) != 0)
+		    __db_c_dup(cp, &jc->j_fdupcurs[i], DB_POSITION)) != 0)
 			goto err;
-
 	}
 
 err:	if (ret != 0)
@@ -532,7 +628,7 @@ samekey:	/*
 		 * Get the key we tried and failed to return last time;
 		 * it should be the current datum of all the secondary cursors.
 		 */
-		if ((ret = jc->j_workcurs[0]->c_real_get(jc->j_workcurs[0],
+		if ((ret = __db_c_get(jc->j_workcurs[0],
 		    &jc->j_key, key_n, DB_CURRENT | opmods)) != 0)
 			return (ret);
 		F_CLR(jc, JOIN_RETRY);
@@ -545,12 +641,13 @@ samekey:	/*
 	 * back into the dbt we were given for the key; call __db_retcopy.
 	 * Otherwise, assert that we do not need to copy anything and proceed.
 	 */
-	DB_ASSERT(F_ISSET(
-	    key_arg, DB_DBT_USERMEM | DB_DBT_MALLOC) || key_n == key_arg);
+	DB_ASSERT(dbenv, F_ISSET(key_arg, DB_DBT_USERMEM | DB_DBT_MALLOC |
+	    DB_DBT_USERCOPY) || key_n == key_arg);
 
-	if (F_ISSET(key_arg, DB_DBT_USERMEM | DB_DBT_MALLOC) &&
-	    (ret = __db_retcopy(
-	    dbp, key_arg, key_n->data, key_n->size, NULL, NULL)) != 0) {
+	if ((F_ISSET(key_arg, DB_DBT_USERMEM | DB_DBT_MALLOC |
+	    DB_DBT_USERCOPY)) &&
+	    (ret = __db_retcopy(dbenv,
+	    key_arg, key_n->data, key_n->size, NULL, NULL)) != 0) {
 		/*
 		 * The retcopy failed, most commonly because we have a user
 		 * buffer for the key which is too small. Set things up to
@@ -563,9 +660,6 @@ samekey:	/*
 	/*
 	 * If DB_JOIN_ITEM is set, we return it; otherwise we do the lookup
 	 * in the primary and then return.
-	 *
-	 * Note that we use key_arg here;  it is safe (and appropriate)
-	 * to do so.
 	 */
 	if (operation == DB_JOIN_ITEM)
 		return (0);
@@ -580,14 +674,19 @@ samekey:	/*
 	 * Instead, use memory that is managed by the join cursor, in
 	 * jc->j_rdata.
 	 */
-	if (!F_ISSET(data_arg, DB_DBT_MALLOC | DB_DBT_REALLOC | DB_DBT_USERMEM))
+	if (!F_ISSET(data_arg, DB_DBT_MALLOC | DB_DBT_REALLOC |
+	    DB_DBT_USERMEM | DB_DBT_USERCOPY))
 		db_manage_data = 1;
 	else
 		db_manage_data = 0;
-	if ((ret = jc->j_primary->get(jc->j_primary,
-	    jc->j_curslist[0]->txn, key_arg,
+	if ((ret = __db_join_primget(jc->j_primary,
+	    jc->j_curslist[0]->txn, jc->j_curslist[0]->locker, key_n,
 	    db_manage_data ? &jc->j_rdata : data_arg, opmods)) != 0) {
-		if (ret == DB_NOTFOUND)
+		if (ret == DB_NOTFOUND) {
+			if (LF_ISSET(DB_READ_UNCOMMITTED) ||
+			    (jc->j_curslist[0]->txn != NULL && F_ISSET(
+			    jc->j_curslist[0]->txn, TXN_READ_UNCOMMITTED)))
+				goto retry;
 			/*
 			 * If ret == DB_NOTFOUND, the primary and secondary
 			 * are out of sync;  every item in each secondary
@@ -596,7 +695,7 @@ samekey:	/*
 			 * Wail.
 			 */
 			ret = __db_secondary_corrupt(jc->j_primary);
-		else
+		} else
 			/*
 			 * The get on the primary failed for some other
 			 * reason, most commonly because we're using a user
@@ -613,7 +712,13 @@ samekey:	/*
 	return (ret);
 }
 
-static int
+/*
+ * __db_join_close --
+ *	DBC->c_close for join cursors.
+ *
+ * PUBLIC: int __db_join_close __P((DBC *));
+ */
+int
 __db_join_close(dbc)
 	DBC *dbc;
 {
@@ -633,9 +738,9 @@ __db_join_close(dbc)
 	 * must happen before any action that can fail and return, or else
 	 * __db_close may loop indefinitely.
 	 */
-	MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
+	MUTEX_LOCK(dbenv, dbp->mutex);
 	TAILQ_REMOVE(&dbp->join_queue, dbc, links);
-	MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+	MUTEX_UNLOCK(dbenv, dbp->mutex);
 
 	PANIC_CHECK(dbenv);
 
@@ -651,23 +756,23 @@ __db_join_close(dbc)
 	 * mucking with.
 	 */
 	for (i = 0; i < jc->j_ncurs; i++) {
-		if (jc->j_workcurs[i] != NULL && (t_ret =
-		    jc->j_workcurs[i]->c_close(jc->j_workcurs[i])) != 0)
+		if (jc->j_workcurs[i] != NULL &&
+		    (t_ret = __db_c_close(jc->j_workcurs[i])) != 0)
 			ret = t_ret;
-		if (jc->j_fdupcurs[i] != NULL && (t_ret =
-		    jc->j_fdupcurs[i]->c_close(jc->j_fdupcurs[i])) != 0)
+		if (jc->j_fdupcurs[i] != NULL &&
+		    (t_ret = __db_c_close(jc->j_fdupcurs[i])) != 0)
 			ret = t_ret;
 	}
 
-	__os_free(dbenv, jc->j_exhausted, 0);
-	__os_free(dbenv, jc->j_curslist, 0);
-	__os_free(dbenv, jc->j_workcurs, 0);
-	__os_free(dbenv, jc->j_fdupcurs, 0);
-	__os_free(dbenv, jc->j_key.data, jc->j_key.ulen);
+	__os_free(dbenv, jc->j_exhausted);
+	__os_free(dbenv, jc->j_curslist);
+	__os_free(dbenv, jc->j_workcurs);
+	__os_free(dbenv, jc->j_fdupcurs);
+	__os_free(dbenv, jc->j_key.data);
 	if (jc->j_rdata.data != NULL)
-		__os_ufree(dbenv, jc->j_rdata.data, 0);
-	__os_free(dbenv, jc, sizeof(JOIN_CURSOR));
-	__os_free(dbenv, dbc, sizeof(DBC));
+		__os_ufree(dbenv, jc->j_rdata.data);
+	__os_free(dbenv, jc);
+	__os_free(dbenv, dbc);
 
 	return (ret);
 }
@@ -709,7 +814,7 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
 		 */
 		memset(&ldata, 0, sizeof(DBT));
 		F_SET(&ldata, DB_DBT_MALLOC);
-		if ((ret = dbc->c_real_get(dbc,
+		if ((ret = __db_c_get(dbc,
 		    key, &ldata, opmods | DB_CURRENT)) != 0)
 			break;
 		cmp = func(dbp, data, &ldata);
@@ -719,10 +824,10 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
 			 * it into data, then free the buffer we malloc'ed
 			 * above.
 			 */
-			if ((ret = __db_retcopy(dbp, data, ldata.data,
+			if ((ret = __db_retcopy(dbp->dbenv, data, ldata.data,
 			    ldata.size, &data->data, &data->size)) != 0)
 				return (ret);
-			__os_ufree(dbp->dbenv, ldata.data, 0);
+			__os_ufree(dbp->dbenv, ldata.data);
 			return (0);
 		}
 
@@ -731,10 +836,10 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
 		 * dups.  We just forget about ldata and free
 		 * its buffer--data contains the value we're searching for.
 		 */
-		__os_ufree(dbp->dbenv, ldata.data, 0);
+		__os_ufree(dbp->dbenv, ldata.data);
 		/* FALLTHROUGH */
 	case 1:
-		ret = dbc->c_real_get(dbc, key, data, opmods | DB_GET_BOTHC);
+		ret = __db_c_get(dbc, key, data, opmods | DB_GET_BOTHC);
 		break;
 	default:
 		ret = EINVAL;
@@ -748,7 +853,6 @@ __db_join_getnext(dbc, key, data, exhausted, opmods)
  * __db_join_cmp --
  *	Comparison function for sorting DBCs in cardinality order.
  */
-
 static int
 __db_join_cmp(a, b)
 	const void *a, *b;
@@ -756,15 +860,84 @@ __db_join_cmp(a, b)
 	DBC *dbca, *dbcb;
 	db_recno_t counta, countb;
 
-	/* In case c_count fails, pretend cursors are equal. */
-	counta = countb = 0;
-
 	dbca = *((DBC * const *)a);
 	dbcb = *((DBC * const *)b);
 
-	if (dbca->c_count(dbca, &counta, 0) != 0 ||
-	    dbcb->c_count(dbcb, &countb, 0) != 0)
+	if (__db_c_count(dbca, &counta) != 0 ||
+	    __db_c_count(dbcb, &countb) != 0)
 		return (0);
 
-	return (counta - countb);
+	return ((long)counta - (long)countb);
+}
+
+/*
+ * __db_join_primget --
+ *	Perform a DB->get in the primary, being careful not to use a new
+ * locker ID if we're doing CDB locking.
+ */
+static int
+__db_join_primget(dbp, txn, lockerid, key, data, flags)
+	DB *dbp;
+	DB_TXN *txn;
+	u_int32_t lockerid;
+	DBT *key, *data;
+	u_int32_t flags;
+{
+	DBC *dbc;
+	u_int32_t rmw;
+	int ret, t_ret;
+
+	if ((ret = __db_cursor_int(dbp,
+	    txn, dbp->type, PGNO_INVALID, 0, lockerid, &dbc)) != 0)
+		return (ret);
+
+	/*
+	 * The only allowable flags here are the two flags copied into "opmods"
+	 * in __db_join_get, DB_RMW and DB_READ_UNCOMMITTED.  The former is an
+	 * op on the c_get call, the latter on the cursor call.  It's a DB bug
+	 * if we allow any other flags down in here.
+	 */
+	rmw = LF_ISSET(DB_RMW);
+	if (LF_ISSET(DB_READ_UNCOMMITTED) ||
+	    (txn != NULL && F_ISSET(txn, TXN_READ_UNCOMMITTED)))
+		F_SET(dbc, DBC_READ_UNCOMMITTED);
+
+	if (LF_ISSET(DB_READ_COMMITTED) ||
+	    (txn != NULL && F_ISSET(txn, TXN_READ_COMMITTED)))
+		F_SET(dbc, DBC_READ_COMMITTED);
+
+	LF_CLR(DB_READ_COMMITTED | DB_READ_UNCOMMITTED | DB_RMW);
+	DB_ASSERT(dbp->dbenv, flags == 0);
+
+	F_SET(dbc, DBC_TRANSIENT);
+
+	/*
+	 * This shouldn't be necessary, thanks to the fact that join cursors
+	 * swap in their own DB_DBT_REALLOC'ed buffers, but just for form's
+	 * sake, we mirror what __db_get does.
+	 */
+	SET_RET_MEM(dbc, dbp);
+
+	ret = __db_c_get(dbc, key, data, DB_SET | rmw);
+
+	if ((t_ret = __db_c_close(dbc)) != 0 && ret == 0)
+		ret = t_ret;
+
+	return (ret);
+}
+
+/*
+ * __db_secondary_corrupt --
+ *	Report that a secondary index appears corrupt, as it has a record
+ * that does not correspond to a record in the primary or vice versa.
+ *
+ * PUBLIC: int __db_secondary_corrupt __P((DB *));
+ */
+int
+__db_secondary_corrupt(dbp)
+	DB *dbp;
+{
+	__db_errx(dbp->dbenv,
+	    "Secondary index corrupt: not consistent with primary");
+	return (DB_SECONDARY_BAD);
 }
