@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1990, 1993, 1994, 1995, 1996
@@ -38,24 +38,16 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: db_overflow.c,v 12.13 2006/08/24 14:45:16 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: db_overflow.c,v 11.36 2001/10/30 19:42:15 ubell Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#endif
-
 #include "db_int.h"
-#include "db_page.h"
-#include "db_am.h"
-#include "db_verify.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_am.h"
+#include "dbinc/mp.h"
 
 /*
  * Big key/data code.
@@ -70,12 +62,13 @@ static const char revid[] = "$Id: db_overflow.c,v 11.36 2001/10/30 19:42:15 ubel
  * __db_goff --
  *	Get an offpage item.
  *
- * PUBLIC: int __db_goff __P((DB *, DBT *,
+ * PUBLIC: int __db_goff __P((DB *, DB_TXN *, DBT *,
  * PUBLIC:     u_int32_t, db_pgno_t, void **, u_int32_t *));
  */
 int
-__db_goff(dbp, dbt, tlen, pgno, bpp, bpsz)
+__db_goff(dbp, txn, dbt, tlen, pgno, bpp, bpsz)
 	DB *dbp;
+	DB_TXN *txn;
 	DBT *dbt;
 	u_int32_t tlen;
 	db_pgno_t pgno;
@@ -112,11 +105,14 @@ __db_goff(dbp, dbt, tlen, pgno, bpp, bpsz)
 		needed = tlen;
 	}
 
+	if (F_ISSET(dbt, DB_DBT_USERCOPY))
+		goto skip_alloc;
+
 	/* Allocate any necessary memory. */
 	if (F_ISSET(dbt, DB_DBT_USERMEM)) {
 		if (needed > dbt->ulen) {
 			dbt->size = needed;
-			return (ENOMEM);
+			return (DB_BUFFER_SMALL);
 		}
 	} else if (F_ISSET(dbt, DB_DBT_MALLOC)) {
 		if ((ret = __os_umalloc(dbenv, needed, &dbt->data)) != 0)
@@ -124,40 +120,56 @@ __db_goff(dbp, dbt, tlen, pgno, bpp, bpsz)
 	} else if (F_ISSET(dbt, DB_DBT_REALLOC)) {
 		if ((ret = __os_urealloc(dbenv, needed, &dbt->data)) != 0)
 			return (ret);
-	} else if (*bpsz == 0 || *bpsz < needed) {
+	} else if (bpsz != NULL && (*bpsz == 0 || *bpsz < needed)) {
 		if ((ret = __os_realloc(dbenv, needed, bpp)) != 0)
 			return (ret);
 		*bpsz = needed;
 		dbt->data = *bpp;
-	} else
+	} else if (bpp != NULL)
 		dbt->data = *bpp;
+	else {
+		DB_ASSERT(dbenv,
+		    F_ISSET(dbt,
+		    DB_DBT_USERMEM | DB_DBT_MALLOC | DB_DBT_REALLOC) ||
+		    bpsz != NULL || bpp != NULL);
+		return (DB_BUFFER_SMALL);
+	}
 
+skip_alloc:
 	/*
 	 * Step through the linked list of pages, copying the data on each
 	 * one into the buffer.  Never copy more than the total data length.
 	 */
 	dbt->size = needed;
 	for (curoff = 0, p = dbt->data; pgno != PGNO_INVALID && needed > 0;) {
-		if ((ret = mpf->get(mpf, &pgno, 0, &h)) != 0)
+		if ((ret = __memp_fget(mpf, &pgno, txn, 0, &h)) != 0)
 			return (ret);
+		DB_ASSERT(dbenv, TYPE(h) == P_OVERFLOW);
 
 		/* Check if we need any bytes from this page. */
 		if (curoff + OV_LEN(h) >= start) {
-			src = (u_int8_t *)h + P_OVERHEAD;
 			bytes = OV_LEN(h);
+			src = (u_int8_t *)h + P_OVERHEAD(dbp);
 			if (start > curoff) {
 				src += start - curoff;
 				bytes -= start - curoff;
 			}
 			if (bytes > needed)
 				bytes = needed;
-			memcpy(p, src, bytes);
+			if (F_ISSET(dbt, DB_DBT_USERCOPY)) {
+				if ((ret = dbenv->dbt_usercopy(dbt, curoff,
+				    src, bytes, DB_USERCOPY_SETDATA)) != 0) {
+					(void)__memp_fput(mpf, h, 0);
+					return (ret);
+				}
+			} else
+				memcpy(p, src, bytes);
 			p += bytes;
 			needed -= bytes;
 		}
 		curoff += OV_LEN(h);
 		pgno = h->next_pgno;
-		(void)mpf->put(mpf, h, 0);
+		(void)__memp_fput(mpf, h, 0);
 	}
 	return (0);
 }
@@ -191,7 +203,7 @@ __db_poff(dbc, dbt, pgnop)
 	 */
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
-	pagespace = P_MAXSPACE(dbp->pgsize);
+	pagespace = P_MAXSPACE(dbp, dbp->pgsize);
 
 	ret = 0;
 	lastp = NULL;
@@ -210,35 +222,33 @@ __db_poff(dbc, dbt, pgnop)
 		 * have a partial record.
 		 */
 		if ((ret = __db_new(dbc, P_OVERFLOW, &pagep)) != 0)
-			return (ret);
-		if (DB_LOGGING(dbc)) {
+			break;
+		if (DBC_LOGGING(dbc)) {
 			tmp_dbt.data = p;
 			tmp_dbt.size = pagespace;
 			ZERO_LSN(null_lsn);
-			if ((ret = __db_big_log(dbp->dbenv, dbc->txn,
-			    &new_lsn, 0, DB_ADD_BIG, dbp->log_fileid,
-			    PGNO(pagep), lastp ? PGNO(lastp) : PGNO_INVALID,
+			if ((ret = __db_big_log(dbp, dbc->txn,
+			    &new_lsn, 0, DB_ADD_BIG, PGNO(pagep),
+			    lastp ? PGNO(lastp) : PGNO_INVALID,
 			    PGNO_INVALID, &tmp_dbt, &LSN(pagep),
 			    lastp == NULL ? &null_lsn : &LSN(lastp),
 			    &null_lsn)) != 0) {
 				if (lastp != NULL)
-					(void)mpf->put(mpf,
-					    lastp, DB_MPOOL_DIRTY);
+					(void)__memp_fput(mpf, lastp, 0);
 				lastp = pagep;
 				break;
 			}
 		} else
 			LSN_NOT_LOGGED(new_lsn);
-		/* Move lsn onto page. */
-		if (lastp)
+
+		/* Move LSN onto page. */
+		if (lastp != NULL)
 			LSN(lastp) = new_lsn;
 		LSN(pagep) = new_lsn;
 
-		P_INIT(pagep, dbp->pgsize,
-		    PGNO(pagep), PGNO_INVALID, PGNO_INVALID, 0, P_OVERFLOW);
 		OV_LEN(pagep) = pagespace;
 		OV_REF(pagep) = 1;
-		memcpy((u_int8_t *)pagep + P_OVERHEAD, p, pagespace);
+		memcpy((u_int8_t *)pagep + P_OVERHEAD(dbp), p, pagespace);
 
 		/*
 		 * If this is the first entry, update the user's info.
@@ -250,26 +260,26 @@ __db_poff(dbc, dbt, pgnop)
 		else {
 			lastp->next_pgno = PGNO(pagep);
 			pagep->prev_pgno = PGNO(lastp);
-			(void)mpf->put(mpf, lastp, DB_MPOOL_DIRTY);
+			(void)__memp_fput(mpf, lastp, 0);
 		}
 		lastp = pagep;
 	}
-	if ((t_ret = mpf->put(mpf, lastp, DB_MPOOL_DIRTY)) != 0 && ret == 0)
+	if (lastp != NULL &&
+	    (t_ret = __memp_fput(mpf, lastp, 0)) != 0 && ret == 0)
 		ret = t_ret;
 	return (ret);
 }
 
 /*
  * __db_ovref --
- *	Increment/decrement the reference count on an overflow page.
+ *	Decrement the reference count on an overflow page.
  *
- * PUBLIC: int __db_ovref __P((DBC *, db_pgno_t, int32_t));
+ * PUBLIC: int __db_ovref __P((DBC *, db_pgno_t));
  */
 int
-__db_ovref(dbc, pgno, adjust)
+__db_ovref(dbc, pgno)
 	DBC *dbc;
 	db_pgno_t pgno;
-	int32_t adjust;
 {
 	DB *dbp;
 	DB_MPOOLFILE *mpf;
@@ -279,24 +289,30 @@ __db_ovref(dbc, pgno, adjust)
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
 
-	if ((ret = mpf->get(mpf, &pgno, 0, &h)) != 0) {
-		(void)__db_pgerr(dbp, pgno);
+	if ((ret = __memp_fget(mpf, &pgno, dbc->txn, DB_MPOOL_DIRTY, &h)) != 0)
 		return (ret);
-	}
 
-	if (DB_LOGGING(dbc)) {
-		if ((ret = __db_ovref_log(dbp->dbenv, dbc->txn,
-		    &LSN(h), 0, dbp->log_fileid, h->pgno, adjust,
-		    &LSN(h))) != 0) {
-			(void)mpf->put(mpf, h, 0);
+	if (DBC_LOGGING(dbc)) {
+		if ((ret = __db_ovref_log(dbp,
+		    dbc->txn, &LSN(h), 0, h->pgno, -1, &LSN(h))) != 0) {
+			(void)__memp_fput(mpf, h, 0);
 			return (ret);
 		}
 	} else
 		LSN_NOT_LOGGED(LSN(h));
-	OV_REF(h) += adjust;
 
-	(void)mpf->put(mpf, h, DB_MPOOL_DIRTY);
-	return (0);
+	/*
+	 * In BDB releases before 4.5, the overflow reference counts were
+	 * incremented when an overflow item was split onto an internal
+	 * page.  There was a lock race in that code, and rather than fix
+	 * the race, we changed BDB to copy overflow items when splitting
+	 * them onto internal pages.  The code to decrement reference
+	 * counts remains so databases already in the field continue to
+	 * work.
+	 */
+	--OV_REF(h);
+
+	return (__memp_fput(mpf, h, 0));
 }
 
 /*
@@ -321,36 +337,40 @@ __db_doff(dbc, pgno)
 	mpf = dbp->mpf;
 
 	do {
-		if ((ret = mpf->get(mpf, &pgno, 0, &pagep)) != 0) {
-			(void)__db_pgerr(dbp, pgno);
+		if ((ret = __memp_fget(mpf, &pgno, dbc->txn, 0, &pagep)) != 0)
 			return (ret);
-		}
 
-		DB_ASSERT(TYPE(pagep) == P_OVERFLOW);
+		DB_ASSERT(dbp->dbenv, TYPE(pagep) == P_OVERFLOW);
 		/*
 		 * If it's referenced by more than one key/data item,
 		 * decrement the reference count and return.
 		 */
 		if (OV_REF(pagep) > 1) {
-			(void)mpf->put(mpf, pagep, 0);
-			return (__db_ovref(dbc, pgno, -1));
+			(void)__memp_fput(mpf, pagep, 0);
+			return (__db_ovref(dbc, pgno));
 		}
 
-		if (DB_LOGGING(dbc)) {
-			tmp_dbt.data = (u_int8_t *)pagep + P_OVERHEAD;
+		if ((ret = __memp_dirty(mpf, &pagep, dbc->txn, 0)) != 0) {
+			(void)__memp_fput(mpf, pagep, 0);
+			return (ret);
+		}
+
+		if (DBC_LOGGING(dbc)) {
+			tmp_dbt.data = (u_int8_t *)pagep + P_OVERHEAD(dbp);
 			tmp_dbt.size = OV_LEN(pagep);
 			ZERO_LSN(null_lsn);
-			if ((ret = __db_big_log(dbp->dbenv, dbc->txn,
-			    &LSN(pagep), 0, DB_REM_BIG, dbp->log_fileid,
+			if ((ret = __db_big_log(dbp, dbc->txn,
+			    &LSN(pagep), 0, DB_REM_BIG,
 			    PGNO(pagep), PREV_PGNO(pagep),
 			    NEXT_PGNO(pagep), &tmp_dbt,
 			    &LSN(pagep), &null_lsn, &null_lsn)) != 0) {
-				(void)mpf->put(mpf, pagep, 0);
+				(void)__memp_fput(mpf, pagep, 0);
 				return (ret);
 			}
 		} else
 			LSN_NOT_LOGGED(LSN(pagep));
 		pgno = pagep->next_pgno;
+		OV_LEN(pagep) = 0;
 		if ((ret = __db_free(dbc, pagep)) != 0)
 			return (ret);
 	} while (pgno != PGNO_INVALID);
@@ -368,12 +388,13 @@ __db_doff(dbc, pgno)
  * specified a comparison function.  In this case, we need to materialize
  * the entire object and call their comparison routine.
  *
- * PUBLIC: int __db_moff __P((DB *, const DBT *, db_pgno_t, u_int32_t,
+ * PUBLIC: int __db_moff __P((DB *, DB_TXN *, const DBT *, db_pgno_t, u_int32_t,
  * PUBLIC:     int (*)(DB *, const DBT *, const DBT *), int *));
  */
 int
-__db_moff(dbp, dbt, pgno, tlen, cmpfunc, cmpp)
+__db_moff(dbp, txn, dbt, pgno, tlen, cmpfunc, cmpp)
 	DB *dbp;
+	DB_TXN *txn;
 	const DBT *dbt;
 	db_pgno_t pgno;
 	u_int32_t tlen;
@@ -398,32 +419,32 @@ __db_moff(dbp, dbt, pgno, tlen, cmpfunc, cmpp)
 		buf = NULL;
 		bufsize = 0;
 
-		if ((ret = __db_goff(dbp,
+		if ((ret = __db_goff(dbp, txn,
 		    &local_dbt, tlen, pgno, &buf, &bufsize)) != 0)
 			return (ret);
 		/* Pass the key as the first argument */
 		*cmpp = cmpfunc(dbp, dbt, &local_dbt);
-		__os_free(dbp->dbenv, buf, bufsize);
+		__os_free(dbp->dbenv, buf);
 		return (0);
 	}
 
 	/* While there are both keys to compare. */
 	for (*cmpp = 0, p1 = dbt->data,
 	    key_left = dbt->size; key_left > 0 && pgno != PGNO_INVALID;) {
-		if ((ret = mpf->get(mpf, &pgno, 0, &pagep)) != 0)
+		if ((ret = __memp_fget(mpf, &pgno, txn, 0, &pagep)) != 0)
 			return (ret);
 
 		cmp_bytes = OV_LEN(pagep) < key_left ? OV_LEN(pagep) : key_left;
 		tlen -= cmp_bytes;
 		key_left -= cmp_bytes;
-		for (p2 =
-		    (u_int8_t *)pagep + P_OVERHEAD; cmp_bytes-- > 0; ++p1, ++p2)
+		for (p2 = (u_int8_t *)pagep + P_OVERHEAD(dbp);
+		    cmp_bytes-- > 0; ++p1, ++p2)
 			if (*p1 != *p2) {
 				*cmpp = (long)*p1 - (long)*p2;
 				break;
 			}
 		pgno = NEXT_PGNO(pagep);
-		if ((ret = mpf->put(mpf, pagep, 0)) != 0)
+		if ((ret = __memp_fput(mpf, pagep, 0)) != 0)
 			return (ret);
 		if (*cmpp != 0)
 			return (0);
@@ -437,287 +458,3 @@ __db_moff(dbp, dbt, pgno, tlen, cmpfunc, cmpp)
 
 	return (0);
 }
-
-/*
- * __db_vrfy_overflow --
- *	Verify overflow page.
- *
- * PUBLIC: int __db_vrfy_overflow __P((DB *, VRFY_DBINFO *, PAGE *, db_pgno_t,
- * PUBLIC:     u_int32_t));
- */
-int
-__db_vrfy_overflow(dbp, vdp, h, pgno, flags)
-	DB *dbp;
-	VRFY_DBINFO *vdp;
-	PAGE *h;
-	db_pgno_t pgno;
-	u_int32_t flags;
-{
-	VRFY_PAGEINFO *pip;
-	int isbad, ret, t_ret;
-
-	isbad = 0;
-	if ((ret = __db_vrfy_getpageinfo(vdp, pgno, &pip)) != 0)
-		return (ret);
-
-	if ((ret = __db_vrfy_datapage(dbp, vdp, h, pgno, flags)) != 0) {
-		if (ret == DB_VERIFY_BAD)
-			isbad = 1;
-		else
-			goto err;
-	}
-
-	pip->refcount = OV_REF(h);
-	if (pip->refcount < 1) {
-		EPRINT((dbp->dbenv,
-		    "Overflow page %lu has zero reference count",
-		    (u_long)pgno));
-		isbad = 1;
-	}
-
-	/* Just store for now. */
-	pip->olen = HOFFSET(h);
-
-err:	if ((t_ret = __db_vrfy_putpageinfo(dbp->dbenv, vdp, pip)) != 0)
-		ret = t_ret;
-	return ((ret == 0 && isbad == 1) ? DB_VERIFY_BAD : ret);
-}
-
-/*
- * __db_vrfy_ovfl_structure --
- *	Walk a list of overflow pages, avoiding cycles and marking
- *	pages seen.
- *
- * PUBLIC: int __db_vrfy_ovfl_structure
- * PUBLIC:     __P((DB *, VRFY_DBINFO *, db_pgno_t, u_int32_t, u_int32_t));
- */
-int
-__db_vrfy_ovfl_structure(dbp, vdp, pgno, tlen, flags)
-	DB *dbp;
-	VRFY_DBINFO *vdp;
-	db_pgno_t pgno;
-	u_int32_t tlen;
-	u_int32_t flags;
-{
-	DB *pgset;
-	VRFY_PAGEINFO *pip;
-	db_pgno_t next, prev;
-	int isbad, p, ret, t_ret;
-	u_int32_t refcount;
-
-	pgset = vdp->pgset;
-	DB_ASSERT(pgset != NULL);
-	isbad = 0;
-
-	/* This shouldn't happen, but just to be sure. */
-	if (!IS_VALID_PGNO(pgno))
-		return (DB_VERIFY_BAD);
-
-	/*
-	 * Check the first prev_pgno;  it ought to be PGNO_INVALID,
-	 * since there's no prev page.
-	 */
-	if ((ret = __db_vrfy_getpageinfo(vdp, pgno, &pip)) != 0)
-		return (ret);
-
-	/* The refcount is stored on the first overflow page. */
-	refcount = pip->refcount;
-
-	if (pip->type != P_OVERFLOW) {
-		EPRINT((dbp->dbenv,
-		    "Overflow page %lu of invalid type",
-		    (u_long)pgno, (u_long)pip->type));
-		ret = DB_VERIFY_BAD;
-		goto err;		/* Unsafe to continue. */
-	}
-
-	prev = pip->prev_pgno;
-	if (prev != PGNO_INVALID) {
-		EPRINT((dbp->dbenv,
-		    "First overflow page %lu has a prev_pgno", (u_long)pgno));
-		isbad = 1;
-	}
-
-	for (;;) {
-		/*
-		 * This is slightly gross.  Btree leaf pages reference
-		 * individual overflow trees multiple times if the overflow page
-		 * is the key to a duplicate set.  The reference count does not
-		 * reflect this multiple referencing.  Thus, if this is called
-		 * during the structure verification of a btree leaf page, we
-		 * check to see whether we've seen it from a leaf page before
-		 * and, if we have, adjust our count of how often we've seen it
-		 * accordingly.
-		 *
-		 * (This will screw up if it's actually referenced--and
-		 * correctly refcounted--from two different leaf pages, but
-		 * that's a very unlikely brokenness that we're not checking for
-		 * anyway.)
-		 */
-
-		if (LF_ISSET(ST_OVFL_LEAF)) {
-			if (F_ISSET(pip, VRFY_OVFL_LEAFSEEN)) {
-				if ((ret =
-				    __db_vrfy_pgset_dec(pgset, pgno)) != 0)
-					goto err;
-			} else
-				F_SET(pip, VRFY_OVFL_LEAFSEEN);
-		}
-
-		if ((ret = __db_vrfy_pgset_get(pgset, pgno, &p)) != 0)
-			goto err;
-
-		/*
-		 * We may have seen this elsewhere, if the overflow entry
-		 * has been promoted to an internal page.
-		 */
-		if ((u_int32_t)p > refcount) {
-			EPRINT((dbp->dbenv,
-			    "Page %lu encountered twice in overflow traversal",
-			    (u_long)pgno));
-			ret = DB_VERIFY_BAD;
-			goto err;
-		}
-		if ((ret = __db_vrfy_pgset_inc(pgset, pgno)) != 0)
-			goto err;
-
-		/* Keep a running tab on how much of the item we've seen. */
-		tlen -= pip->olen;
-
-		/* Send feedback to the application about our progress. */
-		if (!LF_ISSET(DB_SALVAGE))
-			__db_vrfy_struct_feedback(dbp, vdp);
-
-		next = pip->next_pgno;
-
-		/* Are we there yet? */
-		if (next == PGNO_INVALID)
-			break;
-
-		/*
-		 * We've already checked this when we saved it, but just
-		 * to be sure...
-		 */
-		if (!IS_VALID_PGNO(next)) {
-			DB_ASSERT(0);
-			EPRINT((dbp->dbenv,
-			    "Overflow page %lu has bad next_pgno",
-			    (u_long)pgno));
-			ret = DB_VERIFY_BAD;
-			goto err;
-		}
-
-		if ((ret = __db_vrfy_putpageinfo(dbp->dbenv, vdp, pip)) != 0 ||
-		    (ret = __db_vrfy_getpageinfo(vdp, next, &pip)) != 0)
-			return (ret);
-		if (pip->prev_pgno != pgno) {
-			EPRINT((dbp->dbenv,
-			    "Overflow page %lu has bogus prev_pgno value",
-			    (u_long)next));
-			isbad = 1;
-			/*
-			 * It's safe to continue because we have separate
-			 * cycle detection.
-			 */
-		}
-
-		pgno = next;
-	}
-
-	if (tlen > 0) {
-		isbad = 1;
-		EPRINT((dbp->dbenv,
-		    "Overflow item incomplete on page %lu", (u_long)pgno));
-	}
-
-err:	if ((t_ret =
-	    __db_vrfy_putpageinfo(dbp->dbenv, vdp, pip)) != 0 && ret == 0)
-		ret = t_ret;
-	return ((ret == 0 && isbad == 1) ? DB_VERIFY_BAD : ret);
-}
-
-/*
- * __db_safe_goff --
- *	Get an overflow item, very carefully, from an untrusted database,
- *	in the context of the salvager.
- *
- * PUBLIC: int __db_safe_goff __P((DB *, VRFY_DBINFO *, db_pgno_t,
- * PUBLIC:     DBT *, void **, u_int32_t));
- */
-int
-__db_safe_goff(dbp, vdp, pgno, dbt, buf, flags)
-	DB *dbp;
-	VRFY_DBINFO *vdp;
-	db_pgno_t pgno;
-	DBT *dbt;
-	void **buf;
-	u_int32_t flags;
-{
-	DB_MPOOLFILE *mpf;
-	PAGE *h;
-	int ret, t_ret;
-	u_int32_t bytesgot, bytes;
-	u_int8_t *src, *dest;
-
-	mpf = dbp->mpf;
-	h = NULL;
-	ret = t_ret = 0;
-	bytesgot = bytes = 0;
-
-	while ((pgno != PGNO_INVALID) && (IS_VALID_PGNO(pgno))) {
-		/*
-		 * Mark that we're looking at this page;  if we've seen it
-		 * already, quit.
-		 */
-		if ((ret = __db_salvage_markdone(vdp, pgno)) != 0)
-			break;
-
-		if ((ret = mpf->get(mpf, &pgno, 0, &h)) != 0)
-			break;
-
-		/*
-		 * Make sure it's really an overflow page, unless we're
-		 * being aggressive, in which case we pretend it is.
-		 */
-		if (!LF_ISSET(DB_AGGRESSIVE) && TYPE(h) != P_OVERFLOW) {
-			ret = DB_VERIFY_BAD;
-			break;
-		}
-
-		src = (u_int8_t *)h + P_OVERHEAD;
-		bytes = OV_LEN(h);
-
-		if (bytes + P_OVERHEAD > dbp->pgsize)
-			bytes = dbp->pgsize - P_OVERHEAD;
-
-		if ((ret = __os_realloc(dbp->dbenv,
-		    bytesgot + bytes, buf)) != 0)
-			break;
-
-		dest = (u_int8_t *)*buf + bytesgot;
-		bytesgot += bytes;
-
-		memcpy(dest, src, bytes);
-
-		pgno = NEXT_PGNO(h);
-
-		if ((ret = mpf->put(mpf, h, 0)) != 0)
-			break;
-		h = NULL;
-	}
-
-	/*
-	 * If we're being aggressive, salvage a partial datum if there
-	 * was an error somewhere along the way.
-	 */
-	if (ret == 0 || LF_ISSET(DB_AGGRESSIVE)) {
-		dbt->size = bytesgot;
-		dbt->data = *buf;
-	}
-
-	/* If we broke out on error, don't leave pages pinned. */
-	if (h != NULL && (t_ret = mpf->put(mpf, h, 0)) != 0 && ret == 0)
-		ret = t_ret;
-
-	return (ret);
-}
