@@ -1,38 +1,29 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996, 1997, 1998, 1999
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: db_ret.c,v 12.6 2006/08/24 14:45:16 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char sccsid[] = "@(#)db_ret.c	11.1 (Sleepycat) 7/24/99";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <errno.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
-#include "db_page.h"
-#include "btree.h"
-#include "db_am.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_am.h"
 
 /*
  * __db_ret --
  *	Build return DBT.
  *
- * PUBLIC: int __db_ret __P((DB *,
+ * PUBLIC: int __db_ret __P((DB *, DB_TXN *,
  * PUBLIC:    PAGE *, u_int32_t, DBT *, void **, u_int32_t *));
  */
 int
-__db_ret(dbp, h, indx, dbt, memp, memsize)
+__db_ret(dbp, txn, h, indx, dbt, memp, memsize)
 	DB *dbp;
+	DB_TXN *txn;
 	PAGE *h;
 	u_int32_t indx;
 	DBT *dbt;
@@ -48,45 +39,44 @@ __db_ret(dbp, h, indx, dbt, memp, memsize)
 
 	switch (TYPE(h)) {
 	case P_HASH:
-		hk = P_ENTRY(h, indx);
+		hk = P_ENTRY(dbp, h, indx);
 		if (HPAGE_PTYPE(hk) == H_OFFPAGE) {
 			memcpy(&ho, hk, sizeof(HOFFPAGE));
-			return (__db_goff(dbp, dbt,
+			return (__db_goff(dbp, txn, dbt,
 			    ho.tlen, ho.pgno, memp, memsize));
 		}
-		len = LEN_HKEYDATA(h, dbp->pgsize, indx);
+		len = LEN_HKEYDATA(dbp, h, dbp->pgsize, indx);
 		data = HKEYDATA_DATA(hk);
 		break;
-	case P_DUPLICATE:
 	case P_LBTREE:
+	case P_LDUP:
 	case P_LRECNO:
-		bk = GET_BKEYDATA(h, indx);
+		bk = GET_BKEYDATA(dbp, h, indx);
 		if (B_TYPE(bk->type) == B_OVERFLOW) {
 			bo = (BOVERFLOW *)bk;
-			return (__db_goff(dbp, dbt,
+			return (__db_goff(dbp, txn, dbt,
 			    bo->tlen, bo->pgno, memp, memsize));
 		}
 		len = bk->len;
 		data = bk->data;
 		break;
 	default:
-		return (__db_pgfmt(dbp, h->pgno));
+		return (__db_pgfmt(dbp->dbenv, h->pgno));
 	}
 
-	return (__db_retcopy(F_ISSET(dbt,
-	    DB_DBT_INTERNAL) ? NULL : dbp, dbt, data, len, memp, memsize));
+	return (__db_retcopy(dbp->dbenv, dbt, data, len, memp, memsize));
 }
 
 /*
  * __db_retcopy --
  *	Copy the returned data into the user's DBT, handling special flags.
  *
- * PUBLIC: int __db_retcopy __P((DB *, DBT *,
+ * PUBLIC: int __db_retcopy __P((DB_ENV *, DBT *,
  * PUBLIC:    void *, u_int32_t, void **, u_int32_t *));
  */
 int
-__db_retcopy(dbp, dbt, data, len, memp, memsize)
-	DB *dbp;
+__db_retcopy(dbenv, dbt, data, len, memp, memsize)
+	DB_ENV *dbenv;
 	DBT *dbt;
 	void *data;
 	u_int32_t len;
@@ -95,6 +85,8 @@ __db_retcopy(dbp, dbt, data, len, memp, memsize)
 {
 	int ret;
 
+	ret = 0;
+
 	/* If returning a partial record, reset the length. */
 	if (F_ISSET(dbt, DB_DBT_PARTIAL)) {
 		data = (u_int8_t *)data + dbt->doff;
@@ -106,14 +98,6 @@ __db_retcopy(dbp, dbt, data, len, memp, memsize)
 			len = 0;
 	}
 
-	/*
-	 * Return the length of the returned record in the DBT size field.
-	 * This satisfies the requirement that if we're using user memory
-	 * and insufficient memory was provided, return the amount necessary
-	 * in the size field.
-	 */
-	dbt->size = len;
-
 	/*
 	 * Allocate memory to be owned by the application: DB_DBT_MALLOC,
 	 * DB_DBT_REALLOC.
@@ -129,31 +113,42 @@ __db_retcopy(dbp, dbt, data, len, memp, memsize)
 	 * If the length we're going to copy is 0, the application-supplied
 	 * memory pointer is allowed to be NULL.
 	 */
-	if (F_ISSET(dbt, DB_DBT_MALLOC)) {
-		if ((ret = __os_malloc(len,
-		    dbp == NULL ? NULL : dbp->db_malloc, &dbt->data)) != 0)
-			return (ret);
-	} else if (F_ISSET(dbt, DB_DBT_REALLOC)) {
-		if ((ret = __os_realloc(len,
-		    dbp == NULL ? NULL : dbp->db_realloc, &dbt->data)) != 0)
-			return (ret);
+	if (F_ISSET(dbt, DB_DBT_USERCOPY)) {
+		dbt->size = len;
+		return (len == 0 ? 0 : dbenv->dbt_usercopy(dbt, 0, data,
+		    len, DB_USERCOPY_SETDATA));
+
+	} else if (F_ISSET(dbt, DB_DBT_MALLOC))
+		ret = __os_umalloc(dbenv, len, &dbt->data);
+	else if (F_ISSET(dbt, DB_DBT_REALLOC)) {
+		if (dbt->data == NULL || dbt->size == 0 || dbt->size < len)
+			ret = __os_urealloc(dbenv, len, &dbt->data);
 	} else if (F_ISSET(dbt, DB_DBT_USERMEM)) {
 		if (len != 0 && (dbt->data == NULL || dbt->ulen < len))
-			return (ENOMEM);
-	} else if (memp == NULL || memsize == NULL) {
-		return (EINVAL);
-	} else {
+			ret = DB_BUFFER_SMALL;
+	} else if (memp == NULL || memsize == NULL)
+		ret = EINVAL;
+	else {
 		if (len != 0 && (*memsize == 0 || *memsize < len)) {
-			if ((ret = __os_realloc(len, NULL, memp)) != 0) {
+			if ((ret = __os_realloc(dbenv, len, memp)) == 0)
+				*memsize = len;
+			else
 				*memsize = 0;
-				return (ret);
-			}
-			*memsize = len;
 		}
-		dbt->data = *memp;
+		if (ret == 0)
+			dbt->data = *memp;
 	}
 
-	if (len != 0)
+	if (ret == 0 && len != 0)
 		memcpy(dbt->data, data, len);
-	return (0);
+
+	/*
+	 * Return the length of the returned record in the DBT size field.
+	 * This satisfies the requirement that if we're using user memory
+	 * and insufficient memory was provided, return the amount necessary
+	 * in the size field.
+	 */
+	dbt->size = len;
+
+	return (ret);
 }
