@@ -1,15 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1998-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1998-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- *	@(#)db_join.h	11.1 (Sleepycat) 7/25/99
+ * $Id: db_join.h,v 12.5 2006/08/24 14:45:29 bostic Exp $
  */
 
 #ifndef _DB_JOIN_H_
 #define	_DB_JOIN_H_
 
+#if defined(__cplusplus)
+extern "C" {
+#endif
+
 /*
  * Joins use a join cursor that is similar to a regular DB cursor except
  * that it only supports c_get and c_close functionality.  Also, it does
@@ -28,4 +32,7 @@ typedef struct __join_cursor {
 	u_int32_t flags;
 } JOIN_CURSOR;
 
+#if defined(__cplusplus)
+}
+#endif
 #endif /* !_DB_JOIN_H_ */
