@@ -1,15 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: hmac.h,v 1.5 2004/01/28 03:36:02 bostic Exp $
+ * $Id: hmac.h,v 12.4 2006/08/24 14:45:29 bostic Exp $
  */
 
 #ifndef	_DB_HMAC_H_
 #define	_DB_HMAC_H_
 
+#if defined(__cplusplus)
+extern "C" {
+#endif
+
 /*
  * Algorithm specific information.
  */
@@ -28,5 +32,9 @@ typedef struct {
 #define	DB_MAC_MAGIC	"mac derivation key magic value"
 #define	DB_ENC_MAGIC	"encryption and decryption key value magic"
 
+#if defined(__cplusplus)
+}
+#endif
+
 #include "dbinc_auto/hmac_ext.h"
 #endif /* !_DB_HMAC_H_ */
