@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996, 1997, 1998, 1999, 2000
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1990, 1993
@@ -38,29 +38,46 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: dbm.c,v 12.8 2006/08/24 14:45:31 bostic Exp $
  */
 
+#define	DB_DBM_HSEARCH	1
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: dbm.c,v 11.6 2000/05/17 19:21:00 bostic Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <errno.h>
-#include <fcntl.h>
-#include <string.h>
-#endif
-
-#define	DB_DBM_HSEARCH	1
 #include "db_int.h"
 
 /*
  *
  * This package provides dbm and ndbm compatible interfaces to DB.
  *
+ * EXTERN: #if DB_DBM_HSEARCH != 0
+ *
+ * EXTERN: int	 __db_ndbm_clearerr __P((DBM *));
+ * EXTERN: void	 __db_ndbm_close __P((DBM *));
+ * EXTERN: int	 __db_ndbm_delete __P((DBM *, datum));
+ * EXTERN: int	 __db_ndbm_dirfno __P((DBM *));
+ * EXTERN: int	 __db_ndbm_error __P((DBM *));
+ * EXTERN: datum __db_ndbm_fetch __P((DBM *, datum));
+ * EXTERN: datum __db_ndbm_firstkey __P((DBM *));
+ * EXTERN: datum __db_ndbm_nextkey __P((DBM *));
+ * EXTERN: DBM	*__db_ndbm_open __P((const char *, int, int));
+ * EXTERN: int	 __db_ndbm_pagfno __P((DBM *));
+ * EXTERN: int	 __db_ndbm_rdonly __P((DBM *));
+ * EXTERN: int	 __db_ndbm_store __P((DBM *, datum, datum, int));
+ *
+ * EXTERN: int	 __db_dbm_close __P((void));
+ * EXTERN: int	 __db_dbm_delete __P((datum));
+ * EXTERN: datum __db_dbm_fetch __P((datum));
+ * EXTERN: datum __db_dbm_firstkey __P((void));
+ * EXTERN: int	 __db_dbm_init __P((char *));
+ * EXTERN: datum __db_dbm_nextkey __P((datum));
+ * EXTERN: int	 __db_dbm_store __P((datum, datum));
+ *
+ * EXTERN: #endif
+ */
+
+/*
  * The DBM routines, which call the NDBM routines.
  */
 static DBM *__cur_db;
@@ -72,9 +89,9 @@ __db_dbm_init(file)
 	char *file;
 {
 	if (__cur_db != NULL)
-		(void)dbm_close(__cur_db);
+		dbm_close(__cur_db);
 	if ((__cur_db =
-	    dbm_open(file, O_CREAT | O_RDWR, __db_omode("rw----"))) != NULL)
+	    dbm_open(file, O_CREAT | O_RDWR, __db_omode(OWNER_RW))) != NULL)
 		return (0);
 	if ((__cur_db = dbm_open(file, O_RDONLY, 0)) != NULL)
 		return (0);
@@ -99,7 +116,8 @@ __db_dbm_fetch(key)
 
 	if (__cur_db == NULL) {
 		__db_no_open();
-		item.dptr = 0;
+		item.dptr = NULL;
+		item.dsize = 0;
 		return (item);
 	}
 	return (dbm_fetch(__cur_db, key));
@@ -112,7 +130,8 @@ __db_dbm_firstkey()
 
 	if (__cur_db == NULL) {
 		__db_no_open();
-		item.dptr = 0;
+		item.dptr = NULL;
+		item.dsize = 0;
 		return (item);
 	}
 	return (dbm_firstkey(__cur_db));
@@ -128,7 +147,8 @@ __db_dbm_nextkey(key)
 
 	if (__cur_db == NULL) {
 		__db_no_open();
-		item.dptr = 0;
+		item.dptr = NULL;
+		item.dsize = 0;
 		return (item);
 	}
 	return (dbm_nextkey(__cur_db));
@@ -180,7 +200,7 @@ __db_ndbm_open(file, oflags, mode)
 	DB *dbp;
 	DBC *dbc;
 	int ret;
-	char path[MAXPATHLEN];
+	char path[DB_MAXPATHLEN];
 
 	/*
 	 * !!!
@@ -211,7 +231,7 @@ __db_ndbm_open(file, oflags, mode)
 	if ((ret = dbp->set_pagesize(dbp, 4096)) != 0 ||
 	    (ret = dbp->set_h_ffactor(dbp, 40)) != 0 ||
 	    (ret = dbp->set_h_nelem(dbp, 1)) != 0 ||
-	    (ret = dbp->open(dbp,
+	    (ret = dbp->open(dbp, NULL,
 	    path, NULL, DB_HASH, __db_oflags(oflags), mode)) != 0) {
 		__os_set_errno(ret);
 		return (NULL);
@@ -258,10 +278,8 @@ __db_ndbm_fetch(dbm, key)
 
 	dbc = (DBC *)dbm;
 
-	memset(&_key, 0, sizeof(DBT));
+	DB_INIT_DBT(_key, key.dptr, key.dsize);
 	memset(&_data, 0, sizeof(DBT));
-	_key.size = key.dsize;
-	_key.data = key.dptr;
 
 	/*
 	 * Note that we can't simply use the dbc we have to do a c_get/SET,
@@ -270,7 +288,7 @@ __db_ndbm_fetch(dbm, key)
 	 */
 	if ((ret = dbc->dbp->get(dbc->dbp, NULL, &_key, &_data, 0)) == 0) {
 		data.dptr = _data.data;
-		data.dsize = _data.size;
+		data.dsize = (int)_data.size;
 	} else {
 		data.dptr = NULL;
 		data.dsize = 0;
@@ -278,7 +296,7 @@ __db_ndbm_fetch(dbm, key)
 			__os_set_errno(ENOENT);
 		else {
 			__os_set_errno(ret);
-			F_SET(dbc->dbp, DB_DBM_ERROR);
+			F_SET(dbc->dbp, DB_AM_DBM_ERROR);
 		}
 	}
 	return (data);
@@ -305,7 +323,7 @@ __db_ndbm_firstkey(dbm)
 
 	if ((ret = dbc->c_get(dbc, &_key, &_data, DB_FIRST)) == 0) {
 		key.dptr = _key.data;
-		key.dsize = _key.size;
+		key.dsize = (int)_key.size;
 	} else {
 		key.dptr = NULL;
 		key.dsize = 0;
@@ -313,7 +331,7 @@ __db_ndbm_firstkey(dbm)
 			__os_set_errno(ENOENT);
 		else {
 			__os_set_errno(ret);
-			F_SET(dbc->dbp, DB_DBM_ERROR);
+			F_SET(dbc->dbp, DB_AM_DBM_ERROR);
 		}
 	}
 	return (key);
@@ -340,7 +358,7 @@ __db_ndbm_nextkey(dbm)
 
 	if ((ret = dbc->c_get(dbc, &_key, &_data, DB_NEXT)) == 0) {
 		key.dptr = _key.data;
-		key.dsize = _key.size;
+		key.dsize = (int)_key.size;
 	} else {
 		key.dptr = NULL;
 		key.dsize = 0;
@@ -348,7 +366,7 @@ __db_ndbm_nextkey(dbm)
 			__os_set_errno(ENOENT);
 		else {
 			__os_set_errno(ret);
-			F_SET(dbc->dbp, DB_DBM_ERROR);
+			F_SET(dbc->dbp, DB_AM_DBM_ERROR);
 		}
 	}
 	return (key);
@@ -370,9 +388,7 @@ __db_ndbm_delete(dbm, key)
 
 	dbc = (DBC *)dbm;
 
-	memset(&_key, 0, sizeof(DBT));
-	_key.data = key.dptr;
-	_key.size = key.dsize;
+	DB_INIT_DBT(_key, key.dptr, key.dsize);
 
 	if ((ret = dbc->dbp->del(dbc->dbp, NULL, &_key, 0)) == 0)
 		return (0);
@@ -381,7 +397,7 @@ __db_ndbm_delete(dbm, key)
 		__os_set_errno(ENOENT);
 	else {
 		__os_set_errno(ret);
-		F_SET(dbc->dbp, DB_DBM_ERROR);
+		F_SET(dbc->dbp, DB_AM_DBM_ERROR);
 	}
 	return (-1);
 }
@@ -404,13 +420,8 @@ __db_ndbm_store(dbm, key, data, flags)
 
 	dbc = (DBC *)dbm;
 
-	memset(&_key, 0, sizeof(DBT));
-	_key.data = key.dptr;
-	_key.size = key.dsize;
-
-	memset(&_data, 0, sizeof(DBT));
-	_data.data = data.dptr;
-	_data.size = data.dsize;
+	DB_INIT_DBT(_key, key.dptr, key.dsize);
+	DB_INIT_DBT(_data, data.dptr, data.dsize);
 
 	if ((ret = dbc->dbp->put(dbc->dbp, NULL,
 	    &_key, &_data, flags == DBM_INSERT ? DB_NOOVERWRITE : 0)) == 0)
@@ -420,7 +431,7 @@ __db_ndbm_store(dbm, key, data, flags)
 		return (1);
 
 	__os_set_errno(ret);
-	F_SET(dbc->dbp, DB_DBM_ERROR);
+	F_SET(dbc->dbp, DB_AM_DBM_ERROR);
 	return (-1);
 }
 
@@ -432,7 +443,7 @@ __db_ndbm_error(dbm)
 
 	dbc = (DBC *)dbm;
 
-	return (F_ISSET(dbc->dbp, DB_DBM_ERROR));
+	return (F_ISSET(dbc->dbp, DB_AM_DBM_ERROR));
 }
 
 int
@@ -443,7 +454,7 @@ __db_ndbm_clearerr(dbm)
 
 	dbc = (DBC *)dbm;
 
-	F_CLR(dbc->dbp, DB_DBM_ERROR);
+	F_CLR(dbc->dbp, DB_AM_DBM_ERROR);
 	return (0);
 }
 
