@@ -1,20 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: dbreg.c,v 12.12 2005/10/14 14:40:41 bostic Exp $
+ * $Id: dbreg.c,v 12.20 2006/08/24 14:45:31 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/db_page.h"
 #include "dbinc/log.h"
@@ -133,7 +127,7 @@ __dbreg_setup(dbp, name, create_txnid)
 	 * Fill in all the remaining info that we'll need later to register
 	 * the file, if we use it for logging.
 	 */
-	fnp->id = DB_LOGFILEID_INVALID;
+	fnp->id = fnp->old_id = DB_LOGFILEID_INVALID;
 	fnp->s_type = dbp->type;
 	memcpy(fnp->ufid, dbp->fileid, DB_FILE_ID_LEN);
 	fnp->meta_pgno = dbp->meta_pgno;
@@ -145,7 +139,7 @@ __dbreg_setup(dbp, name, create_txnid)
 
 err:	LOG_SYSTEM_UNLOCK(dbenv);
 	if (ret == ENOMEM)
-		__db_err(dbenv,
+		__db_errx(dbenv,
     "Logging region out of memory; you may need to increase its size");
 
 	return (ret);
@@ -178,7 +172,7 @@ __dbreg_teardown(dbp)
 	if (fnp == NULL || F_ISSET(fnp, DB_FNAME_NOTLOGGED))
 		return (0);
 
-	DB_ASSERT(fnp->id == DB_LOGFILEID_INVALID);
+	DB_ASSERT(dbenv, fnp->id == DB_LOGFILEID_INVALID);
 
 	LOG_SYSTEM_LOCK(dbenv);
 	if (fnp->name_off != INVALID_ROFF)
@@ -278,7 +272,7 @@ __dbreg_get_id(dbp, txn, idp)
 	 * Log the registry.  We should only request a new ID in situations
 	 * where logging is reasonable.
 	 */
-	DB_ASSERT(!F_ISSET(dbp, DB_AM_RECOVER));
+	DB_ASSERT(dbenv, !F_ISSET(dbp, DB_AM_RECOVER));
 
 	if ((ret = __dbreg_log_id(dbp, txn, id, 0)) != 0)
 		goto err;
@@ -290,8 +284,8 @@ __dbreg_get_id(dbp, txn, idp)
 	 */
 	fnp->create_txnid = TXN_INVALID;
 
-	DB_ASSERT(dbp->type == fnp->s_type);
-	DB_ASSERT(dbp->meta_pgno == fnp->meta_pgno);
+	DB_ASSERT(dbenv, dbp->type == fnp->s_type);
+	DB_ASSERT(dbenv, dbp->meta_pgno == fnp->meta_pgno);
 
 	if ((ret = __dbreg_add_dbentry(dbenv, dblp, dbp, id)) != 0)
 		goto err;
@@ -339,7 +333,7 @@ __dbreg_assign_id(dbp, id)
 	MUTEX_LOCK(dbenv, lp->mtx_filelist);
 
 	/* We should only call this on DB handles that have no ID. */
-	DB_ASSERT(fnp->id == DB_LOGFILEID_INVALID);
+	DB_ASSERT(dbenv, fnp->id == DB_LOGFILEID_INVALID);
 
 	/*
 	 * Make sure there isn't already a file open with this ID. There can
@@ -427,6 +421,7 @@ __dbreg_revoke_id(dbp, have_lock, force_id)
 	dblp = dbenv->lg_handle;
 	lp = dblp->reginfo.primary;
 	fnp = dbp->log_filename;
+	ret = 0;
 
 	/* If we lack an ID, this is a null-op. */
 	if (fnp == NULL)
@@ -439,14 +434,17 @@ __dbreg_revoke_id(dbp, have_lock, force_id)
 	 */
 	if (force_id != DB_LOGFILEID_INVALID)
 		id = force_id;
-	else if (fnp->id == DB_LOGFILEID_INVALID)
-		return (0);
-	else
+	else if (fnp->id == DB_LOGFILEID_INVALID) {
+		if (fnp->old_id == DB_LOGFILEID_INVALID)
+			return (0);
+		id = fnp->old_id;
+	} else
 		id = fnp->id;
 	if (!have_lock)
 		MUTEX_LOCK(dbenv, lp->mtx_filelist);
 
 	fnp->id = DB_LOGFILEID_INVALID;
+	fnp->old_id = DB_LOGFILEID_INVALID;
 
 	/* Remove the FNAME from the list of open files. */
 	SH_TAILQ_REMOVE(&lp->fq, fnp, q, __fname);
@@ -484,10 +482,8 @@ __dbreg_close_id(dbp, txn, op)
 	DB_TXN *txn;
 	u_int32_t op;
 {
-	DBT fid_dbt, r_name, *dbtp;
 	DB_ENV *dbenv;
 	DB_LOG *dblp;
-	DB_LSN r_unused;
 	FNAME *fnp;
 	LOG *lp;
 	int ret;
@@ -498,11 +494,49 @@ __dbreg_close_id(dbp, txn, op)
 	fnp = dbp->log_filename;
 
 	/* If we lack an ID, this is a null-op. */
-	if (fnp == NULL || fnp->id == DB_LOGFILEID_INVALID)
+	if (fnp == NULL)
 		return (0);
 
+	if (fnp->id == DB_LOGFILEID_INVALID)
+		return (__dbreg_revoke_id(dbp, 0, DB_LOGFILEID_INVALID));
+
 	MUTEX_LOCK(dbenv, lp->mtx_filelist);
 
+	if ((ret = __dbreg_log_close(dbenv, fnp, txn, op)) != 0)
+		goto err;
+	ret = __dbreg_revoke_id(dbp, 1, DB_LOGFILEID_INVALID);
+
+err:	MUTEX_UNLOCK(dbenv, lp->mtx_filelist);
+	return (ret);
+}
+
+/*
+ * __dbreg_log_close --
+ *
+ * Log a close of a database.  Called when closing a file or when a
+ * replication client is becoming a master.  That closes all the
+ * files it previously had open.
+ *
+ * Assumes caller holds the lp->mutex_filelist lock already.
+ *
+ * PUBLIC: int __dbreg_log_close __P((DB_ENV *, FNAME *,
+ * PUBLIC:    DB_TXN *, u_int32_t));
+ */
+int
+__dbreg_log_close(dbenv, fnp, txn, op)
+	DB_ENV *dbenv;
+	FNAME *fnp;
+	DB_TXN *txn;
+	u_int32_t op;
+{
+	DB_LOG *dblp;
+	DBT fid_dbt, r_name, *dbtp;
+	DB_LSN r_unused;
+	int ret;
+
+	dblp = dbenv->lg_handle;
+	ret = 0;
+
 	if (fnp->name_off == INVALID_ROFF)
 		dbtp = NULL;
 	else {
@@ -516,7 +550,7 @@ __dbreg_close_id(dbp, txn, op)
 	fid_dbt.data = fnp->ufid;
 	fid_dbt.size = DB_FILE_ID_LEN;
 	if ((ret = __dbreg_register_log(dbenv, txn, &r_unused,
-	    F_ISSET(dbp, DB_AM_NOT_DURABLE) ? DB_LOG_NOT_DURABLE : 0,
+	    F_ISSET(fnp, DB_FNAME_DURABLE) ? 0 : DB_LOG_NOT_DURABLE,
 	    op, dbtp, &fid_dbt, fnp->id,
 	    fnp->s_type, fnp->meta_pgno, TXN_INVALID)) != 0) {
 		/*
@@ -531,12 +565,7 @@ __dbreg_close_id(dbp, txn, op)
 		 */
 		F_SET(fnp, DB_FNAME_NOTLOGGED);
 		(void)__dbreg_rem_dbentry(dblp, fnp->id);
-		goto err;
 	}
-
-	ret = __dbreg_revoke_id(dbp, 1, DB_LOGFILEID_INVALID);
-
-err:	MUTEX_UNLOCK(dbenv, lp->mtx_filelist);
 	return (ret);
 }
 
@@ -572,8 +601,7 @@ __dbreg_push_id(dbenv, dbp, id)
 	 * If our fid generation in replication has changed, this fid should
 	 * not be pushed back onto the stack.
 	 */
-	if (REP_ON(dbenv) && db_rep->region != NULL &&
-	   ((REP *)db_rep->region)->gen != dbp->fid_gen)
+	if (REP_ON(dbenv) && ((REP *)db_rep->region)->gen != dbp->fid_gen)
 		return (0);
 	/* Check if we have room on the stack. */
 	if (lp->free_fid_stack == INVALID_ROFF ||
