@@ -1,41 +1,25 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2005
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: env_open.c,v 12.36 2005/10/31 02:22:28 bostic Exp $
+ * $Id: env_open.c,v 12.71 2006/08/24 14:45:39 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <ctype.h>
-#include <limits.h>
-#include <stdlib.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/crypto.h"
 #include "dbinc/db_page.h"
-#include "dbinc/db_shash.h"
 #include "dbinc/btree.h"
-#include "dbinc/hash.h"
-#include "dbinc/fop.h"
 #include "dbinc/lock.h"
 #include "dbinc/log.h"
 #include "dbinc/mp.h"
-#include "dbinc/qam.h"
 #include "dbinc/txn.h"
 
-static int __db_parse __P((DB_ENV *, char *));
 static int __db_tmp_open __P((DB_ENV *, u_int32_t, char *, DB_FH **));
-static int __env_config __P((DB_ENV *, const char *, u_int32_t));
 static int __env_refresh __P((DB_ENV *, u_int32_t, int));
-static int __env_remove_int __P((DB_ENV *, const char *, u_int32_t));
 
 /*
  * db_version --
@@ -69,11 +53,7 @@ __env_open_pp(dbenv, db_home, flags, mode)
 	u_int32_t flags;
 	int mode;
 {
-	DB_THREAD_INFO *ip;
-	u_int32_t orig_flags;
-	int need_recovery, ret, t_ret;
-
-	need_recovery = 0;
+	int ret;
 
 #undef	OKFLAGS
 #define	OKFLAGS								\
@@ -92,16 +72,34 @@ __env_open_pp(dbenv, db_home, flags, mode)
 	if ((ret = __db_fcchk(
 	    dbenv, "DB_ENV->open", flags, DB_INIT_CDB, ~OKFLAGS_CDB)) != 0)
 		return (ret);
-	if ((ret = __db_fcchk(dbenv, "DB_ENV->open", flags,
-	    DB_PRIVATE, DB_REGISTER | DB_SYSTEM_MEM)) != 0)
-		return (ret);
+	if (LF_ISSET(DB_REGISTER)) {
+		if (!__os_support_db_register()) {
+			__db_errx(dbenv,
+	     "Berkeley DB library does not support DB_REGISTER on this system");
+			return (EINVAL);
+		}
+		if ((ret = __db_fcchk(dbenv, "DB_ENV->open", flags,
+		    DB_PRIVATE, DB_REGISTER | DB_SYSTEM_MEM)) != 0)
+			return (ret);
+		if (!LF_ISSET(DB_INIT_TXN)) {
+			__db_errx(
+			    dbenv, "registration requires transaction support");
+			return (EINVAL);
+		}
+	}
 	if (LF_ISSET(DB_INIT_REP)) {
+		if (!__os_support_replication()) {
+			__db_errx(dbenv,
+	     "Berkeley DB library does not support replication on this system");
+			return (EINVAL);
+		}
 		if (!LF_ISSET(DB_INIT_LOCK)) {
-			__db_err(dbenv, "replication requires locking support");
+			__db_errx(dbenv,
+			    "replication requires locking support");
 			return (EINVAL);
 		}
 		if (!LF_ISSET(DB_INIT_TXN)) {
-			__db_err(
+			__db_errx(
 			    dbenv, "replication requires transaction support");
 			return (EINVAL);
 		}
@@ -110,31 +108,34 @@ __env_open_pp(dbenv, db_home, flags, mode)
 		if ((ret = __db_fcchk(dbenv,
 		    "DB_ENV->open", flags, DB_RECOVER, DB_RECOVER_FATAL)) != 0)
 			return (ret);
+		if ((ret = __db_fcchk(dbenv,
+		    "DB_ENV->open", flags, DB_REGISTER, DB_RECOVER_FATAL)) != 0)
+			return (ret);
 		if (!LF_ISSET(DB_CREATE)) {
-			__db_err(dbenv, "recovery requires the create flag");
+			__db_errx(dbenv, "recovery requires the create flag");
 			return (EINVAL);
 		}
 		if (!LF_ISSET(DB_INIT_TXN)) {
-			__db_err(
+			__db_errx(
 			    dbenv, "recovery requires transaction support");
 			return (EINVAL);
 		}
 	}
 
+#ifdef HAVE_MUTEX_THREAD_ONLY
 	/*
 	 * Currently we support one kind of mutex that is intra-process only,
 	 * POSIX 1003.1 pthreads, because a variety of systems don't support
 	 * the full pthreads API, and our only alternative is test-and-set.
 	 */
-#ifdef HAVE_MUTEX_THREAD_ONLY
 	if (!LF_ISSET(DB_PRIVATE)) {
-		__db_err(dbenv,
+		__db_errx(dbenv,
 	 "Berkeley DB library configured to support only private environments");
 		return (EINVAL);
 	}
 #endif
 
-#if defined(HAVE_MUTEX_FCNTL)
+#ifdef HAVE_MUTEX_FCNTL
 	/*
 	 * !!!
 	 * We need a file descriptor for fcntl(2) locking.  We use the file
@@ -153,81 +154,13 @@ __env_open_pp(dbenv, db_home, flags, mode)
 	 * process, including all its threads.
 	 */
 	if (F_ISSET(dbenv, DB_ENV_THREAD)) {
-		__db_err(dbenv,
+		__db_errx(dbenv,
 	    "architecture lacks fast mutexes: applications cannot be threaded");
 		return (EINVAL);
 	}
 #endif
 
-	if (LF_ISSET(DB_INIT_REP) && !__os_support_replication()) {
-		__db_err(dbenv,
-	     "Berkeley DB library does not support replication on this system");
-		return (EINVAL);
-	}
-
-	/*
-	 * If we're going to register with the environment, that's the first
-	 * thing we do.
-	 */
-	if (LF_ISSET(DB_REGISTER)) {
-		if (!__os_support_db_register()) {
-			__db_err(dbenv,
-	     "Berkeley DB library does not support DB_REGISTER on this system");
-			return (EINVAL);
-		}
-
-		if ((ret =
-		    __envreg_register(dbenv, db_home, &need_recovery)) != 0)
-			return (ret);
-		if (need_recovery) {
-			if (!LF_ISSET(DB_RECOVER)) {
-				__db_err(dbenv,
-		    "No recovery flag was specified, and recovery is needed");
-				ret = DB_RUNRECOVERY;
-				goto err;
-			}
-		} else
-			LF_CLR(DB_RECOVER | DB_RECOVER_FATAL);
-	}
-
-	/*
-	 * If we're doing recovery, destroy the environment so that we create
-	 * all the regions from scratch.  The major concern I have is if the
-	 * application stomps the environment with a rogue pointer.  We have
-	 * no way of detecting that, and we could be forced into a situation
-	 * where we start up and then crash, repeatedly.
-	 *
-	 * Note that we do not check any flags like DB_PRIVATE before calling
-	 * remove.  We don't care if the current environment was private or
-	 * not, we just want to nail any files that are left-over for whatever
-	 * reason, from whatever session.
-	 */
-	if (LF_ISSET(DB_RECOVER | DB_RECOVER_FATAL)) {
-		orig_flags = dbenv->flags;
-		if ((ret = __env_remove_int(dbenv, db_home, DB_FORCE)) != 0 ||
-		    (ret = __env_refresh(dbenv, orig_flags, 0)) != 0)
-			goto err;
-	}
-
-	ret = __env_open(dbenv, db_home, flags, mode);
-	if (ret == 0 && dbenv->thr_hashtab != NULL &&
-	    (t_ret = __env_set_state(dbenv, &ip, THREAD_OUT)) != 0)
-		ret = t_ret;
-
-err:	if (need_recovery) {
-		/*
-		 * If recovery succeeded, release our exclusive lock, other
-		 * processes can now proceed.
-		 *
-		 * If recovery failed, unregister now.
-		 */
-		if (ret == 0 && (t_ret = __envreg_xunlock(dbenv)) != 0)
-			ret = t_ret;
-		if (ret != 0)
-			(void)__envreg_unregister(dbenv, 1);
-	}
-
-	return (ret);
+	return (__env_open(dbenv, db_home, flags, mode));
 }
 
 /*
@@ -246,15 +179,57 @@ __env_open(dbenv, db_home, flags, mode)
 	DB_THREAD_INFO *ip;
 	REGINFO *infop;
 	u_int32_t init_flags, orig_flags;
-	int rep_check, ret;
+	int register_recovery, rep_check, ret, t_ret;
 
-	orig_flags = dbenv->flags;
-	rep_check = 0;
 	ip = NULL;
+	register_recovery = rep_check = 0;
 
-	/* Initialize the DB_ENV structure. */
-	if ((ret = __env_config(dbenv, db_home, flags)) != 0)
-		goto err;
+	/* Initial configuration. */
+	if ((ret = __env_config(dbenv, db_home, flags, mode)) != 0)
+		return (ret);
+
+	/*
+	 * Save the DB_ENV handle's configuration flags as set by user-called
+	 * configuration methods and the environment directory's DB_CONFIG
+	 * file.  If we use this DB_ENV structure to recover the existing
+	 * environment or to remove an environment we created after failure,
+	 * we'll restore the DB_ENV flags to these values.
+	 */
+	orig_flags = dbenv->flags;
+
+	/*
+	 * If we're going to register with the environment, that's the first
+	 * thing we do.
+	 */
+	if (LF_ISSET(DB_REGISTER)) {
+		if ((ret = __envreg_register(dbenv, &register_recovery)) != 0)
+			goto err;
+		if (register_recovery) {
+			if (!LF_ISSET(DB_RECOVER)) {
+				__db_errx(dbenv,
+	    "The DB_RECOVER flag was not specified, and recovery is needed");
+				ret = DB_RUNRECOVERY;
+				goto err;
+			}
+		} else
+			LF_CLR(DB_RECOVER);
+	}
+
+	/*
+	 * If we're doing recovery, destroy the environment so that we create
+	 * all the regions from scratch.  The major concern I have is if the
+	 * application stomps the environment with a rogue pointer.  We have
+	 * no way of detecting that, and we could be forced into a situation
+	 * where we start up and then crash, repeatedly.
+	 *
+	 * We do not check any flags like DB_PRIVATE before calling remove.
+	 * We don't care if the current environment was private or not, we
+	 * want to remove files left over for any reason, from any session.
+	 */
+	if (LF_ISSET(DB_RECOVER | DB_RECOVER_FATAL))
+		if ((ret = __db_e_remove(dbenv, DB_FORCE)) != 0 ||
+		    (ret = __env_refresh(dbenv, orig_flags, 0)) != 0)
+			goto err;
 
 	/* Convert the DB_ENV->open flags to internal flags. */
 	if (LF_ISSET(DB_CREATE))
@@ -270,9 +245,6 @@ __env_open(dbenv, db_home, flags, mode)
 	if (LF_ISSET(DB_THREAD))
 		F_SET(dbenv, DB_ENV_THREAD);
 
-	/* Default permissions are read-write for both owner and group. */
-	dbenv->db_mode = mode == 0 ? __db_omode("rw-rw----") : mode;
-
 	/*
 	 * Flags saved in the init_flags field of the environment, representing
 	 * flags to DB_ENV->set_flags and DB_ENV->open that need to be set.
@@ -294,6 +266,8 @@ __env_open(dbenv, db_home, flags, mode)
 	init_flags = 0;
 	if (LF_ISSET(DB_INIT_CDB))
 		FLD_SET(init_flags, DB_INITENV_CDB);
+	if (F_ISSET(dbenv, DB_ENV_CDB_ALLDB))
+		FLD_SET(init_flags, DB_INITENV_CDB_ALLDB);
 	if (LF_ISSET(DB_INIT_LOCK))
 		FLD_SET(init_flags, DB_INITENV_LOCK);
 	if (LF_ISSET(DB_INIT_LOG))
@@ -304,8 +278,6 @@ __env_open(dbenv, db_home, flags, mode)
 		FLD_SET(init_flags, DB_INITENV_REP);
 	if (LF_ISSET(DB_INIT_TXN))
 		FLD_SET(init_flags, DB_INITENV_TXN);
-	if (F_ISSET(dbenv, DB_ENV_CDB_ALLDB))
-		FLD_SET(init_flags, DB_INITENV_CDB_ALLDB);
 	if ((ret = __db_e_attach(dbenv, &init_flags)) != 0)
 		goto err;
 
@@ -333,7 +305,7 @@ __env_open(dbenv, db_home, flags, mode)
 		goto err;
 
 	/*
-	 * Save the flags matching the database environment: we've replaced
+	 * Save the flags matching the database environment: we'll replace
 	 * the argument flags with the flags corresponding to the existing,
 	 * underlying set of subsystems.
 	 */
@@ -346,9 +318,9 @@ __env_open(dbenv, db_home, flags, mode)
 	}
 
 	/*
-	 * The DB_ENV structure has been initialized.  This has to be set
-	 * before we start calling into the subsystems, some of them look
-	 * for it.
+	 * The DB_ENV structure has now been initialized.  Turn off further
+	 * use of the DB_ENV structure and most initialization methods, we're
+	 * about to act on the values we currently have.
 	 */
 	F_SET(dbenv, DB_ENV_OPEN_CALLED);
 
@@ -413,29 +385,7 @@ __env_open(dbenv, db_home, flags, mode)
 		 * If the application is running with transactions, initialize
 		 * the function tables.
 		 */
-		if ((ret = __bam_init_recover(dbenv, &dbenv->recover_dtab,
-		    &dbenv->recover_dtab_size)) != 0)
-			goto err;
-		if ((ret = __crdel_init_recover(dbenv, &dbenv->recover_dtab,
-		    &dbenv->recover_dtab_size)) != 0)
-			goto err;
-		if ((ret = __db_init_recover(dbenv, &dbenv->recover_dtab,
-		    &dbenv->recover_dtab_size)) != 0)
-			goto err;
-		if ((ret = __dbreg_init_recover(dbenv, &dbenv->recover_dtab,
-		    &dbenv->recover_dtab_size)) != 0)
-			goto err;
-		if ((ret = __fop_init_recover(dbenv, &dbenv->recover_dtab,
-		    &dbenv->recover_dtab_size)) != 0)
-			goto err;
-		if ((ret = __ham_init_recover(dbenv, &dbenv->recover_dtab,
-		    &dbenv->recover_dtab_size)) != 0)
-			goto err;
-		if ((ret = __qam_init_recover(dbenv, &dbenv->recover_dtab,
-		    &dbenv->recover_dtab_size)) != 0)
-			goto err;
-		if ((ret = __txn_init_recover(dbenv, &dbenv->recover_dtab,
-		    &dbenv->recover_dtab_size)) != 0)
+		if ((ret = __env_init_rec(dbenv, DB_LOGVERSION)) != 0)
 			goto err;
 	}
 
@@ -454,13 +404,13 @@ __env_open(dbenv, db_home, flags, mode)
 	 * region for environments and db handles.  So, the mpool region must
 	 * already be initialized.
 	 */
-	LIST_INIT(&dbenv->dblist);
+	TAILQ_INIT(&dbenv->dblist);
 	if (LF_ISSET(DB_INIT_MPOOL)) {
 		if ((ret = __mutex_alloc(dbenv, MTX_ENV_DBLIST,
-		    DB_MUTEX_THREAD, &dbenv->mtx_dblist)) != 0)
+		    DB_MUTEX_PROCESS_ONLY, &dbenv->mtx_dblist)) != 0)
 			goto err;
-		if ((ret = __mutex_alloc(dbenv,
-		    MTX_TWISTER, DB_MUTEX_THREAD, &dbenv->mtx_mt)) != 0)
+		if ((ret = __mutex_alloc(dbenv, MTX_TWISTER,
+		    DB_MUTEX_PROCESS_ONLY, &dbenv->mtx_mt)) != 0)
 			goto err;
 
 		/* Register DB's pgin/pgout functions.  */
@@ -497,32 +447,45 @@ __env_open(dbenv, db_home, flags, mode)
 	if ((ret = __db_e_golive(dbenv)) != 0)
 		goto err;
 
-	if (rep_check && (ret = __env_db_rep_exit(dbenv)) != 0)
-		goto err;
+	if (rep_check)
+		ret = __env_db_rep_exit(dbenv);
 
-	ENV_LEAVE(dbenv, ip);
-	return (0);
+err:	ENV_LEAVE(dbenv, ip);
 
-err:	/*
-	 * If we fail after creating the regions, panic and remove them.
-	 *
-	 * !!!
-	 * No need to call __env_db_rep_exit, that work is done by the calls to
-	 * __env_refresh.
-	 */
-	infop = dbenv->reginfo;
-	if (infop != NULL && F_ISSET(infop, REGION_CREATE)) {
-		ret = __db_panic(dbenv, ret);
+	if (ret != 0) {
+		/*
+		 * If we fail after creating the regions, panic and remove them.
+		 *
+		 * !!!
+		 * No need to call __env_db_rep_exit, that work is done by the
+		 * calls to __env_refresh.
+		 */
+		infop = dbenv->reginfo;
+		if (infop != NULL && F_ISSET(infop, REGION_CREATE)) {
+			ret = __db_panic(dbenv, ret);
 
-		/* Refresh the DB_ENV so we can use it to call remove. */
-		(void)__env_refresh(dbenv, orig_flags, rep_check);
-		(void)__env_remove_int(dbenv, db_home, DB_FORCE);
-		(void)__env_refresh(dbenv, orig_flags, 0);
-	} else
-		(void)__env_refresh(dbenv, orig_flags, rep_check);
+			/* Refresh the DB_ENV so can use it to call remove. */
+			(void)__env_refresh(dbenv, orig_flags, rep_check);
+			(void)__db_e_remove(dbenv, DB_FORCE);
+			(void)__env_refresh(dbenv, orig_flags, 0);
+		} else
+			(void)__env_refresh(dbenv, orig_flags, rep_check);
+	}
+
+	if (register_recovery) {
+		/*
+		 * If recovery succeeded, release our exclusive lock, other
+		 * processes can now proceed.
+		 *
+		 * If recovery failed, unregister now and let another process
+		 * clean up.
+		 */
+		if (ret == 0 && (t_ret = __envreg_xunlock(dbenv)) != 0)
+			ret = t_ret;
+		if (ret != 0)
+			(void)__envreg_unregister(dbenv, 1);
+	}
 
-	if (ip != NULL)
-		ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -550,7 +513,10 @@ __env_remove(dbenv, db_home, flags)
 
 	ENV_ILLEGAL_AFTER_OPEN(dbenv, "DB_ENV->remove");
 
-	ret = __env_remove_int(dbenv, db_home, flags);
+	if ((ret = __env_config(dbenv, db_home, flags, 0)) != 0)
+		return (ret);
+
+	ret = __db_e_remove(dbenv, flags);
 
 	if ((t_ret = __env_close(dbenv, 0)) != 0 && ret == 0)
 		ret = t_ret;
@@ -558,82 +524,51 @@ __env_remove(dbenv, db_home, flags)
 	return (ret);
 }
 
-/*
- * __env_remove_int --
- *	Discard an environment, internal version.
- */
-static int
-__env_remove_int(dbenv, db_home, flags)
-	DB_ENV *dbenv;
-	const char *db_home;
-	u_int32_t flags;
-{
-	int ret;
-
-	/* Initialize the DB_ENV structure. */
-	if ((ret = __env_config(dbenv, db_home, flags)) != 0)
-		return (ret);
-
-	/* The DB_ENV structure has been initialized. */
-	F_SET(dbenv, DB_ENV_OPEN_CALLED);
-
-	/* Remove the environment. */
-	return (__db_e_remove(dbenv, flags));
-}
-
 /*
  * __env_config --
- *	Initialization of the DB_ENV structure, read the DB_CONFIG file.
+ *	Argument-based initialization.
+ *
+ * PUBLIC: int __env_config __P((DB_ENV *, const char *, u_int32_t, int));
  */
-static int
-__env_config(dbenv, db_home, flags)
+int
+__env_config(dbenv, db_home, flags, mode)
 	DB_ENV *dbenv;
 	const char *db_home;
 	u_int32_t flags;
+	int mode;
 {
-	FILE *fp;
 	int ret;
-	char *p, buf[256];
+	char *home, home_buf[DB_MAXPATHLEN];
 
 	/*
-	 * Set the database home.  Do this before calling __db_appname,
-	 * it uses the home directory.
+	 * Set the database home.
+	 *
+	 * Use db_home by default, this allows utilities to reasonably
+	 * override the environment either explicitly or by using a -h
+	 * option.  Otherwise, use the environment if it's permitted
+	 * and initialized.
 	 */
-	if ((ret = __db_home(dbenv, db_home, flags)) != 0)
+	home = (char *)db_home;
+	if (home == NULL && (LF_ISSET(DB_USE_ENVIRON) ||
+	    (LF_ISSET(DB_USE_ENVIRON_ROOT) && __os_isroot()))) {
+		home = home_buf;
+		if ((ret = __os_getenv(
+		    dbenv, "DB_HOME", &home, sizeof(home_buf))) != 0)
+			return (ret);
+		/*
+		 * home set to NULL if __os_getenv failed to find DB_HOME.
+		 */
+	}
+	if (home != NULL &&
+	    (ret = __os_strdup(dbenv, home, &dbenv->db_home)) != 0)
 		return (ret);
 
-	/* Parse the config file. */
-	p = NULL;
-	if ((ret =
-	    __db_appname(dbenv, DB_APP_NONE, "DB_CONFIG", 0, NULL, &p)) != 0)
+	/* Default permissions are read-write for both owner and group. */
+	dbenv->db_mode = mode == 0 ? __db_omode("rw-rw----") : mode;
+
+	/* Read the DB_CONFIG file. */
+	if ((ret = __env_read_db_config(dbenv)) != 0)
 		return (ret);
-	if (p == NULL)
-		fp = NULL;
-	else {
-		fp = fopen(p, "r");
-		__os_free(dbenv, p);
-	}
-
-	if (fp != NULL) {
-		while (fgets(buf, sizeof(buf), fp) != NULL) {
-			if ((p = strchr(buf, '\n')) != NULL)
-				*p = '\0';
-			else if (strlen(buf) + 1 == sizeof(buf)) {
-				__db_err(dbenv, "DB_CONFIG: line too long");
-				(void)fclose(fp);
-				return (EINVAL);
-			}
-			if (buf[0] == '\0' ||
-			    buf[0] == '#' || isspace((int)buf[0]))
-				continue;
-
-			if ((ret = __db_parse(dbenv, buf)) != 0) {
-				(void)fclose(fp);
-				return (ret);
-			}
-		}
-		(void)fclose(fp);
-	}
 
 	/*
 	 * If no temporary directory path was specified in the config file,
@@ -673,8 +608,20 @@ __env_close_pp(dbenv, flags)
 		ret = t_ret;
 
 	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check && (t_ret = __env_rep_enter(dbenv, 0)) != 0 && ret == 0)
-		ret = t_ret;
+	if (rep_check) {
+#ifdef HAVE_REPLICATION_THREADS
+		/*
+		 * Shut down Replication Manager threads first of all.  This
+		 * must be done before __env_rep_enter to avoid a deadlock that
+		 * could occur if repmgr's background threads try to do a rep
+		 * operation that needs __rep_lockout.
+		 */
+		if ((t_ret = __repmgr_close(dbenv)) != 0 && ret == 0)
+			ret = t_ret;
+#endif
+		if ((t_ret = __env_rep_enter(dbenv, 0)) != 0 && ret == 0)
+			ret = t_ret;
+	}
 
 	if ((t_ret = __env_close(dbenv, rep_check)) != 0 && ret == 0)
 		ret = t_ret;
@@ -707,9 +654,10 @@ __env_close(dbenv, rep_check)
 	if (TXN_ON(dbenv) && (t_ret = __txn_preclose(dbenv)) != 0 && ret == 0)
 		ret = t_ret;
 
-	if (REP_ON(dbenv) &&
-	    (t_ret = __rep_preclose(dbenv)) != 0 && ret == 0)
+#ifdef HAVE_REPLICATION
+	if ((t_ret = __rep_close(dbenv)) != 0 && ret == 0)
 		ret = t_ret;
+#endif
 
 	/*
 	 * Detach from the regions and undo the allocations done by
@@ -718,13 +666,6 @@ __env_close(dbenv, rep_check)
 	if ((t_ret = __env_refresh(dbenv, 0, rep_check)) != 0 && ret == 0)
 		ret = t_ret;
 
-	/* Do per-subsystem close. */
-	if ((t_ret = __lock_dbenv_close(dbenv)) != 0 && ret == 0)
-		ret = t_ret;
-
-	if ((t_ret = __rep_dbenv_close(dbenv)) != 0 && ret == 0)
-		ret = t_ret;
-
 #ifdef HAVE_CRYPTO
 	/*
 	 * Crypto comes last, because higher level close functions need
@@ -733,37 +674,40 @@ __env_close(dbenv, rep_check)
 	if ((t_ret = __crypto_dbenv_close(dbenv)) != 0 && ret == 0)
 		ret = t_ret;
 #endif
-
-	/* Release any string-based configuration parameters we've copied. */
-	if (dbenv->db_log_dir != NULL)
-		__os_free(dbenv, dbenv->db_log_dir);
-	if (dbenv->db_tmp_dir != NULL)
-		__os_free(dbenv, dbenv->db_tmp_dir);
-	if (dbenv->db_data_dir != NULL) {
-		for (p = dbenv->db_data_dir; *p != NULL; ++p)
-			__os_free(dbenv, *p);
-		__os_free(dbenv, dbenv->db_data_dir);
-	}
-
 	/* If we're registered, clean up. */
 	if (dbenv->registry != NULL) {
 		(void)__envreg_unregister(dbenv, 0);
 		dbenv->registry = NULL;
 	}
 
+	/* Release any string-based configuration parameters we've copied. */
+	if (dbenv->db_log_dir != NULL)
+		__os_free(dbenv, dbenv->db_log_dir);
+	dbenv->db_log_dir = NULL;
+	if (dbenv->db_tmp_dir != NULL)
+		__os_free(dbenv, dbenv->db_tmp_dir);
+	dbenv->db_tmp_dir = NULL;
+	if (dbenv->db_data_dir != NULL) {
+		for (p = dbenv->db_data_dir; *p != NULL; ++p)
+			__os_free(dbenv, *p);
+		__os_free(dbenv, dbenv->db_data_dir);
+		dbenv->db_data_dir = NULL;
+		dbenv->data_next = 0;
+	}
+	if (dbenv->db_home != NULL) {
+		__os_free(dbenv, dbenv->db_home);
+		dbenv->db_home = NULL;
+	}
+
 	/* Discard the structure. */
-	memset(dbenv, CLEAR_BYTE, sizeof(DB_ENV));
-	__os_free(NULL, dbenv);
+	__db_env_destroy(dbenv);
 
 	return (ret);
 }
 
 /*
  * __env_refresh --
- *	Refresh the DB_ENV structure, releasing resources allocated by
- * DB_ENV->open, and returning it to the state it was in just before
- * open was called.  (Note that this means that any state set by
- * pre-open configuration functions must be preserved.)
+ *	Refresh the DB_ENV structure.
  */
 static int
 __env_refresh(dbenv, orig_flags, rep_check)
@@ -778,6 +722,10 @@ __env_refresh(dbenv, orig_flags, rep_check)
 	ret = 0;
 
 	/*
+	 * Release resources allocated by DB_ENV->open, and return it to the
+	 * state it was in just before __env_open was called.  (This means
+	 * state set by pre-open configuration functions must be preserved.)
+	 *
 	 * Refresh subsystems, in the reverse order they were opened (txn
 	 * must be first, it may want to discard locks and flush the log).
 	 *
@@ -822,18 +770,17 @@ __env_refresh(dbenv, orig_flags, rep_check)
 	 * log file handles.  Ick.
 	 */
 	if (dbenv->db_ref != 0) {
-		__db_err(dbenv,
+		__db_errx(dbenv,
 		    "Database handles still open at environment close");
-		for (ldbp = LIST_FIRST(&dbenv->dblist);
-		    ldbp != NULL; ldbp = LIST_NEXT(ldbp, dblistlinks))
-			__db_err(dbenv, "Open database handle: %s%s%s",
+		TAILQ_FOREACH(ldbp, &dbenv->dblist, dblistlinks)
+			__db_errx(dbenv, "Open database handle: %s%s%s",
 			    ldbp->fname == NULL ? "unnamed" : ldbp->fname,
 			    ldbp->dname == NULL ? "" : "/",
 			    ldbp->dname == NULL ? "" : ldbp->dname);
 		if (ret == 0)
 			ret = EINVAL;
 	}
-	LIST_INIT(&dbenv->dblist);
+	TAILQ_INIT(&dbenv->dblist);
 
 	if ((t_ret = __mutex_free(dbenv, &dbenv->mtx_dblist)) != 0 && ret == 0)
 		ret = t_ret;
@@ -877,11 +824,12 @@ __env_refresh(dbenv, orig_flags, rep_check)
 	 *
 	 * Must come after we call __env_db_rep_exit above.
 	 */
-	__rep_dbenv_refresh(dbenv);
+	if (REP_ON(dbenv))
+		__rep_dbenv_refresh(dbenv);
 
 	/*
-	 * Mark the thread as out of the env before we get rid
-	 * of the handles needed to do so.
+	 * Mark the thread as out of the env before we get rid of the handles
+	 * needed to do so.
 	 */
 	if (dbenv->thr_hashtab != NULL &&
 	    (t_ret = __env_set_state(dbenv, &ip, THREAD_OUT)) != 0 && ret == 0)
@@ -901,23 +849,11 @@ __env_refresh(dbenv, orig_flags, rep_check)
 		 */
 	}
 
-	/* Undo changes and allocations done by __env_open. */
-	if (dbenv->db_home != NULL) {
-		__os_free(dbenv, dbenv->db_home);
-		dbenv->db_home = NULL;
-	}
-	if (dbenv->db_abshome != NULL) {
-		__os_free(dbenv, dbenv->db_abshome);
-		dbenv->db_abshome = NULL;
-	}
 	if (dbenv->mutex_iq != NULL) {
 		__os_free(dbenv, dbenv->mutex_iq);
 		dbenv->mutex_iq = NULL;
 	}
 
-	dbenv->open_flags = 0;
-	dbenv->db_mode = 0;
-
 	if (dbenv->recover_dtab != NULL) {
 		__os_free(dbenv, dbenv->recover_dtab);
 		dbenv->recover_dtab = NULL;
@@ -985,13 +921,15 @@ __db_appname(dbenv, appname, file, tmp_oflags, fhpp, namep)
 	DB_FH **fhpp;
 	char **namep;
 {
+	enum { TRY_NOTSET, TRY_DATA_DIR, TRY_ENV_HOME, TRY_CREATE } try_state;
 	size_t len, str_len;
 	int data_entry, ret, slash, tmp_create;
 	const char *a, *b;
 	char *p, *str;
 
+	try_state = TRY_NOTSET;
 	a = b = NULL;
-	data_entry = -1;
+	data_entry = 0;
 	tmp_create = 0;
 
 	/*
@@ -1028,11 +966,30 @@ retry:	/*
 	case DB_APP_NONE:
 		break;
 	case DB_APP_DATA:
-		if (dbenv != NULL && dbenv->db_data_dir != NULL &&
-		    (b = dbenv->db_data_dir[++data_entry]) == NULL) {
-			data_entry = -1;
-			b = dbenv->db_data_dir[0];
+		if (dbenv == NULL || dbenv->db_data_dir == NULL) {
+			try_state = TRY_CREATE;
+			break;
 		}
+
+		/*
+		 * First, step through the data_dir entries, if any, looking
+		 * for the file.
+		 */
+		if ((b = dbenv->db_data_dir[data_entry]) != NULL) {
+			++data_entry;
+			try_state = TRY_DATA_DIR;
+			break;
+		}
+
+		/* Second, look in the environment home directory. */
+		if (try_state != TRY_ENV_HOME) {
+			try_state = TRY_ENV_HOME;
+			break;
+		}
+
+		/* Third, try creation in the first data_dir entry. */
+		try_state = TRY_CREATE;
+		b = dbenv->db_data_dir[0];
 		break;
 	case DB_APP_LOG:
 		if (dbenv != NULL)
@@ -1071,7 +1028,8 @@ retry:	/*
 	 * If we're opening a data file, see if it exists.  If it does,
 	 * return it, otherwise, try and find another one to open.
 	 */
-	if (__os_exists(str, NULL) != 0 && data_entry != -1) {
+	if (appname == DB_APP_DATA &&
+	    __os_exists(dbenv, str, NULL) != 0 && try_state != TRY_CREATE) {
 		__os_free(dbenv, str);
 		b = NULL;
 		goto retry;
@@ -1091,423 +1049,6 @@ retry:	/*
 	return (0);
 }
 
-/*
- * __db_home --
- *	Find the database home.
- *
- * PUBLIC:	int __db_home __P((DB_ENV *, const char *, u_int32_t));
- */
-int
-__db_home(dbenv, db_home, flags)
-	DB_ENV *dbenv;
-	const char *db_home;
-	u_int32_t flags;
-{
-	int ret;
-	const char *p;
-	char path[MAXPATHLEN];
-
-	/*
-	 * Use db_home by default, this allows utilities to reasonably
-	 * override the environment either explicitly or by using a -h
-	 * option.  Otherwise, use the environment if it's permitted
-	 * and initialized.
-	 */
-	if ((p = db_home) == NULL &&
-	    (LF_ISSET(DB_USE_ENVIRON) ||
-	    (LF_ISSET(DB_USE_ENVIRON_ROOT) && __os_isroot())) &&
-	    (p = getenv("DB_HOME")) != NULL && p[0] == '\0') {
-		__db_err(dbenv, "illegal DB_HOME environment variable");
-		return (EINVAL);
-	}
-	if (p != NULL && (ret = __os_strdup(dbenv, p, &dbenv->db_home)) != 0)
-		return (ret);
-
-	/*
-	 * Get the absolute pathname of the current directory.  We use this
-	 * to build absolute pathnames when removing log files.
-	 *
-	 * XXX
-	 * Can't trust getcwd(3) to set a valid errno, so don't try to display
-	 * one unless we know it's good.  It's likely a permissions problem:
-	 * use something bland and useless in the default return value, so we
-	 * don't send somebody off in the wrong direction.
-	 */
-	__os_set_errno(0);
-	if ((p = getcwd(path, sizeof(path))) == NULL) {
-		if ((ret = __os_get_errno()) == 0) {
-			__db_err(dbenv,
-			    "no absolute path for the current directory");
-			ret = EAGAIN;
-		} else
-			__db_err(dbenv,
-			    "no absolute path for the current directory: %s",
-			    db_strerror(ret));
-		return (ret);
-	}
-	if (p != NULL && (ret = __os_strdup(dbenv, p, &dbenv->db_abshome)) != 0)
-		return (ret);
-
-	return (0);
-}
-
-#define	__DB_OVFL(v, max)						\
-	if (v > max) {							\
-		__v = v;						\
-		__max = max;						\
-		goto toobig;						\
-	}
-
-/*
- * __db_parse --
- *	Parse a single NAME VALUE pair.
- */
-static int
-__db_parse(dbenv, s)
-	DB_ENV *dbenv;
-	char *s;
-{
-	u_long __max, __v, v1, v2, v3;
-	u_int32_t flags;
-	char *name, *p, *value, v4;
-
-	/*
-	 * !!!
-	 * The constant 40 is hard-coded into format arguments to sscanf
-	 * below, it can't be changed here without changing it there, too.
-	 * The additional bytes are for a trailing nul byte and because we
-	 * are reading user input -- I don't want to risk any off-by-ones.
-	 */
-	char arg[40 + 5];
-
-	/*
-	 * Name/value pairs are parsed as two white-space separated strings.
-	 * Leading and trailing white-space is trimmed from the value, but
-	 * it may contain embedded white-space.  Note: we use the isspace(3)
-	 * macro because it's more portable, but that means that you can use
-	 * characters like form-feed to separate the strings.
-	 */
-	name = s;
-	for (p = name; *p != '\0' && !isspace((int)*p); ++p)
-		;
-	if (*p == '\0' || p == name)
-		goto illegal;
-	*p = '\0';
-	for (++p; isspace((int)*p); ++p)
-		;
-	if (*p == '\0')
-		goto illegal;
-	value = p;
-	for (++p; *p != '\0'; ++p)
-		;
-	for (--p; isspace((int)*p); --p)
-		;
-	++p;
-	if (p == value) {
-illegal:	__db_err(dbenv, "mis-formatted name-value pair: %s", s);
-		return (EINVAL);
-	}
-	*p = '\0';
-
-	if (strcasecmp(name, "mutex_set_align") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__mutex_set_align(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "mutex_set_increment") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__mutex_set_increment(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "mutex_set_max") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__mutex_set_max(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "mutex_set_tas_spins") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__mutex_set_tas_spins(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "rep_set_config") == 0) {
-		if (sscanf(value, "%40s %c", arg, &v4) != 1)
-			goto badarg;
-
-		if (strcasecmp(value, "rep_bulk") == 0)
-			return (__rep_set_config(dbenv,
-			    DB_REP_CONF_BULK, 1));
-		if (strcasecmp(value, "rep_delayclient") == 0)
-			return (__rep_set_config(dbenv,
-			    DB_REP_CONF_DELAYCLIENT, 1));
-		if (strcasecmp(value, "rep_noautoinit") == 0)
-			return (__rep_set_config(dbenv,
-			    DB_REP_CONF_NOAUTOINIT, 1));
-		if (strcasecmp(value, "rep_nowait") == 0)
-			return (__rep_set_config(dbenv, DB_REP_CONF_NOWAIT, 1));
-		goto badarg;
-	}
-
-	if (strcasecmp(name, "set_cachesize") == 0) {
-		if (sscanf(value, "%lu %lu %lu %c", &v1, &v2, &v3, &v4) != 3)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		__DB_OVFL(v2, UINT32_MAX);
-		__DB_OVFL(v3, 10000);
-		return (__memp_set_cachesize(
-		    dbenv, (u_int32_t)v1, (u_int32_t)v2, (int)v3));
-	}
-
-	if (strcasecmp(name, "set_data_dir") == 0 ||
-	    strcasecmp(name, "db_data_dir") == 0)	/* Compatibility. */
-		return (__env_set_data_dir(dbenv, value));
-
-							/* Undocumented. */
-	if (strcasecmp(name, "set_intermediate_dir") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-#ifdef INT_MAX
-		__DB_OVFL(v1, INT_MAX);
-#endif
-		return (__env_set_intermediate_dir(dbenv, (int)v1, 0));
-	}
-
-	if (strcasecmp(name, "set_flags") == 0) {
-		if (sscanf(value, "%40s %c", arg, &v4) != 1)
-			goto badarg;
-
-		if (strcasecmp(value, "db_auto_commit") == 0)
-			return (__env_set_flags(dbenv, DB_AUTO_COMMIT, 1));
-		if (strcasecmp(value, "db_cdb_alldb") == 0)
-			return (__env_set_flags(dbenv, DB_CDB_ALLDB, 1));
-		if (strcasecmp(value, "db_direct_db") == 0)
-			return (__env_set_flags(dbenv, DB_DIRECT_DB, 1));
-		if (strcasecmp(value, "db_direct_log") == 0)
-			return (__env_set_flags(dbenv, DB_DIRECT_LOG, 1));
-		if (strcasecmp(value, "db_dsync_db") == 0)
-			return (__env_set_flags(dbenv, DB_DSYNC_DB, 1));
-		if (strcasecmp(value, "db_dsync_log") == 0)
-			return (__env_set_flags(dbenv, DB_DSYNC_LOG, 1));
-		if (strcasecmp(value, "db_log_autoremove") == 0)
-			return (__env_set_flags(dbenv, DB_LOG_AUTOREMOVE, 1));
-		if (strcasecmp(value, "db_log_inmemory") == 0)
-			return (__env_set_flags(dbenv, DB_LOG_INMEMORY, 1));
-		if (strcasecmp(value, "db_nolocking") == 0)
-			return (__env_set_flags(dbenv, DB_NOLOCKING, 1));
-		if (strcasecmp(value, "db_nommap") == 0)
-			return (__env_set_flags(dbenv, DB_NOMMAP, 1));
-		if (strcasecmp(value, "db_nopanic") == 0)
-			return (__env_set_flags(dbenv, DB_NOPANIC, 1));
-		if (strcasecmp(value, "db_overwrite") == 0)
-			return (__env_set_flags(dbenv, DB_OVERWRITE, 1));
-		if (strcasecmp(value, "db_region_init") == 0)
-			return (__env_set_flags(dbenv, DB_REGION_INIT, 1));
-		if (strcasecmp(value, "db_txn_nosync") == 0)
-			return (__env_set_flags(dbenv, DB_TXN_NOSYNC, 1));
-		if (strcasecmp(value, "db_txn_write_nosync") == 0)
-			return (
-			    __env_set_flags(dbenv, DB_TXN_WRITE_NOSYNC, 1));
-		if (strcasecmp(value, "db_yieldcpu") == 0)
-			return (__env_set_flags(dbenv, DB_YIELDCPU, 1));
-		goto badarg;
-	}
-
-	if (strcasecmp(name, "set_lg_bsize") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__log_set_lg_bsize(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_lg_filemode") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, INT_MAX);
-		return (__log_set_lg_filemode(dbenv, (int)v1));
-	}
-
-	if (strcasecmp(name, "set_lg_max") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__log_set_lg_max(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_lg_regionmax") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__log_set_lg_regionmax(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_lg_dir") == 0 ||
-	    strcasecmp(name, "db_log_dir") == 0)	/* Compatibility. */
-		return (__log_set_lg_dir(dbenv, value));
-
-	if (strcasecmp(name, "set_lk_detect") == 0) {
-		if (sscanf(value, "%40s %c", arg, &v4) != 1)
-			goto badarg;
-		if (strcasecmp(value, "db_lock_default") == 0)
-			flags = DB_LOCK_DEFAULT;
-		else if (strcasecmp(value, "db_lock_expire") == 0)
-			flags = DB_LOCK_EXPIRE;
-		else if (strcasecmp(value, "db_lock_maxlocks") == 0)
-			flags = DB_LOCK_MAXLOCKS;
-		else if (strcasecmp(value, "db_lock_maxwrite") == 0)
-			flags = DB_LOCK_MAXWRITE;
-		else if (strcasecmp(value, "db_lock_minlocks") == 0)
-			flags = DB_LOCK_MINLOCKS;
-		else if (strcasecmp(value, "db_lock_minwrite") == 0)
-			flags = DB_LOCK_MINWRITE;
-		else if (strcasecmp(value, "db_lock_oldest") == 0)
-			flags = DB_LOCK_OLDEST;
-		else if (strcasecmp(value, "db_lock_random") == 0)
-			flags = DB_LOCK_RANDOM;
-		else if (strcasecmp(value, "db_lock_youngest") == 0)
-			flags = DB_LOCK_YOUNGEST;
-		else
-			goto badarg;
-		return (__lock_set_lk_detect(dbenv, flags));
-	}
-
-	if (strcasecmp(name, "set_lk_max") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__lock_set_lk_max(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_lk_max_locks") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__lock_set_lk_max_locks(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_lk_max_lockers") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__lock_set_lk_max_lockers(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_lk_max_objects") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__lock_set_lk_max_objects(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_lock_timeout") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__lock_set_env_timeout(
-		    dbenv, (u_int32_t)v1, DB_SET_LOCK_TIMEOUT));
-	}
-
-	if (strcasecmp(name, "set_mp_max_openfd") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, INT_MAX);
-		return (__memp_set_mp_max_openfd(dbenv, (int)v1));
-	}
-
-	if (strcasecmp(name, "set_mp_max_write") == 0) {
-		if (sscanf(value, "%lu %lu %c", &v1, &v2, &v4) != 2)
-			goto badarg;
-		__DB_OVFL(v1, INT_MAX);
-		__DB_OVFL(v2, INT_MAX);
-		return (__memp_set_mp_max_write(dbenv, (int)v1, (int)v2));
-	}
-
-	if (strcasecmp(name, "set_mp_mmapsize") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__memp_set_mp_mmapsize(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_region_init") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1 || v1 != 1)
-			goto badarg;
-		return (__env_set_flags(
-		    dbenv, DB_REGION_INIT, v1 == 0 ? 0 : 1));
-	}
-
-	if (strcasecmp(name, "set_shm_key") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		return (__env_set_shm_key(dbenv, (long)v1));
-	}
-
-	/*
-	 * The set_tas_spins method has been replaced by mutex_set_tas_spins.
-	 * The set_tas_spins name remains for DB_CONFIG compatibility.
-	 */
-	if (strcasecmp(name, "set_tas_spins") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__mutex_set_tas_spins(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_tmp_dir") == 0 ||
-	    strcasecmp(name, "db_tmp_dir") == 0)	/* Compatibility.*/
-		return (__env_set_tmp_dir(dbenv, value));
-
-	if (strcasecmp(name, "set_tx_max") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__txn_set_tx_max(dbenv, (u_int32_t)v1));
-	}
-
-	if (strcasecmp(name, "set_txn_timeout") == 0) {
-		if (sscanf(value, "%lu %c", &v1, &v4) != 1)
-			goto badarg;
-		__DB_OVFL(v1, UINT32_MAX);
-		return (__lock_set_env_timeout(
-		    dbenv, (u_int32_t)v1, DB_SET_TXN_TIMEOUT));
-	}
-
-	if (strcasecmp(name, "set_verbose") == 0) {
-		if (sscanf(value, "%40s %c", arg, &v4) != 1)
-			goto badarg;
-
-		else if (strcasecmp(value, "db_verb_deadlock") == 0)
-			flags = DB_VERB_DEADLOCK;
-		else if (strcasecmp(value, "db_verb_recovery") == 0)
-			flags = DB_VERB_RECOVERY;
-		else if (strcasecmp(value, "db_verb_register") == 0)
-			flags = DB_VERB_REGISTER;
-		else if (strcasecmp(value, "db_verb_replication") == 0)
-			flags = DB_VERB_REPLICATION;
-		else if (strcasecmp(value, "db_verb_waitsfor") == 0)
-			flags = DB_VERB_WAITSFOR;
-		else
-			goto badarg;
-		return (__env_set_verbose(dbenv, flags, 1));
-	}
-
-	__db_err(dbenv, "unrecognized name-value pair: %s", s);
-	return (EINVAL);
-
-badarg:	__db_err(dbenv, "incorrect arguments for name-value pair: %s", s);
-	return (EINVAL);
-
-toobig:	__db_err(dbenv,
-	    "%s: %lu larger than maximum value %lu", s, __v, __max);
-	return (EINVAL);
-}
-
 /*
  * __db_tmp_open --
  *	Create a temporary file.
@@ -1520,7 +1061,6 @@ __db_tmp_open(dbenv, tmp_oflags, path, fhpp)
 	DB_FH **fhpp;
 {
 	pid_t pid;
-	db_threadid_t tid;
 	int filenum, i, isdir, ret;
 	char *firstx, *trv;
 
@@ -1528,12 +1068,12 @@ __db_tmp_open(dbenv, tmp_oflags, path, fhpp)
 	 * Check the target directory; if you have six X's and it doesn't
 	 * exist, this runs for a *very* long time.
 	 */
-	if ((ret = __os_exists(path, &isdir)) != 0) {
-		__db_err(dbenv, "%s: %s", path, db_strerror(ret));
+	if ((ret = __os_exists(dbenv, path, &isdir)) != 0) {
+		__db_err(dbenv, ret, "%s", path);
 		return (ret);
 	}
 	if (!isdir) {
-		__db_err(dbenv, "%s: %s", path, db_strerror(EINVAL));
+		__db_err(dbenv, EINVAL, "%s", path);
 		return (EINVAL);
 	}
 
@@ -1542,7 +1082,7 @@ __db_tmp_open(dbenv, tmp_oflags, path, fhpp)
 	(void)strcat(path, DB_TRAIL);
 
 	/* Replace the X's with the process ID (in decimal). */
-	__os_id(dbenv, &pid, &tid);
+	__os_id(dbenv, &pid, NULL);
 	for (trv = path + strlen(path); *--trv == 'X'; pid /= 10)
 		*trv = '0' + (u_char)(pid % 10);
 	firstx = trv + 1;
@@ -1562,8 +1102,7 @@ __db_tmp_open(dbenv, tmp_oflags, path, fhpp)
 		 * of other possible errors, we've lost.
 		 */
 		if (ret != EEXIST) {
-			__db_err(dbenv,
-			    "tmp_open: %s: %s", path, db_strerror(ret));
+			__db_err(dbenv, ret, "temporary open: %s", path);
 			return (ret);
 		}
 
