@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1997-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: ex_btrec.c,v 11.14 2001/10/04 18:46:28 sue Exp $
+ * $Id: ex_btrec.c,v 12.3 2006/08/24 14:45:41 bostic Exp $
  */
 
 #include <sys/types.h>
@@ -20,7 +20,7 @@
 int	main __P((void));
 
 int	ex_btrec __P((void));
-void	show __P((char *, DBT *, DBT *));
+void	show __P((const char *, DBT *, DBT *));
 
 int
 main()
@@ -37,7 +37,7 @@ ex_btrec()
 	DB_BTREE_STAT *statp;
 	FILE *fp;
 	db_recno_t recno;
-	u_int32_t len;
+	size_t len;
 	int cnt, ret;
 	char *p, *t, buf[1024], rbuf[1024];
 	const char *progname = "ex_btrec";		/* Program name. */
@@ -68,8 +68,8 @@ ex_btrec()
 		dbp->err(dbp, ret, "set_flags: DB_RECNUM");
 		return (1);
 	}
-	if ((ret =
-	    dbp->open(dbp, DATABASE, NULL, DB_BTREE, DB_CREATE, 0664)) != 0) {
+	if ((ret = dbp->open(dbp,
+	    NULL, DATABASE, NULL, DB_BTREE, DB_CREATE, 0664)) != 0) {
 		dbp->err(dbp, ret, "open: %s", DATABASE);
 		return (1);
 	}
@@ -92,7 +92,7 @@ ex_btrec()
 
 		key.data = buf;
 		data.data = rbuf;
-		data.size = key.size = len - 1;
+		data.size = key.size = (u_int32_t)len - 1;
 
 		if ((ret =
 		    dbp->put(dbp, NULL, &key, &data, DB_NOOVERWRITE)) != 0) {
@@ -106,7 +106,7 @@ ex_btrec()
 	(void)fclose(fp);
 
 	/* Print out the number of records in the database. */
-	if ((ret = dbp->stat(dbp, &statp, 0)) != 0) {
+	if ((ret = dbp->stat(dbp, NULL, &statp, 0)) != 0) {
 		dbp->err(dbp, ret, "DB->stat");
 		goto err1;
 	}
@@ -194,8 +194,8 @@ err1:	(void)dbp->close(dbp, 0);
  */
 void
 show(msg, key, data)
+	const char *msg;
 	DBT *key, *data;
-	char *msg;
 {
 	printf("%s%.*s : %.*s\n", msg,
 	    (int)key->size, (char *)key->data,
