@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1990, 1993, 1994
@@ -34,37 +34,24 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: hash_dup.c,v 12.17 2006/09/07 20:05:29 bostic Exp $
  */
-#include "db_config.h"
-
-#ifndef lint
-static const char revid[] = "$Id: hash_dup.c,v 11.65 2001/08/13 19:11:37 bostic Exp $";
-#endif /* not lint */
 
 /*
  * PACKAGE:  hashing
  *
  * DESCRIPTION:
- *      Manipulation of duplicates for the hash package.
- *
- * ROUTINES:
- *
- * External
- *      __add_dup
- * Internal
+ *	Manipulation of duplicates for the hash package.
  */
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#endif
+#include "db_config.h"
 
 #include "db_int.h"
-#include "db_page.h"
-#include "hash.h"
-#include "btree.h"
-#include "txn.h"
+#include "dbinc/db_page.h"
+#include "dbinc/hash.h"
+#include "dbinc/btree.h"
+#include "dbinc/mp.h"
 
 static int __ham_c_chgpg __P((DBC *,
     db_pgno_t, u_int32_t, db_pgno_t, u_int32_t));
@@ -96,6 +83,7 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 {
 	DB *dbp;
 	DBT pval, tmp_val;
+	DB_ENV *dbenv;
 	DB_MPOOLFILE *mpf;
 	HASH_CURSOR *hcp;
 	u_int32_t add_bytes, new_size;
@@ -103,10 +91,11 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 	u_int8_t *hk;
 
 	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
 	mpf = dbp->mpf;
 	hcp = (HASH_CURSOR *)dbc->internal;
 
-	DB_ASSERT(flags != DB_CURRENT);
+	DB_ASSERT(dbenv, flags != DB_CURRENT);
 
 	add_bytes = nval->size +
 	    (F_ISSET(nval, DB_DBT_PARTIAL) ? nval->doff : 0);
@@ -122,12 +111,12 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 	 * hcp->dndx is the first free ndx or the index of the
 	 * current pointer into the duplicate set.
 	 */
-	hk = H_PAIRDATA(hcp->page, hcp->indx);
+	hk = H_PAIRDATA(dbp, hcp->page, hcp->indx);
 	/* Add the len bytes to the current singleton. */
 	if (HPAGE_PTYPE(hk) != H_DUPLICATE)
 		add_bytes += DUP_SIZE(0);
 	new_size =
-	    LEN_HKEYDATA(hcp->page, dbp->pgsize, H_DATAINDEX(hcp->indx)) +
+	    LEN_HKEYDATA(dbp, hcp->page, dbp->pgsize, H_DATAINDEX(hcp->indx)) +
 	    add_bytes;
 
 	/*
@@ -137,7 +126,7 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 	 */
 	if (HPAGE_PTYPE(hk) != H_OFFDUP &&
 	    (HPAGE_PTYPE(hk) == H_OFFPAGE || ISBIG(hcp, new_size) ||
-	    add_bytes > P_FREESPACE(hcp->page))) {
+	    add_bytes > P_FREESPACE(dbp, hcp->page))) {
 
 		if ((ret = __ham_dup_convert(dbc)) != 0)
 			return (ret);
@@ -150,14 +139,14 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 		if (HPAGE_PTYPE(hk) != H_DUPLICATE) {
 			pval.flags = 0;
 			pval.data = HKEYDATA_DATA(hk);
-			pval.size = LEN_HDATA(hcp->page, dbp->pgsize,
+			pval.size = LEN_HDATA(dbp, hcp->page, dbp->pgsize,
 			    hcp->indx);
-			if ((ret = __ham_make_dup(dbp->dbenv,
+			if ((ret = __ham_make_dup(dbenv,
 			    &pval, &tmp_val, &dbc->my_rdata.data,
 			    &dbc->my_rdata.ulen)) != 0 || (ret =
 			    __ham_replpair(dbc, &tmp_val, 1)) != 0)
 				return (ret);
-			hk = H_PAIRDATA(hcp->page, hcp->indx);
+			hk = H_PAIRDATA(dbp, hcp->page, hcp->indx);
 			HPAGE_PTYPE(hk) = H_DUPLICATE;
 
 			/*
@@ -171,7 +160,7 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 		}
 
 		/* Now make the new entry a duplicate. */
-		if ((ret = __ham_make_dup(dbp->dbenv, nval,
+		if ((ret = __ham_make_dup(dbenv, nval,
 		    &tmp_val, &dbc->my_rdata.data, &dbc->my_rdata.ulen)) != 0)
 			return (ret);
 
@@ -188,7 +177,7 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 				if (cmp == 0)
 					return (__db_duperr(dbp, flags));
 			} else {
-				hcp->dup_tlen = LEN_HDATA(hcp->page,
+				hcp->dup_tlen = LEN_HDATA(dbp, hcp->page,
 				    dbp->pgsize, hcp->indx);
 				hcp->dup_len = nval->size;
 				F_SET(hcp, H_ISDUP);
@@ -205,12 +194,13 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 		case DB_AFTER:
 			tmp_val.doff = hcp->dup_off + DUP_SIZE(hcp->dup_len);
 			break;
+		default:
+			return (__db_unknown_path(dbenv, "__ham_add_dup"));
 		}
+
 		/* Add the duplicate. */
-		ret = __ham_replpair(dbc, &tmp_val, 0);
-		if (ret == 0)
-			ret = mpf->set(mpf, hcp->page, DB_MPOOL_DIRTY);
-		if (ret != 0)
+		if ((ret = __memp_dirty(mpf, &hcp->page, dbc->txn, 0)) != 0 ||
+		    (ret = __ham_replpair(dbc, &tmp_val, 0)) != 0)
 			return (ret);
 
 		/* Now, update the cursor if necessary. */
@@ -218,14 +208,17 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 		case DB_AFTER:
 			hcp->dup_off += DUP_SIZE(hcp->dup_len);
 			hcp->dup_len = nval->size;
-			hcp->dup_tlen += DUP_SIZE(nval->size);
+			hcp->dup_tlen += (db_indx_t)DUP_SIZE(nval->size);
 			break;
+		case DB_BEFORE:
 		case DB_KEYFIRST:
 		case DB_KEYLAST:
-		case DB_BEFORE:
-			hcp->dup_tlen += DUP_SIZE(nval->size);
+		case DB_NODUPDATA:
+			hcp->dup_tlen += (db_indx_t)DUP_SIZE(nval->size);
 			hcp->dup_len = nval->size;
 			break;
+		default:
+			return (__db_unknown_path(dbenv, "__ham_add_dup"));
 		}
 		ret = __ham_c_update(dbc, tmp_val.size, 1, 1);
 		return (ret);
@@ -235,8 +228,8 @@ __ham_add_dup(dbc, nval, flags, pgnop)
 	 * If we get here, then we're on duplicate pages; set pgnop and
 	 * return so the common code can handle it.
 	 */
-	memcpy(pgnop,
-	    HOFFDUP_PGNO(H_PAIRDATA(hcp->page, hcp->indx)), sizeof(db_pgno_t));
+	memcpy(pgnop, HOFFDUP_PGNO(H_PAIRDATA(dbp, hcp->page, hcp->indx)),
+	    sizeof(db_pgno_t));
 
 	return (ret);
 }
@@ -254,6 +247,7 @@ __ham_dup_convert(dbc)
 	DB *dbp;
 	DBC **hcs;
 	DBT dbt;
+	DB_ENV *dbenv;
 	DB_LSN lsn;
 	DB_MPOOLFILE *mpf;
 	HASH_CURSOR *hcp;
@@ -264,6 +258,7 @@ __ham_dup_convert(dbc)
 	u_int8_t *p, *pend;
 
 	dbp = dbc->dbp;
+	dbenv = dbp->dbenv;
 	mpf = dbp->mpf;
 	hcp = (HASH_CURSOR *)dbc->internal;
 
@@ -287,20 +282,20 @@ __ham_dup_convert(dbc)
 	 * Now put the duplicates onto the new page.
 	 */
 	dbt.flags = 0;
-	switch (HPAGE_PTYPE(H_PAIRDATA(hcp->page, hcp->indx))) {
+	switch (HPAGE_PTYPE(H_PAIRDATA(dbp, hcp->page, hcp->indx))) {
 	case H_KEYDATA:
 		/* Simple case, one key on page; move it to dup page. */
-		dbt.size = LEN_HDATA(hcp->page, dbp->pgsize, hcp->indx);
-		dbt.data = HKEYDATA_DATA(H_PAIRDATA(hcp->page, hcp->indx));
+		dbt.size = LEN_HDATA(dbp, hcp->page, dbp->pgsize, hcp->indx);
+		dbt.data = HKEYDATA_DATA(H_PAIRDATA(dbp, hcp->page, hcp->indx));
 		ret = __db_pitem(dbc,
 		    dp, 0, BKEYDATA_SIZE(dbt.size), NULL, &dbt);
 		goto finish;
 	case H_OFFPAGE:
 		/* Simple case, one key on page; move it to dup page. */
-		memcpy(&ho,
-		    P_ENTRY(hcp->page, H_DATAINDEX(hcp->indx)), HOFFPAGE_SIZE);
+		memcpy(&ho, P_ENTRY(dbp, hcp->page, H_DATAINDEX(hcp->indx)),
+		    HOFFPAGE_SIZE);
 		UMRW_SET(bo.unused1);
-		B_TSET(bo.type, ho.type, 0);
+		B_TSET(bo.type, ho.type);
 		UMRW_SET(bo.unused2);
 		bo.pgno = ho.pgno;
 		bo.tlen = ho.tlen;
@@ -309,15 +304,11 @@ __ham_dup_convert(dbc)
 
 		ret = __db_pitem(dbc, dp, 0, dbt.size, &dbt, NULL);
 finish:		if (ret == 0) {
-			if ((ret = mpf->set(mpf, dp, DB_MPOOL_DIRTY)) != 0)
-				break;
-
 			/* Update any other cursors. */
-			if (hcs != NULL && DB_LOGGING(dbc)
-			     && IS_SUBTRANSACTION(dbc->txn)) {
-				if ((ret = __ham_chgpg_log(dbp->dbenv,
-				    dbc->txn, &lsn, 0, dbp->log_fileid,
-				    DB_HAM_DUP, PGNO(hcp->page),
+			if (hcs != NULL && DBC_LOGGING(dbc) &&
+			    IS_SUBTRANSACTION(dbc->txn)) {
+				if ((ret = __ham_chgpg_log(dbp, dbc->txn,
+				    &lsn, 0, DB_HAM_DUP, PGNO(hcp->page),
 				    PGNO(dp), hcp->indx, 0)) != 0)
 					break;
 			}
@@ -328,9 +319,9 @@ finish:		if (ret == 0) {
 		}
 		break;
 	case H_DUPLICATE:
-		p = HKEYDATA_DATA(H_PAIRDATA(hcp->page, hcp->indx));
+		p = HKEYDATA_DATA(H_PAIRDATA(dbp, hcp->page, hcp->indx));
 		pend = p +
-		    LEN_HDATA(hcp->page, dbp->pgsize, hcp->indx);
+		    LEN_HDATA(dbp, hcp->page, dbp->pgsize, hcp->indx);
 
 		/*
 		 * We need to maintain the duplicate cursor position.
@@ -350,11 +341,10 @@ finish:		if (ret == 0) {
 				break;
 
 			/* Update any other cursors */
-			if (hcs != NULL && DB_LOGGING(dbc)
-			     && IS_SUBTRANSACTION(dbc->txn)) {
-				if ((ret = __ham_chgpg_log(dbp->dbenv,
-				    dbc->txn, &lsn, 0, dbp->log_fileid,
-				    DB_HAM_DUP, PGNO(hcp->page),
+			if (hcs != NULL && DBC_LOGGING(dbc) &&
+			    IS_SUBTRANSACTION(dbc->txn)) {
+				if ((ret = __ham_chgpg_log(dbp, dbc->txn,
+				    &lsn, 0, DB_HAM_DUP, PGNO(hcp->page),
 				    PGNO(dp), hcp->indx, i)) != 0)
 					break;
 			}
@@ -367,7 +357,7 @@ finish:		if (ret == 0) {
 		}
 		break;
 	default:
-		ret = __db_pgfmt(dbp->dbenv, (u_long)hcp->pgno);
+		ret = __db_pgfmt(dbenv, hcp->pgno);
 		break;
 	}
 
@@ -375,22 +365,21 @@ finish:		if (ret == 0) {
 	 * Now attach this to the source page in place of the old duplicate
 	 * item.
 	 */
+	if (ret == 0)
+		ret = __memp_dirty(mpf, &hcp->page, dbc->txn, 0);
+
 	if (ret == 0)
 		ret = __ham_move_offpage(dbc, hcp->page,
 		    (u_int32_t)H_DATAINDEX(hcp->indx), PGNO(dp));
 
-err:	if (ret == 0)
-		ret = mpf->set(mpf, hcp->page, DB_MPOOL_DIRTY);
-
-	if ((t_ret =
-	    mpf->put(mpf, dp, ret == 0 ? DB_MPOOL_DIRTY : 0)) != 0 && ret == 0)
+err:	if ((t_ret = __memp_fput(mpf, dp, 0)) != 0 && ret == 0)
 		ret = t_ret;
 
 	if (ret == 0)
 		hcp->dup_tlen = hcp->dup_off = hcp->dup_len = 0;
 
 	if (hcs != NULL)
-		__os_free(dbp->dbenv, hcs, 0);
+		__os_free(dbenv, hcs);
 
 	return (ret);
 }
@@ -466,13 +455,13 @@ __ham_check_move(dbc, add_len)
 	db_pgno_t next_pgno;
 	u_int32_t new_datalen, old_len, rectype;
 	u_int8_t *hk;
-	int ret;
+	int ret, t_ret;
 
 	dbp = dbc->dbp;
 	mpf = dbp->mpf;
 	hcp = (HASH_CURSOR *)dbc->internal;
 
-	hk = H_PAIRDATA(hcp->page, hcp->indx);
+	hk = H_PAIRDATA(dbp, hcp->page, hcp->indx);
 
 	/*
 	 * If the item is already off page duplicates or an offpage item,
@@ -481,8 +470,9 @@ __ham_check_move(dbc, add_len)
 	if (HPAGE_PTYPE(hk) == H_OFFDUP || HPAGE_PTYPE(hk) == H_OFFPAGE)
 		return (0);
 
-	old_len = LEN_HITEM(hcp->page, dbp->pgsize, H_DATAINDEX(hcp->indx));
-	new_datalen = old_len - HKEYDATA_SIZE(0) + add_len;
+	old_len =
+	    LEN_HITEM(dbp, hcp->page, dbp->pgsize, H_DATAINDEX(hcp->indx));
+	new_datalen = (old_len - HKEYDATA_SIZE(0)) + add_len;
 	if (HPAGE_PTYPE(hk) != H_DUPLICATE)
 		new_datalen += DUP_SIZE(0);
 
@@ -495,10 +485,11 @@ __ham_check_move(dbc, add_len)
 	 * If neither of these is true, then we can return.
 	 */
 	if (ISBIG(hcp, new_datalen) && (old_len > HOFFDUP_SIZE ||
-	    HOFFDUP_SIZE - old_len <= P_FREESPACE(hcp->page)))
+	    HOFFDUP_SIZE - old_len <= P_FREESPACE(dbp, hcp->page)))
 		return (0);
 
-	if (!ISBIG(hcp, new_datalen) && add_len <= P_FREESPACE(hcp->page))
+	if (!ISBIG(hcp, new_datalen) &&
+	    (new_datalen - old_len) <= P_FREESPACE(dbp, hcp->page))
 		return (0);
 
 	/*
@@ -510,96 +501,94 @@ __ham_check_move(dbc, add_len)
 
 	new_datalen = ISBIG(hcp, new_datalen) ?
 	    HOFFDUP_SIZE : HKEYDATA_SIZE(new_datalen);
-	new_datalen += LEN_HITEM(hcp->page, dbp->pgsize, H_KEYINDEX(hcp->indx));
+	new_datalen +=
+	    LEN_HITEM(dbp, hcp->page, dbp->pgsize, H_KEYINDEX(hcp->indx));
 
 	next_pagep = NULL;
 	for (next_pgno = NEXT_PGNO(hcp->page); next_pgno != PGNO_INVALID;
 	    next_pgno = NEXT_PGNO(next_pagep)) {
 		if (next_pagep != NULL &&
-		    (ret = mpf->put(mpf, next_pagep, 0)) != 0)
+		    (ret = __memp_fput(mpf, next_pagep, 0)) != 0)
 			return (ret);
 
-		if ((ret = mpf->get(mpf,
-		    &next_pgno, DB_MPOOL_CREATE, &next_pagep)) != 0)
+		if ((ret = __memp_fget(mpf, &next_pgno, dbc->txn,
+		    DB_MPOOL_CREATE, &next_pagep)) != 0)
 			return (ret);
 
-		if (P_FREESPACE(next_pagep) >= new_datalen)
+		if (P_FREESPACE(dbp, next_pagep) >= new_datalen)
 			break;
 	}
 
 	/* No more pages, add one. */
+	if ((ret = __memp_dirty(mpf, &hcp->page, dbc->txn, 0)) != 0)
+		return (ret);
+
 	if (next_pagep == NULL && (ret = __ham_add_ovflpage(dbc,
 	    hcp->page, 0, &next_pagep)) != 0)
 		return (ret);
 
 	/* Add new page at the end of the chain. */
-	if (P_FREESPACE(next_pagep) < new_datalen && (ret =
+	if ((ret = __memp_dirty(mpf, &next_pagep, dbc->txn, 0)) != 0) {
+		(void)__memp_fput(mpf, next_pagep, 0);
+		return (ret);
+	}
+
+	if (P_FREESPACE(dbp, next_pagep) < new_datalen && (ret =
 	    __ham_add_ovflpage(dbc, next_pagep, 1, &next_pagep)) != 0) {
-		(void)mpf->put(mpf, next_pagep, 0);
+		(void)__memp_fput(mpf, next_pagep, 0);
 		return (ret);
 	}
 
 	/* Copy the item to the new page. */
-	if (DB_LOGGING(dbc)) {
+	if (DBC_LOGGING(dbc)) {
 		rectype = PUTPAIR;
 		k.flags = 0;
 		d.flags = 0;
 		if (HPAGE_PTYPE(
-		    H_PAIRKEY(hcp->page, hcp->indx)) == H_OFFPAGE) {
+		    H_PAIRKEY(dbp, hcp->page, hcp->indx)) == H_OFFPAGE) {
 			rectype |= PAIR_KEYMASK;
-			k.data = H_PAIRKEY(hcp->page, hcp->indx);
+			k.data = H_PAIRKEY(dbp, hcp->page, hcp->indx);
 			k.size = HOFFPAGE_SIZE;
 		} else {
 			k.data =
-			    HKEYDATA_DATA(H_PAIRKEY(hcp->page, hcp->indx));
-			k.size = LEN_HKEY(hcp->page, dbp->pgsize, hcp->indx);
+			    HKEYDATA_DATA(H_PAIRKEY(dbp, hcp->page, hcp->indx));
+			k.size =
+			    LEN_HKEY(dbp, hcp->page, dbp->pgsize, hcp->indx);
 		}
 
 		if (HPAGE_PTYPE(hk) == H_OFFPAGE) {
 			rectype |= PAIR_DATAMASK;
-			d.data = H_PAIRDATA(hcp->page, hcp->indx);
+			d.data = H_PAIRDATA(dbp, hcp->page, hcp->indx);
 			d.size = HOFFPAGE_SIZE;
 		} else {
-			if (HPAGE_PTYPE(H_PAIRDATA(hcp->page, hcp->indx))
-			    == H_DUPLICATE)
+			if (HPAGE_PTYPE(H_PAIRDATA(dbp,
+			    hcp->page, hcp->indx)) == H_DUPLICATE)
 				rectype |= PAIR_DUPMASK;
-			d.data =
-			    HKEYDATA_DATA(H_PAIRDATA(hcp->page, hcp->indx));
-			d.size = LEN_HDATA(hcp->page, dbp->pgsize, hcp->indx);
+			d.data = HKEYDATA_DATA(
+			    H_PAIRDATA(dbp, hcp->page, hcp->indx));
+			d.size = LEN_HDATA(dbp, hcp->page,
+			    dbp->pgsize, hcp->indx);
 		}
 
-		if ((ret = __ham_insdel_log(dbp->dbenv,
-		    dbc->txn, &new_lsn, 0, rectype,
-		    dbp->log_fileid, PGNO(next_pagep),
+		if ((ret = __ham_insdel_log(dbp,
+		    dbc->txn, &new_lsn, 0, rectype, PGNO(next_pagep),
 		    (u_int32_t)NUM_ENT(next_pagep), &LSN(next_pagep),
 		    &k, &d)) != 0) {
-			(void)mpf->put(mpf, next_pagep, 0);
+			(void)__memp_fput(mpf, next_pagep, 0);
 			return (ret);
 		}
 	} else
 		LSN_NOT_LOGGED(new_lsn);
 
 	/* Move lsn onto page. */
+	if ((ret = __memp_dirty(mpf, &next_pagep, dbc->txn, 0)) != 0) {
+		(void)__memp_fput(mpf, next_pagep, 0);
+		return (ret);
+	}
 	LSN(next_pagep) = new_lsn;	/* Structure assignment. */
 
-	__ham_copy_item(dbp->pgsize,
-	    hcp->page, H_KEYINDEX(hcp->indx), next_pagep);
-	__ham_copy_item(dbp->pgsize,
-	    hcp->page, H_DATAINDEX(hcp->indx), next_pagep);
-
-	/*
-	 * We've just manually inserted a key and set of data onto
-	 * next_pagep;  however, it's possible that our caller will
-	 * return without further modifying the new page, for instance
-	 * if DB_NODUPDATA is set and our new item is a duplicate duplicate.
-	 * Thus, to be on the safe side, we need to mark the page dirty
-	 * here. [#2996]
-	 *
-	 * Note that __ham_del_pair should dirty the page we're moving
-	 * the items from, so we need only dirty the new page ourselves.
-	 */
-	if ((ret = mpf->set(mpf, next_pagep, DB_MPOOL_DIRTY)) != 0)
-		goto out;
+	__ham_copy_item(dbp, hcp->page, H_KEYINDEX(hcp->indx), next_pagep);
+	__ham_copy_item(dbp, hcp->page, H_DATAINDEX(hcp->indx), next_pagep);
 
 	/* Update all cursors that used to point to this item. */
 	if ((ret = __ham_c_chgpg(dbc, PGNO(hcp->page), H_KEYINDEX(hcp->indx),
@@ -613,12 +602,17 @@ __ham_check_move(dbc, add_len)
 	 * __ham_del_pair decremented nelem.  This is incorrect;  we
 	 * manually copied the element elsewhere, so the total number
 	 * of elements hasn't changed.  Increment it again.
+	 *
+	 * !!!
+	 * Note that we still have the metadata page pinned, and
+	 * __ham_del_pair dirtied it, so we don't need to set the dirty
+	 * flag again.
 	 */
 	if (!STD_LOCKING(dbc))
 		hcp->hdr->nelem++;
 
-out:
-	(void)mpf->put(mpf, hcp->page, DB_MPOOL_DIRTY);
+out:	if ((t_ret = __memp_fput(mpf, hcp->page, 0)) != 0 && ret == 0)
+		ret = t_ret;
 	hcp->page = next_pagep;
 	hcp->pgno = PGNO(hcp->page);
 	hcp->indx = NUM_ENT(hcp->page) - 2;
@@ -649,8 +643,8 @@ __ham_move_offpage(dbc, pagep, ndx, pgno)
 	DBT new_dbt;
 	DBT old_dbt;
 	HOFFDUP od;
-	db_indx_t i;
-	int32_t shrink;
+	db_indx_t i, *inp;
+	int32_t difflen;
 	u_int8_t *src;
 	int ret;
 
@@ -662,34 +656,40 @@ __ham_move_offpage(dbc, pagep, ndx, pgno)
 	od.pgno = pgno;
 	ret = 0;
 
-	if (DB_LOGGING(dbc)) {
+	if (DBC_LOGGING(dbc)) {
 		new_dbt.data = &od;
 		new_dbt.size = HOFFDUP_SIZE;
-		old_dbt.data = P_ENTRY(pagep, ndx);
-		old_dbt.size = LEN_HITEM(pagep, dbp->pgsize, ndx);
-		if ((ret = __ham_replace_log(dbp->dbenv,
-		    dbc->txn, &LSN(pagep), 0, dbp->log_fileid,
+		old_dbt.data = P_ENTRY(dbp, pagep, ndx);
+		old_dbt.size = LEN_HITEM(dbp, pagep, dbp->pgsize, ndx);
+		if ((ret = __ham_replace_log(dbp, dbc->txn, &LSN(pagep), 0,
 		    PGNO(pagep), (u_int32_t)ndx, &LSN(pagep), -1,
 		    &old_dbt, &new_dbt, 0)) != 0)
 			return (ret);
 	} else
 		LSN_NOT_LOGGED(LSN(pagep));
 
-	shrink = LEN_HITEM(pagep, dbp->pgsize, ndx) - HOFFDUP_SIZE;
-
-	if (shrink != 0) {
+	/*
+	 * difflen is the difference in the lengths, and so may be negative.
+	 * We know that the difference between two unsigned lengths from a
+	 * database page will fit into an int32_t.
+	 */
+	difflen =
+	    (int32_t)LEN_HITEM(dbp, pagep, dbp->pgsize, ndx) -
+	    (int32_t)HOFFDUP_SIZE;
+	if (difflen != 0) {
 		/* Copy data. */
+		inp = P_INP(dbp, pagep);
 		src = (u_int8_t *)(pagep) + HOFFSET(pagep);
-		memmove(src + shrink, src, pagep->inp[ndx] - HOFFSET(pagep));
-		HOFFSET(pagep) += shrink;
+		memmove(src + difflen, src, inp[ndx] - HOFFSET(pagep));
+		HOFFSET(pagep) += difflen;
 
 		/* Update index table. */
 		for (i = ndx; i < NUM_ENT(pagep); i++)
-			pagep->inp[i] += shrink;
+			inp[i] += difflen;
 	}
 
 	/* Now copy the offdup entry onto the page. */
-	memcpy(P_ENTRY(pagep, ndx), &od, HOFFDUP_SIZE);
+	memcpy(P_ENTRY(dbp, pagep, ndx), &od, HOFFDUP_SIZE);
 	return (ret);
 }
 
@@ -720,13 +720,13 @@ __ham_dsearch(dbc, dbt, offp, cmpp, flags)
 	func = dbp->dup_compare == NULL ? __bam_defcmp : dbp->dup_compare;
 
 	i = F_ISSET(hcp, H_CONTINUE) ? hcp->dup_off: 0;
-	data = HKEYDATA_DATA(H_PAIRDATA(hcp->page, hcp->indx)) + i;
-	hcp->dup_tlen = LEN_HDATA(hcp->page, dbp->pgsize, hcp->indx);
+	data = HKEYDATA_DATA(H_PAIRDATA(dbp, hcp->page, hcp->indx)) + i;
+	hcp->dup_tlen = LEN_HDATA(dbp, hcp->page, dbp->pgsize, hcp->indx);
+	len = hcp->dup_len;
 	while (i < hcp->dup_tlen) {
 		memcpy(&len, data, sizeof(db_indx_t));
 		data += sizeof(db_indx_t);
-		cur.data = data;
-		cur.size = (u_int32_t)len;
+		DB_SET_DBT(cur, data, len);
 
 		/*
 		 * If we find an exact match, we're done.  If in a sorted
@@ -753,30 +753,6 @@ __ham_dsearch(dbc, dbt, offp, cmpp, flags)
 	F_SET(hcp, H_ISDUP);
 }
 
-#ifdef DEBUG
-/*
- * __ham_cprint --
- *	Display the current cursor list.
- *
- * PUBLIC: void __ham_cprint __P((DBC *));
- */
-void
-__ham_cprint(dbc)
-	DBC *dbc;
-{
-	HASH_CURSOR *cp;
-
-	cp = (HASH_CURSOR *)dbc->internal;
-
-	fprintf(stderr, "%#0lx->%#0lx: page: %lu index: %lu",
-	    P_TO_ULONG(dbc), P_TO_ULONG(cp), (u_long)cp->pgno,
-	    (u_long)cp->indx);
-	if (F_ISSET(cp, H_DELETED))
-		fprintf(stderr, " (deleted)");
-	fprintf(stderr, "\n");
-}
-#endif /* DEBUG */
-
 /*
  * __ham_dcursor --
  *
@@ -789,17 +765,17 @@ __ham_dcursor(dbc, pgno, indx)
 	u_int32_t indx;
 {
 	DB *dbp;
-	DBC *dbc_nopd;
 	HASH_CURSOR *hcp;
 	BTREE_CURSOR *dcp;
 	int ret;
 
 	dbp = dbc->dbp;
+	hcp = (HASH_CURSOR *)dbc->internal;
 
-	if ((ret = __db_c_newopd(dbc, pgno, &dbc_nopd)) != 0)
+	if ((ret = __db_c_newopd(dbc, pgno, hcp->opd, &hcp->opd)) != 0)
 		return (ret);
 
-	dcp = (BTREE_CURSOR *)dbc_nopd->internal;
+	dcp = (BTREE_CURSOR *)hcp->opd->internal;
 	dcp->pgno = pgno;
 	dcp->indx = indx;
 
@@ -816,15 +792,11 @@ __ham_dcursor(dbc, pgno, indx)
 	 * Transfer the deleted flag from the top-level cursor to the
 	 * created one.
 	 */
-	hcp = (HASH_CURSOR *)dbc->internal;
 	if (F_ISSET(hcp, H_DELETED)) {
 		F_SET(dcp, C_DELETED);
 		F_CLR(hcp, H_DELETED);
 	}
 
-	/* Stack the cursors and reset the initial cursor's index. */
-	hcp->opd = dbc_nopd;
-
 	return (0);
 }
 
@@ -855,15 +827,14 @@ __ham_c_chgpg(dbc, old_pgno, old_index, new_pgno, new_index)
 	dbenv = dbp->dbenv;
 
 	my_txn = IS_SUBTRANSACTION(dbc->txn) ? dbc->txn : NULL;
-	found = 0;
 
-	MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
-	for (ldbp = __dblist_get(dbenv, dbp->adj_fileid);
+	MUTEX_LOCK(dbenv, dbenv->mtx_dblist);
+	FIND_FIRST_DB_MATCH(dbenv, dbp, ldbp);
+	for (found = 0;
 	    ldbp != NULL && ldbp->adj_fileid == dbp->adj_fileid;
-	    ldbp = LIST_NEXT(ldbp, dblistlinks)) {
-		MUTEX_THREAD_LOCK(dbenv, dbp->mutexp);
-		for (cp = TAILQ_FIRST(&ldbp->active_queue); cp != NULL;
-		    cp = TAILQ_NEXT(cp, links)) {
+	    ldbp = TAILQ_NEXT(ldbp, dblistlinks)) {
+		MUTEX_LOCK(dbenv, dbp->mutex);
+		TAILQ_FOREACH(cp, &ldbp->active_queue, links) {
 			if (cp == dbc || cp->dbtype != DB_HASH)
 				continue;
 
@@ -877,7 +848,8 @@ __ham_c_chgpg(dbc, old_pgno, old_index, new_pgno, new_index)
 			if (F_ISSET(hcp, H_DELETED))
 				continue;
 
-			if (hcp->pgno == old_pgno) {
+			if (hcp->pgno == old_pgno &&
+			    !MVCC_SKIP_CURADJ(cp, old_pgno)) {
 				if (hcp->indx == old_index) {
 					hcp->pgno = new_pgno;
 					hcp->indx = new_index;
@@ -887,14 +859,13 @@ __ham_c_chgpg(dbc, old_pgno, old_index, new_pgno, new_index)
 					found = 1;
 			}
 		}
-		MUTEX_THREAD_UNLOCK(dbenv, dbp->mutexp);
+		MUTEX_UNLOCK(dbenv, dbp->mutex);
 	}
-	MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
+	MUTEX_UNLOCK(dbenv, dbenv->mtx_dblist);
 
-	if (found != 0 && DB_LOGGING(dbc)) {
-		if ((ret = __ham_chgpg_log(dbenv,
-		     my_txn, &lsn, 0, dbp->log_fileid, DB_HAM_CHGPG,
-		     old_pgno, new_pgno, old_index, new_index)) != 0)
+	if (found != 0 && DBC_LOGGING(dbc)) {
+		if ((ret = __ham_chgpg_log(dbp, my_txn, &lsn, 0, DB_HAM_CHGPG,
+		    old_pgno, new_pgno, old_index, new_index)) != 0)
 			return (ret);
 	}
 	return (0);
