@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996, 1997, 1998, 1999, 2000
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1990, 1993
@@ -38,27 +38,30 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
+ *
+ * $Id: hsearch.c,v 12.7 2006/08/24 14:46:06 bostic Exp $
  */
 
+#define	DB_DBM_HSEARCH	1
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: hsearch.c,v 11.4 2000/02/14 02:59:53 bostic Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <errno.h>
-#include <string.h>
-#endif
-
-#define	DB_DBM_HSEARCH	1
 #include "db_int.h"
 
 static DB	*dbp;
 static ENTRY	 retval;
 
+/*
+ * Translate HSEARCH calls into DB calls so that DB doesn't step on the
+ * application's name space.
+ *
+ * EXTERN: #if DB_DBM_HSEARCH != 0
+ *
+ * EXTERN: int __db_hcreate __P((size_t));
+ * EXTERN: ENTRY *__db_hsearch __P((ENTRY, ACTION));
+ * EXTERN: void __db_hdestroy __P((void));
+ *
+ * EXTERN: #endif
+ */
 int
 __db_hcreate(nel)
 	size_t nel;
@@ -72,9 +75,9 @@ __db_hcreate(nel)
 
 	if ((ret = dbp->set_pagesize(dbp, 512)) != 0 ||
 	    (ret = dbp->set_h_ffactor(dbp, 16)) != 0 ||
-	    (ret = dbp->set_h_nelem(dbp, nel)) != 0 ||
-	    (ret = dbp->open(dbp,
-	    NULL, NULL, DB_HASH, DB_CREATE, __db_omode("rw----"))) != 0)
+	    (ret = dbp->set_h_nelem(dbp, (u_int32_t)nel)) != 0 ||
+	    (ret = dbp->open(dbp, NULL,
+	    NULL, NULL, DB_HASH, DB_CREATE, __db_omode(OWNER_RW))) != 0)
 		__os_set_errno(ret);
 
 	/*
@@ -96,15 +99,12 @@ __db_hsearch(item, action)
 		__os_set_errno(EINVAL);
 		return (NULL);
 	}
-	memset(&key, 0, sizeof(key));
+	DB_INIT_DBT(key, item.key, strlen(item.key) + 1);
 	memset(&val, 0, sizeof(val));
-	key.data = item.key;
-	key.size = strlen(item.key) + 1;
 
 	switch (action) {
 	case ENTER:
-		val.data = item.data;
-		val.size = strlen(item.data) + 1;
+		DB_SET_DBT(val, item.data, strlen(item.data) + 1);
 
 		/*
 		 * Try and add the key to the database.  If we fail because
