@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2000-2004
- *      Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2000-2006
+ *      Oracle Corporation.  All rights reserved.
  *
- * $Id: DataView.java,v 1.4 2004/08/02 18:52:05 mjc Exp $
+ * $Id: DataView.java,v 12.4 2006/08/31 18:14:07 bostic Exp $
  */
 
 package com.sleepycat.collections;
@@ -12,6 +12,7 @@ package com.sleepycat.collections;
 import com.sleepycat.bind.EntityBinding;
 import com.sleepycat.bind.EntryBinding;
 import com.sleepycat.compat.DbCompat;
+import com.sleepycat.db.CursorConfig;
 import com.sleepycat.db.Database;
 import com.sleepycat.db.DatabaseConfig;
 import com.sleepycat.db.DatabaseEntry;
@@ -24,6 +25,8 @@ import com.sleepycat.db.SecondaryDatabase;
 import com.sleepycat.db.SecondaryKeyCreator;
 import com.sleepycat.db.Transaction;
 import com.sleepycat.util.RuntimeExceptionWrapper;
+import com.sleepycat.util.keyrange.KeyRange;
+import com.sleepycat.util.keyrange.KeyRangeException;
 
 /**
  * Represents a Berkeley DB database and adds support for indices, bindings and
@@ -45,19 +48,29 @@ final class DataView implements Cloneable {
     EntityBinding entityBinding;
     PrimaryKeyAssigner keyAssigner;
     SecondaryKeyCreator secKeyCreator;
-    boolean writeAllowed;       // Read-write view
-    boolean ordered;            // Not a HASH Db
-    boolean recNumAllowed;      // QUEUE, RECNO, or BTREE-RECNUM Db
-    boolean recNumAccess;       // recNumAllowed && using a rec num binding
-    boolean btreeRecNumDb;      // BTREE-RECNUM Db
-    boolean btreeRecNumAccess;  // recNumAccess && BTREE-RECNUM Db
-    boolean recNumRenumber;     // RECNO-RENUM Db
-    boolean keysRenumbered;     // recNumRenumber || btreeRecNumAccess
-    boolean dupsAllowed;        // Dups configured
-    boolean dupsOrdered;        // Sorted dups configured
-    boolean transactional;      // Db is transactional
-    boolean dirtyReadAllowed;   // Dirty-read is optional in DB-CORE
-    boolean dirtyReadEnabled;   // This view is a dirty-ready view
+    CursorConfig cursorConfig;      // Used for all operations via this view
+    boolean writeAllowed;           // Read-write view
+    boolean ordered;                // Not a HASH Db
+    boolean recNumAllowed;          // QUEUE, RECNO, or BTREE-RECNUM Db
+    boolean recNumAccess;           // recNumAllowed && using a rec num binding
+    boolean btreeRecNumDb;          // BTREE-RECNUM Db
+    boolean btreeRecNumAccess;      // recNumAccess && BTREE-RECNUM Db
+    boolean recNumRenumber;         // RECNO-RENUM Db
+    boolean keysRenumbered;         // recNumRenumber || btreeRecNumAccess
+    boolean dupsAllowed;            // Dups configured
+    boolean dupsOrdered;            // Sorted dups configured
+    boolean transactional;          // Db is transactional
+    boolean readUncommittedAllowed; // Read-uncommited is optional in DB-CORE
+
+    /*
+     * If duplicatesView is called, dupsView will be true and dupsKey will be
+     * the secondary key used as the "single key" range.  dupRange will be set
+     * as the range of the primary key values if subRange is subsequently
+     * called, to further narrow the view.
+     */
+    DatabaseEntry dupsKey;
+    boolean dupsView;
+    KeyRange dupsRange;
 
     /**
      * Creates a view for a given database and bindings.  The initial key range
@@ -94,7 +107,7 @@ final class DataView implements Cloneable {
             dupsOrdered = DbCompat.getSortedDuplicates(dbConfig);
             transactional = currentTxn.isTxnMode() &&
                             dbConfig.getTransactional();
-            dirtyReadAllowed = DbCompat.getDirtyRead(dbConfig);
+            readUncommittedAllowed = DbCompat.getReadUncommitted(dbConfig);
             btreeRecNumDb = recNumAllowed && DbCompat.isTypeBtree(dbConfig);
             range = new KeyRange(dbConfig.getBtreeComparator());
         } catch (DatabaseException e) {
@@ -105,6 +118,7 @@ final class DataView implements Cloneable {
         this.valueBinding = valueBinding;
         this.entityBinding = entityBinding;
         this.keyAssigner = keyAssigner;
+        cursorConfig = CursorConfig.DEFAULT;
 
         if (valueBinding != null && entityBinding != null)
             throw new IllegalArgumentException(
@@ -189,7 +203,7 @@ final class DataView implements Cloneable {
          * Must do subRange before valueSetView since the latter clears the
          * key binding needed for the former.
          */
-        KeyRange singleKeyRange = subRange(singleKey);
+        KeyRange singleKeyRange = subRange(range, singleKey);
         DataView view = valueSetView();
         view.range = singleKeyRange;
         return view;
@@ -211,14 +225,35 @@ final class DataView implements Cloneable {
     }
 
     /**
-     * Returns a new view with a specified dirtyRead setting.
+     * Return a new duplicates view for a given secondary key.
      */
-    DataView dirtyReadView(boolean enable) {
+    DataView duplicatesView(Object secondaryKey,
+                            EntryBinding primaryKeyBinding)
+        throws DatabaseException, KeyRangeException {
 
-        if (!dirtyReadAllowed)
-            return this;
+        if (!isSecondary()) {
+            throw new UnsupportedOperationException
+                ("Only allowed for maps on secondary databases");
+        }
+        if (dupsView) {
+            throw new IllegalStateException();
+        }
         DataView view = cloneView();
-        view.dirtyReadEnabled = enable;
+        view.range = subRange(view.range, secondaryKey);
+        view.dupsKey = view.range.getSingleKey();
+        view.dupsView = true;
+        view.keyBinding = primaryKeyBinding;
+        return view;
+    }
+
+    /**
+     * Returns a new view with a specified cursor configuration.
+     */
+    DataView configuredView(CursorConfig config) {
+
+        DataView view = cloneView();
+        view.cursorConfig = (config != null) ?
+            DbCompat.cloneCursorConfig(config) : CursorConfig.DEFAULT;
         return view;
     }
 
@@ -238,7 +273,14 @@ final class DataView implements Cloneable {
                           Object endKey, boolean endInclusive)
         throws DatabaseException, KeyRangeException {
 
-        range = subRange(beginKey, beginInclusive, endKey, endInclusive);
+        KeyRange useRange = useSubRange();
+        useRange = subRange
+            (useRange, beginKey, beginInclusive, endKey, endInclusive);
+        if (dupsView) {
+            dupsRange = useRange;
+        } else {
+            range = useRange;
+        }
     }
 
     /**
@@ -458,7 +500,7 @@ final class DataView implements Cloneable {
                 throw new IllegalStateException(
                     "EntityBinding required to derive key from value");
             }
-            if (isSecondary()) {
+            if (!dupsView && isSecondary()) {
                 DatabaseEntry primaryKeyThang = new DatabaseEntry();
                 entityBinding.objectToKey(value, primaryKeyThang);
                 DatabaseEntry valueThang = new DatabaseEntry();
@@ -514,7 +556,7 @@ final class DataView implements Cloneable {
                     "non-null value with null value/entity binding");
             }
         } else {
-            valueThang.setData(new byte[0]);
+            valueThang.setData(KeyRange.ZERO_LENGTH_BYTE_ARRAY);
             valueThang.setOffset(0);
             valueThang.setSize(0);
         }
@@ -523,18 +565,24 @@ final class DataView implements Cloneable {
     /**
      * Converts a key entry to a key object.
      */
-    Object makeKey(DatabaseEntry keyThang)
-        throws DatabaseException {
+    Object makeKey(DatabaseEntry keyThang, DatabaseEntry priKeyThang) {
 
-        if (keyThang.getSize() == 0) return null;
-        return keyBinding.entryToObject(keyThang);
+        if (keyBinding == null) {
+            throw new UnsupportedOperationException();
+        } else {
+            DatabaseEntry thang = dupsView ? priKeyThang : keyThang;
+            if (thang.getSize() == 0) {
+                return null;
+            } else {
+                return keyBinding.entryToObject(thang);
+            }
+        }
     }
 
     /**
      * Converts a key-value entry pair to a value object.
      */
-    Object makeValue(DatabaseEntry primaryKeyThang, DatabaseEntry valueThang)
-        throws DatabaseException {
+    Object makeValue(DatabaseEntry primaryKeyThang, DatabaseEntry valueThang) {
 
         Object value;
         if (valueBinding != null) {
@@ -552,21 +600,22 @@ final class DataView implements Cloneable {
     /**
      * Intersects the given key and the current range.
      */
-    KeyRange subRange(Object singleKey)
+    KeyRange subRange(KeyRange useRange, Object singleKey)
         throws DatabaseException, KeyRangeException {
 
-        return range.subRange(makeRangeKey(singleKey));
+        return useRange.subRange(makeRangeKey(singleKey));
     }
 
     /**
      * Intersects the given range and the current range.
      */
-    KeyRange subRange(Object beginKey, boolean beginInclusive,
+    KeyRange subRange(KeyRange useRange,
+                      Object beginKey, boolean beginInclusive,
                       Object endKey, boolean endInclusive)
         throws DatabaseException, KeyRangeException {
 
         if (beginKey == endKey && beginInclusive && endInclusive) {
-            return subRange(beginKey);
+            return subRange(useRange, beginKey);
         }
         if (!ordered) {
             throw new UnsupportedOperationException(
@@ -577,8 +626,30 @@ final class DataView implements Cloneable {
         DatabaseEntry endThang =
             (endKey != null) ? makeRangeKey(endKey) : null;
 
-        return range.subRange(beginThang, beginInclusive,
-                              endThang, endInclusive);
+        return useRange.subRange(beginThang, beginInclusive,
+                                 endThang, endInclusive);
+    }
+
+    /**
+     * Returns the range to use for sub-ranges.  Returns range if this is not a
+     * dupsView, or the dupsRange if this is a dupsView, creating dupsRange if
+     * necessary.
+     */
+    KeyRange useSubRange()
+        throws DatabaseException {
+
+        if (dupsView) {
+            synchronized (this) {
+                if (dupsRange == null) {
+                    DatabaseConfig config =
+                        secDb.getPrimaryDatabase().getConfig();
+                    dupsRange = new KeyRange(config.getBtreeComparator());
+                }
+            }
+            return dupsRange;
+        } else {
+            return range;
+        }
     }
 
     /**
