@@ -1,18 +1,17 @@
-/*
- *  -
- *  See the file LICENSE for redistribution information.
+/*-
+ * See the file LICENSE for redistribution information.
  *
- *  Copyright (c) 1999-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1999-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- *  $Id: DeadlockException.java,v 1.1 2004/04/06 20:43:40 mjc Exp $
+ * $Id: DeadlockException.java,v 12.4 2006/08/24 14:46:07 bostic Exp $
  */
 package com.sleepycat.db;
 
 import com.sleepycat.db.internal.DbEnv;
 
 public class DeadlockException extends DatabaseException {
-    protected DeadlockException(final String s,
+    /* package */ DeadlockException(final String s,
                                 final int errno,
                                 final DbEnv dbenv) {
         super(s, errno, dbenv);
