@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2001-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2001-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: Lock.java,v 1.1 2004/04/06 20:43:40 mjc Exp $
+ * $Id: Lock.java,v 12.4 2006/08/24 14:46:08 bostic Exp $
  */
 
 package com.sleepycat.db;
@@ -14,9 +14,9 @@ import com.sleepycat.db.internal.DbLock;
 public final class Lock {
     private DbLock dbLock;
 
-    private Lock(final DbLock dblock) {
-        this.dbLock = dbLock;
-        dbLock.wrapper = this;
+    private Lock(final DbLock inLock) {
+        this.dbLock = inLock;
+        inLock.wrapper = this;
     }
 
     /* package */
