@@ -1,10 +1,10 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2000-2004
- *      Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2000-2006
+ *      Oracle Corporation.  All rights reserved.
  *
- * $Id: RuntimeExceptionWrapper.java,v 1.1 2004/04/09 16:34:10 mark Exp $
+ * $Id: RuntimeExceptionWrapper.java,v 12.4 2006/08/31 18:14:09 bostic Exp $
  */
 
 package com.sleepycat.util;
@@ -25,8 +25,16 @@ public class RuntimeExceptionWrapper extends RuntimeException
         this.e = e;
     }
 
+    /**
+     * @deprecated replaced by {@link #getCause}.
+     */
     public Throwable getDetail() {
 
         return e;
     }
+
+    public Throwable getCause() {
+
+        return e;
+    }
 }
