@@ -1,20 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: log_get.c,v 11.110 2004/09/17 22:00:31 mjc Exp $
+ * $Id: log_get.c,v 12.35 2006/09/07 20:05:32 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/crypto.h"
 #include "dbinc/db_page.h"
@@ -30,13 +24,14 @@ static int __log_c_get_int __P((DB_LOGC *, DB_LSN *, DBT *, u_int32_t));
 static int __log_c_hdrchk __P((DB_LOGC *, DB_LSN *, HDR *, int *));
 static int __log_c_incursor __P((DB_LOGC *, DB_LSN *, HDR *, u_int8_t **));
 static int __log_c_inregion __P((DB_LOGC *,
-	       DB_LSN *, RLOCK *, DB_LSN *, HDR *, u_int8_t **));
+	       DB_LSN *, RLOCK *, DB_LSN *, HDR *, u_int8_t **, int *));
 static int __log_c_io __P((DB_LOGC *,
 	       u_int32_t, u_int32_t, void *, size_t *, int *));
 static int __log_c_ondisk __P((DB_LOGC *,
 	       DB_LSN *, DB_LSN *, u_int32_t, HDR *, u_int8_t **, int *));
 static int __log_c_set_maxrec __P((DB_LOGC *, char *));
 static int __log_c_shortread __P((DB_LOGC *, DB_LSN *, int));
+static int __log_c_version_pp __P((DB_LOGC *, u_int32_t *, u_int32_t));
 
 /*
  * __log_cursor_pp --
@@ -50,7 +45,8 @@ __log_cursor_pp(dbenv, logcp, flags)
 	DB_LOGC **logcp;
 	u_int32_t flags;
 {
-	int rep_check, ret;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	PANIC_CHECK(dbenv);
 	ENV_REQUIRES_CONFIG(dbenv,
@@ -60,12 +56,9 @@ __log_cursor_pp(dbenv, logcp, flags)
 	if ((ret = __db_fchk(dbenv, "DB_ENV->log_cursor", flags, 0)) != 0)
 		return (ret);
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __log_cursor(dbenv, logcp);
-	if (rep_check)
-		__env_db_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__log_cursor(dbenv, logcp)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -89,7 +82,7 @@ __log_cursor(dbenv, logcp)
 	if ((ret = __os_calloc(dbenv, 1, sizeof(DB_LOGC), &logc)) != 0)
 		return (ret);
 
-	logc->bp_size = DB_LOGC_BUF_SIZE;
+	logc->bp_size = LG_CURSOR_BUF_SIZE;
 	/*
 	 * Set this to something positive.
 	 */
@@ -102,6 +95,7 @@ __log_cursor(dbenv, logcp)
 	logc->dbenv = dbenv;
 	logc->close = __log_c_close_pp;
 	logc->get = __log_c_get_pp;
+	logc->version = __log_c_version_pp;
 
 	*logcp = logc;
 	return (0);
@@ -116,8 +110,9 @@ __log_c_close_pp(logc, flags)
 	DB_LOGC *logc;
 	u_int32_t flags;
 {
+	DB_THREAD_INFO *ip;
 	DB_ENV *dbenv;
-	int rep_check, ret;
+	int ret;
 
 	dbenv = logc->dbenv;
 
@@ -125,12 +120,9 @@ __log_c_close_pp(logc, flags)
 	if ((ret = __db_fchk(dbenv, "DB_LOGC->close", flags, 0)) != 0)
 		return (ret);
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __log_c_close(logc);
-	if (rep_check)
-		__env_db_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__log_c_close(logc)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -162,6 +154,88 @@ __log_c_close(logc)
 	return (0);
 }
 
+/*
+ * __log_c_version_pp --
+ *	DB_LOGC->version.
+ */
+static int
+__log_c_version_pp(logc, versionp, flags)
+	DB_LOGC *logc;
+	u_int32_t *versionp;
+	u_int32_t flags;
+{
+	DB_ENV *dbenv;
+	DB_THREAD_INFO *ip;
+	int ret;
+
+	dbenv = logc->dbenv;
+
+	PANIC_CHECK(dbenv);
+	if ((ret = __db_fchk(dbenv, "DB_LOGC->version", flags, 0)) != 0)
+		return (ret);
+
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__log_c_version(logc, versionp)), ret);
+	ENV_LEAVE(dbenv, ip);
+	return (ret);
+}
+
+/*
+ * __log_c_version --
+ *	DB_LOGC->version.
+ *
+ * PUBLIC: int __log_c_version __P((DB_LOGC *, u_int32_t *));
+ */
+int
+__log_c_version(logc, versionp)
+	DB_LOGC *logc;
+	u_int32_t *versionp;
+{
+	DB_ENV *dbenv;
+	DB_LSN plsn;
+	DB_LOGC *plogc;
+	DBT hdrdbt;
+	LOGP *persist;
+	int ret, t_ret;
+
+	dbenv = logc->dbenv;
+	if (IS_ZERO_LSN(logc->c_lsn)) {
+		__db_errx(dbenv, "DB_LOGC->get: unset cursor");
+		return (EINVAL);
+	}
+	ret = 0;
+	/*
+	 * Check if the persist info we have is for the same file
+	 * as the current cursor position.  If we already have the
+	 * information, then we're done.  If not, we open a new
+	 * log cursor and get the header.
+	 *
+	 * Since most users walk forward through the log when
+	 * using this feature (i.e. printlog) we're likely to
+	 * have the information we need.
+	 */
+	if (logc->c_lsn.file != logc->p_lsn.file) {
+		if ((ret = __log_cursor(dbenv, &plogc)) != 0)
+			return (ret);
+		plsn.file = logc->c_lsn.file;
+		plsn.offset = 0;
+		plogc->c_lsn = plsn;
+		memset(&hdrdbt, 0, sizeof(DBT));
+		if ((ret = __log_c_get_int(plogc,
+		    &plsn, &hdrdbt, DB_SET)) == 0) {
+			persist = (LOGP *)hdrdbt.data;
+			logc->p_lsn = logc->c_lsn;
+			logc->p_version = persist->version;
+		}
+		if ((t_ret = __log_c_close(plogc)) != 0 && ret == 0)
+			ret = t_ret;
+	}
+	/* Return the version. */
+	if (ret == 0)
+		*versionp = logc->p_version;
+	return (ret);
+}
+
 /*
  * __log_c_get_pp --
  *	DB_LOGC->get pre/post processing.
@@ -174,7 +248,8 @@ __log_c_get_pp(logc, alsn, dbt, flags)
 	u_int32_t flags;
 {
 	DB_ENV *dbenv;
-	int rep_check, ret;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	dbenv = logc->dbenv;
 
@@ -190,7 +265,7 @@ __log_c_get_pp(logc, alsn, dbt, flags)
 		break;
 	case DB_SET:
 		if (IS_ZERO_LSN(*alsn)) {
-			__db_err(dbenv, "DB_LOGC->get: invalid LSN: %lu/%lu",
+			__db_errx(dbenv, "DB_LOGC->get: invalid LSN: %lu/%lu",
 			    (u_long)alsn->file, (u_long)alsn->offset);
 			return (EINVAL);
 		}
@@ -199,12 +274,9 @@ __log_c_get_pp(logc, alsn, dbt, flags)
 		return (__db_ferr(dbenv, "DB_LOGC->get", 1));
 	}
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __log_c_get(logc, alsn, dbt, flags);
-	if (rep_check)
-		__env_db_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__log_c_get(logc, alsn, dbt, flags)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -223,6 +295,7 @@ __log_c_get(logc, alsn, dbt, flags)
 {
 	DB_ENV *dbenv;
 	DB_LSN saved_lsn;
+	LOGP *persist;
 	int ret;
 
 	dbenv = logc->dbenv;
@@ -242,7 +315,6 @@ __log_c_get(logc, alsn, dbt, flags)
 	 * put in there.
 	 */
 	saved_lsn = *alsn;
-
 	/*
 	 * If we get one of the log's header records as a result of doing a
 	 * DB_FIRST, DB_NEXT, DB_LAST or DB_PREV, repeat the operation, log
@@ -266,6 +338,13 @@ __log_c_get(logc, alsn, dbt, flags)
 		default:
 			break;
 		}
+		/*
+		 * If we're walking the log and we find a persist header
+		 * then store so that we may use it later if needed.
+		 */
+		persist = (LOGP *)dbt->data;
+		logc->p_lsn = *alsn;
+		logc->p_version = persist->version;
 		if (F_ISSET(dbt, DB_DBT_MALLOC)) {
 			__os_free(dbenv, dbt->data);
 			dbt->data = NULL;
@@ -300,9 +379,10 @@ __log_c_get_int(logc, alsn, dbt, flags)
 	logfile_validity status;
 	u_int32_t cnt;
 	u_int8_t *rp;
-	int eof, is_hmac, ret;
+	int eof, is_hmac, need_cksum, ret;
 
 	dbenv = logc->dbenv;
+	db_cipher = dbenv->crypto_handle;
 	dblp = dbenv->lg_handle;
 	lp = dblp->reginfo.primary;
 	is_hmac = 0;
@@ -346,7 +426,7 @@ __log_c_get_int(logc, alsn, dbt, flags)
 		 */
 		switch (status) {
 		case DB_LV_INCOMPLETE:
-			DB_ASSERT(lp->lsn.file == cnt);
+			DB_ASSERT(dbenv, lp->lsn.file == cnt);
 			/* FALLTHROUGH */
 		case DB_LV_NORMAL:
 		case DB_LV_OLD_READABLE:
@@ -354,11 +434,11 @@ __log_c_get_int(logc, alsn, dbt, flags)
 			break;
 		case DB_LV_NONEXISTENT:
 			nlsn.file = 1;
-			DB_ASSERT(lp->lsn.file == nlsn.file);
+			DB_ASSERT(dbenv, lp->lsn.file == nlsn.file);
 			break;
 		case DB_LV_OLD_UNREADABLE:
 			nlsn.file = cnt + 1;
-			DB_ASSERT(lp->lsn.file == nlsn.file);
+			DB_ASSERT(dbenv, lp->lsn.file == nlsn.file);
 			break;
 		}
 		nlsn.offset = 0;
@@ -375,7 +455,7 @@ __log_c_get_int(logc, alsn, dbt, flags)
 				}
 				if ((!lp->db_log_inmemory &&
 				    (__log_valid(dblp, nlsn.file - 1, 0, NULL,
-				    0, &status) != 0 ||
+				    0, &status, NULL) != 0 ||
 				    (status != DB_LV_NORMAL &&
 				    status != DB_LV_OLD_READABLE)))) {
 					ret = DB_NOTFOUND;
@@ -391,7 +471,7 @@ __log_c_get_int(logc, alsn, dbt, flags)
 	case DB_LAST:				/* Last log record. */
 		if (rlock == L_NONE) {
 			rlock = L_ACQUIRED;
-			R_LOCK(dbenv, &dblp->reginfo);
+			LOG_SYSTEM_LOCK(dbenv);
 		}
 		nlsn.file = lp->lsn.file;
 		nlsn.offset = lp->lsn.offset - lp->len;
@@ -400,8 +480,7 @@ __log_c_get_int(logc, alsn, dbt, flags)
 		nlsn = *alsn;
 		break;
 	default:
-		DB_ASSERT(0);
-		ret = EINVAL;
+		ret = __db_unknown_path(dbenv, "__log_c_get_int");
 		goto err;
 	}
 
@@ -422,7 +501,11 @@ next_file:	++nlsn.file;
 		hdr.size = HDR_NORMAL_SZ;
 		is_hmac = 0;
 	}
-	/* Check to see if the record is in the cursor's buffer. */
+
+	/*
+	 * Check to see if the record is in the cursor's buffer -- if so,
+	 * we'll need to checksum it.
+	 */
 	if ((ret = __log_c_incursor(logc, &nlsn, &hdr, &rp)) != 0)
 		goto err;
 	if (rp != NULL)
@@ -434,7 +517,7 @@ next_file:	++nlsn.file;
 	 * buffer.  Else, check the region's buffer.
 	 *
 	 * If the record isn't in the region's buffer, then either logs are
-	 * in-memory, and we're done, or  we're going to have to read the
+	 * in-memory, and we're done, or we're going to have to read the
 	 * record from disk.  We want to make a point of not reading past the
 	 * end of the logical log (after recovery, there may be data after the
 	 * end of the logical log, not to mention the log file may have been
@@ -445,14 +528,22 @@ next_file:	++nlsn.file;
 	 */
 	ZERO_LSN(last_lsn);
 	if (!F_ISSET(logc, DB_LOG_DISK) ||
-	    log_compare(&nlsn, &logc->c_lsn) > 0) {
+	    LOG_COMPARE(&nlsn, &logc->c_lsn) > 0) {
 		F_CLR(logc, DB_LOG_DISK);
 
 		if ((ret = __log_c_inregion(logc,
-		    &nlsn, &rlock, &last_lsn, &hdr, &rp)) != 0)
+		    &nlsn, &rlock, &last_lsn, &hdr, &rp, &need_cksum)) != 0)
 			goto err;
-		if (rp != NULL)
-			goto cksum;
+		if (rp != NULL) {
+			/*
+			 * If we read the entire record from the in-memory log
+			 * buffer, we don't need to checksum it, nor do we need
+			 * to worry about vtruncate issues.
+			 */
+			if (need_cksum)
+				goto cksum;
+			goto from_memory;
+		}
 		if (lp->db_log_inmemory)
 			goto nohdr;
 	}
@@ -467,7 +558,7 @@ next_file:	++nlsn.file;
 	 */
 	if (rlock == L_ACQUIRED) {
 		rlock = L_NONE;
-		R_UNLOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_UNLOCK(dbenv);
 	}
 	if ((ret = __log_c_ondisk(
 	    logc, &nlsn, &last_lsn, flags, &hdr, &rp, &eof)) != 0)
@@ -485,13 +576,13 @@ next_file:	++nlsn.file;
 
 cksum:	/*
 	 * Discard the region lock if we're still holding it.  (The path to
-	 * get here is that we acquired the lock because of the caller's
-	 * flag argument, but we found the record in the cursor's buffer.
-	 * Improbable, but it's easy to avoid.
+	 * get here is we acquired the region lock because of the caller's
+	 * flag argument, but we found the record in the in-memory or cursor
+	 * buffers.  Improbable, but it's easy to avoid.)
 	 */
 	if (rlock == L_ACQUIRED) {
 		rlock = L_NONE;
-		R_UNLOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_UNLOCK(dbenv);
 	}
 
 	/*
@@ -500,17 +591,16 @@ cksum:	/*
 	 * OK if we're searching for the end of the log, and very, very bad
 	 * if we're reading random log records.
 	 */
-	db_cipher = dbenv->crypto_handle;
-	if ((ret = __db_check_chksum(dbenv, db_cipher,
+	if ((ret = __db_check_chksum(dbenv, &hdr, db_cipher,
 	    hdr.chksum, rp + hdr.size, hdr.len - hdr.size, is_hmac)) != 0) {
 		if (F_ISSET(logc, DB_LOG_SILENT_ERR)) {
 			if (ret == 0 || ret == -1)
 				ret = EIO;
 		} else if (ret == -1) {
-			__db_err(dbenv,
+			__db_errx(dbenv,
 		    "DB_LOGC->get: log record LSN %lu/%lu: checksum mismatch",
 			    (u_long)nlsn.file, (u_long)nlsn.offset);
-			__db_err(dbenv,
+			__db_errx(dbenv,
 		    "DB_LOGC->get: catastrophic recovery may be required");
 			ret = __db_panic(dbenv, DB_RUNRECOVERY);
 		}
@@ -538,9 +628,8 @@ nohdr:		switch (flags) {
 			 * the first record in that new file should have its
 			 * prev field set correctly.
 			 */
-			__db_err(dbenv,
+			__db_errx(dbenv,
 		"Encountered zero length records while traversing backwards");
-			DB_ASSERT(0);
 			ret = __db_panic(dbenv, DB_RUNRECOVERY);
 			goto err;
 		case DB_SET:
@@ -550,6 +639,18 @@ nohdr:		switch (flags) {
 		}
 	}
 
+from_memory:
+	/*
+	 * Discard the region lock if we're still holding it.  (The path to
+	 * get here is we acquired the region lock because of the caller's
+	 * flag argument, but we found the record in the in-memory or cursor
+	 * buffers.  Improbable, but it's easy to avoid.)
+	 */
+	if (rlock == L_ACQUIRED) {
+		rlock = L_NONE;
+		LOG_SYSTEM_UNLOCK(dbenv);
+	}
+
 	/* Copy the record into the user's DBT. */
 	if ((ret = __db_retcopy(dbenv, dbt, rp + hdr.size,
 	    (u_int32_t)(hdr.len - hdr.size),
@@ -579,7 +680,7 @@ nohdr:		switch (flags) {
 	logc->c_prev = hdr.prev;
 
 err:	if (rlock == L_ACQUIRED)
-		R_UNLOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_UNLOCK(dbenv);
 
 	return (ret);
 }
@@ -644,12 +745,13 @@ __log_c_incursor(logc, lsn, hdr, pp)
  *	Check to see if the requested record is in the region's buffer.
  */
 static int
-__log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
+__log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp, need_cksump)
 	DB_LOGC *logc;
 	DB_LSN *lsn, *last_lsn;
 	RLOCK *rlockp;
 	HDR *hdr;
 	u_int8_t **pp;
+	int *need_cksump;
 {
 	DB_ENV *dbenv;
 	DB_LOG *dblp;
@@ -661,16 +763,17 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 
 	dbenv = logc->dbenv;
 	dblp = dbenv->lg_handle;
-	lp = ((DB_LOG *)logc->dbenv->lg_handle)->reginfo.primary;
+	lp = dbenv->lg_handle->reginfo.primary;
 
 	ret = 0;
 	b_region = 0;
 	*pp = NULL;
+	*need_cksump = 0;
 
 	/* If we haven't yet acquired the log region lock, do so. */
 	if (*rlockp == L_NONE) {
 		*rlockp = L_ACQUIRED;
-		R_LOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_LOCK(dbenv);
 	}
 
 	/*
@@ -711,12 +814,12 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	 */
 	if (IS_ZERO_LSN(lp->lsn))
 		return (0);
-	if (log_compare(lsn, &lp->lsn) >= 0)
+	if (LOG_COMPARE(lsn, &lp->lsn) >= 0)
 		return (DB_NOTFOUND);
 	else if (lp->db_log_inmemory) {
 		if ((ret = __log_inmem_lsnoff(dblp, lsn, &b_region)) != 0)
 			return (ret);
-	} else if (lp->b_off == 0 || log_compare(lsn, &lp->f_lsn) < 0)
+	} else if (lp->b_off == 0 || LOG_COMPARE(lsn, &lp->f_lsn) < 0)
 		return (0);
 
 	/*
@@ -737,10 +840,10 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	 *
 	 * There is one case where the header check can fail: on a scan through
 	 * in-memory logs, when we reach the end of a file we can read an empty
-	 * heady.  In that case, it's safe to return zero, here: it will be
+	 * header.  In that case, it's safe to return zero, here: it will be
 	 * caught in our caller.  Otherwise, the LSN is bogus.  Fail hard.
 	 */
-	if (lp->db_log_inmemory || log_compare(lsn, &lp->f_lsn) > 0) {
+	if (lp->db_log_inmemory || LOG_COMPARE(lsn, &lp->f_lsn) > 0) {
 		if (!lp->db_log_inmemory)
 			b_region = lsn->offset - lp->w_off;
 		__log_inmem_copyout(dblp, b_region, hdr, hdr->size);
@@ -754,7 +857,7 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 		} else if (lsn->offset + hdr->len > lp->w_off + lp->buffer_size)
 			return (DB_NOTFOUND);
 		if (logc->bp_size <= hdr->len) {
-			len = (size_t)DB_ALIGN(hdr->len * 2, 128);
+			len = (size_t)DB_ALIGN((uintmax_t)hdr->len * 2, 128);
 			if ((ret =
 			    __os_realloc(logc->dbenv, len, &logc->bp)) != 0)
 				 return (ret);
@@ -765,7 +868,7 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 		return (0);
 	}
 
-	DB_ASSERT(!lp->db_log_inmemory);
+	DB_ASSERT(dbenv, !lp->db_log_inmemory);
 
 	/*
 	 * There's a partial record, that is, the requested record starts
@@ -803,7 +906,7 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	 * of waiting.
 	 */
 	if (logc->bp_size <= b_region + b_disk) {
-		len = (size_t)DB_ALIGN((b_region + b_disk) * 2, 128);
+		len = (size_t)DB_ALIGN((uintmax_t)(b_region + b_disk) * 2, 128);
 		if ((ret = __os_realloc(logc->dbenv, len, &logc->bp)) != 0)
 			return (ret);
 		logc->bp_size = (u_int32_t)len;
@@ -816,7 +919,7 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 	/* Release the region lock. */
 	if (*rlockp == L_ACQUIRED) {
 		*rlockp = L_NONE;
-		R_UNLOCK(dbenv, &dblp->reginfo);
+		LOG_SYSTEM_UNLOCK(dbenv);
 	}
 
 	/*
@@ -831,6 +934,9 @@ __log_c_inregion(logc, lsn, rlockp, last_lsn, hdr, pp)
 			return (ret);
 		if (nr < b_disk)
 			return (__log_c_shortread(logc, lsn, 0));
+
+		/* We read bytes from the disk, we'll need to checksum them. */
+		*need_cksump = 1;
 	}
 
 	/* Copy the header information into the caller's structure. */
@@ -898,7 +1004,7 @@ __log_c_ondisk(logc, lsn, last_lsn, flags, hdr, pp, eofp)
 	 * Make sure we have enough space.
 	 */
 	if (logc->bp_size <= hdr->len) {
-		len = (size_t)DB_ALIGN(hdr->len * 2, 128);
+		len = (size_t)DB_ALIGN((uintmax_t)hdr->len * 2, 128);
 		if ((ret = __os_realloc(dbenv, len, &logc->bp)) != 0)
 			return (ret);
 		logc->bp_size = (u_int32_t)len;
@@ -1012,7 +1118,7 @@ __log_c_hdrchk(logc, lsn, hdr, eofp)
 		 * check information was wrong.  Get it again, to be sure.
 		 */
 		if ((ret = __log_c_set_maxrec(logc, NULL)) != 0) {
-			__db_err(dbenv, "DB_LOGC->get: %s", db_strerror(ret));
+			__db_err(dbenv, ret, "DB_LOGC->get");
 			return (ret);
 		}
 		if (logc->bp_maxrec != 0 && hdr->len > logc->bp_maxrec)
@@ -1021,7 +1127,7 @@ __log_c_hdrchk(logc, lsn, hdr, eofp)
 	return (0);
 
 err:	if (!F_ISSET(logc, DB_LOG_SILENT_ERR))
-		__db_err(dbenv,
+		__db_errx(dbenv,
 		    "DB_LOGC->get: LSN %lu/%lu: invalid log record header",
 		    (u_long)lsn->file, (u_long)lsn->offset);
 	return (EIO);
@@ -1041,11 +1147,13 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 {
 	DB_ENV *dbenv;
 	DB_LOG *dblp;
+	LOG *lp;
 	int ret;
 	char *np;
 
 	dbenv = logc->dbenv;
 	dblp = dbenv->lg_handle;
+	lp = dblp->reginfo.primary;
 
 	/*
 	 * If we've switched files, discard the current file handle and acquire
@@ -1070,15 +1178,13 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 				*eofp = 1;
 				ret = 0;
 			} else if (!F_ISSET(logc, DB_LOG_SILENT_ERR))
-				__db_err(dbenv, "DB_LOGC->get: %s: %s",
-				    np, db_strerror(ret));
+				__db_err(dbenv, ret, "DB_LOGC->get: %s", np);
 			__os_free(dbenv, np);
 			return (ret);
 		}
 
 		if ((ret = __log_c_set_maxrec(logc, np)) != 0) {
-			__db_err(dbenv,
-			    "DB_LOGC->get: %s: %s", np, db_strerror(ret));
+			__db_err(dbenv, ret, "DB_LOGC->get: %s", np);
 			__os_free(dbenv, np);
 			return (ret);
 		}
@@ -1087,22 +1193,14 @@ __log_c_io(logc, fnum, offset, p, nrp, eofp)
 		logc->bp_lsn.file = fnum;
 	}
 
-	/* Seek to the record's offset. */
-	if ((ret = __os_seek(dbenv,
-	    logc->c_fhp, 0, 0, offset, 0, DB_OS_SEEK_SET)) != 0) {
+	++lp->stat.st_rcount;
+	/* Seek to the record's offset and read the data. */
+	if ((ret = __os_io(dbenv, DB_IO_READ,
+	    logc->c_fhp, 0, 0, offset, (u_int32_t)*nrp, p, nrp)) != 0) {
 		if (!F_ISSET(logc, DB_LOG_SILENT_ERR))
-			__db_err(dbenv,
-			    "DB_LOGC->get: LSN: %lu/%lu: seek: %s",
-			    (u_long)fnum, (u_long)offset, db_strerror(ret));
-		return (ret);
-	}
-
-	/* Read the data. */
-	if ((ret = __os_read(dbenv, logc->c_fhp, p, *nrp, nrp)) != 0) {
-		if (!F_ISSET(logc, DB_LOG_SILENT_ERR))
-			__db_err(dbenv,
-			    "DB_LOGC->get: LSN: %lu/%lu: read: %s",
-			    (u_long)fnum, (u_long)offset, db_strerror(ret));
+			__db_err(dbenv, ret,
+			    "DB_LOGC->get: LSN: %lu/%lu: read",
+			    (u_long)fnum, (u_long)offset);
 		return (ret);
 	}
 
@@ -1120,7 +1218,7 @@ __log_c_shortread(logc, lsn, check_silent)
 	int check_silent;
 {
 	if (!check_silent || !F_ISSET(logc, DB_LOG_SILENT_ERR))
-		__db_err(logc->dbenv, "DB_LOGC->get: LSN: %lu/%lu: short read",
+		__db_errx(logc->dbenv, "DB_LOGC->get: LSN: %lu/%lu: short read",
 		    (u_long)lsn->file, (u_long)lsn->offset);
 	return (EIO);
 }
@@ -1178,3 +1276,132 @@ __log_c_set_maxrec(logc, np)
 
 	return (0);
 }
+
+#ifdef HAVE_REPLICATION
+/*
+ * __log_rep_split --
+ *	- Split a log buffer into individual records.
+ *
+ * This is used by a replication client to process a bulk log message from the
+ * master and convert it into individual __rep_apply requests.
+ *
+ * PUBLIC: int __log_rep_split __P((DB_ENV *, REP_CONTROL *, DBT *, DB_LSN *,
+ * PUBLIC:     DB_LSN *));
+ */
+int
+__log_rep_split(dbenv, rp, rec, ret_lsnp, last_lsnp)
+	DB_ENV *dbenv;
+	REP_CONTROL *rp;
+	DBT *rec;
+	DB_LSN *ret_lsnp;
+	DB_LSN *last_lsnp;
+{
+	DB_LSN save_lsn, tmp_lsn;
+	DBT logrec;
+	REP_CONTROL tmprp;
+	u_int32_t len;
+	int is_dup, is_perm, ret, save_ret;
+	u_int8_t *p, *ep;
+#ifdef DIAGNOSTIC
+	DB_MSGBUF mb;
+#endif
+
+	memset(&logrec, 0, sizeof(logrec));
+	memset(&save_lsn, 0, sizeof(save_lsn));
+	memset(&tmp_lsn, 0, sizeof(tmp_lsn));
+	/*
+	 * We're going to be modifying the rp LSN contents so make
+	 * our own private copy to play with.
+	 */
+	memcpy(&tmprp, rp, sizeof(tmprp));
+	/*
+	 * We send the bulk buffer on a PERM record, so often we will have
+	 * DB_LOG_PERM set.  However, we only want to mark the last LSN
+	 * we have as a PERM record.  So clear it here, and when we're on
+	 * the last record below, set it.
+	 */
+	is_perm = F_ISSET(rp, REPCTL_PERM);
+	F_CLR(&tmprp, REPCTL_PERM);
+	ret = save_ret = 0;
+	for (ep = (u_int8_t *)rec->data + rec->size, p = (u_int8_t *)rec->data;
+	    p < ep; ) {
+		/*
+		 * First thing in the buffer is the length.  Then the LSN
+		 * of this record, then the record itself.
+		 */
+		/*
+		 * XXX
+		 * If/when we add architecture neutral log files we may want
+		 * to send/receive these lengths in network byte order.
+		 */
+		memcpy(&len, p, sizeof(len));
+		p += sizeof(len);
+		memcpy(&tmprp.lsn, p, sizeof(DB_LSN));
+		p += sizeof(DB_LSN);
+		logrec.data = p;
+		logrec.size = len;
+		RPRINT(dbenv, (dbenv, &mb,
+		    "log_rep_split: Processing LSN [%lu][%lu]",
+		    (u_long)tmprp.lsn.file, (u_long)tmprp.lsn.offset));
+		RPRINT(dbenv, (dbenv, &mb,
+    "log_rep_split: p %#lx ep %#lx logrec data %#lx, size %lu (%#lx)",
+		    P_TO_ULONG(p), P_TO_ULONG(ep), P_TO_ULONG(logrec.data),
+		    (u_long)logrec.size, (u_long)logrec.size));
+		is_dup = 0;
+		p += len;
+		if (p >= ep && is_perm)
+			F_SET(&tmprp, REPCTL_PERM);
+		ret = __rep_apply(dbenv,
+		    &tmprp, &logrec, &tmp_lsn, &is_dup, last_lsnp);
+		RPRINT(dbenv, (dbenv, &mb,
+		    "log_split: rep_apply ret %d, tmp_lsn [%lu][%lu]",
+		    ret, (u_long)tmp_lsn.file, (u_long)tmp_lsn.offset));
+#if 0
+		/*
+		 * This buffer may be old and we've already gotten these
+		 * records.  Short-circuit processing this buffer.
+		 */
+		if (is_dup)
+			goto out;
+#endif
+		switch (ret) {
+		/*
+		 * If we received the pieces we need for running recovery,
+		 * short-circuit because recovery will truncate the log to
+		 * the LSN we want anyway.
+		 */
+		case DB_REP_LOGREADY:
+			goto out;
+		/*
+		 * If we just handled a special record, retain that information.
+		 */
+		case DB_REP_ISPERM:
+		case DB_REP_NOTPERM:
+			save_ret = ret;
+			save_lsn = tmp_lsn;
+			ret = 0;
+			break;
+		/*
+		 * Normal processing, do nothing, just continue.
+		 */
+		case 0:
+			break;
+		/*
+		 * If we get an error, then stop immediately.
+		 */
+		default:
+			goto out;
+		}
+	}
+out:
+	/*
+	 * If we finish processing successfully, set our return values
+	 * based on what we saw.
+	 */
+	if (ret == 0) {
+		ret = save_ret;
+		*ret_lsnp = save_lsn;
+	}
+	return (ret);
+}
+#endif
