@@ -1,24 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: mp_alloc.c,v 11.47 2004/10/15 16:59:42 bostic Exp $
+ * $Id: mp_alloc.c,v 12.20 2006/09/07 15:11:26 mjc Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
-#include "dbinc/db_shash.h"
 #include "dbinc/mp.h"
+#include "dbinc/txn.h"
 
-static void __memp_bad_buffer __P((DB_MPOOL_HASH *));
+static void __memp_bad_buffer __P((DB_ENV *, DB_MPOOL_HASH *));
 
 /*
  * __memp_alloc --
@@ -36,16 +31,18 @@ __memp_alloc(dbmp, infop, mfp, len, offsetp, retp)
 	roff_t *offsetp;
 	void *retp;
 {
-	BH *bhp;
+	BH *bhp, *oldest_bhp, *tbhp;
+	BH_FROZEN_PAGE *frozen_bhp;
 	DB_ENV *dbenv;
 	DB_MPOOL_HASH *dbht, *hp, *hp_end, *hp_tmp;
-	DB_MUTEX *mutexp;
 	MPOOL *c_mp;
 	MPOOLFILE *bh_mfp;
 	size_t freed_space;
-	u_int32_t buckets, buffers, high_priority, priority, put_counter;
-	u_int32_t total_buckets;
-	int aggressive, giveup, ret;
+	db_mutex_t mutex;
+	u_int32_t buckets, buffers, high_priority, priority;
+	u_int32_t put_counter, total_buckets;
+	int aggressive, alloc_freeze, giveup, got_oldest, ret;
+	u_int8_t *endp;
 	void *p;
 
 	dbenv = dbmp->dbenv;
@@ -54,7 +51,7 @@ __memp_alloc(dbmp, infop, mfp, len, offsetp, retp)
 	hp_end = &dbht[c_mp->htab_buckets];
 
 	buckets = buffers = put_counter = total_buckets = 0;
-	aggressive = giveup = 0;
+	aggressive = alloc_freeze = giveup = got_oldest = 0;
 	hp_tmp = NULL;
 
 	c_mp->stat.st_alloc++;
@@ -66,10 +63,14 @@ __memp_alloc(dbmp, infop, mfp, len, offsetp, retp)
 	 * NULL, we'll compare the underlying page sizes of the two buffers
 	 * before free-ing and re-allocating buffers.
 	 */
-	if (mfp != NULL)
-		len = (sizeof(BH) - sizeof(u_int8_t)) + mfp->stat.st_pagesize;
+	if (mfp != NULL) {
+		len = SSZA(BH, buf) + mfp->stat.st_pagesize;
+		/* Add space for alignment padding for MVCC diagnostics. */
+		MVCC_BHSIZE(mfp, len);
+	}
+
+	MPOOL_REGION_LOCK(dbenv, infop);
 
-	R_LOCK(dbenv, infop);
 	/*
 	 * Anything newer than 1/10th of the buffer pool is ignored during
 	 * allocation (unless allocation starts failing).
@@ -85,10 +86,15 @@ __memp_alloc(dbmp, infop, mfp, len, offsetp, retp)
 	 * we need in the hopes it will coalesce into a contiguous chunk of the
 	 * right size.  In the latter case we branch back here and try again.
 	 */
-alloc:	if ((ret = __db_shalloc(infop, len, MUTEX_ALIGN, &p)) == 0) {
+alloc:	if ((ret = __db_shalloc(infop, len, 0, &p)) == 0) {
 		if (mfp != NULL)
 			c_mp->stat.st_pages++;
-		R_UNLOCK(dbenv, infop);
+		MPOOL_REGION_UNLOCK(dbenv, infop);
+		/*
+		 * For MVCC diagnostics, align the pointer so that the buffer
+		 * starts on a page boundary.
+		 */
+		MVCC_BHALIGN(mfp, p);
 
 found:		if (offsetp != NULL)
 			*offsetp = R_OFFSET(infop, p);
@@ -113,12 +119,13 @@ found:		if (offsetp != NULL)
 		}
 		return (0);
 	} else if (giveup || c_mp->stat.st_pages == 0) {
-		R_UNLOCK(dbenv, infop);
+		MPOOL_REGION_UNLOCK(dbenv, infop);
 
-		__db_err(dbenv,
+		__db_errx(dbenv,
 		    "unable to allocate space from the buffer cache");
 		return (ret);
 	}
+	ret = 0;
 
 	/*
 	 * We re-attempt the allocation every time we've freed 3 times what
@@ -187,7 +194,7 @@ found:		if (offsetp != NULL)
 		if ((++buckets % c_mp->htab_buckets) == 0) {
 			if (freed_space > 0)
 				goto alloc;
-			R_UNLOCK(dbenv, infop);
+			MPOOL_REGION_UNLOCK(dbenv, infop);
 
 			switch (++aggressive) {
 			case 1:
@@ -211,7 +218,7 @@ found:		if (offsetp != NULL)
 				break;
 			}
 
-			R_LOCK(dbenv, infop);
+			MPOOL_REGION_LOCK(dbenv, infop);
 			goto alloc;
 		}
 
@@ -239,12 +246,12 @@ found:		if (offsetp != NULL)
 		priority = hp->hash_priority;
 
 		/* Unlock the region and lock the hash bucket. */
-		R_UNLOCK(dbenv, infop);
-		mutexp = &hp->hash_mutex;
-		MUTEX_LOCK(dbenv, mutexp);
+		MPOOL_REGION_UNLOCK(dbenv, infop);
+		mutex = hp->mtx_hash;
+		MUTEX_LOCK(dbenv, mutex);
 
 #ifdef DIAGNOSTIC
-		__memp_check_order(hp);
+		__memp_check_order(dbenv, hp);
 #endif
 		/*
 		 * The lowest priority page is first in the bucket, as they are
@@ -255,8 +262,7 @@ found:		if (offsetp != NULL)
 		 * we have to restart.  We will still take the first buffer on
 		 * the bucket's list, though, if it has a low enough priority.
 		 */
-		if ((bhp = SH_TAILQ_FIRST(&hp->hash_bucket, __bh)) == NULL ||
-		    bhp->ref != 0 || bhp->priority > priority)
+this_hb:	if ((bhp = SH_TAILQ_FIRST(&hp->hash_bucket, __bh)) == NULL)
 			goto next_hb;
 
 		buffers++;
@@ -264,6 +270,41 @@ found:		if (offsetp != NULL)
 		/* Find the associated MPOOLFILE. */
 		bh_mfp = R_ADDR(dbmp->reginfo, bhp->mf_offset);
 
+		/* Select the lowest priority buffer in the chain. */
+		for (oldest_bhp = bhp, tbhp = SH_CHAIN_PREV(bhp, vc, __bh);
+		    tbhp != NULL;
+		    oldest_bhp = tbhp, tbhp = SH_CHAIN_PREV(tbhp, vc, __bh))
+			if (tbhp->ref <= bhp->ref &&
+			    tbhp->priority <= bhp->priority)
+				bhp = tbhp;
+
+		/*
+		 * Prefer the last buffer in the chain.
+		 *
+		 * If the oldest buffer isn't obsolete with respect to the
+		 * cached old reader LSN, recalculate the oldest reader LSN
+		 * and retry.  There is a tradeoff here, because if we had the
+		 * LSN earlier, we might have found pages to evict, but to get
+		 * it, we need to lock the transaction region.
+		 */
+		if (oldest_bhp != bhp && oldest_bhp->ref == 0) {
+			if (F_ISSET(bhp, BH_FROZEN) &&
+			    !F_ISSET(oldest_bhp, BH_FROZEN))
+				bhp = oldest_bhp;
+			else if (BH_OBSOLETE(oldest_bhp, hp->old_reader))
+				bhp = oldest_bhp;
+			else if (!got_oldest &&
+			    __txn_oldest_reader(dbenv, &hp->old_reader) == 0) {
+				got_oldest = 1;
+				if (BH_OBSOLETE(oldest_bhp, hp->old_reader))
+					bhp = oldest_bhp;
+			}
+		}
+
+		if (bhp->ref != 0 || (bhp != oldest_bhp &&
+		    !aggressive && bhp->priority > priority))
+			goto next_hb;
+
 		/* If the page is dirty, pin it and write it. */
 		ret = 0;
 		if (F_ISSET(bhp, BH_DIRTY)) {
@@ -275,6 +316,32 @@ found:		if (offsetp != NULL)
 		} else
 			++c_mp->stat.st_ro_evict;
 
+		/*
+		 * Freeze this buffer, if necessary.  That is, if the buffer
+		 * itself or the next version created could be read by the
+		 * oldest reader in the system.
+		 */
+		if (ret == 0 && bh_mfp->multiversion) {
+			if (!got_oldest && !SH_CHAIN_HASPREV(bhp, vc) &&
+			    !BH_OBSOLETE(bhp, hp->old_reader)) {
+				(void)__txn_oldest_reader(dbenv,
+				    &hp->old_reader);
+				got_oldest = 1;
+			}
+			if (SH_CHAIN_HASPREV(bhp, vc) ||
+			    !BH_OBSOLETE(bhp, hp->old_reader)) {
+				/*
+				 * Before freezing, double-check that we have
+				 * an up-to-date old_reader LSN.
+				 */
+				if (!aggressive ||
+				    F_ISSET(bhp, BH_FROZEN) || bhp->ref != 0)
+					goto next_hb;
+				ret = __memp_bh_freeze(dbmp,
+				    infop, hp, bhp, &alloc_freeze);
+			}
+		}
+
 		/*
 		 * If a write fails for any reason, we can't proceed.
 		 *
@@ -288,25 +355,61 @@ found:		if (offsetp != NULL)
 		 */
 		if (ret != 0 || bhp->ref != 0) {
 			if (ret != 0 && aggressive)
-				__memp_bad_buffer(hp);
+				__memp_bad_buffer(dbenv, hp);
 			goto next_hb;
 		}
 
 		/*
+		 * If we need some empty buffer headers for freezing, turn the
+		 * buffer we've found into frozen headers and put them on the
+		 * free list.  Only reset alloc_freeze if we've actually
+		 * allocated some frozen buffer headers.
+		 *
 		 * Check to see if the buffer is the size we're looking for.
-		 * If so, we can simply reuse it.  Else, free the buffer and
-		 * its space and keep looking.
+		 * If so, we can simply reuse it.  Otherwise, free the buffer
+		 * and its space and keep looking.
 		 */
-		if (mfp != NULL &&
-		    mfp->stat.st_pagesize == bh_mfp->stat.st_pagesize) {
-			__memp_bhfree(dbmp, hp, bhp, 0);
+		if (F_ISSET(bhp, BH_FROZEN)) {
+			++bhp->ref;
+			if ((ret = __memp_bh_thaw(dbmp, infop, hp,
+			    bhp, NULL)) != 0) {
+				MUTEX_UNLOCK(dbenv, mutex);
+				return (ret);
+			}
+			alloc_freeze = 0;
+			goto this_hb;
+		} else if (alloc_freeze) {
+			if ((ret = __memp_bhfree(dbmp, hp, bhp, 0)) != 0)
+				return (ret);
+			MVCC_MPROTECT(bhp->buf, bh_mfp->stat.st_pagesize,
+			    PROT_READ | PROT_WRITE | PROT_EXEC);
 
+			MPOOL_REGION_LOCK(dbenv, infop);
+			SH_TAILQ_INSERT_TAIL(&c_mp->alloc_frozen,
+			    (BH_FROZEN_ALLOC *)bhp, links);
+			frozen_bhp = (BH_FROZEN_PAGE *)
+			    ((BH_FROZEN_ALLOC *)bhp + 1);
+			endp = (u_int8_t *)bhp->buf + bh_mfp->stat.st_pagesize;
+			while ((u_int8_t *)(frozen_bhp + 1) < endp) {
+				SH_TAILQ_INSERT_TAIL(&c_mp->free_frozen,
+				    (BH *)frozen_bhp, hq);
+				frozen_bhp++;
+			}
+			alloc_freeze = 0;
+			continue;
+		} else if (mfp != NULL &&
+		    mfp->stat.st_pagesize == bh_mfp->stat.st_pagesize) {
+			if ((ret = __memp_bhfree(dbmp, hp, bhp, 0)) != 0)
+				return (ret);
 			p = bhp;
 			goto found;
+		} else {
+			freed_space += __db_shalloc_sizeof(bhp);
+			if ((ret = __memp_bhfree(dbmp,
+			    hp, bhp, BH_FREE_FREEMEM)) != 0)
+				return (ret);
 		}
 
-		freed_space += __db_shalloc_sizeof(bhp);
-		__memp_bhfree(dbmp, hp, bhp, BH_FREE_FREEMEM);
 		if (aggressive > 1)
 			aggressive = 1;
 
@@ -316,9 +419,9 @@ found:		if (offsetp != NULL)
 		 * hash bucket lock has already been discarded.
 		 */
 		if (0) {
-next_hb:		MUTEX_UNLOCK(dbenv, mutexp);
+next_hb:		MUTEX_UNLOCK(dbenv, mutex);
 		}
-		R_LOCK(dbenv, infop);
+		MPOOL_REGION_LOCK(dbenv, infop);
 
 		/*
 		 * Retry the allocation as soon as we've freed up sufficient
@@ -332,38 +435,70 @@ next_hb:		MUTEX_UNLOCK(dbenv, mutexp);
 	/* NOTREACHED */
 }
 
+/*
+ * __memp_free --
+ *	Free some space from a cache region.
+ *
+ * PUBLIC: void __memp_free __P((REGINFO *, MPOOLFILE *, void *));
+ */
+void
+__memp_free(infop, mfp, buf)
+	REGINFO *infop;
+	MPOOLFILE *mfp;
+	void *buf;
+{
+	MVCC_BHUNALIGN(mfp, buf);
+	COMPQUIET(mfp, NULL);
+	__db_shalloc_free(infop, buf);
+}
+
 /*
  * __memp_bad_buffer --
  *	Make the first buffer in a hash bucket the least desirable buffer.
  */
 static void
-__memp_bad_buffer(hp)
+__memp_bad_buffer(dbenv, hp)
+	DB_ENV *dbenv;
 	DB_MPOOL_HASH *hp;
 {
-	BH *bhp;
+	BH *bhp, *last_bhp;
 	u_int32_t priority;
 
-	/* Remove the first buffer from the bucket. */
+	/*
+	 * Get the first buffer from the bucket.  If it is also the last buffer
+	 * (in other words, it is the only buffer in the bucket), we're done.
+	 */
 	bhp = SH_TAILQ_FIRST(&hp->hash_bucket, __bh);
+	last_bhp = SH_TAILQ_LASTP(&hp->hash_bucket, hq, __bh);
+	if (bhp == last_bhp)
+		return;
+
+	/* There are multiple buffers in the bucket, remove the first one. */
 	SH_TAILQ_REMOVE(&hp->hash_bucket, bhp, hq, __bh);
 
 	/*
 	 * Find the highest priority buffer in the bucket.  Buffers are
 	 * sorted by priority, so it's the last one in the bucket.
 	 */
-	priority = bhp->priority;
-	if (!SH_TAILQ_EMPTY(&hp->hash_bucket))
-	  priority = SH_TAILQ_LAST(&hp->hash_bucket, hq, __bh)->priority;
+	priority = BH_PRIORITY(last_bhp);
 
 	/*
-	 * Set our buffer's priority to be just as bad, and append it to
-	 * the bucket.
+	 * Append our buffer to the bucket and set its priority to be just as
+	 * bad.
 	 */
-	bhp->priority = priority;
 	SH_TAILQ_INSERT_TAIL(&hp->hash_bucket, bhp, hq);
+	for (; bhp != NULL ; bhp = SH_CHAIN_PREV(bhp, vc, __bh))
+		bhp->priority = priority;
 
 	/* Reset the hash bucket's priority. */
-	hp->hash_priority = SH_TAILQ_FIRST(&hp->hash_bucket, __bh)->priority;
+	hp->hash_priority = BH_PRIORITY(
+	    SH_TAILQ_FIRSTP(&hp->hash_bucket, __bh));
+
+#ifdef DIAGNOSTIC
+	__memp_check_order(dbenv, hp);
+#else
+	COMPQUIET(dbenv, NULL);
+#endif
 }
 
 #ifdef DIAGNOSTIC
@@ -372,27 +507,45 @@ __memp_bad_buffer(hp)
  *	Verify the priority ordering of a hash bucket chain.
  *
  * PUBLIC: #ifdef DIAGNOSTIC
- * PUBLIC: void __memp_check_order __P((DB_MPOOL_HASH *));
+ * PUBLIC: void __memp_check_order __P((DB_ENV *, DB_MPOOL_HASH *));
  * PUBLIC: #endif
  */
 void
-__memp_check_order(hp)
+__memp_check_order(dbenv, hp)
+	DB_ENV *dbenv;
 	DB_MPOOL_HASH *hp;
 {
-	BH *bhp;
-	u_int32_t priority;
+	BH *bhp, *first_bhp, *tbhp;
+	u_int32_t priority, last_priority;
 
 	/*
 	 * Assumes the hash bucket is locked.
 	 */
-	if ((bhp = SH_TAILQ_FIRST(&hp->hash_bucket, __bh)) == NULL)
-		return;
+	last_priority = hp->hash_priority;
+	for (bhp = first_bhp = SH_TAILQ_FIRST(&hp->hash_bucket, __bh);
+	    bhp != NULL; bhp = SH_TAILQ_NEXT(bhp, hq, __bh)) {
+		DB_ASSERT(dbenv, !SH_CHAIN_HASNEXT(bhp, vc));
 
-	DB_ASSERT(bhp->priority == hp->hash_priority);
+		priority = BH_PRIORITY(bhp);
+		DB_ASSERT(dbenv, (bhp == first_bhp) ?
+		    priority == last_priority : priority >= last_priority);
+		last_priority = priority;
 
-	for (priority = bhp->priority;
-	    (bhp = SH_TAILQ_NEXT(bhp, hq, __bh)) != NULL;
-	    priority = bhp->priority)
-		DB_ASSERT(priority <= bhp->priority);
+		/* Chains have referential integrity. */
+		for (tbhp = SH_CHAIN_PREV(bhp, vc, __bh); tbhp != NULL;
+		    tbhp = SH_CHAIN_PREV(tbhp, vc, __bh))
+			DB_ASSERT(dbenv, tbhp == SH_CHAIN_PREV(
+			    SH_CHAIN_NEXT(tbhp, vc, __bh), vc, __bh));
+
+		/*
+		 * No repeats.
+		 * XXX This is O(N**2) where N is the number of buffers in the
+		 * bucket, but we generally assume small buckets.
+		 */
+		for (tbhp = SH_TAILQ_NEXT(bhp, hq, __bh); tbhp != NULL;
+		    tbhp = SH_TAILQ_NEXT(tbhp, hq, __bh))
+			DB_ASSERT(dbenv, bhp->pgno != tbhp->pgno ||
+			    bhp->mf_offset != tbhp->mf_offset);
+	}
 }
 #endif
