@@ -1,29 +1,20 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2002
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1997-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: os_alloc.c,v 12.8 2006/08/24 14:46:17 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_alloc.c,v 11.32 2002/08/06 04:57:07 bostic Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
 
 #ifdef DIAGNOSTIC
 static void __os_guard __P((DB_ENV *));
 
-union __db_alloc {
+union __db_allocinfo {
 	size_t size;
 	double align;
 };
@@ -41,13 +32,6 @@ union __db_alloc {
  * !!!
  * Correct for systems that don't set errno when malloc and friends fail.
  *
- * !!!
- * There is no circumstance in which we can call __os_umalloc, __os_urealloc
- * or __os_ufree without an environment handle, as we need one to determine
- * whether or not to use an application-specified malloc function.  If we
- * don't have an environment handle, we should be calling __os_XXX instead.
- * Make DIAGNOSTIC blow up if we get this wrong.
- *
  *	Out of memory.
  *	We wish to hold the whole sky,
  *	But we never will.
@@ -55,9 +39,11 @@ union __db_alloc {
 
 /*
  * __os_umalloc --
- *	A malloc(3) function that will use, in order of preference,
- *	the allocation function specified to the DB handle, the DB_ENV
- *	handle, or __os_malloc.
+ *	Allocate memory to be used by the application.
+ *
+ *	Use, in order of preference, the allocation function specified to the
+ *	DB_ENV handle, the allocation function specified as a replacement for
+ *	the library malloc, or the library malloc().
  *
  * PUBLIC: int __os_umalloc __P((DB_ENV *, size_t, void *));
  */
@@ -69,9 +55,6 @@ __os_umalloc(dbenv, size, storep)
 {
 	int ret;
 
-	/* Require an environment handle. */
-	DB_ASSERT(dbenv != NULL);
-
 	/* Never allocate 0 bytes -- some C libraries don't like it. */
 	if (size == 0)
 		++size;
@@ -85,19 +68,19 @@ __os_umalloc(dbenv, size, storep)
 			/*
 			 *  Correct error return, see __os_malloc.
 			 */
-			if ((ret = __os_get_errno()) == 0) {
+			if ((ret = __os_get_errno_ret_zero()) == 0) {
 				ret = ENOMEM;
 				__os_set_errno(ENOMEM);
 			}
-			__db_err(dbenv,
-			    "malloc: %s: %lu", strerror(ret), (u_long)size);
+			__db_err(dbenv, ret, "malloc: %lu", (u_long)size);
 			return (ret);
 		}
 		return (0);
 	}
 
 	if ((*(void **)storep = dbenv->db_malloc(size)) == NULL) {
-		__db_err(dbenv, "User-specified malloc function returned NULL");
+		__db_errx(dbenv,
+		    "user-specified malloc function returned NULL");
 		return (ENOMEM);
 	}
 
@@ -106,7 +89,9 @@ __os_umalloc(dbenv, size, storep)
 
 /*
  * __os_urealloc --
- *	realloc(3) counterpart to __os_umalloc.
+ *	Allocate memory to be used by the application.
+ *
+ *	A realloc(3) counterpart to __os_umalloc's malloc(3).
  *
  * PUBLIC: int __os_urealloc __P((DB_ENV *, size_t, void *));
  */
@@ -121,9 +106,6 @@ __os_urealloc(dbenv, size, storep)
 
 	ptr = *(void **)storep;
 
-	/* Require an environment handle. */
-	DB_ASSERT(dbenv != NULL);
-
 	/* Never allocate 0 bytes -- some C libraries don't like it. */
 	if (size == 0)
 		++size;
@@ -140,19 +122,18 @@ __os_urealloc(dbenv, size, storep)
 			/*
 			 * Correct errno, see __os_realloc.
 			 */
-			if ((ret = __os_get_errno()) == 0) {
+			if ((ret = __os_get_errno_ret_zero()) == 0) {
 				ret = ENOMEM;
 				__os_set_errno(ENOMEM);
 			}
-			__db_err(dbenv,
-			    "realloc: %s: %lu", strerror(ret), (u_long)size);
+			__db_err(dbenv, ret, "realloc: %lu", (u_long)size);
 			return (ret);
 		}
 		return (0);
 	}
 
 	if ((*(void **)storep = dbenv->db_realloc(ptr, size)) == NULL) {
-		__db_err(dbenv,
+		__db_errx(dbenv,
 		    "User-specified realloc function returned NULL");
 		return (ENOMEM);
 	}
@@ -162,26 +143,23 @@ __os_urealloc(dbenv, size, storep)
 
 /*
  * __os_ufree --
- *	free(3) counterpart to __os_umalloc.
+ *	Free memory used by the application.
  *
- * PUBLIC: int __os_ufree __P((DB_ENV *, void *));
+ *	A free(3) counterpart to __os_umalloc's malloc(3).
+ *
+ * PUBLIC: void __os_ufree __P((DB_ENV *, void *));
  */
-int
+void
 __os_ufree(dbenv, ptr)
 	DB_ENV *dbenv;
 	void *ptr;
 {
-	/* Require an environment handle. */
-	DB_ASSERT(dbenv != NULL);
-
 	if (dbenv != NULL && dbenv->db_free != NULL)
 		dbenv->db_free(ptr);
 	else if (DB_GLOBAL(j_free) != NULL)
 		DB_GLOBAL(j_free)(ptr);
 	else
 		free(ptr);
-
-	return (0);
 }
 
 /*
@@ -260,7 +238,7 @@ __os_malloc(dbenv, size, storep)
 
 #ifdef DIAGNOSTIC
 	/* Add room for size and a guard byte. */
-	size += sizeof(union __db_alloc) + 1;
+	size += sizeof(union __db_allocinfo) + 1;
 #endif
 
 	if (DB_GLOBAL(j_malloc) != NULL)
@@ -274,16 +252,18 @@ __os_malloc(dbenv, size, storep)
 		 * but it turns out that setting errno is quite expensive on
 		 * Windows/NT in an MT environment.
 		 */
-		if ((ret = __os_get_errno()) == 0) {
+		if ((ret = __os_get_errno_ret_zero()) == 0) {
 			ret = ENOMEM;
 			__os_set_errno(ENOMEM);
 		}
-		__db_err(dbenv,
-		    "malloc: %s: %lu", strerror(ret), (u_long)size);
+		__db_err(dbenv, ret, "malloc: %lu", (u_long)size);
 		return (ret);
 	}
 
 #ifdef DIAGNOSTIC
+	/* Overwrite memory. */
+	memset(p, CLEAR_BYTE, size);
+
 	/*
 	 * Guard bytes: if #DIAGNOSTIC is defined, we allocate an additional
 	 * byte after the memory and set it to a special value that we check
@@ -291,8 +271,8 @@ __os_malloc(dbenv, size, storep)
 	 */
 	((u_int8_t *)p)[size - 1] = CLEAR_BYTE;
 
-	((union __db_alloc *)p)->size = size;
-	p = &((union __db_alloc *)p)[1];
+	((union __db_allocinfo *)p)->size = size;
+	p = &((union __db_allocinfo *)p)[1];
 #endif
 	*(void **)storep = p;
 
@@ -326,10 +306,18 @@ __os_realloc(dbenv, size, storep)
 
 #ifdef DIAGNOSTIC
 	/* Add room for size and a guard byte. */
-	size += sizeof(union __db_alloc) + 1;
+	size += sizeof(union __db_allocinfo) + 1;
 
-	/* Back up to the real begining */
-	ptr = &((union __db_alloc *)ptr)[-1];
+	/* Back up to the real beginning */
+	ptr = &((union __db_allocinfo *)ptr)[-1];
+
+	{
+		size_t s;
+
+		s = ((union __db_allocinfo *)ptr)->size;
+		if (((u_int8_t *)ptr)[s - 1] != CLEAR_BYTE)
+			 __os_guard(dbenv);
+	}
 #endif
 
 	/*
@@ -347,19 +335,18 @@ __os_realloc(dbenv, size, storep)
 		 * but it turns out that setting errno is quite expensive on
 		 * Windows/NT in an MT environment.
 		 */
-		if ((ret = __os_get_errno()) == 0) {
+		if ((ret = __os_get_errno_ret_zero()) == 0) {
 			ret = ENOMEM;
 			__os_set_errno(ENOMEM);
 		}
-		__db_err(dbenv,
-		    "realloc: %s: %lu", strerror(ret), (u_long)size);
+		__db_err(dbenv, ret, "realloc: %lu", (u_long)size);
 		return (ret);
 	}
 #ifdef DIAGNOSTIC
 	((u_int8_t *)p)[size - 1] = CLEAR_BYTE;	/* Initialize guard byte. */
 
-	((union __db_alloc *)p)->size = size;
-	p = &((union __db_alloc *)p)[1];
+	((union __db_allocinfo *)p)->size = size;
+	p = &((union __db_allocinfo *)p)[1];
 #endif
 
 	*(void **)storep = p;
@@ -379,7 +366,7 @@ __os_free(dbenv, ptr)
 	void *ptr;
 {
 #ifdef DIAGNOSTIC
-	int size;
+	size_t size;
 	/*
 	 * Check that the guard byte (one past the end of the memory) is
 	 * still CLEAR_BYTE.
@@ -387,12 +374,12 @@ __os_free(dbenv, ptr)
 	if (ptr == NULL)
 		return;
 
-	ptr = &((union __db_alloc *)ptr)[-1];
-	size = ((union __db_alloc *)ptr)->size;
+	ptr = &((union __db_allocinfo *)ptr)[-1];
+	size = ((union __db_allocinfo *)ptr)->size;
 	if (((u_int8_t *)ptr)[size - 1] != CLEAR_BYTE)
 		 __os_guard(dbenv);
 
-	/* Clear memory. */
+	/* Overwrite memory. */
 	if (size != 0)
 		memset(ptr, CLEAR_BYTE, size);
 #endif
@@ -413,7 +400,7 @@ static void
 __os_guard(dbenv)
 	DB_ENV *dbenv;
 {
-	__db_err(dbenv, "Guard byte incorrect during free");
+	__db_errx(dbenv, "Guard byte incorrect during free");
 	abort();
 	/* NOTREACHED */
 }
