@@ -1,31 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2001-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2001-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: os_clock.c,v 1.15 2004/07/06 17:33:14 bostic Exp $
+ * $Id: os_clock.c,v 12.11 2006/08/24 14:46:17 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#if TIME_WITH_SYS_TIME
-#include <sys/time.h>
-#include <time.h>
-#else
-#if HAVE_SYS_TIME_H
-#include <sys/time.h>
-#else
-#include <time.h>
-#endif /* HAVE_SYS_TIME_H */
-#endif /* TIME_WITH SYS_TIME */
-
-#include <string.h>
-#endif
-
 #include "db_int.h"
 
 /*
@@ -86,6 +69,6 @@ __os_clock(dbenv, secsp, usecsp)
 #endif
 	return;
 
-err:	__db_err(dbenv, "%s: %s", sc, strerror(ret));
-	(void)__db_panic(dbenv, ret);
+err:	__db_syserr(dbenv, ret, "%s", sc);
+	(void)__db_panic(dbenv, __os_posix_err(ret));
 }
