@@ -1,18 +1,15 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997, 1998, 1999, 2000
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1997-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: os_dir.c,v 12.8 2006/08/24 14:46:17 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_dir.c,v 11.7.2.1 2000/06/27 17:52:57 bostic Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
+#include "db_int.h"
 
 #if HAVE_DIRENT_H
 # include <dirent.h>
@@ -31,11 +28,6 @@ static const char revid[] = "$Id: os_dir.c,v 11.7.2.1 2000/06/27 17:52:57 bostic
 # endif
 #endif
 
-#endif
-
-#include "db_int.h"
-#include "os_jump.h"
-
 /*
  * __os_dirlist --
  *	Return a list of the files in a directory.
@@ -51,11 +43,12 @@ __os_dirlist(dbenv, dir, namesp, cntp)
 {
 	struct dirent *dp;
 	DIR *dirp;
+	struct stat sb;
 	int arraysz, cnt, ret;
-	char **names;
+	char **names, buf[DB_MAXPATHLEN];
 
-	if (__db_jump.j_dirlist != NULL)
-		return (__db_jump.j_dirlist(dir, namesp, cntp));
+	if (DB_GLOBAL(j_dirlist) != NULL)
+		return (DB_GLOBAL(j_dirlist)(dir, namesp, cntp));
 
 #ifdef HAVE_VXWORKS
 	if ((dirp = opendir((char *)dir)) == NULL)
@@ -64,15 +57,32 @@ __os_dirlist(dbenv, dir, namesp, cntp)
 #endif
 		return (__os_get_errno());
 	names = NULL;
-	for (arraysz = cnt = 0; (dp = readdir(dirp)) != NULL; ++cnt) {
+	for (arraysz = cnt = 0; (dp = readdir(dirp)) != NULL;) {
+		snprintf(buf, sizeof(buf), "%s/%s", dir, dp->d_name);
+
+		/*
+		 * We're only interested in regular files, skip the rest.
+		 * Other parts of Berkeley DB depend on this, for example,
+		 * the db_hotbackup utility doesn't want to try and copy
+		 * directories.
+		 */
+		RETRY_CHK(stat(buf, &sb), ret);
+		if (ret != 0) {
+			ret = __os_posix_err(ret);
+			goto err;
+		}
+		if (!S_ISREG(sb.st_mode))
+			continue;
+
 		if (cnt >= arraysz) {
 			arraysz += 100;
 			if ((ret = __os_realloc(dbenv,
-			    arraysz * sizeof(names[0]), NULL, &names)) != 0)
-				goto nomem;
+			    (u_int)arraysz * sizeof(names[0]), &names)) != 0)
+				goto err;
 		}
 		if ((ret = __os_strdup(dbenv, dp->d_name, &names[cnt])) != 0)
-			goto nomem;
+			goto err;
+		cnt++;
 	}
 	(void)closedir(dirp);
 
@@ -80,8 +90,8 @@ __os_dirlist(dbenv, dir, namesp, cntp)
 	*cntp = cnt;
 	return (0);
 
-nomem:	if (names != NULL)
-		__os_dirfree(names, cnt);
+err:	if (names != NULL)
+		__os_dirfree(dbenv, names, cnt);
 	if (dirp != NULL)
 		(void)closedir(dirp);
 	return (ret);
@@ -91,18 +101,19 @@ nomem:	if (names != NULL)
  * __os_dirfree --
  *	Free the list of files.
  *
- * PUBLIC: void __os_dirfree __P((char **, int));
+ * PUBLIC: void __os_dirfree __P((DB_ENV *, char **, int));
  */
 void
-__os_dirfree(names, cnt)
+__os_dirfree(dbenv, names, cnt)
+	DB_ENV *dbenv;
 	char **names;
 	int cnt;
 {
-	if (__db_jump.j_dirfree != NULL)
-		__db_jump.j_dirfree(names, cnt);
+	if (DB_GLOBAL(j_dirfree) != NULL)
+		DB_GLOBAL(j_dirfree)(names, cnt);
 	else {
 		while (cnt > 0)
-			__os_free(names[--cnt], 0);
-		__os_free(names, 0);
+			__os_free(dbenv, names[--cnt]);
+		__os_free(dbenv, names);
 	}
 }
