@@ -1,21 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2004-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: os_truncate.c,v 11.7 2004/09/17 22:00:31 mjc Exp $
+ * $Id: os_truncate.c,v 12.7 2006/08/24 14:46:18 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#include <unistd.h>
-#endif
-
 #include "db_int.h"
 
 /*
@@ -50,9 +43,10 @@ __os_truncate(dbenv, fhp, pgno, pgsize)
 #endif
 	}
 
-	if (ret != 0)
-		__db_err(dbenv,
-		    "ftruncate: %lu: %s", (u_long)offset, strerror(ret));
+	if (ret != 0) {
+		__db_syserr(dbenv, ret, "ftruncate: %lu", (u_long)offset);
+		ret = __os_posix_err(ret);
+	}
 
 	return (ret);
 }
