@@ -1,16 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1997-2003
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1997-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: os_vx_abs.c,v 12.3 2006/08/24 14:46:20 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: os_vx_abs.c,v 1.8 2003/01/08 05:32:23 bostic Exp $";
-#endif /* not lint */
-
 #include "db_int.h"
 #include "iosLib.h"
 
