@@ -1,29 +1,42 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1999-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: qam_rec.c,v 12.20 2006/09/09 14:28:24 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: qam_rec.c,v 11.55 2001/10/04 21:28:35 bostic Exp $";
-#endif /* not lint */
-
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#endif
-
 #include "db_int.h"
-#include "db_page.h"
-#include "db_shash.h"
-#include "lock.h"
-#include "db_am.h"
-#include "qam.h"
-#include "log.h"
+#include "dbinc/db_page.h"
+#include "dbinc/db_am.h"
+#include "dbinc/lock.h"
+#include "dbinc/log.h"
+#include "dbinc/mp.h"
+#include "dbinc/qam.h"
+#include "dbinc/txn.h"
+
+/*
+ * LSNs in queue data pages are advisory.  They do not have to be accurate
+ * as all operations are idempotent on records.  They should not be rolled
+ * forward during recovery as committed transaction may obscure updates from
+ * an incomplete transaction that updates the same page.  The incomplete
+ * transaction may be completed during a later hot backup cycle.
+ */
+
+/* Determine if we are restoring prepared transactions from __txn_recover. */
+#define	IS_IN_RESTORE(dbenv)						\
+	(((DB_TXNREGION *)						\
+	     (dbenv)->tx_handle->reginfo.primary)->stat.st_nrestores != 0)
+
+/* Queue version of REC_DIRTY -- needs to probe the correct file. */
+#define	QAM_DIRTY(dbp, pgno, pagep)					\
+	if ((ret = __qam_dirty((dbp), pgno, pagep, NULL, 0)) != 0) {	\
+		ret = __db_pgerr((dbp), (pgno), ret);			\
+		goto out;						\
+	}
 
 /*
  * __qam_incfirst_recover --
@@ -44,25 +57,29 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 	DB *file_dbp;
 	DBC *dbc;
 	DB_LOCK lock;
+	DB_LSN trunc_lsn;
 	DB_MPOOLFILE *mpf;
 	QMETA *meta;
 	QUEUE_CURSOR *cp;
 	db_pgno_t metapg;
-	int exact, modified, ret, rec_ext;
+	u_int32_t rec_ext;
+	int exact, ret, t_ret;
 
-	COMPQUIET(info, NULL);
+	LOCK_INIT(lock);
+	COMPQUIET(meta, NULL);
 	REC_PRINT(__qam_incfirst_print);
-	REC_INTRO(__qam_incfirst_read, 1);
+	REC_INTRO(__qam_incfirst_read, 1, 1);
 
 	metapg = ((QUEUE *)file_dbp->q_internal)->q_meta;
 
 	if ((ret = __db_lget(dbc,
 	    LCK_ROLLBACK, metapg,  DB_LOCK_WRITE, 0, &lock)) != 0)
 		goto done;
-	if ((ret = mpf->get(mpf, &metapg, 0, &meta)) != 0) {
+	if ((ret = __memp_fget(mpf, &metapg, NULL,
+	    0, &meta)) != 0) {
 		if (DB_REDO(op)) {
-			if ((ret = mpf->get(mpf,
-			    &metapg, DB_MPOOL_CREATE, &meta)) != 0) {
+			if ((ret = __memp_fget(mpf, &metapg, NULL,
+			    DB_MPOOL_CREATE, &meta)) != 0) {
 				(void)__LPUT(dbc, lock);
 				goto out;
 			}
@@ -70,14 +87,11 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 			meta->dbmeta.type = P_QAMMETA;
 		} else {
 			*lsnp = argp->prev_lsn;
-			ret = 0;
-			(void)__LPUT(dbc, lock);
+			ret = __LPUT(dbc, lock);
 			goto out;
 		}
 	}
 
-	modified = 0;
-
 	/*
 	 * Only move first_recno backwards so we pick up the aborted delete.
 	 * When going forward we need to be careful since
@@ -85,27 +99,37 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 	 */
 	if (DB_UNDO(op)) {
 		if (QAM_BEFORE_FIRST(meta, argp->recno)) {
+			REC_DIRTY(mpf, &meta);
 			meta->first_recno = argp->recno;
-			modified = 1;
+		}
+
+		trunc_lsn = ((DB_TXNHEAD *)info)->trunc_lsn;
+		/* if we are truncating, update the LSN */
+		if (!IS_ZERO_LSN(trunc_lsn) &&
+		    LOG_COMPARE(&LSN(meta), &trunc_lsn) > 0) {
+			REC_DIRTY(mpf, &meta);
+			LSN(meta) = trunc_lsn;
 		}
 	} else {
-		if (log_compare(&LSN(meta), lsnp) < 0) {
+		if (LOG_COMPARE(&LSN(meta), lsnp) < 0) {
+			REC_DIRTY(mpf, &meta);
 			LSN(meta) = *lsnp;
-			modified = 1;
 		}
-		rec_ext = 0;
-		if (meta->page_ext != 0)
+		if (meta->page_ext == 0)
+			rec_ext = 0;
+		else
 			rec_ext = meta->page_ext * meta->rec_page;
 		cp = (QUEUE_CURSOR *)dbc->internal;
 		if (meta->first_recno == RECNO_OOB)
 			meta->first_recno++;
-		while (meta->first_recno != meta->cur_recno
-		    && !QAM_BEFORE_FIRST(meta, argp->recno + 1)) {
+		while (meta->first_recno != meta->cur_recno &&
+		    !QAM_BEFORE_FIRST(meta, argp->recno + 1)) {
 			if ((ret = __qam_position(dbc,
-			    &meta->first_recno, QAM_READ, &exact)) != 0)
+			    &meta->first_recno, DB_LOCK_READ, 0, &exact)) != 0)
+				goto err;
+			if (cp->page != NULL && (ret =
+			    __qam_fput(file_dbp, cp->pgno, cp->page, 0)) != 0)
 				goto err;
-			if (cp->page != NULL)
-				__qam_fput(file_dbp, cp->pgno, cp->page, 0);
 
 			if (exact == 1)
 				break;
@@ -114,24 +138,25 @@ __qam_incfirst_recover(dbenv, dbtp, lsnp, op, info)
 				if ((ret =
 				    __qam_fremove(file_dbp, cp->pgno)) != 0)
 					goto err;
+			REC_DIRTY(mpf, &meta);
 			meta->first_recno++;
 			if (meta->first_recno == RECNO_OOB)
 				meta->first_recno++;
-			modified = 1;
 		}
 	}
 
-	if ((ret = mpf->put(mpf, meta, modified ? DB_MPOOL_DIRTY : 0)) != 0)
-		goto err1;
-
-	(void)__LPUT(dbc, lock);
+	ret = __memp_fput(mpf, meta, 0);
+	if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+		ret = t_ret;
+	if (ret != 0)
+		goto out;
 
 done:	*lsnp = argp->prev_lsn;
 	ret = 0;
 
 	if (0) {
-err:		(void)mpf->put(mpf, meta, 0);
-err1:		(void)__LPUT(dbc, lock);
+err:		(void)__memp_fput(mpf, meta, 0);
+		(void)__LPUT(dbc, lock);
 	}
 
 out:	REC_CLOSE;
@@ -155,25 +180,27 @@ __qam_mvptr_recover(dbenv, dbtp, lsnp, op, info)
 	__qam_mvptr_args *argp;
 	DB *file_dbp;
 	DBC *dbc;
+	DB_LSN trunc_lsn;
 	DB_LOCK lock;
 	DB_MPOOLFILE *mpf;
 	QMETA *meta;
+	QUEUE_CURSOR *cp;
 	db_pgno_t metapg;
-	int cmp_n, cmp_p, modified, ret;
+	int cmp_n, cmp_p, exact, ret;
 
-	COMPQUIET(info, NULL);
 	REC_PRINT(__qam_mvptr_print);
-	REC_INTRO(__qam_mvptr_read, 1);
+	REC_INTRO(__qam_mvptr_read, 1, 1);
 
 	metapg = ((QUEUE *)file_dbp->q_internal)->q_meta;
 
 	if ((ret = __db_lget(dbc,
 	    LCK_ROLLBACK, metapg,  DB_LOCK_WRITE, 0, &lock)) != 0)
 		goto done;
-	if ((ret = mpf->get(mpf, &metapg, 0, &meta)) != 0) {
+	if ((ret = __memp_fget(mpf, &metapg, NULL,
+	    0, &meta)) != 0) {
 		if (DB_REDO(op)) {
-			if ((ret = mpf->get(mpf,
-			    &metapg, DB_MPOOL_CREATE, &meta)) != 0) {
+			if ((ret = __memp_fget(mpf, &metapg, NULL,
+			    DB_MPOOL_CREATE, &meta)) != 0) {
 				(void)__LPUT(dbc, lock);
 				goto out;
 			}
@@ -181,49 +208,93 @@ __qam_mvptr_recover(dbenv, dbtp, lsnp, op, info)
 			meta->dbmeta.type = P_QAMMETA;
 		} else {
 			*lsnp = argp->prev_lsn;
-			ret = 0;
-			(void)__LPUT(dbc, lock);
+			ret = __LPUT(dbc, lock);
 			goto out;
 		}
 	}
 
-	modified = 0;
-	cmp_n = log_compare(lsnp, &LSN(meta));
-	cmp_p = log_compare(&LSN(meta), &argp->metalsn);
+	cmp_n = LOG_COMPARE(lsnp, &LSN(meta));
+	cmp_p = LOG_COMPARE(&LSN(meta), &argp->metalsn);
 
 	/*
 	 * Under normal circumstances, we never undo a movement of one of
 	 * the pointers.  Just move them along regardless of abort/commit.
+	 * When going forward we need to verify that this is really where
+	 * the pointer belongs.  A transaction may roll back and reinsert
+	 * a record that was missing at the time of this action.
 	 *
 	 * If we're undoing a truncate, we need to reset the pointers to
 	 * their state before the truncate.
 	 */
-	if (DB_UNDO(op) && (argp->opcode & QAM_TRUNCATE)) {
-		if (cmp_n == 0) {
+	if (DB_UNDO(op)) {
+		if ((argp->opcode & QAM_TRUNCATE) && cmp_n <= 0) {
+			REC_DIRTY(mpf, &meta);
 			meta->first_recno = argp->old_first;
 			meta->cur_recno = argp->old_cur;
-			modified = 1;
-			meta->dbmeta.lsn = argp->metalsn;
+			LSN(meta) = argp->metalsn;
+		}
+		/* If the page lsn is beyond the truncate point, move it back */
+		trunc_lsn = ((DB_TXNHEAD *)info)->trunc_lsn;
+		if (!IS_ZERO_LSN(trunc_lsn) &&
+		    LOG_COMPARE(&trunc_lsn, &LSN(meta)) < 0) {
+			REC_DIRTY(mpf, &meta);
+			LSN(meta) = argp->metalsn;
+		}
+	} else if (op == DB_TXN_APPLY || cmp_p == 0) {
+		REC_DIRTY(mpf, &meta);
+		cp = (QUEUE_CURSOR *)dbc->internal;
+		if ((argp->opcode & QAM_SETFIRST) &&
+		    meta->first_recno == argp->old_first) {
+			if (argp->old_first > argp->new_first)
+				meta->first_recno = argp->new_first;
+			else {
+				if ((ret = __qam_position(dbc,
+				    &meta->first_recno, DB_LOCK_READ, 0,
+				    &exact)) != 0)
+					goto err;
+				if (!exact)
+					meta->first_recno = argp->new_first;
+				if (cp->page != NULL &&
+				    (ret = __qam_fput(file_dbp,
+				    cp->pgno, cp->page, 0)) != 0)
+					goto err;
+			}
 		}
-	} else if (cmp_p == 0) {
-		if (argp->opcode & QAM_SETFIRST)
-			meta->first_recno = argp->new_first;
 
-		if (argp->opcode & QAM_SETCUR)
-			meta->cur_recno = argp->new_cur;
+		if ((argp->opcode & QAM_SETCUR) &&
+		    meta->cur_recno == argp->old_cur) {
+			if (argp->old_cur < argp->new_cur)
+				meta->cur_recno = argp->new_cur;
+			else {
+				if ((ret = __qam_position(dbc, &meta->cur_recno,
+				    DB_LOCK_READ, 0, &exact)) != 0)
+					goto err;
+				if (!exact)
+					meta->cur_recno = argp->new_cur;
+				if (cp->page != NULL &&
+				    (ret = __qam_fput(file_dbp,
+				    cp->pgno, cp->page, 0)) != 0)
+					goto err;
+			}
+		}
 
-		modified = 1;
 		meta->dbmeta.lsn = *lsnp;
 	}
 
-	if ((ret = mpf->put(mpf, meta, modified ? DB_MPOOL_DIRTY : 0)) != 0)
+	if ((ret = __memp_fput(mpf, meta, 0)) != 0)
 		goto out;
 
-	(void)__LPUT(dbc, lock);
+	if ((ret = __LPUT(dbc, lock)) != 0)
+		goto out;
 
 done:	*lsnp = argp->prev_lsn;
 	ret = 0;
 
+	if (0) {
+err:		(void)__memp_fput(mpf, meta, 0);
+		(void)__LPUT(dbc, lock);
+	}
+
 out:	REC_CLOSE;
 }
 
@@ -252,24 +323,24 @@ __qam_del_recover(dbenv, dbtp, lsnp, op, info)
 	QMETA *meta;
 	QPAGE *pagep;
 	db_pgno_t metapg;
-	int cmp_n, modified, ret;
+	int cmp_n, ret, t_ret;
 
 	COMPQUIET(info, NULL);
+	COMPQUIET(pagep, NULL);
 	REC_PRINT(__qam_del_print);
-	REC_INTRO(__qam_del_read, 1);
+	REC_INTRO(__qam_del_read, 1, 1);
 
-	if ((ret = __qam_fget(file_dbp,
-	     &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0)
+	if ((ret = __qam_fget(file_dbp, &argp->pgno, NULL,
+	    DB_MPOOL_CREATE, &pagep)) != 0)
 		goto out;
 
-	modified = 0;
 	if (pagep->pgno == PGNO_INVALID) {
+		QAM_DIRTY(file_dbp, argp->pgno, &pagep);
 		pagep->pgno = argp->pgno;
 		pagep->type = P_QAMDATA;
-		modified = 1;
 	}
 
-	cmp_n = log_compare(lsnp, &LSN(pagep));
+	cmp_n = LOG_COMPARE(lsnp, &LSN(pagep));
 
 	if (DB_UNDO(op)) {
 		/* make sure first is behind us */
@@ -277,22 +348,28 @@ __qam_del_recover(dbenv, dbtp, lsnp, op, info)
 		if ((ret = __db_lget(dbc,
 		    LCK_ROLLBACK, metapg, DB_LOCK_WRITE, 0, &lock)) != 0)
 			goto err;
-		if ((ret = mpf->get(mpf, &metapg, 0, &meta)) != 0) {
+		if ((ret = __memp_fget(mpf, &metapg, NULL,
+		    DB_MPOOL_EDIT, &meta)) != 0) {
 			(void)__LPUT(dbc, lock);
 			goto err;
 		}
 		if (meta->first_recno == RECNO_OOB ||
-		    (QAM_BEFORE_FIRST(meta, argp->recno)
-		    && (meta->first_recno <= meta->cur_recno
-		    || meta->first_recno -
+		    (QAM_BEFORE_FIRST(meta, argp->recno) &&
+		    (meta->first_recno <= meta->cur_recno ||
+		    meta->first_recno -
 		    argp->recno < argp->recno - meta->cur_recno))) {
+			REC_DIRTY(mpf, &meta);
 			meta->first_recno = argp->recno;
-			(void)mpf->put(mpf, meta, DB_MPOOL_DIRTY);
+			ret = __memp_fput(mpf, meta, 0);
 		} else
-			(void)mpf->put(mpf, meta, 0);
-		(void)__LPUT(dbc, lock);
+			ret = __memp_fput(mpf, meta, 0);
+		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
+			goto err;
 
 		/* Need to undo delete - mark the record as present */
+		QAM_DIRTY(file_dbp, pagep->pgno, &pagep);
 		qp = QAM_GET_RECORD(file_dbp, pagep, argp->indx);
 		F_SET(qp, QAM_VALID);
 
@@ -303,19 +380,28 @@ __qam_del_recover(dbenv, dbtp, lsnp, op, info)
 		 * foul up a concurrent put.  Having too late an LSN
 		 * is harmless in queue except when we're determining
 		 * what we need to roll forward during recovery.  [#2588]
+		 * If we are aborting a restored transaction then it
+		 * might get rolled forward later so the LSN needs to
+		 * be correct in that case too. [#12181]
 		 */
-		if (op == DB_TXN_BACKWARD_ROLL && cmp_n <= 0)
+		if (cmp_n <= 0 &&
+		      (op == DB_TXN_BACKWARD_ROLL || IS_IN_RESTORE(dbenv)))
 			LSN(pagep) = argp->lsn;
-		modified = 1;
-	} else if (cmp_n > 0 && DB_REDO(op)) {
+	} else if (op == DB_TXN_APPLY || (cmp_n > 0 && DB_REDO(op))) {
 		/* Need to redo delete - clear the valid bit */
+		QAM_DIRTY(file_dbp, pagep->pgno, &pagep);
 		qp = QAM_GET_RECORD(file_dbp, pagep, argp->indx);
 		F_CLR(qp, QAM_VALID);
-		LSN(pagep) = *lsnp;
-		modified = 1;
+		/*
+		 * We only move the LSN forward during replication.
+		 * During recovery we could obscure an update from
+		 * a partially completed transaction while processing
+		 * a hot backup.  [#13823]
+		 */
+		if (op == DB_TXN_APPLY)
+			LSN(pagep) = *lsnp;
 	}
-	if ((ret = __qam_fput(file_dbp,
-	    argp->pgno, pagep, modified ? DB_MPOOL_DIRTY : 0)) != 0)
+	if ((ret = __qam_fput(file_dbp, argp->pgno, pagep, 0)) != 0)
 		goto out;
 
 done:	*lsnp = argp->prev_lsn;
@@ -351,34 +437,31 @@ __qam_delext_recover(dbenv, dbtp, lsnp, op, info)
 	QMETA *meta;
 	QPAGE *pagep;
 	db_pgno_t metapg;
-	int cmp_n, modified, ret;
+	int cmp_n, ret, t_ret;
 
 	COMPQUIET(info, NULL);
+	COMPQUIET(pagep, NULL);
 	REC_PRINT(__qam_delext_print);
-	REC_INTRO(__qam_delext_read, 1);
+	REC_INTRO(__qam_delext_read, 1, 1);
 
-	if ((ret = __qam_fget(file_dbp, &argp->pgno, 0, &pagep)) != 0) {
-		if (ret != DB_PAGE_NOTFOUND && ret != ENOENT)
-			goto out;
+	if ((ret = __qam_fget(file_dbp, &argp->pgno, NULL,
+	     DB_REDO(op) ? 0 : DB_MPOOL_CREATE, &pagep)) != 0) {
 		/*
 		 * If we are redoing a delete and the page is not there
 		 * we are done.
 		 */
-		if (DB_REDO(op))
+		if (DB_REDO(op) && (ret == DB_PAGE_NOTFOUND || ret == ENOENT))
 			goto done;
-		if ((ret = __qam_fget(file_dbp,
-		    &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-			goto out;
+		goto out;
 	}
 
-	modified = 0;
 	if (pagep->pgno == PGNO_INVALID) {
+		QAM_DIRTY(file_dbp, argp->pgno, &pagep);
 		pagep->pgno = argp->pgno;
 		pagep->type = P_QAMDATA;
-		modified = 1;
 	}
 
-	cmp_n = log_compare(lsnp, &LSN(pagep));
+	cmp_n = LOG_COMPARE(lsnp, &LSN(pagep));
 
 	if (DB_UNDO(op)) {
 		/* make sure first is behind us */
@@ -386,21 +469,26 @@ __qam_delext_recover(dbenv, dbtp, lsnp, op, info)
 		if ((ret = __db_lget(dbc,
 		    LCK_ROLLBACK, metapg, DB_LOCK_WRITE, 0, &lock)) != 0)
 			goto err;
-		if ((ret = mpf->get(mpf, &metapg, 0, &meta)) != 0) {
+		if ((ret = __memp_fget(mpf, &metapg, NULL,
+		    DB_MPOOL_EDIT, &meta)) != 0) {
 			(void)__LPUT(dbc, lock);
 			goto err;
 		}
 		if (meta->first_recno == RECNO_OOB ||
-		    (QAM_BEFORE_FIRST(meta, argp->recno)
-		    && (meta->first_recno <= meta->cur_recno
-		    || meta->first_recno -
+		    (QAM_BEFORE_FIRST(meta, argp->recno) &&
+		    (meta->first_recno <= meta->cur_recno ||
+		    meta->first_recno -
 		    argp->recno < argp->recno - meta->cur_recno))) {
 			meta->first_recno = argp->recno;
-			(void)mpf->put(mpf, meta, DB_MPOOL_DIRTY);
+			ret = __memp_fput(mpf, meta, 0);
 		} else
-			(void)mpf->put(mpf, meta, 0);
-		(void)__LPUT(dbc, lock);
+			ret = __memp_fput(mpf, meta, 0);
+		if ((t_ret = __LPUT(dbc, lock)) != 0 && ret == 0)
+			ret = t_ret;
+		if (ret != 0)
+			goto err;
 
+		QAM_DIRTY(file_dbp, pagep->pgno, &pagep);
 		if ((ret = __qam_pitem(dbc, pagep,
 		    argp->indx, argp->recno, &argp->data)) != 0)
 			goto err;
@@ -413,18 +501,24 @@ __qam_delext_recover(dbenv, dbtp, lsnp, op, info)
 		 * is harmless in queue except when we're determining
 		 * what we need to roll forward during recovery.  [#2588]
 		 */
-		if (op == DB_TXN_BACKWARD_ROLL && cmp_n <= 0)
+		if (cmp_n <= 0 &&
+		      (op == DB_TXN_BACKWARD_ROLL || IS_IN_RESTORE(dbenv)))
 			LSN(pagep) = argp->lsn;
-		modified = 1;
-	} else if (cmp_n > 0 && DB_REDO(op)) {
+	} else if (op == DB_TXN_APPLY || (cmp_n > 0 && DB_REDO(op))) {
+		QAM_DIRTY(file_dbp, pagep->pgno, &pagep);
 		/* Need to redo delete - clear the valid bit */
 		qp = QAM_GET_RECORD(file_dbp, pagep, argp->indx);
 		F_CLR(qp, QAM_VALID);
-		LSN(pagep) = *lsnp;
-		modified = 1;
+		/*
+		 * We only move the LSN forward during replication.
+		 * During recovery we could obscure an update from
+		 * a partially completed transaction while processing
+		 * a hot backup.  [#13823]
+		 */
+		if (op == DB_TXN_APPLY)
+			LSN(pagep) = *lsnp;
 	}
-	if ((ret = __qam_fput(file_dbp,
-	    argp->pgno, pagep, modified ? DB_MPOOL_DIRTY : 0)) != 0)
+	if ((ret = __qam_fput(file_dbp, argp->pgno, pagep, 0)) != 0)
 		goto out;
 
 done:	*lsnp = argp->prev_lsn;
@@ -459,59 +553,66 @@ __qam_add_recover(dbenv, dbtp, lsnp, op, info)
 	QMETA *meta;
 	QPAGE *pagep;
 	db_pgno_t metapg;
-	int cmp_n, meta_dirty, modified, ret;
+	int cmp_n, ret;
 
 	COMPQUIET(info, NULL);
+	COMPQUIET(pagep, NULL);
 	REC_PRINT(__qam_add_print);
-	REC_INTRO(__qam_add_read, 1);
+	REC_INTRO(__qam_add_read, 1, 1);
 
-	modified = 0;
-	if ((ret = __qam_fget(file_dbp, &argp->pgno, 0, &pagep)) != 0) {
-		if (ret != DB_PAGE_NOTFOUND && ret != ENOENT)
-			goto out;
+	if ((ret = __qam_fget(file_dbp, &argp->pgno, NULL,
+	     DB_UNDO(op) ? 0 : DB_MPOOL_CREATE, &pagep)) != 0) {
 		/*
 		 * If we are undoing an append and the page is not there
 		 * we are done.
 		 */
-		if (DB_UNDO(op))
+		if (DB_UNDO(op) && (ret == DB_PAGE_NOTFOUND || ret == ENOENT))
 			goto done;
-		if ((ret = __qam_fget(file_dbp,
-		    &argp->pgno, DB_MPOOL_CREATE, &pagep)) != 0)
-			goto out;
+		goto out;
 	}
 
 	if (pagep->pgno == PGNO_INVALID) {
+		QAM_DIRTY(file_dbp, argp->pgno, &pagep);
 		pagep->pgno = argp->pgno;
 		pagep->type = P_QAMDATA;
-		modified = 1;
 	}
 
-	cmp_n = log_compare(lsnp, &LSN(pagep));
+	cmp_n = LOG_COMPARE(lsnp, &LSN(pagep));
 
-	if (cmp_n > 0 && DB_REDO(op)) {
-		/* Need to redo add - put the record on page */
-		if ((ret = __qam_pitem(dbc,
-		    pagep, argp->indx, argp->recno, &argp->data)) != 0)
-			goto err;
-		LSN(pagep) = *lsnp;
-		modified = 1;
-		/* Make sure pointers include this record. */
+	if (DB_REDO(op)) {
+		/* Fix meta-data page. */
 		metapg = ((QUEUE *)file_dbp->q_internal)->q_meta;
-		if ((ret = mpf->get(mpf, &metapg, 0, &meta)) != 0)
+		if ((ret = __memp_fget(mpf, &metapg, NULL,
+		    0, &meta)) != 0)
 			goto err;
-		meta_dirty = 0;
 		if (QAM_BEFORE_FIRST(meta, argp->recno)) {
+			REC_DIRTY(mpf, &meta);
 			meta->first_recno = argp->recno;
-			meta_dirty = 1;
 		}
 		if (argp->recno == meta->cur_recno ||
-		   QAM_AFTER_CURRENT(meta, argp->recno)) {
+		    QAM_AFTER_CURRENT(meta, argp->recno)) {
+			REC_DIRTY(mpf, &meta);
 			meta->cur_recno = argp->recno + 1;
-			meta_dirty = 1;
 		}
-		if ((ret =
-		    mpf->put(mpf, meta, meta_dirty? DB_MPOOL_DIRTY : 0)) != 0)
+		if ((ret = __memp_fput(mpf, meta, 0)) != 0)
 			goto err;
+
+		/* Now update the actual page if necessary. */
+		if (op == DB_TXN_APPLY || cmp_n > 0) {
+			QAM_DIRTY(file_dbp, pagep->pgno, &pagep);
+			/* Need to redo add - put the record on page */
+			if ((ret = __qam_pitem(dbc,
+			    pagep, argp->indx, argp->recno, &argp->data)) != 0)
+				goto err;
+			/*
+			 * We only move the LSN forward during replication.
+			 * During recovery we could obscure an update from
+			 * a partially completed transaction while processing
+			 * a hot backup.  [#13823]
+			 */
+			if (op == DB_TXN_APPLY)
+				LSN(pagep) = *lsnp;
+		}
 	} else if (DB_UNDO(op)) {
 		/*
 		 * Need to undo add
@@ -519,6 +620,7 @@ __qam_add_recover(dbenv, dbtp, lsnp, op, info)
 		 *	Otherwise just clear the valid bit
 		 */
 		if (argp->olddata.size != 0) {
+			QAM_DIRTY(file_dbp, pagep->pgno, &pagep);
 			if ((ret = __qam_pitem(dbc, pagep,
 			    argp->indx, argp->recno, &argp->olddata)) != 0)
 				goto err;
@@ -528,11 +630,10 @@ __qam_add_recover(dbenv, dbtp, lsnp, op, info)
 				    file_dbp, pagep, argp->indx);
 				F_CLR(qp, QAM_VALID);
 			}
-			modified = 1;
 		} else {
+			QAM_DIRTY(file_dbp, pagep->pgno, &pagep);
 			qp = QAM_GET_RECORD(file_dbp, pagep, argp->indx);
 			qp->flags = 0;
-			modified = 1;
 		}
 
 		/*
@@ -543,12 +644,13 @@ __qam_add_recover(dbenv, dbtp, lsnp, op, info)
 		 * is harmless in queue except when we're determining
 		 * what we need to roll forward during recovery.  [#2588]
 		 */
-		if (op == DB_TXN_BACKWARD_ROLL && cmp_n <= 0)
+		if (cmp_n <= 0 &&
+		      (op == DB_TXN_BACKWARD_ROLL || IS_IN_RESTORE(dbenv)))
 			LSN(pagep) = argp->lsn;
 	}
 
 	if ((ret = __qam_fput(file_dbp,
-	    argp->pgno, pagep, modified ? DB_MPOOL_DIRTY : 0)) != 0)
+	    argp->pgno, pagep, 0)) != 0)
 		goto out;
 
 done:	*lsnp = argp->prev_lsn;
@@ -560,149 +662,3 @@ err:		(void)__qam_fput(file_dbp, argp->pgno, pagep, 0);
 
 out:	REC_CLOSE;
 }
-
-/*
- * __qam_delete_recover --
- *	Recovery function for delete of an extent.
- *
- * PUBLIC: int __qam_delete_recover
- * PUBLIC:   __P((DB_ENV *, DBT *, DB_LSN *, db_recops, void *));
- */
-int
-__qam_delete_recover(dbenv, dbtp, lsnp, op, info)
-	DB_ENV *dbenv;
-	DBT *dbtp;
-	DB_LSN *lsnp;
-	db_recops op;
-	void *info;
-{
-	__qam_delete_args *argp;
-	int ret;
-	char *backup, *real_back, *real_name;
-
-	COMPQUIET(info, NULL);
-
-	REC_PRINT(__qam_delete_print);
-
-	backup = real_back = real_name = NULL;
-	if ((ret = __qam_delete_read(dbenv, dbtp->data, &argp)) != 0)
-		goto out;
-
-	if (DB_REDO(op)) {
-		/*
-		 * On a recovery, as we recreate what was going on, we
-		 * recreate the creation of the file.  And so, even though
-		 * it committed, we need to delete it.  Try to delete it,
-		 * but it is not an error if that delete fails.
-		 */
-		if ((ret = __db_appname(dbenv, DB_APP_DATA,
-		    NULL, argp->name.data, 0, NULL, &real_name)) != 0)
-			goto out;
-		if (__os_exists(real_name, NULL) == 0) {
-			if ((ret = __os_unlink(dbenv, real_name)) != 0)
-				goto out;
-		}
-	} else if (DB_UNDO(op)) {
-		/*
-		 * Trying to undo.  File may or may not have been deleted.
-		 * Try to move the backup to the original.  If the backup
-		 * exists, then this is right.  If it doesn't exist, then
-		 * nothing will happen and that's OK.
-		 */
-		if ((ret = __db_backup_name(dbenv, argp->name.data,
-		    &backup, &argp->lsn)) != 0)
-			goto out;
-		if ((ret = __db_appname(dbenv,
-		    DB_APP_DATA, NULL, backup, 0, NULL, &real_back)) != 0)
-			goto out;
-		if ((ret = __db_appname(dbenv, DB_APP_DATA,
-		    NULL, argp->name.data, 0, NULL, &real_name)) != 0)
-			goto out;
-		if (__os_exists(real_back, NULL) == 0)
-			if ((ret =
-			     __os_rename(dbenv, real_back, real_name)) != 0)
-				goto out;
-	}
-	*lsnp = argp->prev_lsn;
-	ret = 0;
-
-out:	if (argp != NULL)
-		__os_free(dbenv, argp, 0);
-	if (backup != NULL)
-		__os_freestr(dbenv, backup);
-	if (real_back != NULL)
-		__os_freestr(dbenv, real_back);
-	if (real_name != NULL)
-		__os_freestr(dbenv, real_name);
-	return (ret);
-}
-
-/*
- * __qam_rename_recover --
- *	Recovery function for rename.
- *
- * PUBLIC: int __qam_rename_recover
- * PUBLIC:   __P((DB_ENV *, DBT *, DB_LSN *, db_recops, void *));
- */
-int
-__qam_rename_recover(dbenv, dbtp, lsnp, op, info)
-	DB_ENV *dbenv;
-	DBT *dbtp;
-	DB_LSN *lsnp;
-	db_recops op;
-	void *info;
-{
-	__qam_rename_args *argp;
-	char *new_name, *real_name;
-	int ret;
-
-	COMPQUIET(info, NULL);
-
-	REC_PRINT(__qam_rename_print);
-
-	new_name = real_name = NULL;
-
-	if ((ret = __qam_rename_read(dbenv, dbtp->data, &argp)) != 0)
-		goto out;
-
-	if (DB_REDO(op)) {
-		if ((ret = __db_appname(dbenv, DB_APP_DATA,
-		    NULL, argp->name.data, 0, NULL, &real_name)) != 0)
-			goto out;
-		if (__os_exists(real_name, NULL) == 0) {
-			if ((ret = __db_appname(dbenv, DB_APP_DATA,
-			    NULL, argp->newname.data, 0, NULL, &new_name)) != 0)
-				goto out;
-			if ((ret =
-			    __os_rename(dbenv, real_name, new_name)) != 0)
-				goto out;
-		}
-	} else {
-		if ((ret = __db_appname(dbenv, DB_APP_DATA,
-		    NULL, argp->newname.data, 0, NULL, &new_name)) != 0)
-			goto out;
-		if (__os_exists(new_name, NULL) == 0) {
-			if ((ret = __db_appname(dbenv,
-			    DB_APP_DATA, NULL, argp->name.data,
-			    0, NULL, &real_name)) != 0)
-				goto out;
-			if ((ret = __os_rename(dbenv,
-			    new_name, real_name)) != 0)
-				goto out;
-		}
-	}
-
-	*lsnp = argp->prev_lsn;
-	ret = 0;
-
-out:	if (argp != NULL)
-		__os_free(dbenv, argp, 0);
-
-	if (new_name != NULL)
-		__os_free(dbenv, new_name, 0);
-
-	if (real_name != NULL)
-		__os_free(dbenv, real_name, 0);
-
-	return (ret);
-}
