@@ -1,30 +1,31 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 2004-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: sequence.c,v 1.26 2004/10/25 17:59:28 bostic Exp $
+ * $Id: sequence.c,v 12.38 2006/08/24 14:46:31 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
-#include "dbinc/db_shash.h"
 #include "dbinc/db_page.h"
-#include "dbinc/db_swap.h"
 #include "dbinc/db_am.h"
+#include "dbinc/lock.h"
 #include "dbinc/mp.h"
+#include "dbinc/txn.h"
 #include "dbinc_auto/sequence_ext.h"
 
-#ifdef HAVE_SEQUENCE
+#ifdef HAVE_RPC
+#ifndef NO_SYSTEM_INCLUDES
+#include <rpc/rpc.h>
+#endif
+#include "db_server.h"
+#include "dbinc_auto/rpc_client_ext.h"
+#endif
+
+#ifdef HAVE_64BIT_TYPES
 #define	SEQ_ILLEGAL_AFTER_OPEN(seq, name)				\
 	if (seq->seq_key.data != NULL)					\
 		return (__db_mi_open((seq)->seq_dbp->dbenv, name, 1));
@@ -33,16 +34,28 @@
 	if (seq->seq_key.data == NULL)					\
 		return (__db_mi_open((seq)->seq_dbp->dbenv, name, 0));
 
-#define SEQ_SWAP(rp)	\
-	do {							\
-		M_32_SWAP((rp)->seq_version);			\
-		M_32_SWAP((rp)->flags);				\
-		M_64_SWAP((rp)->seq_value);			\
-		M_64_SWAP((rp)->seq_max);			\
-		M_64_SWAP((rp)->seq_min);			\
+#define	SEQ_IS_OPEN(seq)	((seq)->seq_key.data != NULL)
+
+/*
+ * Sequences must be architecture independent but they are stored as user
+ * data in databases so the code here must handle the byte ordering.  We
+ * store them in little-endian byte ordering.  If we are on a big-endian
+ * machine we swap in and out when we read from the database. seq->seq_rp
+ * always points to the record in native ordering.
+ *
+ * Version 1 always stored things in native format so if we detect this we
+ * upgrade on the fly and write the record back at open time.
+ */
+#define	SEQ_SWAP(rp)							\
+	do {								\
+		M_32_SWAP((rp)->seq_version);				\
+		M_32_SWAP((rp)->flags);					\
+		M_64_SWAP((rp)->seq_value);				\
+		M_64_SWAP((rp)->seq_max);				\
+		M_64_SWAP((rp)->seq_min);				\
 	} while (0)
 
-#define SEQ_SWAP_IN(seq) \
+#define	SEQ_SWAP_IN(seq) \
 	do {								\
 		if (__db_isbigendian()) {				\
 			memcpy(&seq->seq_record, seq->seq_data.data,	\
@@ -50,8 +63,8 @@
 			SEQ_SWAP(&seq->seq_record);			\
 		}							\
 	} while (0)
-		
-#define SEQ_SWAP_OUT(seq) \
+
+#define	SEQ_SWAP_OUT(seq) \
 	do {								\
 		if (__db_isbigendian()) {				\
 			memcpy(seq->seq_data.data,			\
@@ -59,22 +72,22 @@
 			SEQ_SWAP((DB_SEQ_RECORD*)seq->seq_data.data);	\
 		}							\
 	} while (0)
-		
 
+static int __seq_chk_cachesize __P((DB_ENV *, int32_t, db_seq_t, db_seq_t));
 static int __seq_close __P((DB_SEQUENCE *, u_int32_t));
-static int __seq_get __P((DB_SEQUENCE *,
-			    DB_TXN *, int32_t,  db_seq_t *, u_int32_t));
+static int __seq_get
+	       __P((DB_SEQUENCE *, DB_TXN *, int32_t,  db_seq_t *, u_int32_t));
 static int __seq_get_cachesize __P((DB_SEQUENCE *, int32_t *));
+static int __seq_get_db __P((DB_SEQUENCE *, DB **));
 static int __seq_get_flags __P((DB_SEQUENCE *, u_int32_t *));
 static int __seq_get_key __P((DB_SEQUENCE *, DBT *));
 static int __seq_get_range __P((DB_SEQUENCE *, db_seq_t *, db_seq_t *));
-static int __seq_set_range __P((DB_SEQUENCE *, db_seq_t, db_seq_t));
-static int __seq_get_db __P((DB_SEQUENCE *, DB **));
 static int __seq_initial_value __P((DB_SEQUENCE *, db_seq_t));
-static int __seq_open __P((DB_SEQUENCE *, DB_TXN *, DBT *, u_int32_t));
+static int __seq_open_pp __P((DB_SEQUENCE *, DB_TXN *, DBT *, u_int32_t));
 static int __seq_remove __P((DB_SEQUENCE *, DB_TXN *, u_int32_t));
 static int __seq_set_cachesize __P((DB_SEQUENCE *, int32_t));
 static int __seq_set_flags __P((DB_SEQUENCE *, u_int32_t));
+static int __seq_set_range __P((DB_SEQUENCE *, db_seq_t, db_seq_t));
 static int __seq_update __P((DB_SEQUENCE *, DB_TXN *, int32_t, u_int32_t));
 
 /*
@@ -95,6 +108,12 @@ db_sequence_create(seqp, dbp, flags)
 
 	dbenv = dbp->dbenv;
 
+	DB_ILLEGAL_BEFORE_OPEN(dbp, "db_sequence_create");
+#ifdef HAVE_RPC
+	if (RPC_ON(dbenv))
+		return (__dbcl_dbenv_illegal(dbenv));
+#endif
+
 	/* Check for invalid function flags. */
 	switch (flags) {
 	case 0:
@@ -103,8 +122,6 @@ db_sequence_create(seqp, dbp, flags)
 		return (__db_ferr(dbenv, "db_sequence_create", 0));
 	}
 
-	DB_ILLEGAL_BEFORE_OPEN(dbp, "db_sequence_create");
-
 	/* Allocate the sequence. */
 	if ((ret = __os_calloc(dbenv, 1, sizeof(*seq), &seq)) != 0)
 		return (ret);
@@ -119,7 +136,7 @@ db_sequence_create(seqp, dbp, flags)
 	seq->get_key = __seq_get_key;
 	seq->get_range = __seq_get_range;
 	seq->initial_value = __seq_initial_value;
-	seq->open = __seq_open;
+	seq->open = __seq_open_pp;
 	seq->remove = __seq_remove;
 	seq->set_flags = __seq_set_flags;
 	seq->set_range = __seq_set_range;
@@ -137,7 +154,7 @@ db_sequence_create(seqp, dbp, flags)
  *
  */
 static int
-__seq_open(seq, txn, keyp, flags)
+__seq_open_pp(seq, txn, keyp, flags)
 	DB_SEQUENCE *seq;
 	DB_TXN *txn;
 	DBT *keyp;
@@ -145,39 +162,56 @@ __seq_open(seq, txn, keyp, flags)
 {
 	DB *dbp;
 	DB_ENV *dbenv;
-	DB_MPOOL *dbmp;
 	DB_SEQ_RECORD *rp;
+	DB_THREAD_INFO *ip;
 	u_int32_t tflags;
-	int txn_local, ret;
-#define	SEQ_OPEN_FLAGS (DB_AUTO_COMMIT | DB_CREATE | DB_EXCL | DB_THREAD)
+	int handle_check, txn_local, ret, t_ret;
+#define	SEQ_OPEN_FLAGS	(DB_CREATE | DB_EXCL | DB_THREAD)
 
 	dbp = seq->seq_dbp;
 	dbenv = dbp->dbenv;
 	txn_local = 0;
 
+	STRIP_AUTO_COMMIT(flags);
 	SEQ_ILLEGAL_AFTER_OPEN(seq, "DB_SEQUENCE->open");
-	if (keyp->size == 0) {
-		__db_err(dbenv, "Zero length sequence key specified");
-		return (EINVAL);
+
+	ENV_ENTER(dbenv, ip);
+
+	/* Check for replication block. */
+	handle_check = IS_ENV_REPLICATED(dbenv);
+	if (handle_check &&
+	    (ret = __db_rep_enter(dbp, 1, 0, txn != NULL)) != 0) {
+		handle_check = 0;
+		goto err;
 	}
 
-	if (LF_ISSET(~SEQ_OPEN_FLAGS))
-		return (__db_ferr(dbenv, "DB_SEQUENCE->open", 0));
+	if ((ret = __db_fchk(dbenv,
+	    "DB_SEQUENCE->open", flags, SEQ_OPEN_FLAGS)) != 0)
+		goto err;
 
-	if ((ret = dbp->get_flags(dbp, &tflags)) != 0)
-		return (ret);
+	if (keyp->size == 0) {
+		__db_errx(dbenv, "Zero length sequence key specified");
+		goto err;
+	}
+
+	if ((ret = __db_get_flags(dbp, &tflags)) != 0)
+		goto err;
 
 	if (FLD_ISSET(tflags, DB_DUP)) {
-		__db_err(dbenv,
+		__db_errx(dbenv,
 	"Sequences not supported in databases configured for duplicate data");
-		return (EINVAL);
+		goto err;
 	}
 
 	if (LF_ISSET(DB_THREAD)) {
-		dbmp = dbenv->mp_handle;
-		if ((ret = __db_mutex_setup(dbenv, dbmp->reginfo,
-		     &seq->seq_mutexp, MUTEX_ALLOC | MUTEX_THREAD)) != 0)
-			return (ret);
+		if (RPC_ON(dbenv)) {
+			__db_errx(dbenv,
+		    "DB_SEQUENCE->open: DB_THREAD not supported with RPC");
+			goto err;
+		}
+		if ((ret = __mutex_alloc(dbenv,
+		    MTX_SEQUENCE, DB_MUTEX_PROCESS_ONLY, &seq->mtx_seq)) != 0)
+			goto err;
 	}
 
 	memset(&seq->seq_data, 0, sizeof(DBT));
@@ -196,14 +230,12 @@ __seq_open(seq, txn, keyp, flags)
 
 	memset(&seq->seq_key, 0, sizeof(DBT));
 	if ((ret = __os_malloc(dbenv, keyp->size, &seq->seq_key.data)) != 0)
-		return (ret);
+		goto err;
 	memcpy(seq->seq_key.data, keyp->data, keyp->size);
 	seq->seq_key.size = seq->seq_key.ulen = keyp->size;
 	seq->seq_key.flags = DB_DBT_USERMEM;
 
-		
-
-retry:	if ((ret = dbp->get(dbp, txn, &seq->seq_key, &seq->seq_data, 0)) != 0) {
+retry:	if ((ret = __db_get(dbp, txn, &seq->seq_key, &seq->seq_data, 0)) != 0) {
 		if (ret == DB_BUFFER_SMALL &&
 		    seq->seq_data.size > sizeof(seq->seq_record)) {
 			seq->seq_data.flags = DB_DBT_REALLOC;
@@ -216,8 +248,6 @@ retry:	if ((ret = dbp->get(dbp, txn, &seq->seq_key, &seq->seq_data, 0)) != 0) {
 		ret = 0;
 
 		rp = &seq->seq_record;
-		tflags = DB_NOOVERWRITE;
-		tflags |= LF_ISSET(DB_AUTO_COMMIT);
 		if (!F_ISSET(rp, DB_SEQ_RANGE_SET)) {
 			rp->seq_max = INT64_MAX;
 			rp->seq_min = INT64_MIN;
@@ -230,14 +260,22 @@ retry:	if ((ret = dbp->get(dbp, txn, &seq->seq_key, &seq->seq_data, 0)) != 0) {
 
 		if (rp->seq_value > rp->seq_max ||
 		    rp->seq_value < rp->seq_min) {
-			__db_err(dbenv, "Sequence value out of range");
+			__db_errx(dbenv, "Sequence value out of range");
 			ret = EINVAL;
 			goto err;
 		} else {
 			SEQ_SWAP_OUT(seq);
-			if ((ret = dbp->put(dbp, txn,
-			    &seq->seq_key, &seq->seq_data, tflags)) != 0) {
-				__db_err(dbenv, "Sequence create failed");
+			/* Create local transaction as necessary. */
+			if (IS_DB_AUTO_COMMIT(dbp, txn)) {
+				if ((ret =
+				    __txn_begin(dbenv, NULL, &txn, 0)) != 0)
+					goto err;
+				txn_local = 1;
+			}
+
+			if ((ret = __db_put(dbp, txn, &seq->seq_key,
+			     &seq->seq_data, DB_NOOVERWRITE)) != 0) {
+				__db_errx(dbenv, "Sequence create failed");
 				goto err;
 			}
 		}
@@ -245,7 +283,7 @@ retry:	if ((ret = dbp->get(dbp, txn, &seq->seq_key, &seq->seq_data, 0)) != 0) {
 		ret = EEXIST;
 		goto err;
 	} else if (seq->seq_data.size < sizeof(seq->seq_record)) {
-		__db_err(dbenv, "Bad sequence record format");
+		__db_errx(dbenv, "Bad sequence record format");
 		ret = EINVAL;
 		goto err;
 	}
@@ -255,25 +293,23 @@ retry:	if ((ret = dbp->get(dbp, txn, &seq->seq_key, &seq->seq_data, 0)) != 0) {
 
 	/*
 	 * The first release was stored in native mode.
-	 * Check the verison number before swapping.
+	 * Check the version number before swapping.
 	 */
 	rp = seq->seq_data.data;
 	if (rp->seq_version == DB_SEQUENCE_OLDVER) {
 oldver:		rp->seq_version = DB_SEQUENCE_VERSION;
 		if (__db_isbigendian()) {
-			if (IS_AUTO_COMMIT(dbp, txn, flags)) {
+			if (IS_DB_AUTO_COMMIT(dbp, txn)) {
 				if ((ret =
-				     __db_txn_auto_init(dbenv, &txn)) != 0)
-					return (ret);
+				    __txn_begin(dbenv, NULL, &txn, 0)) != 0)
+					goto err;
 				txn_local = 1;
-				LF_CLR(DB_AUTO_COMMIT);
 				goto retry;
-			} else
-				txn_local = 0;
+			}
 			memcpy(&seq->seq_record, rp, sizeof(seq->seq_record));
 			SEQ_SWAP_OUT(seq);
 		}
-		if ((ret = dbp->put(dbp,
+		if ((ret = __db_put(dbp,
 		     txn, &seq->seq_key, &seq->seq_data, 0)) != 0)
 			goto err;
 	}
@@ -299,8 +335,8 @@ oldver:		rp->seq_version = DB_SEQUENCE_VERSION;
 			goto oldver;
 		}
 		M_32_SWAP(rp->seq_version);
-		__db_err(dbenv,
-		     "Unknown sequence version: %d", rp->seq_version);
+		__db_errx(dbenv,
+		     "Unsupported sequence version: %d", rp->seq_version);
 		goto err;
 	}
 
@@ -310,11 +346,26 @@ oldver:		rp->seq_version = DB_SEQUENCE_VERSION;
 	else
 		seq->seq_last_value++;
 
-err:	if (ret != 0) {
+	/*
+	 * It's an error to specify a cache larger than the range of sequences.
+	 */
+	if (seq->seq_cache_size != 0 && (ret = __seq_chk_cachesize(
+	    dbenv, seq->seq_cache_size, rp->seq_max, rp->seq_min)) != 0)
+		goto err;
+
+err:	if (txn_local &&
+	    (t_ret = __db_txn_auto_resolve(dbenv, txn, 0, ret)) && ret == 0)
+		ret = t_ret;
+	if (ret != 0) {
 		__os_free(dbenv, seq->seq_key.data);
 		seq->seq_key.data = NULL;
 	}
-	return (txn_local ? __db_txn_auto_resolve(dbenv, txn, 0, ret) : ret);
+	/* Release replication block. */
+	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+		ret = t_ret;
+
+	ENV_LEAVE(dbenv, ip);
+	return (ret);
 }
 
 /*
@@ -343,11 +394,23 @@ __seq_set_cachesize(seq, cachesize)
 	DB_SEQUENCE *seq;
 	int32_t cachesize;
 {
+	DB_ENV *dbenv;
+	int ret;
+
+	dbenv = seq->seq_dbp->dbenv;
+
 	if (cachesize < 0) {
-		__db_err(seq->seq_dbp->dbenv,
-		     "Illegal cache size: %d", cachesize);
+		__db_errx(dbenv, "Cache size must be >= 0");
 		return (EINVAL);
 	}
+
+	/*
+	 * It's an error to specify a cache larger than the range of sequences.
+	 */
+	if (SEQ_IS_OPEN(seq) && (ret = __seq_chk_cachesize(dbenv,
+	    cachesize, seq->seq_rp->seq_max, seq->seq_rp->seq_min)) != 0)
+		return (ret);
+
 	seq->seq_cache_size = cachesize;
 	return (0);
 }
@@ -385,11 +448,12 @@ __seq_set_flags(seq, flags)
 
 	dbenv = seq->seq_dbp->dbenv;
 	rp = seq->seq_rp;
+
 	SEQ_ILLEGAL_AFTER_OPEN(seq, "DB_SEQUENCE->set_flags");
 
-	if (LF_ISSET(~SEQ_SET_FLAGS))
-		return (__db_ferr(dbenv, "DB_SEQUENCE->set_flags", 0));
-
+	if ((ret = __db_fchk(
+	    dbenv, "DB_SEQUENCE->set_flags", flags, SEQ_SET_FLAGS)) != 0)
+		return (ret);
 	if ((ret = __db_fcchk(dbenv,
 	     "DB_SEQUENCE->set_flags", flags, DB_SEQ_DEC, DB_SEQ_INC)) != 0)
 		return (ret);
@@ -403,7 +467,7 @@ __seq_set_flags(seq, flags)
 
 /*
  * __seq_initial_value --
- *	DB_SEQUENCE->init_value.
+ *	DB_SEQUENCE->initial_value.
  *
  */
 static int
@@ -415,12 +479,12 @@ __seq_initial_value(seq, value)
 	DB_SEQ_RECORD *rp;
 
 	dbenv = seq->seq_dbp->dbenv;
-	SEQ_ILLEGAL_AFTER_OPEN(seq, "DB_SEQUENCE->init_value");
+	SEQ_ILLEGAL_AFTER_OPEN(seq, "DB_SEQUENCE->initial_value");
 
 	rp = seq->seq_rp;
 	if (F_ISSET(rp, DB_SEQ_RANGE_SET) &&
 	     (value > rp->seq_max || value < rp->seq_min)) {
-		__db_err(dbenv, "Sequence value out of range");
+		__db_errx(dbenv, "Sequence value out of range");
 		return (EINVAL);
 	}
 
@@ -441,7 +505,6 @@ __seq_get_range(seq, minp, maxp)
 {
 	SEQ_ILLEGAL_BEFORE_OPEN(seq, "DB_SEQUENCE->get_range");
 
-	F_SET(seq->seq_rp, DB_SEQ_RANGE_SET);
 	*minp = seq->seq_rp->seq_min;
 	*maxp = seq->seq_rp->seq_max;
 	return (0);
@@ -463,7 +526,8 @@ __seq_set_range(seq, min, max)
 	SEQ_ILLEGAL_AFTER_OPEN(seq, "DB_SEQUENCE->set_range");
 
 	if (min >= max) {
-		__db_err(dbenv, "Illegal sequence range");
+		__db_errx(dbenv,
+	    "Minimum sequence value must be less than maximum sequence value");
 		return (EINVAL);
 	}
 
@@ -485,16 +549,28 @@ __seq_update(seq, txn, delta, flags)
 	DB_ENV *dbenv;
 	DB_SEQ_RECORD *rp;
 	int32_t adjust;
-	int ret;
+	int ret, txn_local;
 
 	dbp = seq->seq_dbp;
 	dbenv = dbp->dbenv;
 
-	if (LF_ISSET(DB_AUTO_COMMIT) &&
-	    (ret = __db_txn_auto_init(dbenv, &txn)) != 0)
-		return (ret);
-retry:
-	if ((ret = dbp->get(dbp, txn, &seq->seq_key, &seq->seq_data, 0)) != 0) {
+	/*
+	 * Create a local transaction as necessary, check for consistent
+	 * transaction usage, and, if we have no transaction but do have
+	 * locking on, acquire a locker id for the handle lock acquisition.
+	 */
+	if (IS_DB_AUTO_COMMIT(dbp, txn)) {
+		if ((ret = __txn_begin(dbenv, NULL, &txn, 0)) != 0)
+			return (ret);
+		txn_local = 1;
+	} else
+		txn_local = 0;
+
+	/* Check for consistent transaction usage. */
+	if ((ret = __db_check_txn(dbp, txn, DB_LOCK_INVALIDID, 0)) != 0)
+		goto err;
+
+retry:	if ((ret = __db_get(dbp, txn, &seq->seq_key, &seq->seq_data, 0)) != 0) {
 		if (ret == DB_BUFFER_SMALL &&
 		    seq->seq_data.size > sizeof(seq->seq_record)) {
 			seq->seq_data.flags = DB_DBT_REALLOC;
@@ -509,8 +585,11 @@ retry:
 	SEQ_SWAP_IN(seq);
 	rp = seq->seq_rp;
 
+	if (F_ISSET(rp, DB_SEQ_WRAPPED))
+		goto overflow;
+
 	if (seq->seq_data.size < sizeof(seq->seq_record)) {
-		__db_err(dbenv, "Bad sequence record format");
+		__db_errx(dbenv, "Bad sequence record format");
 		ret = EINVAL;
 		goto err;
 	}
@@ -522,42 +601,56 @@ retry:
 	 *
 	 * The sequence minimum and maximum values can be INT64_MIN and
 	 * INT64_MAX, so we need to do the test carefully to cope with
-	 * arithmetic overflow.  That means we need to check whether the value
-	 * is in a range, we can't get away with a single comparison.
-	 *
-	 * For example, if seq_value == -1 and seq_max == INT64_MAX, the first
-	 * test below will be true, since -1 - (INT64_MAX + 1) == INT64_MAX.
-	 * The second part of the test makes sure that seq_value is close
-	 * enough to the maximum to really cause wrapping.
+	 * arithmetic overflow.  The first part of the test below checks
+	 * whether we will hit the end of the 64-bit range.  The second part
+	 * checks whether we hit the end of the sequence.
 	 */
-	if (F_ISSET(rp, DB_SEQ_INC)) {
-		if (rp->seq_value - ((rp->seq_max - adjust) + 2) >= 0 &&
-		    (rp->seq_max + 1) - rp->seq_value >= 0) {
+again:	if (F_ISSET(rp, DB_SEQ_INC)) {
+		if (rp->seq_value + adjust - 1 < rp->seq_value ||
+		     rp->seq_value + adjust - 1 > rp->seq_max) {
+			/* Don't wrap just to fill the cache. */
+			if (adjust > delta) {
+				adjust = delta;
+				goto again;
+			}
 			if (F_ISSET(rp, DB_SEQ_WRAP))
 				rp->seq_value = rp->seq_min;
 			else {
-overflow:			__db_err(dbenv, "Sequence overflow");
+overflow:			__db_errx(dbenv, "Sequence overflow");
 				ret = EINVAL;
 				goto err;
 			}
 		}
+		/* See if we are at the end of the 64 bit range. */
+		if (!F_ISSET(rp, DB_SEQ_WRAP) &&
+		    rp->seq_value + adjust < rp->seq_value)
+			F_SET(rp, DB_SEQ_WRAPPED);
 	} else {
-		if (rp->seq_value - (rp->seq_min - 1) >= 0 &&
-		    (rp->seq_min + adjust - 2) - rp->seq_value >= 0) {
+		if ((rp->seq_value - adjust) + 1 > rp->seq_value ||
+		   (rp->seq_value - adjust) + 1 < rp->seq_min) {
+			/* Don't wrap just to fill the cache. */
+			if (adjust > delta) {
+				adjust = delta;
+				goto again;
+			}
 			if (F_ISSET(rp, DB_SEQ_WRAP))
 				rp->seq_value = rp->seq_max;
 			else
 				goto overflow;
 		}
+		/* See if we are at the end of the 64 bit range. */
+		if (!F_ISSET(rp, DB_SEQ_WRAP) &&
+		    rp->seq_value - adjust > rp->seq_value)
+			F_SET(rp, DB_SEQ_WRAPPED);
 		adjust = -adjust;
 	}
 
 	rp->seq_value += adjust;
 	SEQ_SWAP_OUT(seq);
-	ret = dbp->put(dbp, txn, &seq->seq_key, &seq->seq_data, 0);
+	ret = __db_put(dbp, txn, &seq->seq_key, &seq->seq_data, 0);
 	rp->seq_value -= adjust;
 	if (ret != 0) {
-		__db_err(dbenv, "Sequence update failed");
+		__db_errx(dbenv, "Sequence update failed");
 		goto err;
 	}
 	seq->seq_last_value = rp->seq_value + adjust;
@@ -566,11 +659,8 @@ overflow:			__db_err(dbenv, "Sequence overflow");
 	else
 		seq->seq_last_value++;
 
-err:	if (LF_ISSET(DB_AUTO_COMMIT))
-		ret = __db_txn_auto_resolve(dbenv,
-		    txn, LF_ISSET(DB_TXN_NOSYNC), ret);
-	return (ret);
-
+err:	return (txn_local ? __db_txn_auto_resolve(
+	    dbenv, txn, LF_ISSET(DB_TXN_NOSYNC), ret) : ret);
 }
 
 static int
@@ -584,23 +674,39 @@ __seq_get(seq, txn, delta, retp, flags)
 	DB *dbp;
 	DB_ENV *dbenv;
 	DB_SEQ_RECORD *rp;
-	int ret;
+	DB_THREAD_INFO *ip;
+	int handle_check, ret, t_ret;
 
 	dbp = seq->seq_dbp;
 	dbenv = dbp->dbenv;
 	rp = seq->seq_rp;
 	ret = 0;
 
+	STRIP_AUTO_COMMIT(flags);
 	SEQ_ILLEGAL_BEFORE_OPEN(seq, "DB_SEQUENCE->get");
 
 	if (delta <= 0) {
-		__db_err(dbenv, "Sequence delta must be greater than 0");
+		__db_errx(dbenv, "Sequence delta must be greater than 0");
 		return (EINVAL);
 	}
-	MUTEX_THREAD_LOCK(dbenv, seq->seq_mutexp);
+
+	if (seq->seq_cache_size != 0 && txn != NULL) {
+		__db_errx(dbenv,
+	    "Sequence with non-zero cache may not specify transaction handle");
+		return (EINVAL);
+	}
+
+	ENV_ENTER(dbenv, ip);
+
+	/* Check for replication block. */
+	handle_check = IS_ENV_REPLICATED(dbenv);
+	if (handle_check && (ret = __db_rep_enter(dbp, 1, 0, txn != NULL)) != 0)
+		return (ret);
+
+	MUTEX_LOCK(dbenv, seq->mtx_seq);
 
 	if (rp->seq_min + delta > rp->seq_max) {
-		__db_err(dbenv, "Sequence overflow");
+		__db_errx(dbenv, "Sequence overflow");
 		ret = EINVAL;
 		goto err;
 	}
@@ -623,14 +729,19 @@ __seq_get(seq, txn, delta, retp, flags)
 		rp->seq_value -= delta;
 	}
 
-err:	MUTEX_THREAD_UNLOCK(dbenv, seq->seq_mutexp);
+err:	MUTEX_UNLOCK(dbenv, seq->mtx_seq);
 
+	/* Release replication block. */
+	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+		ret = t_ret;
+
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
 /*
  * __seq_get_db --
- *	Accessor for dbp passed into DB_SEQUENCE->open call
+ *	Accessor for dbp passed into db_sequence_create call
  *
  */
 static int
@@ -638,8 +749,6 @@ __seq_get_db(seq, dbpp)
 	DB_SEQUENCE *seq;
 	DB **dbpp;
 {
-	SEQ_ILLEGAL_BEFORE_OPEN(seq, "DB_SEQUENCE->get_db");
-
 	*dbpp = seq->seq_dbp;
 	return (0);
 }
@@ -673,25 +782,27 @@ __seq_close(seq, flags)
 	u_int32_t flags;
 {
 	DB_ENV *dbenv;
-	DB_MPOOL *dbmp;
-	int ret;
+	int ret, t_ret;
 
 	ret = 0;
 	dbenv = seq->seq_dbp->dbenv;
 
 	if (flags != 0)
 		ret = __db_ferr(dbenv, "DB_SEQUENCE->close", 0);
-	if (seq->seq_mutexp != NULL) {
-		dbmp = dbenv->mp_handle;
-		__db_mutex_free(dbenv, dbmp->reginfo, seq->seq_mutexp);
-	}
+
+	if ((t_ret = __mutex_free(dbenv, &seq->mtx_seq)) != 0 && ret == 0)
+		ret = t_ret;
+
 	if (seq->seq_key.data != NULL)
 		__os_free(dbenv, seq->seq_key.data);
-	if (seq->seq_data.data != &seq->seq_record)
+	if (seq->seq_data.data != NULL &&
+	    seq->seq_data.data != &seq->seq_record)
 		__os_ufree(dbenv, seq->seq_data.data);
 	seq->seq_key.data = NULL;
+
 	memset(seq, CLEAR_BYTE, sizeof(*seq));
 	__os_free(dbenv, seq);
+
 	return (ret);
 }
 
@@ -707,29 +818,63 @@ __seq_remove(seq, txn, flags)
 {
 	DB *dbp;
 	DB_ENV *dbenv;
-	int ret, t_ret;
+	DB_THREAD_INFO *ip;
+	int handle_check, ret, t_ret;
 
 	dbp = seq->seq_dbp;
 	dbenv = dbp->dbenv;
 
 	SEQ_ILLEGAL_BEFORE_OPEN(seq, "DB_SEQUENCE->remove");
+	ENV_ENTER(dbenv, ip);
 
-	if (LF_ISSET(DB_AUTO_COMMIT) &&
-	    (ret = __db_txn_auto_init(dbenv, &txn)) != 0)
+	/* Check for replication block. */
+	handle_check = IS_ENV_REPLICATED(dbenv);
+	if (handle_check &&
+	    (ret = __db_rep_enter(dbp, 1, 0, txn != NULL)) != 0) {
+		handle_check = 0;
 		goto err;
+	}
+	if (flags != 0)
+		ret = __db_ferr(dbenv, "DB_SEQUENCE->remove", 0);
 
-	ret = dbp->del(dbp, txn, &seq->seq_key, 0);
+	ret = __db_del(dbp, txn, &seq->seq_key, 0);
 
-	if (LF_ISSET(DB_AUTO_COMMIT))
-		ret = __db_txn_auto_resolve(dbenv,
-		    txn, LF_ISSET(DB_TXN_NOSYNC), ret);
-
-err: if ((t_ret = __seq_close(seq, 0)) != 0 && ret == 0)
+	if ((t_ret = __seq_close(seq, 0)) != 0 && ret == 0)
 		ret = t_ret;
+
+	/* Release replication block. */
+	if (handle_check && (t_ret = __env_db_rep_exit(dbenv)) != 0 && ret == 0)
+		ret = t_ret;
+err:	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
-#else /* !HAVE_SEQUENCE */
+/*
+ * __seq_chk_cachesize --
+ *	Validate the cache size vs. the range.
+ */
+static int
+__seq_chk_cachesize(dbenv, cachesize, max, min)
+	DB_ENV *dbenv;
+	int32_t cachesize;
+	db_seq_t max, min;
+{
+	/*
+	 * It's an error to specify caches larger than the sequence range.
+	 *
+	 * The min and max of the range can be either positive or negative,
+	 * the difference will fit in an unsigned variable of the same type.
+	 * Assume a 2's complement machine, and simply subtract.
+	 */
+	if ((u_int32_t)cachesize > (u_int64_t)max - (u_int64_t)min) {
+		__db_errx(dbenv,
+    "Number of items to be cached is larger than the sequence range");
+		return (EINVAL);
+	}
+	return (0);
+}
+
+#else /* !HAVE_64BIT_TYPES */
 
 int
 db_sequence_create(seqp, dbp, flags)
@@ -739,8 +884,8 @@ db_sequence_create(seqp, dbp, flags)
 {
 	COMPQUIET(seqp, NULL);
 	COMPQUIET(flags, 0);
-	__db_err(dbp->dbenv,
+	__db_errx(dbp->dbenv,
 	    "library build did not include support for sequences");
 	return (DB_OPNOTSUP);
 }
-#endif /* HAVE_SEQUENCE */
+#endif /* HAVE_64BIT_TYPES */
