@@ -1,27 +1,22 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1999-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: tcl_log.c,v 12.10 2006/08/24 14:46:33 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: tcl_log.c,v 11.35 2001/11/01 13:50:59 krinsky Exp $";
-#endif /* not lint */
-
+#include "db_int.h"
 #ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
 #include <tcl.h>
 #endif
+#include "dbinc/log.h"
+#include "dbinc/tcl_db.h"
 
-#include "db_int.h"
-#include "tcl_db.h"
-
+#ifdef CONFIG_TEST
 static int tcl_LogcGet __P((Tcl_Interp *, int, Tcl_Obj * CONST*, DB_LOGC *));
 
 /*
@@ -37,12 +32,12 @@ tcl_LogArchive(interp, objc, objv, envp)
 	Tcl_Obj *CONST objv[];		/* The argument objects */
 	DB_ENV *envp;			/* Environment pointer */
 {
-	static char *archopts[] = {
-		"-arch_abs",	"-arch_data",	"-arch_log",
+	static const char *archopts[] = {
+		"-arch_abs",	"-arch_data",	"-arch_log",	"-arch_remove",
 		NULL
 	};
 	enum archopts {
-		ARCH_ABS,	ARCH_DATA,	ARCH_LOG
+		ARCH_ABS,	ARCH_DATA,	ARCH_LOG,	ARCH_REMOVE
 	};
 	Tcl_Obj *fileobj, *res;
 	u_int32_t flag;
@@ -71,16 +66,19 @@ tcl_LogArchive(interp, objc, objv, envp)
 		case ARCH_LOG:
 			flag |= DB_ARCH_LOG;
 			break;
+		case ARCH_REMOVE:
+			flag |= DB_ARCH_REMOVE;
+			break;
 		}
 	}
 	_debug_check();
 	list = NULL;
 	ret = envp->log_archive(envp, &list, flag);
-	result = _ReturnSetup(interp, ret, "log archive");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "log archive");
 	if (result == TCL_OK) {
 		res = Tcl_NewListObj(0, NULL);
 		for (file = list; file != NULL && *file != NULL; file++) {
-			fileobj = Tcl_NewStringObj(*file, strlen(*file));
+			fileobj = NewStringObj(*file, strlen(*file));
 			result = Tcl_ListObjAppendElement(interp, res, fileobj);
 			if (result != TCL_OK)
 				break;
@@ -88,7 +86,7 @@ tcl_LogArchive(interp, objc, objv, envp)
 		Tcl_SetObjResult(interp, res);
 	}
 	if (list != NULL)
-		__os_free(envp, list, 0);
+		__os_ufree(envp, list);
 	return (result);
 }
 
@@ -168,7 +166,7 @@ tcl_LogFile(interp, objc, objv, envp)
 	name = NULL;
 	while (ret == ENOMEM) {
 		if (name != NULL)
-			__os_free(envp, name, len/2);
+			__os_free(envp, name);
 		ret = __os_malloc(envp, len, &name);
 		if (ret != 0) {
 			Tcl_SetResult(interp, db_strerror(ret), TCL_STATIC);
@@ -178,14 +176,14 @@ tcl_LogFile(interp, objc, objv, envp)
 		ret = envp->log_file(envp, &lsn, name, len);
 		len *= 2;
 	}
-	result = _ReturnSetup(interp, ret, "log_file");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "log_file");
 	if (ret == 0) {
-		res = Tcl_NewStringObj(name, strlen(name));
+		res = NewStringObj(name, strlen(name));
 		Tcl_SetObjResult(interp, res);
 	}
 
 	if (name != NULL)
-		__os_free(envp, name, len/2);
+		__os_free(envp, name);
 
 	return (result);
 }
@@ -225,7 +223,7 @@ tcl_LogFlush(interp, objc, objv, envp)
 
 	_debug_check();
 	ret = envp->log_flush(envp, lsnp);
-	result = _ReturnSetup(interp, ret, "log_flush");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "log_flush");
 	return (result);
 }
 
@@ -264,21 +262,23 @@ tcl_LogPut(interp, objc, objv, envp)
 	Tcl_Obj *CONST objv[];		/* The argument objects */
 	DB_ENV *envp;			/* Environment pointer */
 {
-	static char *logputopts[] = {
-		"-checkpoint",	"-curlsn",	"-flush",
+	static const char *logputopts[] = {
+		"-flush",
 		NULL
 	};
 	enum logputopts {
-		LOGPUT_CKP,	LOGPUT_CUR,	LOGPUT_FLUSH
+		LOGPUT_FLUSH
 	};
 	DB_LSN lsn;
 	DBT data;
 	Tcl_Obj *intobj, *res;
+	void *dtmp;
 	u_int32_t flag;
-	int itmp, optindex, result, ret;
+	int freedata, optindex, result, ret;
 
 	result = TCL_OK;
 	flag = 0;
+	freedata = 0;
 	if (objc < 3) {
 		Tcl_WrongNumArgs(interp, 2, objv, "?-args? record");
 		return (TCL_ERROR);
@@ -288,8 +288,14 @@ tcl_LogPut(interp, objc, objv, envp)
 	 * Data/record must be the last arg.
 	 */
 	memset(&data, 0, sizeof(data));
-	data.data = Tcl_GetByteArrayFromObj(objv[objc-1], &itmp);
-	data.size = itmp;
+	ret = _CopyObjBytes(interp, objv[objc-1], &dtmp,
+	    &data.size, &freedata);
+	if (ret != 0) {
+		result = _ReturnSetup(interp, ret,
+		    DB_RETOK_STD(ret), "log put");
+		return (result);
+	}
+	data.data = dtmp;
 
 	/*
 	 * Get the command name index from the object based on the options
@@ -301,12 +307,6 @@ tcl_LogPut(interp, objc, objv, envp)
 			return (IS_HELP(objv[2]));
 		}
 		switch ((enum logputopts)optindex) {
-		case LOGPUT_CKP:
-			flag = DB_CHECKPOINT;
-			break;
-		case LOGPUT_CUR:
-			flag = DB_CURLSN;
-			break;
 		case LOGPUT_FLUSH:
 			flag = DB_FLUSH;
 			break;
@@ -318,68 +318,19 @@ tcl_LogPut(interp, objc, objv, envp)
 
 	_debug_check();
 	ret = envp->log_put(envp, &lsn, &data, flag);
-	result = _ReturnSetup(interp, ret, "log_put");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "log_put");
 	if (result == TCL_ERROR)
 		return (result);
 	res = Tcl_NewListObj(0, NULL);
-	intobj = Tcl_NewLongObj((long)lsn.file);
+	intobj = Tcl_NewWideIntObj((Tcl_WideInt)lsn.file);
 	result = Tcl_ListObjAppendElement(interp, res, intobj);
-	intobj = Tcl_NewLongObj((long)lsn.offset);
+	intobj = Tcl_NewWideIntObj((Tcl_WideInt)lsn.offset);
 	result = Tcl_ListObjAppendElement(interp, res, intobj);
 	Tcl_SetObjResult(interp, res);
+	if (freedata)
+		__os_free(NULL, dtmp);
 	return (result);
 }
-
-/*
- * tcl_LogRegister --
- *
- * PUBLIC: int tcl_LogRegister __P((Tcl_Interp *, int,
- * PUBLIC:    Tcl_Obj * CONST*, DB_ENV *));
- */
-int
-tcl_LogRegister(interp, objc, objv, envp)
-	Tcl_Interp *interp;		/* Interpreter */
-	int objc;			/* How many arguments? */
-	Tcl_Obj *CONST objv[];		/* The argument objects */
-	DB_ENV *envp;			/* Environment pointer */
-{
-	DB *dbp;
-	Tcl_Obj *res;
-	int result, ret;
-	char *arg, msg[MSG_SIZE];
-
-	result = TCL_OK;
-	if (objc != 4) {
-		Tcl_WrongNumArgs(interp, 2, objv, "db filename");
-		return (TCL_ERROR);
-	}
-	/*
-	 * First comes the DB.
-	 */
-	arg = Tcl_GetStringFromObj(objv[2], NULL);
-	dbp = NAME_TO_DB(arg);
-	if (dbp == NULL) {
-		snprintf(msg, MSG_SIZE,
-		    "LogRegister: Invalid db: %s\n", arg);
-		Tcl_SetResult(interp, msg, TCL_VOLATILE);
-		return (TCL_ERROR);
-	}
-
-	/*
-	 * Next is the filename.
-	 */
-	arg = Tcl_GetStringFromObj(objv[3], NULL);
-
-	_debug_check();
-	ret = envp->log_register(envp, dbp, arg);
-	result = _ReturnSetup(interp, ret, "log_register");
-	if (result == TCL_OK) {
-		res = Tcl_NewLongObj((long)dbp->log_fileid);
-		Tcl_SetObjResult(interp, res);
-	}
-	return (result);
-}
-
 /*
  * tcl_LogStat --
  *
@@ -407,7 +358,7 @@ tcl_LogStat(interp, objc, objv, envp)
 	}
 	_debug_check();
 	ret = envp->log_stat(envp, &sp, 0);
-	result = _ReturnSetup(interp, ret, "log stat");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "log stat");
 	if (result == TCL_ERROR)
 		return (result);
 
@@ -424,7 +375,8 @@ tcl_LogStat(interp, objc, objv, envp)
 	MAKE_STAT_LIST("Region size", sp->st_regsize);
 	MAKE_STAT_LIST("Log file mode", sp->st_mode);
 	MAKE_STAT_LIST("Log record cache size", sp->st_lg_bsize);
-	MAKE_STAT_LIST("Maximum log file size", sp->st_lg_max);
+	MAKE_STAT_LIST("Current log file size", sp->st_lg_size);
+	MAKE_STAT_LIST("Log file records written", sp->st_record);
 	MAKE_STAT_LIST("Mbytes written", sp->st_w_mbytes);
 	MAKE_STAT_LIST("Bytes written (over Mb)", sp->st_w_bytes);
 	MAKE_STAT_LIST("Mbytes written since checkpoint", sp->st_wc_mbytes);
@@ -433,59 +385,19 @@ tcl_LogStat(interp, objc, objv, envp)
 	MAKE_STAT_LIST("Times log written", sp->st_wcount);
 	MAKE_STAT_LIST("Times log written because cache filled up",
 	    sp->st_wcount_fill);
-	MAKE_STAT_LIST("Times log flushed", sp->st_scount);
+	MAKE_STAT_LIST("Times log read from disk", sp->st_rcount);
+	MAKE_STAT_LIST("Times log flushed to disk", sp->st_scount);
 	MAKE_STAT_LIST("Current log file number", sp->st_cur_file);
 	MAKE_STAT_LIST("Current log file offset", sp->st_cur_offset);
 	MAKE_STAT_LIST("On-disk log file number", sp->st_disk_file);
 	MAKE_STAT_LIST("On-disk log file offset", sp->st_disk_offset);
 	MAKE_STAT_LIST("Max commits in a log flush", sp->st_maxcommitperflush);
 	MAKE_STAT_LIST("Min commits in a log flush", sp->st_mincommitperflush);
-	MAKE_STAT_LIST("Log flushes containing a commit", sp->st_flushcommit);
 	MAKE_STAT_LIST("Number of region lock waits", sp->st_region_wait);
 	MAKE_STAT_LIST("Number of region lock nowaits", sp->st_region_nowait);
 	Tcl_SetObjResult(interp, res);
 error:
-	__os_free(envp, sp, sizeof(*sp));
-	return (result);
-}
-
-/*
- * tcl_LogUnregister --
- *
- * PUBLIC: int tcl_LogUnregister __P((Tcl_Interp *, int,
- * PUBLIC:    Tcl_Obj * CONST*, DB_ENV *));
- */
-int
-tcl_LogUnregister(interp, objc, objv, envp)
-	Tcl_Interp *interp;		/* Interpreter */
-	int objc;			/* How many arguments? */
-	Tcl_Obj *CONST objv[];		/* The argument objects */
-	DB_ENV *envp;			/* Environment pointer */
-{
-	DB *dbp;
-	char *arg, msg[MSG_SIZE];
-	int result, ret;
-
-	result = TCL_OK;
-	/*
-	 * 1 arg for this.  Error if more or less.
-	 */
-	if (objc != 3) {
-		Tcl_WrongNumArgs(interp, 2, objv, NULL);
-		return (TCL_ERROR);
-	}
-	arg = Tcl_GetStringFromObj(objv[2], NULL);
-	dbp = NAME_TO_DB(arg);
-	if (dbp == NULL) {
-		snprintf(msg, MSG_SIZE,
-		    "log_unregister: Invalid db identifier: %s\n", arg);
-		Tcl_SetResult(interp, msg, TCL_VOLATILE);
-		return (TCL_ERROR);
-	}
-	_debug_check();
-	ret = envp->log_unregister(envp, dbp);
-	result = _ReturnSetup(interp, ret, "log_unregister");
-
+	__os_ufree(envp, sp);
 	return (result);
 }
 
@@ -497,22 +409,26 @@ tcl_LogUnregister(interp, objc, objv, envp)
  */
 int
 logc_Cmd(clientData, interp, objc, objv)
-	ClientData clientData;          /* Cursor handle */
-	Tcl_Interp *interp;             /* Interpreter */
-	int objc;                       /* How many arguments? */
-	Tcl_Obj *CONST objv[];          /* The argument objects */
+	ClientData clientData;		/* Cursor handle */
+	Tcl_Interp *interp;		/* Interpreter */
+	int objc;			/* How many arguments? */
+	Tcl_Obj *CONST objv[];		/* The argument objects */
 {
-	static char *logccmds[] = {
+	static const char *logccmds[] = {
 		"close",
 		"get",
+		"version",
 		NULL
 	};
 	enum logccmds {
 		LOGCCLOSE,
-		LOGCGET
+		LOGCGET,
+		LOGCVERSION
 	};
 	DB_LOGC *logc;
 	DBTCL_INFO *logcip;
+	Tcl_Obj *res;
+	u_int32_t version;
 	int cmdindex, result, ret;
 
 	Tcl_ResetResult(interp);
@@ -551,7 +467,8 @@ logc_Cmd(clientData, interp, objc, objv)
 		}
 		_debug_check();
 		ret = logc->close(logc, 0);
-		result = _ReturnSetup(interp, ret, "logc close");
+		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "logc close");
 		if (result == TCL_OK) {
 			(void)Tcl_DeleteCommand(interp, logcip->i_name);
 			_DeleteInfo(logcip);
@@ -560,7 +477,24 @@ logc_Cmd(clientData, interp, objc, objv)
 	case LOGCGET:
 		result = tcl_LogcGet(interp, objc, objv, logc);
 		break;
+	case LOGCVERSION:
+		/*
+		 * No args for this.  Error if there are some.
+		 */
+		if (objc > 2) {
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
+			return (TCL_ERROR);
+		}
+		_debug_check();
+		ret = logc->version(logc, &version, 0);
+		if ((result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "logc version")) == TCL_OK) {
+			res = Tcl_NewIntObj((int)version);
+			Tcl_SetObjResult(interp, res);
+		}
+		break;
 	}
+
 	return (result);
 }
 
@@ -571,8 +505,7 @@ tcl_LogcGet(interp, objc, objv, logc)
 	Tcl_Obj * CONST *objv;
 	DB_LOGC *logc;
 {
-	static char *logcgetopts[] = {
-		"-checkpoint",
+	static const char *logcgetopts[] = {
 		"-current",
 		"-first",
 		"-last",
@@ -582,7 +515,6 @@ tcl_LogcGet(interp, objc, objv, logc)
 		NULL
 	};
 	enum logcgetopts {
-		LOGCGET_CHECKPOINT,
 		LOGCGET_CURRENT,
 		LOGCGET_FIRST,
 		LOGCGET_LAST,
@@ -590,7 +522,6 @@ tcl_LogcGet(interp, objc, objv, logc)
 		LOGCGET_PREV,
 		LOGCGET_SET
 	};
-	DB_ENV *envp;
 	DB_LSN lsn;
 	DBT data;
 	Tcl_Obj *dataobj, *lsnlist, *myobjv[2], *res;
@@ -600,7 +531,6 @@ tcl_LogcGet(interp, objc, objv, logc)
 	result = TCL_OK;
 	res = NULL;
 	flag = 0;
-	envp = logc->dbenv;
 
 	if (objc < 3) {
 		Tcl_WrongNumArgs(interp, 2, objv, "?-args? lsn");
@@ -618,10 +548,6 @@ tcl_LogcGet(interp, objc, objv, logc)
 			return (IS_HELP(objv[i]));
 		i++;
 		switch ((enum logcgetopts)optindex) {
-		case LOGCGET_CHECKPOINT:
-			FLAG_CHECK(flag);
-			flag |= DB_CHECKPOINT;
-			break;
 		case LOGCGET_CURRENT:
 			FLAG_CHECK(flag);
 			flag |= DB_CURRENT;
@@ -673,20 +599,21 @@ tcl_LogcGet(interp, objc, objv, logc)
 		 * is a sublist {file offset}.
 		 */
 		myobjc = 2;
-		myobjv[0] = Tcl_NewLongObj((long)lsn.file);
-		myobjv[1] = Tcl_NewLongObj((long)lsn.offset);
+		myobjv[0] = Tcl_NewWideIntObj((Tcl_WideInt)lsn.file);
+		myobjv[1] = Tcl_NewWideIntObj((Tcl_WideInt)lsn.offset);
 		lsnlist = Tcl_NewListObj(myobjc, myobjv);
 		if (lsnlist == NULL)
 			goto memerr;
 
 		result = Tcl_ListObjAppendElement(interp, res, lsnlist);
-		dataobj = Tcl_NewStringObj(data.data, data.size);
+		dataobj = NewStringObj(data.data, data.size);
 		if (dataobj == NULL) {
 			goto memerr;
 		}
 		result = Tcl_ListObjAppendElement(interp, res, dataobj);
 	} else
-		result = _ReturnSetup(interp, ret, "DB_LOGC->get");
+		result = _ReturnSetup(interp, ret, DB_RETOK_LGGET(ret),
+		    "DB_LOGC->get");
 
 	Tcl_SetObjResult(interp, res);
 
@@ -698,3 +625,4 @@ memerr:		if (res != NULL)
 
 	return (result);
 }
+#endif
