@@ -1,26 +1,19 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1999-2001
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1999-2006
+ *	Oracle Corporation.  All rights reserved.
+ *
+ * $Id: tcl_txn.c,v 12.16 2006/09/11 14:53:42 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef lint
-static const char revid[] = "$Id: tcl_txn.c,v 11.44 2001/10/11 18:32:29 ubell Exp $";
-#endif /* not lint */
-
+#include "db_int.h"
 #ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <stdlib.h>
-#include <string.h>
 #include <tcl.h>
 #endif
-
-#include "db_int.h"
-#include "tcl_db.h"
+#include "dbinc/tcl_db.h"
 
 static int tcl_TxnCommit __P((Tcl_Interp *,
 	       int, Tcl_Obj * CONST *, DB_TXN *, DBTCL_INFO *));
@@ -37,7 +30,7 @@ static int txn_Cmd __P((ClientData, Tcl_Interp *, int, Tcl_Obj * CONST *));
  */
 void
 _TxnInfoDelete(interp, txnip)
-	Tcl_Interp *interp;             /* Interpreter */
+	Tcl_Interp *interp;		/* Interpreter */
 	DBTCL_INFO *txnip;		/* Info for txn */
 {
 	DBTCL_INFO *nextp, *p;
@@ -48,7 +41,7 @@ _TxnInfoDelete(interp, txnip)
 		 * txn.  Remove its commands and info structure.
 		 */
 		nextp = LIST_NEXT(p, entries);
-		 if (p->i_parent == txnip && p->i_type == I_TXN) {
+		if (p->i_parent == txnip && p->i_type == I_TXN) {
 			_TxnInfoDelete(interp, p);
 			(void)Tcl_DeleteCommand(interp, p->i_name);
 			_DeleteInfo(p);
@@ -69,16 +62,22 @@ tcl_TxnCheckpoint(interp, objc, objv, envp)
 	Tcl_Obj *CONST objv[];		/* The argument objects */
 	DB_ENV *envp;			/* Environment pointer */
 {
-	static char *txnckpopts[] = {
-		"-kbyte",	"-min",
+	static const char *txnckpopts[] = {
+		"-force",
+		"-kbyte",
+		"-min",
 		NULL
 	};
 	enum txnckpopts {
-		TXNCKP_KB,	TXNCKP_MIN
+		TXNCKP_FORCE,
+		TXNCKP_KB,
+		TXNCKP_MIN
 	};
+	u_int32_t flags;
 	int i, kb, min, optindex, result, ret;
 
 	result = TCL_OK;
+	flags = 0;
 	kb = min = 0;
 
 	/*
@@ -93,6 +92,9 @@ tcl_TxnCheckpoint(interp, objc, objv, envp)
 		}
 		i++;
 		switch ((enum txnckpopts)optindex) {
+		case TXNCKP_FORCE:
+			flags = DB_FORCE;
+			break;
 		case TXNCKP_KB:
 			if (i == objc) {
 				Tcl_WrongNumArgs(interp, 2, objv,
@@ -113,8 +115,9 @@ tcl_TxnCheckpoint(interp, objc, objv, envp)
 		}
 	}
 	_debug_check();
-	ret = envp->txn_checkpoint(envp, (u_int32_t)kb, (u_int32_t)min, 0);
-	result = _ReturnSetup(interp, ret, "txn checkpoint");
+	ret = envp->txn_checkpoint(envp, (u_int32_t)kb, (u_int32_t)min, flags);
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+	    "txn checkpoint");
 	return (result);
 }
 
@@ -132,40 +135,57 @@ tcl_Txn(interp, objc, objv, envp, envip)
 	DB_ENV *envp;			/* Environment pointer */
 	DBTCL_INFO *envip;		/* Info pointer */
 {
-	static char *txnopts[] = {
-		"-dirty",
+	static const char *txnopts[] = {
+#ifdef CONFIG_TEST
 		"-lock_timeout",
+		"-read_committed",
+		"-read_uncommitted",
+		"-txn_timeout",
+#endif
 		"-nosync",
 		"-nowait",
 		"-parent",
+		"-snapshot",
 		"-sync",
-		"-txn_timeout",
+		"-wrnosync",
 		NULL
 	};
 	enum txnopts {
-		TXNDIRTY,
-		TXN_LOCK_TIMEOUT,
+#ifdef CONFIG_TEST
+		TXNLOCK_TIMEOUT,
+		TXNREAD_COMMITTED,
+		TXNREAD_UNCOMMITTED,
+		TXNTIMEOUT,
+#endif
 		TXNNOSYNC,
 		TXNNOWAIT,
 		TXNPARENT,
+		TXNSNAPSHOT,
 		TXNSYNC,
-		TXN_TIMEOUT
+		TXNWRNOSYNC
 	};
 	DBTCL_INFO *ip;
 	DB_TXN *parent;
 	DB_TXN *txn;
 	Tcl_Obj *res;
-	db_timeout_t lk_time, tx_time;
-	u_int32_t flag, lk_timeflag, tx_timeflag;
+	u_int32_t flag;
 	int i, optindex, result, ret;
 	char *arg, msg[MSG_SIZE], newname[MSG_SIZE];
+#ifdef CONFIG_TEST
+	db_timeout_t lk_time, tx_time;
+	u_int32_t lk_timeflag, tx_timeflag;
+#endif
 
 	result = TCL_OK;
 	memset(newname, 0, MSG_SIZE);
 
 	parent = NULL;
 	flag = 0;
+#ifdef CONFIG_TEST
+	COMPQUIET(tx_time, 0);
+	COMPQUIET(lk_time, 0);
 	lk_timeflag = tx_timeflag = 0;
+#endif
 	i = 2;
 	while (i < objc) {
 		if (Tcl_GetIndexFromObj(interp, objv[i],
@@ -174,6 +194,36 @@ tcl_Txn(interp, objc, objv, envp, envip)
 		}
 		i++;
 		switch ((enum txnopts)optindex) {
+#ifdef CONFIG_TEST
+		case TXNLOCK_TIMEOUT:
+			lk_timeflag = DB_SET_LOCK_TIMEOUT;
+			goto get_timeout;
+		case TXNTIMEOUT:
+			tx_timeflag = DB_SET_TXN_TIMEOUT;
+get_timeout:		if (i >= objc) {
+				Tcl_WrongNumArgs(interp, 2, objv,
+				    "?-txn_timestamp time?");
+				return (TCL_ERROR);
+			}
+			result = Tcl_GetLongFromObj(interp, objv[i++], (long *)
+			    ((enum txnopts)optindex == TXNLOCK_TIMEOUT ?
+			    &lk_time : &tx_time));
+			if (result != TCL_OK)
+				return (TCL_ERROR);
+			break;
+		case TXNREAD_COMMITTED:
+			flag |= DB_READ_COMMITTED;
+			break;
+		case TXNREAD_UNCOMMITTED:
+			flag |= DB_READ_UNCOMMITTED;
+			break;
+#endif
+		case TXNNOSYNC:
+			flag |= DB_TXN_NOSYNC;
+			break;
+		case TXNNOWAIT:
+			flag |= DB_TXN_NOWAIT;
+			break;
 		case TXNPARENT:
 			if (i == objc) {
 				Tcl_WrongNumArgs(interp, 2, objv,
@@ -191,37 +241,14 @@ tcl_Txn(interp, objc, objv, envp, envip)
 				return (TCL_ERROR);
 			}
 			break;
-		case TXNDIRTY:
-			flag |= DB_DIRTY_READ;
-			break;
-		case TXNNOWAIT:
-			FLAG_CHECK2(flag, DB_DIRTY_READ);
-			flag |= DB_TXN_NOWAIT;
+		case TXNSNAPSHOT:
+			flag |= DB_TXN_SNAPSHOT;
 			break;
 		case TXNSYNC:
-			FLAG_CHECK2(flag, DB_DIRTY_READ);
 			flag |= DB_TXN_SYNC;
 			break;
-		case TXNNOSYNC:
-			FLAG_CHECK2(flag, DB_DIRTY_READ);
-			flag |= DB_TXN_NOSYNC;
-			break;
-		case TXN_LOCK_TIMEOUT:
-			lk_timeflag = DB_SET_LOCK_TIMEOUT;
-			goto getit;
-		case TXN_TIMEOUT:
-			tx_timeflag = DB_SET_TXN_TIMEOUT;
-getit:
-			if (i >= objc) {
-				Tcl_WrongNumArgs(interp, 2, objv,
-				    "?-txn_timestamp time?");
-				return (TCL_ERROR);
-			}
-			result = Tcl_GetLongFromObj(interp, objv[i++],
-			    (long *)(optindex == TXN_LOCK_TIMEOUT ?
-			    &lk_time : &tx_time));
-			if (result != TCL_OK)
-				return (TCL_ERROR);
+		case TXNWRNOSYNC:
+			flag |= DB_TXN_WRITE_NOSYNC;
 			break;
 		}
 	}
@@ -235,7 +262,8 @@ getit:
 	}
 	_debug_check();
 	ret = envp->txn_begin(envp, parent, &txn, flag);
-	result = _ReturnSetup(interp, ret, "txn");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+	    "txn");
 	if (result == TCL_ERROR)
 		_DeleteInfo(ip);
 	else {
@@ -249,15 +277,17 @@ getit:
 		else
 			ip->i_parent = envip;
 		_SetInfoData(ip, txn);
-		Tcl_CreateObjCommand(interp, newname,
+		(void)Tcl_CreateObjCommand(interp, newname,
 		    (Tcl_ObjCmdProc *)txn_Cmd, (ClientData)txn, NULL);
-		res = Tcl_NewStringObj(newname, strlen(newname));
+		res = NewStringObj(newname, strlen(newname));
 		Tcl_SetObjResult(interp, res);
+#ifdef CONFIG_TEST
 		if (tx_timeflag != 0) {
 			ret = txn->set_timeout(txn, tx_time, tx_timeflag);
 			if (ret != 0) {
 				result =
-				    _ReturnSetup(interp, ret, "set_timeout");
+				    _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+					"set_timeout");
 				_DeleteInfo(ip);
 			}
 		}
@@ -265,10 +295,69 @@ getit:
 			ret = txn->set_timeout(txn, lk_time, lk_timeflag);
 			if (ret != 0) {
 				result =
-				    _ReturnSetup(interp, ret, "set_timeout");
+				    _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+					"set_timeout");
 				_DeleteInfo(ip);
 			}
 		}
+#endif
+	}
+	return (result);
+}
+
+/*
+ * tcl_CDSGroup --
+ *
+ * PUBLIC: int tcl_CDSGroup __P((Tcl_Interp *, int,
+ * PUBLIC:    Tcl_Obj * CONST*, DB_ENV *, DBTCL_INFO *));
+ */
+int
+tcl_CDSGroup(interp, objc, objv, envp, envip)
+	Tcl_Interp *interp;		/* Interpreter */
+	int objc;			/* How many arguments? */
+	Tcl_Obj *CONST objv[];		/* The argument objects */
+	DB_ENV *envp;			/* Environment pointer */
+	DBTCL_INFO *envip;		/* Info pointer */
+{
+	DBTCL_INFO *ip;
+	DB_TXN *txn;
+	Tcl_Obj *res;
+	int result, ret;
+	char newname[MSG_SIZE];
+
+	if (objc != 2) {
+		Tcl_WrongNumArgs(interp, 1, objv, "env cdsgroup");
+		return (TCL_ERROR);
+	}
+
+	result = TCL_OK;
+	memset(newname, 0, MSG_SIZE);
+
+	snprintf(newname, sizeof(newname), "%s.txn%d",
+	    envip->i_name, envip->i_envtxnid);
+	ip = _NewInfo(interp, NULL, newname, I_TXN);
+	if (ip == NULL) {
+		Tcl_SetResult(interp, "Could not set up info",
+		    TCL_STATIC);
+		return (TCL_ERROR);
+	}
+	_debug_check();
+	ret = envp->cdsgroup_begin(envp, &txn);
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "cdsgroup");
+	if (result == TCL_ERROR)
+		_DeleteInfo(ip);
+	else {
+		/*
+		 * Success.  Set up return.  Set up new info
+		 * and command widget for this txn.
+		 */
+		envip->i_envtxnid++;
+		ip->i_parent = envip;
+		_SetInfoData(ip, txn);
+		(void)Tcl_CreateObjCommand(interp, newname,
+		    (Tcl_ObjCmdProc *)txn_Cmd, (ClientData)txn, NULL);
+		res = NewStringObj(newname, strlen(newname));
+		Tcl_SetObjResult(interp, res);
 	}
 	return (result);
 }
@@ -286,21 +375,6 @@ tcl_TxnStat(interp, objc, objv, envp)
 	Tcl_Obj *CONST objv[];		/* The argument objects */
 	DB_ENV *envp;			/* Environment pointer */
 {
-#define	MAKE_STAT_LSN(s, lsn)						\
-do {									\
-	myobjc = 2;							\
-	myobjv[0] = Tcl_NewLongObj((long)(lsn)->file);			\
-	myobjv[1] = Tcl_NewLongObj((long)(lsn)->offset);		\
-	lsnlist = Tcl_NewListObj(myobjc, myobjv);			\
-	myobjc = 2;							\
-	myobjv[0] = Tcl_NewStringObj((s), strlen(s));			\
-	myobjv[1] = lsnlist;						\
-	thislist = Tcl_NewListObj(myobjc, myobjv);			\
-	result = Tcl_ListObjAppendElement(interp, res, thislist);	\
-	if (result != TCL_OK)						\
-		goto error;						\
-} while (0);
-
 	DBTCL_INFO *ip;
 	DB_TXN_ACTIVE *p;
 	DB_TXN_STAT *sp;
@@ -318,7 +392,8 @@ do {									\
 	}
 	_debug_check();
 	ret = envp->txn_stat(envp, &sp, 0);
-	result = _ReturnSetup(interp, ret, "txn stat");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+	    "txn stat");
 	if (result == TCL_ERROR)
 		return (result);
 
@@ -332,21 +407,21 @@ do {									\
 	 */
 	MAKE_STAT_LIST("Region size", sp->st_regsize);
 	MAKE_STAT_LSN("LSN of last checkpoint", &sp->st_last_ckp);
-	MAKE_STAT_LSN("LSN of pending checkpoint", &sp->st_pending_ckp);
 	MAKE_STAT_LIST("Time of last checkpoint", sp->st_time_ckp);
 	MAKE_STAT_LIST("Last txn ID allocated", sp->st_last_txnid);
-	MAKE_STAT_LIST("Max Txns", sp->st_maxtxns);
+	MAKE_STAT_LIST("Maximum txns", sp->st_maxtxns);
 	MAKE_STAT_LIST("Number aborted txns", sp->st_naborts);
-	MAKE_STAT_LIST("Number active txns", sp->st_nactive);
-	MAKE_STAT_LIST("Maximum  active txns", sp->st_maxnactive);
 	MAKE_STAT_LIST("Number txns begun", sp->st_nbegins);
 	MAKE_STAT_LIST("Number committed txns", sp->st_ncommits);
+	MAKE_STAT_LIST("Number active txns", sp->st_nactive);
+	MAKE_STAT_LIST("Number of snapshot txns", sp->st_nsnapshot);
 	MAKE_STAT_LIST("Number restored txns", sp->st_nrestores);
+	MAKE_STAT_LIST("Maximum active txns", sp->st_maxnactive);
+	MAKE_STAT_LIST("Maximum snapshot txns", sp->st_maxnsnapshot);
 	MAKE_STAT_LIST("Number of region lock waits", sp->st_region_wait);
 	MAKE_STAT_LIST("Number of region lock nowaits", sp->st_region_nowait);
 	for (i = 0, p = sp->st_txnarray; i < sp->st_nactive; i++, p++)
-		for (ip = LIST_FIRST(&__db_infohead); ip != NULL;
-		    ip = LIST_NEXT(ip, entries)) {
+		LIST_FOREACH(ip, &__db_infohead, entries) {
 			if (ip->i_type != I_TXN)
 				continue;
 			if (ip->i_type == I_TXN &&
@@ -362,7 +437,7 @@ do {									\
 		}
 	Tcl_SetObjResult(interp, res);
 error:
-	__os_free(envp, sp, 0);
+	__os_ufree(envp, sp);
 	return (result);
 }
 
@@ -394,7 +469,8 @@ tcl_TxnTimeout(interp, objc, objv, envp)
 		return (result);
 	_debug_check();
 	ret = envp->set_timeout(envp, (u_int32_t)timeout, DB_SET_TXN_TIMEOUT);
-	result = _ReturnSetup(interp, ret, "lock timeout");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+	    "lock timeout");
 	return (result);
 }
 
@@ -404,31 +480,45 @@ tcl_TxnTimeout(interp, objc, objv, envp)
  */
 static int
 txn_Cmd(clientData, interp, objc, objv)
-	ClientData clientData;          /* Txn handle */
-	Tcl_Interp *interp;             /* Interpreter */
-	int objc;                       /* How many arguments? */
-	Tcl_Obj *CONST objv[];          /* The argument objects */
+	ClientData clientData;		/* Txn handle */
+	Tcl_Interp *interp;		/* Interpreter */
+	int objc;			/* How many arguments? */
+	Tcl_Obj *CONST objv[];		/* The argument objects */
 {
-	static char *txncmds[] = {
-		"abort",
-		"commit",
+	static const char *txncmds[] = {
+#ifdef CONFIG_TEST
 		"discard",
+		"getname",
 		"id",
 		"prepare",
+		"setname",
+#endif
+		"abort",
+		"commit",
+		"getname",
+		"setname",
 		NULL
 	};
 	enum txncmds {
-		TXNABORT,
-		TXNCOMMIT,
+#ifdef CONFIG_TEST
 		TXNDISCARD,
+		TXNGETNAME,
 		TXNID,
-		TXNPREPARE
+		TXNPREPARE,
+		TXNSETNAME,
+#endif
+		TXNABORT,
+		TXNCOMMIT
 	};
 	DBTCL_INFO *txnip;
 	DB_TXN *txnp;
 	Tcl_Obj *res;
 	int cmdindex, result, ret;
-	u_int8_t *gid;
+#ifdef CONFIG_TEST
+	u_int8_t *gid, garray[DB_XIDDATASIZE];
+	int length;
+	const char *name;
+#endif
 
 	Tcl_ResetResult(interp);
 	txnp = (DB_TXN *)clientData;
@@ -453,14 +543,27 @@ txn_Cmd(clientData, interp, objc, objv)
 
 	res = NULL;
 	switch ((enum txncmds)cmdindex) {
+#ifdef CONFIG_TEST
+	case TXNDISCARD:
+		if (objc != 2) {
+			Tcl_WrongNumArgs(interp, 1, objv, NULL);
+			return (TCL_ERROR);
+		}
+		_debug_check();
+		ret = txnp->discard(txnp, 0);
+		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "txn discard");
+		_TxnInfoDelete(interp, txnip);
+		(void)Tcl_DeleteCommand(interp, txnip->i_name);
+		_DeleteInfo(txnip);
+		break;
 	case TXNID:
 		if (objc != 2) {
 			Tcl_WrongNumArgs(interp, 1, objv, NULL);
 			return (TCL_ERROR);
 		}
 		_debug_check();
-		ret = txnp->id(txnp);
-		res = Tcl_NewIntObj(ret);
+		res = Tcl_NewIntObj((int)txnp->id(txnp));
 		break;
 	case TXNPREPARE:
 		if (objc != 3) {
@@ -468,8 +571,9 @@ txn_Cmd(clientData, interp, objc, objv)
 			return (TCL_ERROR);
 		}
 		_debug_check();
-		gid = (u_int8_t *)Tcl_GetByteArrayFromObj(objv[2], NULL);
-		ret = txnp->prepare(txnp, gid);
+		gid = (u_int8_t *)Tcl_GetByteArrayFromObj(objv[2], &length);
+		memcpy(garray, gid, (size_t)length);
+		ret = txnp->prepare(txnp, garray);
 		/*
 		 * !!!
 		 * DB_TXN->prepare commits all outstanding children.  But it
@@ -478,26 +582,31 @@ txn_Cmd(clientData, interp, objc, objv)
 		 * we do not call _DeleteInfo on ourselves.
 		 */
 		_TxnInfoDelete(interp, txnip);
-		result = _ReturnSetup(interp, ret, "txn prepare");
+		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "txn prepare");
 		break;
-	case TXNCOMMIT:
-		result = tcl_TxnCommit(interp, objc, objv, txnp, txnip);
-		_TxnInfoDelete(interp, txnip);
-		(void)Tcl_DeleteCommand(interp, txnip->i_name);
-		_DeleteInfo(txnip);
-		break;
-	case TXNDISCARD:
+	case TXNGETNAME:
 		if (objc != 2) {
-			Tcl_WrongNumArgs(interp, 1, objv, NULL);
+			Tcl_WrongNumArgs(interp, 2, objv, NULL);
 			return (TCL_ERROR);
 		}
 		_debug_check();
-		ret = txnp->discard(txnp, 0);
-		result = _ReturnSetup(interp, ret, "txn discard");
-		_TxnInfoDelete(interp, txnip);
-		(void)Tcl_DeleteCommand(interp, txnip->i_name);
-		_DeleteInfo(txnip);
+		ret = txnp->get_name(txnp, &name);
+		if ((result = _ReturnSetup(
+		    interp, ret, DB_RETOK_STD(ret), "txn getname")) == TCL_OK)
+			res = NewStringObj(name, strlen(name));
 		break;
+	case TXNSETNAME:
+		if (objc != 3) {
+			Tcl_WrongNumArgs(interp, 2, objv, "name");
+			return (TCL_ERROR);
+		}
+		_debug_check();
+		ret = txnp->set_name(txnp, Tcl_GetStringFromObj(objv[2], NULL));
+		result =
+		    _ReturnSetup(interp, ret, DB_RETOK_STD(ret), "setname");
+		break;
+#endif
 	case TXNABORT:
 		if (objc != 2) {
 			Tcl_WrongNumArgs(interp, 1, objv, NULL);
@@ -505,7 +614,14 @@ txn_Cmd(clientData, interp, objc, objv)
 		}
 		_debug_check();
 		ret = txnp->abort(txnp);
-		result = _ReturnSetup(interp, ret, "txn abort");
+		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "txn abort");
+		_TxnInfoDelete(interp, txnip);
+		(void)Tcl_DeleteCommand(interp, txnip->i_name);
+		_DeleteInfo(txnip);
+		break;
+	case TXNCOMMIT:
+		result = tcl_TxnCommit(interp, objc, objv, txnp, txnip);
 		_TxnInfoDelete(interp, txnip);
 		(void)Tcl_DeleteCommand(interp, txnip->i_name);
 		_DeleteInfo(txnip);
@@ -522,20 +638,22 @@ txn_Cmd(clientData, interp, objc, objv)
 
 static int
 tcl_TxnCommit(interp, objc, objv, txnp, txnip)
-	Tcl_Interp *interp;             /* Interpreter */
-	int objc;                       /* How many arguments? */
-	Tcl_Obj *CONST objv[];          /* The argument objects */
+	Tcl_Interp *interp;		/* Interpreter */
+	int objc;			/* How many arguments? */
+	Tcl_Obj *CONST objv[];		/* The argument objects */
 	DB_TXN *txnp;			/* Transaction pointer */
 	DBTCL_INFO *txnip;		/* Info pointer */
 {
-	static char *commitopt[] = {
+	static const char *commitopt[] = {
 		"-nosync",
 		"-sync",
+		"-wrnosync",
 		NULL
 	};
 	enum commitopt {
+		COMNOSYNC,
 		COMSYNC,
-		COMNOSYNC
+		COMWRNOSYNC
 	};
 	u_int32_t flag;
 	int optindex, result, ret;
@@ -554,22 +672,25 @@ tcl_TxnCommit(interp, objc, objv, txnp, txnip)
 			return (IS_HELP(objv[2]));
 		switch ((enum commitopt)optindex) {
 		case COMSYNC:
-			FLAG_CHECK(flag);
 			flag = DB_TXN_SYNC;
 			break;
 		case COMNOSYNC:
-			FLAG_CHECK(flag);
 			flag = DB_TXN_NOSYNC;
 			break;
+		case COMWRNOSYNC:
+			flag = DB_TXN_WRITE_NOSYNC;
+			break;
 		}
 	}
 
 	_debug_check();
 	ret = txnp->commit(txnp, flag);
-	result = _ReturnSetup(interp, ret, "txn commit");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+	    "txn commit");
 	return (result);
 }
 
+#ifdef CONFIG_TEST
 /*
  * tcl_TxnRecover --
  *
@@ -598,7 +719,7 @@ for (i = 0; i < count; i++) {						\
 	ip->i_parent = envip;						\
 	p = &prep[i];							\
 	_SetInfoData(ip, p->txn);					\
-	Tcl_CreateObjCommand(interp, newname,				\
+	(void)Tcl_CreateObjCommand(interp, newname,			\
 	    (Tcl_ObjCmdProc *)txn_Cmd, (ClientData)p->txn, NULL);	\
 	result = _SetListElem(interp, res, newname, strlen(newname),	\
 	    p->gid, DB_XIDDATASIZE);					\
@@ -623,7 +744,8 @@ for (i = 0; i < count; i++) {						\
 	}
 	_debug_check();
 	ret = envp->txn_recover(envp, prep, DBTCL_PREP, &count, DB_FIRST);
-	result = _ReturnSetup(interp, ret, "txn recover");
+	result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+	    "txn recover");
 	if (result == TCL_ERROR)
 		return (result);
 	res = Tcl_NewObj();
@@ -636,7 +758,8 @@ for (i = 0; i < count; i++) {						\
 	while (count == DBTCL_PREP) {
 		ret = envp->txn_recover(
 		    envp, prep, DBTCL_PREP, &count, DB_NEXT);
-		result = _ReturnSetup(interp, ret, "txn recover");
+		result = _ReturnSetup(interp, ret, DB_RETOK_STD(ret),
+		    "txn recover");
 		if (result == TCL_ERROR)
 			return (result);
 		DO_PREPLIST(count);
@@ -645,3 +768,4 @@ for (i = 0; i < count; i++) {						\
 error:
 	return (result);
 }
+#endif
