@@ -1,9 +1,9 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2001
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2006
+#	Oracle Corporation.  All rights reserved.
 #
-# $Id: dead002.tcl,v 11.20 2001/10/11 16:15:30 sandstro Exp $
+# $Id: dead002.tcl,v 12.4 2006/08/24 14:46:34 bostic Exp $
 #
 # TEST	dead002
 # TEST	Same test as dead001, but use "detect on every collision" instead
@@ -12,7 +12,7 @@ proc dead002 { { procs "2 4 10" } {tests "ring clump" } \
     {timeout 0} {tnum 002} } {
 	source ./include.tcl
 
-	puts "Dead$tnum: Deadlock detector tests"
+	puts "Dead$tnum: Deadlock detector tests (detect on every collision)"
 
 	env_cleanup $testdir
 
@@ -22,14 +22,14 @@ proc dead002 { { procs "2 4 10" } {tests "ring clump" } \
 	if { $timeout != 0 } {
 		set lmode "expire"
 	}
-	set env [berkdb env \
+	set env [berkdb_env \
 	    -create -mode 0644 -home $testdir \
-	    -lock -txn_timeout $timeout -lock_detect $lmode]
+	    -lock -lock_timeout $timeout -lock_detect $lmode]
 	error_check_good lock_env:open [is_valid_env $env] TRUE
 
 	foreach t $tests {
-		set pidlist ""
 		foreach n $procs {
+			set pidlist ""
 			sentinel_init
 
 			# Fire off the tests
@@ -37,15 +37,19 @@ proc dead002 { { procs "2 4 10" } {tests "ring clump" } \
 			for { set i 0 } { $i < $n } { incr i } {
 				set locker [$env lock_id]
 				puts "$tclsh_path $test_path/wrap.tcl \
-				    $testdir/dead$tnum.log.$i \
-				    ddscript.tcl $testdir $t $locker $i $n"
-				set p [exec $tclsh_path \
-					$test_path/wrap.tcl \
+				    ddscript.tcl $testdir/dead$tnum.log.$i \
+				    $testdir $t $locker $i $n"
+				set p [exec $tclsh_path $test_path/wrap.tcl \
 					ddscript.tcl $testdir/dead$tnum.log.$i \
 					$testdir $t $locker $i $n &]
 				lappend pidlist $p
+				# If we're running with timeouts, pause so that
+				# locks will have a chance to time out.
+				if { $timeout != 0 } {
+					tclsleep 2
+				}
 			}
-			watch_procs 5
+			watch_procs $pidlist 5
 
 			# Now check output
 			set dead 0
@@ -62,6 +66,8 @@ proc dead002 { { procs "2 4 10" } {tests "ring clump" } \
 				}
 				close $did
 			}
+
+			puts "\tDead$tnum: dead check ..."
 			dead_check $t $n $timeout $dead $clean $other
 		}
 	}
