@@ -1,21 +1,21 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999, 2000
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1999-2006
+#	Oracle Corporation.  All rights reserved.
 #
-#	$Id: env002.tcl,v 11.10 2000/05/22 12:51:36 bostic Exp $
+# $Id: env002.tcl,v 12.3 2006/08/24 14:46:35 bostic Exp $
 #
-# Env Test 002
-# Test set_lg_dir and env name resolution
-# With an environment path specified using -home, and then again
-# with it specified by the environment variable DB_HOME:
-#	1) Make sure that the set_lg_dir option is respected
-#		a) as a relative pathname.
-#		b) as an absolute pathname.
-#	2) Make sure that the DB_LOG_DIR db_config argument is respected,
-#		again as relative and absolute pathnames.
-#	3) Make sure that if -both- db_config and a file are present,
-#		only the file is respected (see doc/env/naming.html).
+# TEST	env002
+# TEST	Test of DB_LOG_DIR and env name resolution.
+# TEST	With an environment path specified using -home, and then again
+# TEST	with it specified by the environment variable DB_HOME:
+# TEST	1) Make sure that the set_lg_dir option is respected
+# TEST		a) as a relative pathname.
+# TEST		b) as an absolute pathname.
+# TEST	2) Make sure that the DB_LOG_DIR db_config argument is respected,
+# TEST		again as relative and absolute pathnames.
+# TEST	3) Make sure that if -both- db_config and a file are present,
+# TEST		only the file is respected (see doc/env/naming.html).
 proc env002 { } {
 	#   env002 is essentially just a small driver that runs
 	# env002_body--formerly the entire test--twice;  once, it
@@ -30,7 +30,7 @@ proc env002 { } {
 
 	puts "Env002: set_lg_dir test."
 
-	puts "\tEnv002: Running with -home argument to berkdb env."
+	puts "\tEnv002: Running with -home argument to berkdb_env."
 	env002_body "-home $testdir"
 
 	puts "\tEnv002: Running with environment variable DB_HOME set."
@@ -51,7 +51,7 @@ proc env002 { } {
 proc env002_body { home_arg } {
 	source ./include.tcl
 
-	cleanup $testdir
+	env_cleanup $testdir
 	set logdir "logs_in_here"
 
 	file mkdir $testdir/$logdir
@@ -68,7 +68,7 @@ proc env002_body { home_arg } {
 	env002_run_test a 1 "relative path, config file" $home_arg \
 		$testdir/$logdir
 
-	cleanup $testdir
+	env_cleanup $testdir
 
 	file mkdir $fulllogdir
 	env002_make_config $fulllogdir
@@ -77,7 +77,7 @@ proc env002_body { home_arg } {
 	env002_run_test a 2 "absolute path, config file" $home_arg \
 		$fulllogdir
 
-	cleanup $testdir
+	env_cleanup $testdir
 
 	# Now we try without a config file, but instead with db_config
 	# relative paths
@@ -86,7 +86,7 @@ proc env002_body { home_arg } {
 		-log_dir $logdir -data_dir ." \
 		$testdir/$logdir
 
-	cleanup $testdir
+	env_cleanup $testdir
 
 	# absolute
 	file mkdir $fulllogdir
@@ -94,7 +94,7 @@ proc env002_body { home_arg } {
 		-log_dir $fulllogdir -data_dir ." \
 		$fulllogdir
 
-	cleanup $testdir
+	env_cleanup $testdir
 
 	# Now, set db_config -and- have a # DB_CONFIG file, and make
 	# sure only the latter is honored.
@@ -106,7 +106,7 @@ proc env002_body { home_arg } {
 	env002_run_test c 1 "relative path, both db_config and file" \
 		"$home_arg -log_dir $testdir/bogus \
 		-data_dir ." $testdir/$logdir
-	cleanup $testdir
+	env_cleanup $testdir
 
 	file mkdir $fulllogdir
 	env002_make_config $fulllogdir
@@ -125,8 +125,8 @@ proc env002_run_test { major minor msg env_args log_path} {
 
 	# Create an environment, with logging, and scribble some
 	# stuff in a [btree] database in it.
-	# puts [concat {berkdb env -create -log -private} $env_args]
-	set dbenv [eval {berkdb env -create -log -private} $env_args]
+	# puts [concat {berkdb_env -create -log -private} $env_args]
+	set dbenv [eval {berkdb_env -create -log -private} $env_args]
 	error_check_good env_open [is_valid_env $dbenv] TRUE
 	set db [berkdb_open -env $dbenv -create -btree -mode 0644 $testfile]
 	error_check_good db_open [is_valid_db $db] TRUE
