@@ -1,9 +1,9 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2001
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2006
+#	Oracle Corporation.  All rights reserved.
 #
-# $Id: mdbscript.tcl,v 11.26 2001/10/30 21:40:37 dda Exp $
+# $Id: mdbscript.tcl,v 12.3 2006/08/24 14:46:36 bostic Exp $
 #
 # Process script for the multi-process db tester.
 
@@ -84,8 +84,8 @@ set klock NOLOCK
 # flushes in the main part of the loop below.
 flush stdout
 
-set dbenv [berkdb env -create -cdb -home $dir]
-#set dbenv [berkdb env -create -cdb -log -home $dir]
+set dbenv [berkdb_env -create -cdb -home $dir]
+#set dbenv [berkdb_env -create -cdb -log -home $dir]
 error_check_good dbenv [is_valid_env $dbenv] TRUE
 
 set locker [ $dbenv lock_id ]
@@ -96,7 +96,6 @@ error_check_good dbopen [is_valid_db $db] TRUE
 # Init globals (no data)
 set nkeys [db_init $db 0]
 puts "Initial number of keys: $nkeys"
-error_check_good db_init $nkeys $nentries
 tclsleep 5
 
 proc get_lock { k } {
@@ -124,6 +123,23 @@ proc get_lock { k } {
 	return 0
 }
 
+# If we are renumbering, then each time we delete an item, the number of
+# items in the file is temporarily decreased, so the highest record numbers
+# do not exist.  To make sure this doesn't happen, we never generate the
+# highest few record numbers as keys.
+#
+# For record-based methods, record numbers begin at 1, while for other keys,
+# we begin at 0 to index into an array.
+proc rand_key { method nkeys renum procs} {
+	if { $renum == 1 } {
+		return [berkdb random_int 1 [expr $nkeys - $procs]]
+	} elseif { [is_record_based $method] == 1 } {
+		return [berkdb random_int 1 $nkeys]
+	} else {
+		return [berkdb random_int 0 [expr $nkeys - 1]]
+	}
+}
+
 # On each iteration we're going to randomly pick a key.
 # 1. We'll either get it (verifying that its contents are reasonable).
 # 2. Put it (using an overwrite to make the data be datastr:ID).
@@ -366,13 +382,9 @@ for { set i 0 } { $i < $iter } { incr i } {
 	}
 }
 
-if {[catch {$db close} ret] != 0 } {
-	error_check_good close [is_substr $errorInfo "DB_INCOMPLETE"] 1
-	puts "Warning: sync incomplete on close ([pid])"
-} else  {
-	error_check_good close $ret 0
-}
-$dbenv close
+error_check_good db_close_catch [catch {$db close} ret] 0
+error_check_good db_close $ret 0
+error_check_good dbenv_close [$dbenv close] 0
 
 flush stdout
 exit
