@@ -1,11 +1,13 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996, 1997, 1998, 1999, 2000
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2006
+#	Oracle Corporation.  All rights reserved.
 #
-#	$Id: recd002.tcl,v 11.22 2000/12/11 17:24:54 sue Exp $
+# $Id: recd002.tcl,v 12.5 2006/08/24 14:46:36 bostic Exp $
 #
-# Recovery Test #2.  Verify that splits can be recovered.
+# TEST	recd002
+# TEST	Split recovery tests.  For every known split log message, makes sure
+# TEST	that we exercise redo, undo, and do-nothing condition.
 proc recd002 { method {select 0} args} {
 	source ./include.tcl
 	global rand_init
@@ -33,11 +35,10 @@ proc recd002 { method {select 0} args} {
 	env_cleanup $testdir
 	set testfile recd002.db
 	set testfile2 recd002-2.db
-	set eflags \
-	    "-create -txn -lock_max 2000 -home $testdir"
+	set eflags "-create -txn -lock_max_locks 2000 -home $testdir"
 
 	puts "\tRecd002.a: creating environment"
-	set env_cmd "berkdb env $eflags"
+	set env_cmd "berkdb_env $eflags"
 	set dbenv [eval $env_cmd]
 	error_check_bad dbenv $dbenv NULL
 
@@ -80,9 +81,14 @@ proc recd002 { method {select 0} args} {
 		}
 		op_recover abort $testdir $env_cmd $testfile $cmd $msg
 		op_recover commit $testdir $env_cmd $testfile $cmd $msg
-		op_recover prepare $testdir $env_cmd $testfile2 $cmd $msg
+		#
+		# Note that since prepare-discard ultimately aborts
+		# the txn, it must come before prepare-commit.
+		#
 		op_recover prepare-abort $testdir $env_cmd $testfile2 \
 			$cmd $msg
+		op_recover prepare-discard $testdir $env_cmd $testfile2 \
+			$cmd $msg
 		op_recover prepare-commit $testdir $env_cmd $testfile2 \
 			$cmd $msg
 	}
