@@ -1,33 +1,51 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999, 2000
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1999-2006
+#	Oracle Corporation.  All rights reserved.
 #
-#	$Id: sdb001.tcl,v 11.11 2000/04/21 18:36:22 krinsky Exp $
+# $Id: sdb001.tcl,v 12.3 2006/08/24 14:46:38 bostic Exp $
 #
-# Sub DB Test 1 {access method}
-# Test non-subdb and subdb operations
-# Test naming (filenames begin with -)
-# Test existence (cannot create subdb of same name with -excl)
-proc subdb001 { method args } {
+# TEST	sdb001	Tests mixing db and subdb operations
+# TEST	Tests mixing db and subdb operations
+# TEST	Create a db, add data, try to create a subdb.
+# TEST	Test naming db and subdb with a leading - for correct parsing
+# TEST	Existence check -- test use of -excl with subdbs
+# TEST
+# TEST	Test non-subdb and subdb operations
+# TEST	Test naming (filenames begin with -)
+# TEST	Test existence (cannot create subdb of same name with -excl)
+proc sdb001 { method args } {
 	source ./include.tcl
+	global errorInfo
 
 	set args [convert_args $method $args]
 	set omethod [convert_method $method]
 
+	if { [is_queue $method] == 1 } {
+		puts "Subdb001: skipping for method $method"
+		return
+	}
 	puts "Subdb001: $method ($args) subdb and non-subdb tests"
 
-	# Create the database and open the dictionary
 	set testfile $testdir/subdb001.db
+	set eindex [lsearch -exact $args "-env"]
+	if { $eindex != -1 } {
+		set env NULL
+		incr eindex
+		set env [lindex $args $eindex]
+		puts "Subdb001 skipping for env $env"
+		return
+	}
+	# Create the database and open the dictionary
 	set subdb subdb0
-	cleanup $testdir
+	cleanup $testdir NULL
 	puts "\tSubdb001.a: Non-subdb database and subdb operations"
 	#
 	# Create a db with no subdbs.  Add some data.  Close.  Try to
 	# open/add with a subdb.  Should fail.
 	#
 	puts "\tSubdb001.a.0: Create db, add data, close, try subdb"
-	set db [eval {berkdb_open -create -truncate -mode 0644} \
+	set db [eval {berkdb_open -create -mode 0644} \
 	    $args {$omethod $testfile}]
 	error_check_good dbopen [is_valid_db $db] TRUE
 
@@ -35,7 +53,6 @@ proc subdb001 { method args } {
 
 	set pflags ""
 	set gflags ""
-	set txn ""
 	set count 0
 
 	if { [is_record_based $method] == 1 } {
@@ -51,7 +68,7 @@ proc subdb001 { method args } {
 			set key $str
 		}
 		set ret [eval \
-		    {$db put} $txn $pflags {$key [chop_data $method $str]}]
+		    {$db put} $pflags {$key [chop_data $method $str]}]
 		error_check_good put $ret 0
 
 		set ret [eval {$db get} $gflags {$key}]
@@ -68,8 +85,14 @@ proc subdb001 { method args } {
 	# Create a db with no subdbs.  Add no data.  Close.  Try to
 	# open/add with a subdb.  Should fail.
 	#
-	cleanup $testdir
+	set testfile $testdir/subdb001a.db
 	puts "\tSubdb001.a.1: Create db, close, try subdb"
+	#
+	# !!!
+	# Using -truncate is illegal when opening for subdbs, but we
+	# can use it here because we are not using subdbs for this
+	# create.
+	#
 	set db [eval {berkdb_open -create -truncate -mode 0644} $args \
 	    {$omethod $testfile}]
 	error_check_good dbopen [is_valid_db $db] TRUE
@@ -88,7 +111,6 @@ proc subdb001 { method args } {
 	# Test naming, db and subdb names beginning with -.
 	#
 	puts "\tSubdb001.b: Naming"
-	cleanup $testdir
 	set cwd [pwd]
 	cd $testdir
 	set testfile1 -subdb001.db
@@ -109,9 +131,9 @@ proc subdb001 { method args } {
 	# Create 1 db with 1 subdb.  Try to create another subdb of
 	# the same name.  Should fail.
 	#
+
 	puts "\tSubdb001.c: Existence check"
-	cleanup $testdir
-	set testfile $testdir/subdb001.db
+	set testfile $testdir/subdb001d.db
 	set subdb subdb
 	set ret [catch {eval {berkdb_open -create -excl -mode 0644} $args \
 	    {$omethod $testfile $subdb}} db]
