@@ -1,9 +1,9 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2003
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2006
+#	Oracle Corporation.  All rights reserved.
 #
-# $Id: test006.tcl,v 11.23 2003/09/04 23:41:16 bostic Exp $
+# $Id: test006.tcl,v 12.3 2006/08/24 14:46:39 bostic Exp $
 #
 # TEST	test006
 # TEST	Small keys/medium data
@@ -31,7 +31,7 @@ proc test006 { method {nentries 10000} {reopen 0} {tnum "006"} \
 
 proc test006_body { method {nentries 10000} {reopen 0} {tnum "006"} \
     {ndups 5} sort flags {largs ""} } {
-
+	global is_je_test
 	source ./include.tcl
 
 	set do_renumber [is_rrecno $method]
@@ -54,6 +54,10 @@ proc test006_body { method {nentries 10000} {reopen 0} {tnum "006"} \
 		set basename $dbname
 		incr eindex
 		set env [lindex $largs $eindex]
+		if { $is_je_test && $sort == "unsorted" } {
+			puts "Test$tnum skipping $sort duplicates for JE"
+			return
+		}
 		set txnenv [is_txnenv $env]
 		if { $txnenv == 1 } {
 			append largs " -auto_commit "
