@@ -1,19 +1,24 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1996-2001
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1996-2006
+#	Oracle Corporation.  All rights reserved.
 #
-# $Id: test011.tcl,v 11.21 2001/01/25 18:23:09 bostic Exp $
+# $Id: test011.tcl,v 12.3 2006/08/24 14:46:39 bostic Exp $
 #
-# DB Test 11 {access method}
-# Use the first 10,000 entries from the dictionary.
-# Insert each with self as key and data; add duplicate
-# records for each.
-# Then do some key_first/key_last add_before, add_after operations.
-# This does not work for recno
-# To test if dups work when they fall off the main page, run this with
-# a very tiny page size.
-proc test011 { method {nentries 10000} {ndups 5} {tnum 11} args } {
+# TEST	test011
+# TEST	Duplicate test
+# TEST		Small key/data pairs.
+# TEST		Test DB_KEYFIRST, DB_KEYLAST, DB_BEFORE and DB_AFTER.
+# TEST		To test off-page duplicates, run with small pagesize.
+# TEST
+# TEST	Use the first 10,000 entries from the dictionary.
+# TEST	Insert each with self as key and data; add duplicate records for each.
+# TEST	Then do some key_first/key_last add_before, add_after operations.
+# TEST	This does not work for recno
+# TEST
+# TEST	To test if dups work when they fall off the main page, run this with
+# TEST	a very tiny page size.
+proc test011 { method {nentries 10000} {ndups 5} {tnum "011"} args } {
 	global dlist
 	global rand_init
 	source ./include.tcl
@@ -21,15 +26,12 @@ proc test011 { method {nentries 10000} {ndups 5} {tnum 11} args } {
 	set dlist ""
 
 	if { [is_rbtree $method] == 1 } {
-		puts "Test0$tnum skipping for method $method"
+		puts "Test$tnum skipping for method $method"
 		return
 	}
 	if { [is_record_based $method] == 1 } {
 		test011_recno $method $nentries $tnum $args
 		return
-	} else {
-		puts -nonewline "Test0$tnum: $method $nentries small dup "
-		puts "key/data pairs, cursor ops"
 	}
 	if {$ndups < 5} {
 		set ndups 5
@@ -41,24 +43,42 @@ proc test011 { method {nentries 10000} {ndups 5} {tnum 11} args } {
 	berkdb srand $rand_init
 
 	# Create the database and open the dictionary
+	set txnenv 0
 	set eindex [lsearch -exact $args "-env"]
 	#
 	# If we are using an env, then testfile should just be the db name.
 	# Otherwise it is the test directory and the name.
 	if { $eindex == -1 } {
-		set testfile $testdir/test0$tnum.db
+		set testfile $testdir/test$tnum.db
 		set env NULL
 	} else {
-		set testfile test0$tnum.db
+		set testfile test$tnum.db
 		incr eindex
 		set env [lindex $args $eindex]
+		set txnenv [is_txnenv $env]
+		if { $txnenv == 1 } {
+			append args " -auto_commit "
+			#
+			# If we are using txns and running with the
+			# default, set the default down a bit.
+			#
+			if { $nentries == 10000 } {
+				set nentries 100
+			}
+			reduce_dups nentries ndups
+		}
+		set testdir [get_home $env]
 	}
+
+	puts -nonewline "Test$tnum: $method $nentries small $ndups dup "
+	puts "key/data pairs, cursor ops"
+
 	set t1 $testdir/t1
 	set t2 $testdir/t2
 	set t3 $testdir/t3
 	cleanup $testdir $env
 
-	set db [eval {berkdb_open -create -truncate \
+	set db [eval {berkdb_open -create \
 	    -mode 0644} [concat $args "-dup"] {$omethod $testfile}]
 	error_check_good dbopen [is_valid_db $db] TRUE
 
@@ -73,8 +93,7 @@ proc test011 { method {nentries 10000} {ndups 5} {tnum 11} args } {
 	# We will add dups with values 1, 3, ... $ndups.  Then we'll add
 	# 0 and $ndups+1 using keyfirst/keylast.  We'll add 2 and 4 using
 	# add before and add after.
-	puts "\tTest0$tnum.a: put and get duplicate keys."
-	set dbc [eval {$db cursor} $txn]
+	puts "\tTest$tnum.a: put and get duplicate keys."
 	set i ""
 	for { set i 1 } { $i <= $ndups } { incr i 2 } {
 		lappend dlist $i
@@ -83,12 +102,26 @@ proc test011 { method {nentries 10000} {ndups 5} {tnum 11} args } {
 	while { [gets $did str] != -1 && $count < $nentries } {
 		for { set i 1 } { $i <= $ndups } { incr i 2 } {
 			set datastr $i:$str
+			if { $txnenv == 1 } {
+				set t [$env txn]
+				error_check_good txn [is_valid_txn $t $env] TRUE
+				set txn "-txn $t"
+			}
 			set ret [eval {$db put} $txn $pflags {$str $datastr}]
 			error_check_good put $ret 0
+			if { $txnenv == 1 } {
+				error_check_good txn [$t commit] 0
+			}
 		}
 
 		# Now retrieve all the keys matching this key
 		set x 1
+		if { $txnenv == 1 } {
+			set t [$env txn]
+			error_check_good txn [is_valid_txn $t $env] TRUE
+			set txn "-txn $t"
+		}
+		set dbc [eval {$db cursor} $txn]
 		for {set ret [$dbc get "-set" $str ]} \
 		    {[llength $ret] != 0} \
 		    {set ret [$dbc get "-next"] } {
@@ -102,22 +135,33 @@ proc test011 { method {nentries 10000} {ndups 5} {tnum 11} args } {
 			set datastr [lindex [lindex $ret 0] 1]
 			set d [data_of $datastr]
 
-			error_check_good Test0$tnum:put $d $str
+			error_check_good Test$tnum:put $d $str
 			set id [ id_of $datastr ]
-			error_check_good Test0$tnum:dup# $id $x
+			error_check_good Test$tnum:dup# $id $x
 			incr x 2
 		}
-		error_check_good Test0$tnum:numdups $x $maxodd
+		error_check_good Test$tnum:numdups $x $maxodd
+		error_check_good curs_close [$dbc close] 0
+		if { $txnenv == 1 } {
+			error_check_good txn [$t commit] 0
+		}
 		incr count
 	}
-	error_check_good curs_close [$dbc close] 0
 	close $did
 
 	# Now we will get each key from the DB and compare the results
 	# to the original.
-	puts "\tTest0$tnum.b: \
+	puts "\tTest$tnum.b: \
 	    traverse entire file checking duplicates before close."
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	dup_check $db $txn $t1 $dlist
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 
 	# Now compare the keys to see if they match the dictionary entries
 	set q q
@@ -125,7 +169,7 @@ proc test011 { method {nentries 10000} {ndups 5} {tnum 11} args } {
 	filesort $t3 $t2
 	filesort $t1 $t3
 
-	error_check_good Test0$tnum:diff($t3,$t2) \
+	error_check_good Test$tnum:diff($t3,$t2) \
 	    [filecmp $t3 $t2] 0
 
 	error_check_good db_close [$db close] 0
@@ -133,34 +177,74 @@ proc test011 { method {nentries 10000} {ndups 5} {tnum 11} args } {
 	set db [eval {berkdb_open} $args $testfile]
 	error_check_good dbopen [is_valid_db $db] TRUE
 
-	puts "\tTest0$tnum.c: \
+	puts "\tTest$tnum.c: \
 	    traverse entire file checking duplicates after close."
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	dup_check $db $txn $t1 $dlist
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 
 	# Now compare the keys to see if they match the dictionary entries
 	filesort $t1 $t3
-	error_check_good Test0$tnum:diff($t3,$t2) \
+	error_check_good Test$tnum:diff($t3,$t2) \
 	    [filecmp $t3 $t2] 0
 
-	puts "\tTest0$tnum.d: Testing key_first functionality"
+	puts "\tTest$tnum.d: Testing key_first functionality"
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	add_dup $db $txn $nentries "-keyfirst" 0 0
 	set dlist [linsert $dlist 0 0]
 	dup_check $db $txn $t1 $dlist
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 
-	puts "\tTest0$tnum.e: Testing key_last functionality"
+	puts "\tTest$tnum.e: Testing key_last functionality"
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	add_dup $db $txn $nentries "-keylast" [expr $maxodd - 1] 0
 	lappend dlist [expr $maxodd - 1]
 	dup_check $db $txn $t1 $dlist
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 
-	puts "\tTest0$tnum.f: Testing add_before functionality"
+	puts "\tTest$tnum.f: Testing add_before functionality"
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	add_dup $db $txn $nentries "-before" 2 3
 	set dlist [linsert $dlist 2 2]
 	dup_check $db $txn $t1 $dlist
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 
-	puts "\tTest0$tnum.g: Testing add_after functionality"
+	puts "\tTest$tnum.g: Testing add_after functionality"
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	add_dup $db $txn $nentries "-after" 4 4
 	set dlist [linsert $dlist 4 4]
 	dup_check $db $txn $t1 $dlist
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 
 	error_check_good db_close [$db close] 0
 }
@@ -193,7 +277,7 @@ proc add_dup {db txn nentries flag dataval iter} {
 	$dbc close
 }
 
-proc test011_recno { method {nentries 10000} {tnum 11} largs } {
+proc test011_recno { method {nentries 10000} {tnum "011"} largs } {
 	global dlist
 	source ./include.tcl
 
@@ -201,7 +285,7 @@ proc test011_recno { method {nentries 10000} {tnum 11} largs } {
 	set omethod [convert_method $method]
 	set renum [is_rrecno $method]
 
-	puts "Test0$tnum: \
+	puts "Test$tnum: \
 	    $method ($largs) $nentries test cursor insert functionality"
 
 	# Create the database and open the dictionary
@@ -209,13 +293,26 @@ proc test011_recno { method {nentries 10000} {tnum 11} largs } {
 	#
 	# If we are using an env, then testfile should just be the db name.
 	# Otherwise it is the test directory and the name.
+	set txnenv 0
 	if { $eindex == -1 } {
-		set testfile $testdir/test0$tnum.db
+		set testfile $testdir/test$tnum.db
 		set env NULL
 	} else {
-		set testfile test0$tnum.db
+		set testfile test$tnum.db
 		incr eindex
 		set env [lindex $largs $eindex]
+		set txnenv [is_txnenv $env]
+		if { $txnenv == 1 } {
+			append largs " -auto_commit "
+			#
+			# If we are using txns and running with the
+			# default, set the default down a bit.
+			#
+			if { $nentries == 10000 } {
+				set nentries 100
+			}
+		}
+		set testdir [get_home $env]
 	}
 	set t1 $testdir/t1
 	set t2 $testdir/t2
@@ -226,7 +323,7 @@ proc test011_recno { method {nentries 10000} {tnum 11} largs } {
 		append largs " -renumber"
 	}
 	set db [eval {berkdb_open \
-	     -create -truncate -mode 0644} $largs {$omethod $testfile}]
+	     -create -mode 0644} $largs {$omethod $testfile}]
 	error_check_good dbopen [is_valid_db $db] TRUE
 
 	set did [open $dict]
@@ -247,15 +344,28 @@ proc test011_recno { method {nentries 10000} {tnum 11} largs } {
 
 	# Seed the database with an initial record
 	gets $did str
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	set ret [eval {$db put} $txn {1 [chop_data $method $str]}]
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 	error_check_good put $ret 0
 	set count 1
 
 	set dlist "NULL $str"
 
 	# Open a cursor
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	set dbc [eval {$db cursor} $txn]
-	puts "\tTest0$tnum.a: put and get entries"
+	puts "\tTest$tnum.a: put and get entries"
 	while { [gets $did str] != -1 && $count < $nentries } {
 		# Pick a random key
 		set key [berkdb random_int 1 $count]
@@ -312,6 +422,9 @@ proc test011_recno { method {nentries 10000} {tnum 11} largs } {
 	}
 	close $did
 	error_check_good cclose [$dbc close] 0
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 
 	# Create  check key file.
 	set oid [open $t2 w]
@@ -320,25 +433,33 @@ proc test011_recno { method {nentries 10000} {tnum 11} largs } {
 	}
 	close $oid
 
-	puts "\tTest0$tnum.b: dump file"
+	puts "\tTest$tnum.b: dump file"
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	dump_file $db $txn $t1 test011_check
-	error_check_good Test0$tnum:diff($t2,$t1) \
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
+	error_check_good Test$tnum:diff($t2,$t1) \
 	    [filecmp $t2 $t1] 0
 
 	error_check_good db_close [$db close] 0
 
-	puts "\tTest0$tnum.c: close, open, and dump file"
-	open_and_dump_file $testfile $env $txn $t1 test011_check \
+	puts "\tTest$tnum.c: close, open, and dump file"
+	open_and_dump_file $testfile $env $t1 test011_check \
 	    dump_file_direction "-first" "-next"
-	error_check_good Test0$tnum:diff($t2,$t1) \
+	error_check_good Test$tnum:diff($t2,$t1) \
 	    [filecmp $t2 $t1] 0
 
-	puts "\tTest0$tnum.d: close, open, and dump file in reverse direction"
-	open_and_dump_file $testfile $env $txn $t1 test011_check \
+	puts "\tTest$tnum.d: close, open, and dump file in reverse direction"
+	open_and_dump_file $testfile $env $t1 test011_check \
 	    dump_file_direction "-last" "-prev"
 
 	filesort $t1 $t3 -n
-	error_check_good Test0$tnum:diff($t2,$t3) \
+	error_check_good Test$tnum:diff($t2,$t3) \
 	    [filecmp $t2 $t3] 0
 }
 
