@@ -1,9 +1,9 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999-2001
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1999-2006
+#	Oracle Corporation.  All rights reserved.
 #
-# $Id: test070.tcl,v 11.22 2001/08/03 16:39:45 bostic Exp $
+# $Id: test070.tcl,v 12.3 2006/08/24 14:46:41 bostic Exp $
 #
 # TEST	test070
 # TEST	Test of DB_CONSUME (Four consumers, 1000 items.)
@@ -13,9 +13,10 @@
 # TEST	the consumers will each get 10000.
 # TEST	Then, verify that no record was lost or retrieved twice.
 proc test070 { method {nconsumers 4} {nproducers 2} \
-    {nitems 1000} {mode CONSUME } {start 0} {txn -txn} {tnum 70} args } {
+    {nitems 1000} {mode CONSUME } {start 0} {txn -txn} {tnum "070"} args } {
 	source ./include.tcl
 	global alphabet
+	global encrypt
 
 	#
 	# If we are using an env, then skip this test.  It needs its own.
@@ -23,28 +24,32 @@ proc test070 { method {nconsumers 4} {nproducers 2} \
 	if { $eindex != -1 } {
 		incr eindex
 		set env [lindex $args $eindex]
-		puts "Test0$tnum skipping for env $env"
+		puts "Test$tnum skipping for env $env"
 		return
 	}
 	set omethod [convert_method $method]
 	set args [convert_args $method $args]
+	if { $encrypt != 0 } {
+		puts "Test$tnum skipping for security"
+		return
+	}
 
-	puts "Test0$tnum: $method ($args) Test of DB_$mode flag to DB->get."
+	puts "Test$tnum: $method ($args) Test of DB_$mode flag to DB->get."
 	puts "\tUsing $txn environment."
 
 	error_check_good enough_consumers [expr $nconsumers > 0] 1
 	error_check_good enough_producers [expr $nproducers > 0] 1
 
 	if { [is_queue $method] != 1 } {
-		puts "\tSkipping Test0$tnum for method $method."
+		puts "\tSkipping Test$tnum for method $method."
 		return
 	}
 
 	env_cleanup $testdir
-	set testfile test0$tnum.db
+	set testfile test$tnum.db
 
 	# Create environment
-	set dbenv [eval {berkdb env -create $txn -home } $testdir]
+	set dbenv [eval {berkdb_env -create $txn -home } $testdir]
 	error_check_good dbenv_create [is_valid_env $dbenv] TRUE
 
 	# Create database
@@ -54,7 +59,7 @@ proc test070 { method {nconsumers 4} {nproducers 2} \
 
 	if { $start != 0 } {
 		error_check_good set_seed [$db put $start "consumer data"] 0
-		puts "\tStarting at $start."
+		puts "\tTest$tnum: starting at $start."
 	} else {
 		incr start
 	}
@@ -88,34 +93,28 @@ proc test070 { method {nconsumers 4} {nproducers 2} \
 	}
 
 	# Wait for all children.
-	watch_procs 10
+	watch_procs $pidlist 10
 
 	# Verify: slurp all record numbers into list, sort, and make
 	# sure each appears exactly once.
-	puts "\tTest0$tnum: Verifying results."
+	puts "\tTest$tnum: Verifying results."
 	set reclist {}
 	for { set ndx 0 } { $ndx < $nconsumers } { incr ndx } {
 		set input $consumerlog$ndx
 		set iid [open $input r]
 		while { [gets $iid str] != -1 } {
-			# Convert high ints to negative ints, to
-			# simulate Tcl's behavior on a 32-bit machine
-			# even if we're on a 64-bit one.
-			if { $str > 0x7fffffff } {
-				set str [expr $str - 1 - 0xffffffff]
-			}
 			lappend reclist $str
 		}
 		close $iid
 	}
-	set sortreclist [lsort -integer $reclist]
+	set sortreclist [lsort -command int32_compare $reclist]
 
 	set nitems [expr $start + $nitems]
-	for { set ndx $start } { $ndx < $nitems } { incr ndx } {
-		# Convert high ints to negative ints, to simulate
-		# 32-bit behavior on 64-bit platforms.
-		if { $ndx > 0x7fffffff } {
-			set cmp [expr $ndx - 1 - 0xffffffff]
+	for { set ndx $start } { $ndx < $nitems } { set ndx [expr $ndx + 1] } {
+		# Wrap if $ndx goes beyond 32 bits because our
+		# recno wrapped if it did.
+		if { $ndx > 0xffffffff } {
+			set cmp [expr $ndx - 0xffffffff]
 		} else {
 			set cmp [expr $ndx + 0]
 		}
@@ -133,5 +132,5 @@ proc test070 { method {nconsumers 4} {nproducers 2} \
 	error_check_good db_close [$db close] 0
 	error_check_good dbenv_close [$dbenv close] 0
 
-	puts "\tTest0$tnum completed successfully."
+	puts "\tTest$tnum completed successfully."
 }
