@@ -1,9 +1,9 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999-2001
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1999-2006
+#	Oracle Corporation.  All rights reserved.
 #
-# $Id: test073.tcl,v 11.21 2001/08/03 16:39:45 bostic Exp $
+# $Id: test073.tcl,v 12.3 2006/08/24 14:46:41 bostic Exp $
 #
 # TEST	test073
 # TEST	Test of cursor stability on duplicate pages.
@@ -22,30 +22,37 @@
 # TEST	f. Ditto for the two sequence tests, only doing a
 # TEST	   DBC->c_put(DB_CURRENT) of a larger datum instead of adding a
 # TEST	   new one.
-proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
+proc test073 { method {pagesize 512} {ndups 50} {tnum "073"} args } {
 	source ./include.tcl
 	global alphabet
 
 	set omethod [convert_method $method]
 	set args [convert_args $method $args]
 
+	set txnenv 0
 	set eindex [lsearch -exact $args "-env"]
 	#
 	# If we are using an env, then testfile should just be the db name.
 	# Otherwise it is the test directory and the name.
 	if { $eindex == -1 } {
-		set testfile $testdir/test0$tnum.db
+		set testfile $testdir/test$tnum.db
 		set env NULL
 	} else {
-		set testfile test0$tnum.db
+		set testfile test$tnum.db
 		incr eindex
 		set env [lindex $args $eindex]
+		set txnenv [is_txnenv $env]
+		if { $txnenv == 1 } {
+			append args " -auto_commit "
+		}
+		set testdir [get_home $env]
 	}
 	cleanup $testdir $env
 
 	set key "the key"
+	set txn ""
 
-	puts -nonewline "Test0$tnum $omethod ($args): "
+	puts -nonewline "Test$tnum $omethod ($args): "
 	if { [is_record_based $method] || [is_rbtree $method] } {
 		puts "Skipping for method $method."
 		return
@@ -67,22 +74,36 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 	# Number of outstanding keys.
 	set keys 0
 
-	puts "\tTest0$tnum.a.1: Initializing put loop; $ndups dups, short data."
+	puts "\tTest$tnum.a.1: Initializing put loop; $ndups dups, short data."
 
 	for { set i 0 } { $i < $ndups } { incr i } {
 		set datum [makedatum_t73 $i 0]
 
-		error_check_good "db put ($i)" [$db put $key $datum] 0
+		if { $txnenv == 1 } {
+			set t [$env txn]
+			error_check_good txn [is_valid_txn $t $env] TRUE
+			set txn "-txn $t"
+		}
+		set ret [eval {$db put} $txn {$key $datum}]
+		error_check_good "db put ($i)" $ret 0
+		if { $txnenv == 1 } {
+			error_check_good txn [$t commit] 0
+		}
 
 		set is_long($i) 0
 		incr keys
 	}
 
-	puts "\tTest0$tnum.a.2: Initializing cursor get loop; $keys dups."
+	puts "\tTest$tnum.a.2: Initializing cursor get loop; $keys dups."
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
 	for { set i 0 } { $i < $keys } { incr i } {
 		set datum [makedatum_t73 $i 0]
 
-		set dbc($i) [$db cursor]
+		set dbc($i) [eval {$db cursor} $txn]
 		error_check_good "db cursor ($i)"\
 		    [is_valid_cursor $dbc($i) $db] TRUE
 		error_check_good "dbc get -get_both ($i)"\
@@ -90,7 +111,7 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 		    [list [list $key $datum]]
 	}
 
-	puts "\tTest0$tnum.b: Cursor put (DB_KEYLAST); $ndups new dups,\
+	puts "\tTest$tnum.b: Cursor put (DB_KEYLAST); $ndups new dups,\
 	    short data."
 
 	for { set i 0 } { $i < $ndups } { incr i } {
@@ -98,7 +119,7 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 		# to be added (since they start from zero)
 
 		set datum [makedatum_t73 $keys 0]
-		set curs [$db cursor]
+		set curs [eval {$db cursor} $txn]
 		error_check_good "db cursor create" [is_valid_cursor $curs $db]\
 		    TRUE
 		error_check_good "c_put(DB_KEYLAST, $keys)"\
@@ -111,7 +132,7 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 		verify_t73 is_long dbc $keys $key
 	}
 
-	puts "\tTest0$tnum.c: Cursor put (DB_KEYFIRST); $ndups new dups,\
+	puts "\tTest$tnum.c: Cursor put (DB_KEYFIRST); $ndups new dups,\
 	    short data."
 
 	for { set i 0 } { $i < $ndups } { incr i } {
@@ -119,7 +140,7 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 		# to be added (since they start from zero)
 
 		set datum [makedatum_t73 $keys 0]
-		set curs [$db cursor]
+		set curs [eval {$db cursor} $txn]
 		error_check_good "db cursor create" [is_valid_cursor $curs $db]\
 		    TRUE
 		error_check_good "c_put(DB_KEYFIRST, $keys)"\
@@ -132,14 +153,14 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 		verify_t73 is_long dbc $keys $key
 	}
 
-	puts "\tTest0$tnum.d: Cursor put (DB_AFTER) first to last;\
+	puts "\tTest$tnum.d: Cursor put (DB_AFTER) first to last;\
 	    $keys new dups, short data"
 	# We want to add a datum after each key from 0 to the current
 	# value of $keys, which we thus need to save.
 	set keysnow $keys
 	for { set i 0 } { $i < $keysnow } { incr i } {
 		set datum [makedatum_t73 $keys 0]
-		set curs [$db cursor]
+		set curs [eval {$db cursor} $txn]
 		error_check_good "db cursor create" [is_valid_cursor $curs $db]\
 		    TRUE
 
@@ -158,12 +179,12 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 		verify_t73 is_long dbc $keys $key
 	}
 
-	puts "\tTest0$tnum.e: Cursor put (DB_BEFORE) last to first;\
+	puts "\tTest$tnum.e: Cursor put (DB_BEFORE) last to first;\
 	    $keys new dups, short data"
 
 	for { set i [expr $keys - 1] } { $i >= 0 } { incr i -1 } {
 		set datum [makedatum_t73 $keys 0]
-		set curs [$db cursor]
+		set curs [eval {$db cursor} $txn]
 		error_check_good "db cursor create" [is_valid_cursor $curs $db]\
 		    TRUE
 
@@ -185,13 +206,13 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 	}
 	verify_t73 is_long dbc $keys $key
 
-	puts "\tTest0$tnum.f: Cursor put (DB_CURRENT), first to last,\
+	puts "\tTest$tnum.f: Cursor put (DB_CURRENT), first to last,\
 	    growing $keys data."
 	set keysnow $keys
 	for { set i 0 } { $i < $keysnow } { incr i } {
 		set olddatum [makedatum_t73 $i 0]
 		set newdatum [makedatum_t73 $i 1]
-		set curs [$db cursor]
+		set curs [eval {$db cursor} $txn]
 		error_check_good "db cursor create" [is_valid_cursor $curs $db]\
 		    TRUE
 
@@ -212,10 +233,13 @@ proc test073 { method {pagesize 512} {ndups 50} {tnum 73} args } {
 	verify_t73 is_long dbc $keys $key
 
 	# Close cursors.
-	puts "\tTest0$tnum.g: Closing cursors."
+	puts "\tTest$tnum.g: Closing cursors."
 	for { set i 0 } { $i < $keys } { incr i } {
 		error_check_good "dbc close ($i)" [$dbc($i) close] 0
 	}
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 	error_check_good "db close" [$db close] 0
 }
 
