@@ -1,14 +1,16 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 1999, 2000
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 1999-2006
+#	Oracle Corporation.  All rights reserved.
 #
-#	$Id: test074.tcl,v 11.10 2000/08/25 14:21:58 sue Exp $
+# $Id: test074.tcl,v 12.3 2006/08/24 14:46:41 bostic Exp $
 #
-# DB Test 74: Test of DB_NEXT_NODUP.
-proc test074 { method {dir -nextnodup} {pagesize 512} {nitems 100} {tnum 74} args } {
+# TEST	test074
+# TEST	Test of DB_NEXT_NODUP.
+proc test074 { method {dir -nextnodup} {nitems 100} {tnum "074"} args } {
 	source ./include.tcl
 	global alphabet
+	global is_je_test
 	global rand_init
 
 	set omethod [convert_method $method]
@@ -24,32 +26,39 @@ proc test074 { method {dir -nextnodup} {pagesize 512} {nitems 100} {tnum 74} arg
 		set globaldata [repeat $alphabet 4]
 	}
 
-	puts "Test0$tnum $omethod ($args): Test of $dir"
+	puts "Test$tnum $omethod ($args): Test of $dir"
 
 	# First, test non-dup (and not-very-interesting) case with
 	# all db types.
 
-	puts "\tTest0$tnum.a: No duplicates."
+	puts "\tTest$tnum.a: No duplicates."
 
+	set txnenv 0
 	set eindex [lsearch -exact $args "-env"]
 	#
 	# If we are using an env, then testfile should just be the db name.
 	# Otherwise it is the test directory and the name.
 	if { $eindex == -1 } {
-		set testfile $testdir/test0$tnum-nodup.db
+		set testfile $testdir/test$tnum-nodup.db
 		set env NULL
 	} else {
-		set testfile test0$tnum-nodup.db
+		set testfile test$tnum-nodup.db
 		incr eindex
 		set env [lindex $args $eindex]
+		set txnenv [is_txnenv $env]
+		if { $txnenv == 1 } {
+			append args " -auto_commit "
+		}
+		set testdir [get_home $env]
 	}
 	cleanup $testdir $env
-	set db [eval {berkdb_open -create -truncate -mode 0644} $omethod\
+	set db [eval {berkdb_open -create -mode 0644} $omethod\
 	    $args {$testfile}]
 	error_check_good db_open [is_valid_db $db] TRUE
+	set txn ""
 
 	# Insert nitems items.
-	puts "\t\tTest0$tnum.a.1: Put loop."
+	puts "\t\tTest$tnum.a.1: Put loop."
 	for {set i 1} {$i <= $nitems} {incr i} {
 		#
 		# If record based, set key to $i * 2 to leave
@@ -61,14 +70,28 @@ proc test074 { method {dir -nextnodup} {pagesize 512} {nitems 100} {tnum 74} arg
 			set key "key$i"
 		}
 		set data "$globaldata$i"
-		error_check_good put($i) [$db put $key\
-		    [chop_data $method $data]] 0
+		if { $txnenv == 1 } {
+			set t [$env txn]
+			error_check_good txn [is_valid_txn $t $env] TRUE
+			set txn "-txn $t"
+		}
+		set ret [eval {$db put} $txn {$key \
+		    [chop_data $method $data]}]
+		error_check_good put($i) $ret 0
+		if { $txnenv == 1 } {
+			error_check_good txn [$t commit] 0
+		}
 	}
 
-	puts "\t\tTest0$tnum.a.2: Get($dir)"
+	puts "\t\tTest$tnum.a.2: Get($dir)"
 
 	# foundarray($i) is set when key number i is found in the database
-	set dbc [$db cursor]
+	if { $txnenv == 1 } {
+		set t [$env txn]
+		error_check_good txn [is_valid_txn $t $env] TRUE
+		set txn "-txn $t"
+	}
+	set dbc [eval {$db cursor} $txn]
 	error_check_good db_cursor [is_valid_cursor $dbc $db] TRUE
 
 	# Initialize foundarray($i) to zero for all $i
@@ -96,54 +119,68 @@ proc test074 { method {dir -nextnodup} {pagesize 512} {nitems 100} {tnum 74} arg
 		set foundarray($num) 1
 	}
 
-	puts "\t\tTest0$tnum.a.3: Final key."
+	puts "\t\tTest$tnum.a.3: Final key."
 	error_check_good last_db_get [$dbc get $dir] [list]
 
-	puts "\t\tTest0$tnum.a.4: Verify loop."
+	puts "\t\tTest$tnum.a.4: Verify loop."
 	for { set i 1 } { $i <= $nitems } { incr i } {
 		error_check_good found_key($i) $foundarray($i) 1
 	}
 
 	error_check_good dbc_close(nodup) [$dbc close] 0
+	if { $txnenv == 1 } {
+		error_check_good txn [$t commit] 0
+	}
 
 	# If we are a method that doesn't allow dups, verify that
 	# we get an empty list if we try to use DB_NEXT_DUP
 	if { [is_record_based $method] == 1 || [is_rbtree $method] == 1 } {
-		puts "\t\tTest0$tnum.a.5: Check DB_NEXT_DUP for $method."
-		set dbc [$db cursor]
+		if { $txnenv == 1 } {
+			set t [$env txn]
+			error_check_good txn [is_valid_txn $t $env] TRUE
+			set txn "-txn $t"
+		}
+		puts "\t\tTest$tnum.a.5: Check DB_NEXT_DUP for $method."
+		set dbc [eval {$db cursor} $txn]
 		error_check_good db_cursor [is_valid_cursor $dbc $db] TRUE
 
 		set dbt [$dbc get $dir]
 		error_check_good $method:nextdup [$dbc get -nextdup] [list]
 		error_check_good dbc_close(nextdup) [$dbc close] 0
+		if { $txnenv == 1 } {
+			error_check_good txn [$t commit] 0
+		}
 	}
 	error_check_good db_close(nodup) [$db close] 0
 
 	# Quit here if we're a method that won't allow dups.
 	if { [is_record_based $method] == 1 || [is_rbtree $method] == 1 } {
-		puts "\tTest0$tnum: Skipping remainder for method $method."
+		puts "\tTest$tnum: Skipping remainder for method $method."
 		return
 	}
 
 	foreach opt { "-dup" "-dupsort" } {
+		if { $is_je_test && $opt == "-dup" } {
+			continue
+		}
 
 		#
 		# If we are using an env, then testfile should just be the
 		# db name.  Otherwise it is the test directory and the name.
 		if { $eindex == -1 } {
-			set testfile $testdir/test0$tnum$opt.db
+			set testfile $testdir/test$tnum$opt.db
 		} else {
-			set testfile test0$tnum$opt.db
+			set testfile test$tnum$opt.db
 		}
 
 		if { [string compare $opt "-dupsort"] == 0 } {
 			set opt "-dup -dupsort"
 		}
 
-		puts "\tTest0$tnum.b: Duplicates ($opt)."
+		puts "\tTest$tnum.b: Duplicates ($opt)."
 
-		puts "\t\tTest0$tnum.b.1 ($opt): Put loop."
-		set db [eval {berkdb_open -create -truncate -mode 0644}\
+		puts "\t\tTest$tnum.b.1 ($opt): Put loop."
+		set db [eval {berkdb_open -create -mode 0644}\
 		    $opt $omethod $args {$testfile}]
 		error_check_good db_open [is_valid_db $db] TRUE
 
@@ -160,8 +197,17 @@ proc test074 { method {dir -nextnodup} {pagesize 512} {nitems 100} {tnum 74} arg
 					set data "$globaldata$j"
 				}
 
-				error_check_good put($i,$j) \
-				    [$db put $key $data] 0
+				if { $txnenv == 1 } {
+					set t [$env txn]
+					error_check_good txn \
+					    [is_valid_txn $t $env] TRUE
+					set txn "-txn $t"
+				}
+				set ret [eval {$db put} $txn {$key $data}]
+				error_check_good put($i,$j) $ret 0
+				if { $txnenv == 1 } {
+					error_check_good txn [$t commit] 0
+				}
 			}
 		}
 
@@ -173,9 +219,14 @@ proc test074 { method {dir -nextnodup} {pagesize 512} {nitems 100} {tnum 74} arg
 
 		# Get loop--after each get, move forward a random increment
 		# within the duplicate set.
-		puts "\t\tTest0$tnum.b.2 ($opt): Get loop."
+		puts "\t\tTest$tnum.b.2 ($opt): Get loop."
 		set one "001"
-		set dbc [$db cursor]
+		if { $txnenv == 1 } {
+			set t [$env txn]
+			error_check_good txn [is_valid_txn $t $env] TRUE
+			set txn "-txn $t"
+		}
+		set dbc [eval {$db cursor} $txn]
 		error_check_good dbc($opt) [is_valid_cursor $dbc $db] TRUE
 		for { set i 1 } { $i <= $nitems } { incr i } {
 			set dbt [$dbc get $dir]
@@ -206,16 +257,19 @@ proc test074 { method {dir -nextnodup} {pagesize 512} {nitems 100} {tnum 74} arg
 			}
 		}
 
-		puts "\t\tTest0$tnum.b.3 ($opt): Final key."
+		puts "\t\tTest$tnum.b.3 ($opt): Final key."
 		error_check_good last_db_get($opt) [$dbc get $dir] [list]
 
 		# Verify
-		puts "\t\tTest0$tnum.b.4 ($opt): Verify loop."
+		puts "\t\tTest$tnum.b.4 ($opt): Verify loop."
 		for { set i 1 } { $i <= $nitems } { incr i } {
 			error_check_good found_key($i) $foundarray($i) 1
 		}
 
 		error_check_good dbc_close [$dbc close] 0
+		if { $txnenv == 1 } {
+			error_check_good txn [$t commit] 0
+		}
 		error_check_good db_close [$db close] 0
 	}
 }
