@@ -1,17 +1,20 @@
 # See the file LICENSE for redistribution information.
 #
-# Copyright (c) 2000-2001
-#	Sleepycat Software.  All rights reserved.
+# Copyright (c) 2000-2006
+#	Oracle Corporation.  All rights reserved.
 #
-# $Id: test076.tcl,v 1.12 2001/08/03 16:39:46 bostic Exp $
+# $Id: test076.tcl,v 12.4 2006/08/24 14:46:41 bostic Exp $
 #
 # TEST	test076
 # TEST	Test creation of many small databases in a single environment. [#1528].
-proc test076 { method { ndbs 1000  } { tnum 76 } args } {
+proc test076 { method { ndbs 1000 } { tnum "076" } args } {
+	global is_qnx_test
 	source ./include.tcl
 
-	set omethod [convert_method $method]
 	set args [convert_args $method $args]
+	set encargs ""
+	set args [split_encargs $args encargs]
+	set omethod [convert_method $method]
 
 	if { [is_record_based $method] == 1 } {
 		set key ""
@@ -20,34 +23,60 @@ proc test076 { method { ndbs 1000  } { tnum 76 } args } {
 	}
 	set data "datamoredatamoredata"
 
-	puts -nonewline "Test0$tnum $method ($args): "
-	puts -nonewline "Create $ndbs"
-	puts " small databases in one env."
-
 	# Create an env if we weren't passed one.
+	set txnenv 0
 	set eindex [lsearch -exact $args "-env"]
 	if { $eindex == -1 } {
 		set deleteenv 1
 		env_cleanup $testdir
-		set env [eval {berkdb env -create -home} $testdir]
+		set env [eval {berkdb_env -create -home} $testdir $encargs]
 		error_check_good env [is_valid_env $env] TRUE
 		set args "$args -env $env"
 	} else {
 		set deleteenv 0
 		incr eindex
 		set env [lindex $args $eindex]
+		set txnenv [is_txnenv $env]
+		if { $txnenv == 1 } {
+			append args " -auto_commit "
+			if { $ndbs == 1000 } {
+				set ndbs 100
+			}
+		}
+		set testdir [get_home $env]
 	}
+	if { $is_qnx_test && $ndbs > 100 } {
+		set ndbs 100
+	}
+	if { [is_queueext $method] } {
+		set ndbs 500
+	}
+
+	puts -nonewline "Test$tnum $method ($args): "
+	puts -nonewline "Create $ndbs"
+	puts " small databases in one env."
+
 	cleanup $testdir $env
+	set txn ""
 
 	for { set i 1 } { $i <= $ndbs } { incr i } {
-		set testfile test0$tnum.$i.db
+		set testfile test$tnum.$i.db
 
 		set db [eval {berkdb_open -create -mode 0644}\
 		    $args $omethod $testfile]
 		error_check_good db_open($i) [is_valid_db $db] TRUE
 
-		error_check_good db_put($i) [$db put $key$i \
-		    [chop_data $method $data$i]] 0
+		if { $txnenv == 1 } {
+			set t [$env txn]
+			error_check_good txn [is_valid_txn $t $env] TRUE
+			set txn "-txn $t"
+		}
+		set ret [eval {$db put} $txn {$key$i \
+		    [chop_data $method $data$i]}]
+		error_check_good db_put($i) $ret 0
+		if { $txnenv == 1 } {
+			error_check_good txn [$t commit] 0
+		}
 		error_check_good db_close($i) [$db close] 0
 	}
 
@@ -55,5 +84,5 @@ proc test076 { method { ndbs 1000  } { tnum 76 } args } {
 		error_check_good env_close [$env close] 0
 	}
 
-	puts "\tTest0$tnum passed."
+	puts "\tTest$tnum passed."
 }
