@@ -1,8 +1,8 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  */
 /*
  * Copyright (c) 1996
@@ -32,17 +32,11 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: txn_rec.c,v 11.64 2004/09/22 17:41:10 bostic Exp $
+ * $Id: txn_rec.c,v 12.14 2006/08/24 14:46:53 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/db_page.h"
 #include "dbinc/txn.h"
@@ -90,28 +84,28 @@ __txn_regop_recover(dbenv, dbtp, lsnp, op, info)
 		 * that's OK.  Ignore the return code from remove.
 		 */
 		if ((ret = __db_txnlist_remove(dbenv,
-		    info, argp->txnid->txnid)) != DB_NOTFOUND && ret != 0)
+		    info, argp->txnp->txnid)) != DB_NOTFOUND && ret != 0)
 			goto err;
 	} else if ((dbenv->tx_timestamp != 0 &&
 	    argp->timestamp > (int32_t)dbenv->tx_timestamp) ||
 	    (!IS_ZERO_LSN(headp->trunc_lsn) &&
-	    log_compare(&headp->trunc_lsn, lsnp) < 0)) {
+	    LOG_COMPARE(&headp->trunc_lsn, lsnp) < 0)) {
 		/*
 		 * We failed either the timestamp check or the trunc_lsn check,
 		 * so we treat this as an abort even if it was a commit record.
 		 */
 		if ((ret = __db_txnlist_update(dbenv, info,
-		    argp->txnid->txnid, TXN_ABORT, NULL, &status, 1)) != 0)
+		    argp->txnp->txnid, TXN_ABORT, NULL, &status, 1)) != 0)
 			goto err;
 		else if (status != TXN_IGNORE && status != TXN_OK)
 			goto err;
 	} else {
 		/* This is a normal commit; mark it appropriately. */
 		if ((ret = __db_txnlist_update(dbenv,
-		    info, argp->txnid->txnid, argp->opcode, lsnp,
+		    info, argp->txnp->txnid, argp->opcode, lsnp,
 		    &status, 0)) == DB_NOTFOUND) {
 			if ((ret = __db_txnlist_add(dbenv,
-			    info, argp->txnid->txnid,
+			    info, argp->txnp->txnid,
 			    argp->opcode == TXN_ABORT ?
 			    TXN_IGNORE : argp->opcode, lsnp)) != 0)
 				goto err;
@@ -124,9 +118,9 @@ __txn_regop_recover(dbenv, dbtp, lsnp, op, info)
 		*lsnp = argp->prev_lsn;
 
 	if (0) {
-err:		__db_err(dbenv,
+err:		__db_errx(dbenv,
 		    "txnid %lx commit record found, already on commit list",
-		    (u_long)argp->txnid->txnid);
+		    (u_long)argp->txnp->txnid);
 		ret = EINVAL;
 	}
 	__os_free(dbenv, argp);
@@ -171,7 +165,7 @@ __txn_xa_regop_recover(dbenv, dbtp, lsnp, op, info)
 	 * cases below, we overwrite the ret value so we return
 	 * appropriately.
 	 */
-	ret = __db_txnlist_find(dbenv, info, argp->txnid->txnid, &status);
+	ret = __db_txnlist_find(dbenv, info, argp->txnp->txnid, &status);
 
 	/*
 	 * If we are rolling forward, then an aborted prepare
@@ -182,7 +176,7 @@ __txn_xa_regop_recover(dbenv, dbtp, lsnp, op, info)
 
 	if (op == DB_TXN_FORWARD_ROLL) {
 		if ((ret = __db_txnlist_remove(dbenv,
-		    info, argp->txnid->txnid)) != 0)
+		    info, argp->txnp->txnid)) != 0)
 			goto txn_err;
 	} else if (op == DB_TXN_BACKWARD_ROLL && status == TXN_PREPARE) {
 		/*
@@ -198,7 +192,7 @@ __txn_xa_regop_recover(dbenv, dbtp, lsnp, op, info)
 		 */
 		if (argp->opcode == TXN_ABORT) {
 			if ((ret = __db_txnlist_update(dbenv,
-			     info, argp->txnid->txnid,
+			     info, argp->txnp->txnid,
 			     TXN_ABORT, NULL, &status, 0)) != 0 &&
 			     status != TXN_PREPARE)
 				goto txn_err;
@@ -212,12 +206,13 @@ __txn_xa_regop_recover(dbenv, dbtp, lsnp, op, info)
 		 * after recovery (see txn_recover).
 		 */
 		else if ((ret = __db_txnlist_remove(dbenv,
-		    info, argp->txnid->txnid)) != 0) {
-txn_err:		__db_err(dbenv,
-			    "Transaction not in list %x", argp->txnid->txnid);
+		    info, argp->txnp->txnid)) != 0) {
+txn_err:		__db_errx(dbenv,
+			    "transaction not in list %lx",
+			    (u_long)argp->txnp->txnid);
 			ret = DB_NOTFOUND;
 		} else if ((ret = __db_txnlist_add(dbenv,
-		   info, argp->txnid->txnid, TXN_COMMIT, lsnp)) == 0)
+		   info, argp->txnp->txnid, TXN_COMMIT, lsnp)) == 0)
 			ret = __txn_restore_txn(dbenv, lsnp, argp);
 	} else
 		ret = 0;
@@ -298,22 +293,20 @@ __txn_child_recover(dbenv, dbtp, lsnp, op, info)
 
 	/*
 	 * This is a record in a PARENT's log trail indicating that a
-	 * child committed.  If we are aborting, we need to update the
-	 * parent's LSN array.  If we are in recovery, then if the
+	 * child committed.  If we are aborting, return the childs last
+	 * record's LSN.  If we are in recovery, then if the
 	 * parent is committing, we set ourselves up to commit, else
 	 * we do nothing.
 	 */
 	if (op == DB_TXN_ABORT) {
-		/* Note that __db_txnlist_lsnadd rewrites its LSN
-		 * parameter, so you cannot reuse the argp->c_lsn field.
-		 */
-		ret = __db_txnlist_lsnadd(dbenv,
-		    info, &argp->c_lsn, TXNLIST_NEW);
+		*lsnp = argp->c_lsn;
+		ret = __db_txnlist_lsnadd(dbenv, info, &argp->prev_lsn);
+		goto out;
 	} else if (op == DB_TXN_BACKWARD_ROLL) {
 		/* Child might exist -- look for it. */
 		ret = __db_txnlist_find(dbenv, info, argp->child, &c_stat);
 		t_ret =
-		    __db_txnlist_find(dbenv, info, argp->txnid->txnid, &p_stat);
+		    __db_txnlist_find(dbenv, info, argp->txnp->txnid, &p_stat);
 		if (ret != 0 && ret != DB_NOTFOUND)
 			goto out;
 		if (t_ret != 0 && t_ret != DB_NOTFOUND) {
@@ -375,13 +368,13 @@ __txn_child_recover(dbenv, dbtp, lsnp, op, info)
 		if ((ret = __db_txnlist_find(dbenv,
 		    info, argp->child, &c_stat)) == DB_NOTFOUND)
 			ret = __db_txnlist_update(dbenv, info,
-			     argp->txnid->txnid, TXN_IGNORE,
+			     argp->txnp->txnid, TXN_IGNORE,
 			     NULL, &p_stat, 1);
 	} else if (DB_REDO(op)) {
 		/* Forward Roll */
 		if ((ret =
 		    __db_txnlist_remove(dbenv, info, argp->child)) != 0)
-			__db_err(dbenv,
+			__db_errx(dbenv,
 			    "Transaction not in list %x", argp->child);
 	}
 
@@ -422,36 +415,42 @@ __txn_restore_txn(dbenv, lsnp, argp)
 
 	mgr = dbenv->tx_handle;
 	region = mgr->reginfo.primary;
-	R_LOCK(dbenv, &mgr->reginfo);
+	TXN_SYSTEM_LOCK(dbenv);
 
 	/* Allocate a new transaction detail structure. */
 	if ((ret =
 	    __db_shalloc(&mgr->reginfo, sizeof(TXN_DETAIL), 0, &td)) != 0) {
-		R_UNLOCK(dbenv, &mgr->reginfo);
+		TXN_SYSTEM_UNLOCK(dbenv);
 		return (ret);
 	}
 
 	/* Place transaction on active transaction list. */
 	SH_TAILQ_INSERT_HEAD(&region->active_txn, td, links, __txn_detail);
 
-	td->txnid = argp->txnid->txnid;
-	td->begin_lsn = argp->begin_lsn;
+	td->txnid = argp->txnp->txnid;
+	__os_id(dbenv, &td->pid, &td->tid);
 	td->last_lsn = *lsnp;
-	td->parent = 0;
+	td->begin_lsn = argp->begin_lsn;
+	td->parent = INVALID_ROFF;
+	td->name = INVALID_ROFF;
+	SH_TAILQ_INIT(&td->kids);
+	MAX_LSN(td->read_lsn);
+	MAX_LSN(td->visible_lsn);
+	td->mvcc_ref = 0;
+	td->mvcc_mtx = MUTEX_INVALID;
 	td->status = TXN_PREPARED;
+	td->flags = TXN_DTL_RESTORED;
 	td->xa_status = TXN_XA_PREPARED;
 	memcpy(td->xid, argp->xid.data, argp->xid.size);
 	td->bqual = argp->bqual;
 	td->gtrid = argp->gtrid;
 	td->format = argp->formatID;
-	td->flags = 0;
-	F_SET(td, TXN_DTL_RESTORED);
 
 	region->stat.st_nrestores++;
 	region->stat.st_nactive++;
 	if (region->stat.st_nactive > region->stat.st_maxnactive)
 		region->stat.st_maxnactive = region->stat.st_nactive;
-	R_UNLOCK(dbenv, &mgr->reginfo);
+	TXN_SYSTEM_UNLOCK(dbenv);
 	return (0);
 }
 
@@ -489,3 +488,130 @@ __txn_recycle_recover(dbenv, dbtp, lsnp, op, info)
 
 	return (0);
 }
+
+/*
+ * PUBLIC: int __txn_regop_42_recover
+ * PUBLIC:    __P((DB_ENV *, DBT *, DB_LSN *, db_recops, void *));
+ *
+ * These records are only ever written for commits.  Normally, we redo any
+ * committed transaction, however if we are doing recovery to a timestamp, then
+ * we may treat transactions that committed after the timestamp as aborted.
+ */
+int
+__txn_regop_42_recover(dbenv, dbtp, lsnp, op, info)
+	DB_ENV *dbenv;
+	DBT *dbtp;
+	DB_LSN *lsnp;
+	db_recops op;
+	void *info;
+{
+	DB_TXNHEAD *headp;
+	__txn_regop_42_args *argp;
+	int ret;
+	u_int32_t status;
+
+#ifdef DEBUG_RECOVER
+	(void)__txn_regop_42_print(dbenv, dbtp, lsnp, op, info);
+#endif
+
+	if ((ret = __txn_regop_42_read(dbenv, dbtp->data, &argp)) != 0)
+		return (ret);
+
+	headp = info;
+	/*
+	 * We are only ever called during FORWARD_ROLL or BACKWARD_ROLL.
+	 * We check for the former explicitly and the last two clauses
+	 * apply to the BACKWARD_ROLL case.
+	 */
+
+	if (op == DB_TXN_FORWARD_ROLL) {
+		/*
+		 * If this was a 2-phase-commit transaction, then it
+		 * might already have been removed from the list, and
+		 * that's OK.  Ignore the return code from remove.
+		 */
+		if ((ret = __db_txnlist_remove(dbenv,
+		    info, argp->txnp->txnid)) != DB_NOTFOUND && ret != 0)
+			goto err;
+	} else if ((dbenv->tx_timestamp != 0 &&
+	    argp->timestamp > (int32_t)dbenv->tx_timestamp) ||
+	    (!IS_ZERO_LSN(headp->trunc_lsn) &&
+	    LOG_COMPARE(&headp->trunc_lsn, lsnp) < 0)) {
+		/*
+		 * We failed either the timestamp check or the trunc_lsn check,
+		 * so we treat this as an abort even if it was a commit record.
+		 */
+		if ((ret = __db_txnlist_update(dbenv, info,
+		    argp->txnp->txnid, TXN_ABORT, NULL, &status, 1)) != 0)
+			goto err;
+		else if (status != TXN_IGNORE && status != TXN_OK)
+			goto err;
+	} else {
+		/* This is a normal commit; mark it appropriately. */
+		if ((ret = __db_txnlist_update(dbenv,
+		    info, argp->txnp->txnid, argp->opcode, lsnp,
+		    &status, 0)) == DB_NOTFOUND) {
+			if ((ret = __db_txnlist_add(dbenv,
+			    info, argp->txnp->txnid,
+			    argp->opcode == TXN_ABORT ?
+			    TXN_IGNORE : argp->opcode, lsnp)) != 0)
+				goto err;
+		} else if (ret != 0 ||
+		    (status != TXN_IGNORE && status != TXN_OK))
+			goto err;
+	}
+
+	if (ret == 0)
+		*lsnp = argp->prev_lsn;
+
+	if (0) {
+err:		__db_errx(dbenv,
+		    "txnid %lx commit record found, already on commit list",
+		    (u_long)argp->txnp->txnid);
+		ret = EINVAL;
+	}
+	__os_free(dbenv, argp);
+
+	return (ret);
+}
+
+/*
+ * PUBLIC: int __txn_ckp_42_recover
+ * PUBLIC: __P((DB_ENV *, DBT *, DB_LSN *, db_recops, void *));
+ */
+int
+__txn_ckp_42_recover(dbenv, dbtp, lsnp, op, info)
+	DB_ENV *dbenv;
+	DBT *dbtp;
+	DB_LSN *lsnp;
+	db_recops op;
+	void *info;
+{
+	DB_REP *db_rep;
+	REP *rep;
+	__txn_ckp_42_args *argp;
+	int ret;
+
+#ifdef DEBUG_RECOVER
+	__txn_ckp_42_print(dbenv, dbtp, lsnp, op, info);
+#endif
+	if ((ret = __txn_ckp_42_read(dbenv, dbtp->data, &argp)) != 0)
+		return (ret);
+
+	if (op == DB_TXN_BACKWARD_ROLL)
+		__db_txnlist_ckp(dbenv, info, lsnp);
+
+	if (op == DB_TXN_FORWARD_ROLL) {
+		/* Record the max generation number that we've seen. */
+		if (REP_ON(dbenv)) {
+			db_rep = dbenv->rep_handle;
+			rep = db_rep->region;
+			if (argp->rep_gen > rep->recover_gen)
+				rep->recover_gen = argp->rep_gen;
+		}
+	}
+
+	*lsnp = argp->last_ckp;
+	__os_free(dbenv, argp);
+	return (DB_TXN_CKP);
+}
