@@ -1,32 +1,14 @@
 /*-
  * See the file LICENSE for redistribution information.
  *
- * Copyright (c) 1996-2004
- *	Sleepycat Software.  All rights reserved.
+ * Copyright (c) 1996-2006
+ *	Oracle Corporation.  All rights reserved.
  *
- * $Id: txn_stat.c,v 11.37 2004/10/15 16:59:44 bostic Exp $
+ * $Id: txn_stat.c,v 12.18 2006/08/24 14:46:53 bostic Exp $
  */
 
 #include "db_config.h"
 
-#ifndef NO_SYSTEM_INCLUDES
-#include <sys/types.h>
-
-#if TIME_WITH_SYS_TIME
-#include <sys/time.h>
-#include <time.h>
-#else
-#if HAVE_SYS_TIME_H
-#include <sys/time.h>
-#else
-#include <time.h>
-#endif
-#endif
-
-#include <stdlib.h>
-#include <string.h>
-#endif
-
 #include "db_int.h"
 #include "dbinc/db_page.h"
 #include "dbinc/db_am.h"
@@ -38,7 +20,11 @@ static int  __txn_compare __P((const void *, const void *));
 static int  __txn_print_all __P((DB_ENV *, u_int32_t));
 static int  __txn_print_stats __P((DB_ENV *, u_int32_t));
 static int  __txn_stat __P((DB_ENV *, DB_TXN_STAT **, u_int32_t));
-static void __txn_xid_stats __P((DB_ENV *, DB_MSGBUF *, DB_TXN_ACTIVE *));
+static char *__txn_status __P((DB_TXN_ACTIVE *));
+static void __txn_gid __P((DB_ENV *, DB_MSGBUF *, DB_TXN_ACTIVE *));
+
+#define	XID_FIELD_IS_SET(p)						\
+	((p)->xa_status != 0 || (p)->status == TXN_PREPARED)
 
 /*
  * __txn_stat_pp --
@@ -52,7 +38,8 @@ __txn_stat_pp(dbenv, statp, flags)
 	DB_TXN_STAT **statp;
 	u_int32_t flags;
 {
-	int rep_check, ret;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	PANIC_CHECK(dbenv);
 	ENV_REQUIRES_CONFIG(dbenv,
@@ -62,12 +49,9 @@ __txn_stat_pp(dbenv, statp, flags)
 	    "DB_ENV->txn_stat", flags, DB_STAT_CLEAR)) != 0)
 		return (ret);
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __txn_stat(dbenv, statp, flags);
-	if (rep_check)
-		__env_db_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__txn_stat(dbenv, statp, flags)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -84,7 +68,7 @@ __txn_stat(dbenv, statp, flags)
 	DB_TXNMGR *mgr;
 	DB_TXNREGION *region;
 	DB_TXN_STAT *stats;
-	TXN_DETAIL *txnp;
+	TXN_DETAIL *td;
 	size_t nbytes;
 	u_int32_t maxtxn, ndx;
 	int ret;
@@ -108,7 +92,7 @@ __txn_stat(dbenv, statp, flags)
 	if ((ret = __os_umalloc(dbenv, nbytes, &stats)) != 0)
 		return (ret);
 
-	R_LOCK(dbenv, &mgr->reginfo);
+	TXN_SYSTEM_LOCK(dbenv);
 	memcpy(stats, &region->stat, sizeof(*stats));
 	stats->st_last_txnid = region->last_txnid;
 	stats->st_last_ckp = region->last_ckp;
@@ -116,35 +100,50 @@ __txn_stat(dbenv, statp, flags)
 	stats->st_txnarray = (DB_TXN_ACTIVE *)&stats[1];
 
 	for (ndx = 0,
-	    txnp = SH_TAILQ_FIRST(&region->active_txn, __txn_detail);
-	    txnp != NULL && ndx < maxtxn;
-	    txnp = SH_TAILQ_NEXT(txnp, links, __txn_detail), ++ndx) {
-		stats->st_txnarray[ndx].txnid = txnp->txnid;
-		if (txnp->parent == INVALID_ROFF)
+	    td = SH_TAILQ_FIRST(&region->active_txn, __txn_detail);
+	    td != NULL && ndx < maxtxn;
+	    td = SH_TAILQ_NEXT(td, links, __txn_detail), ++ndx) {
+		stats->st_txnarray[ndx].txnid = td->txnid;
+		if (td->parent == INVALID_ROFF)
 			stats->st_txnarray[ndx].parentid = TXN_INVALID;
 		else
 			stats->st_txnarray[ndx].parentid =
 			    ((TXN_DETAIL *)R_ADDR(&mgr->reginfo,
-			    txnp->parent))->txnid;
-		stats->st_txnarray[ndx].lsn = txnp->begin_lsn;
-		if ((stats->st_txnarray[ndx].xa_status = txnp->xa_status) != 0)
+			    td->parent))->txnid;
+		stats->st_txnarray[ndx].pid = td->pid;
+		stats->st_txnarray[ndx].tid = td->tid;
+		stats->st_txnarray[ndx].lsn = td->begin_lsn;
+		stats->st_txnarray[ndx].read_lsn = td->read_lsn;
+		stats->st_txnarray[ndx].mvcc_ref = td->mvcc_ref;
+		stats->st_txnarray[ndx].status = td->status;
+		stats->st_txnarray[ndx].xa_status = td->xa_status;
+		if (XID_FIELD_IS_SET(td))
 			memcpy(stats->st_txnarray[ndx].xid,
-			    txnp->xid, DB_XIDDATASIZE);
+			    td->xid, sizeof(td->xid));
+		if (td->name != INVALID_ROFF) {
+			(void)strncpy(stats->st_txnarray[ndx].name,
+			    R_ADDR(&mgr->reginfo, td->name),
+			    sizeof(stats->st_txnarray[ndx].name) - 1);
+			stats->st_txnarray[ndx].name[
+			    sizeof(stats->st_txnarray[ndx].name) - 1] = '\0';
+		} else
+			stats->st_txnarray[ndx].name[0] = '\0';
 	}
 
-	stats->st_region_wait = mgr->reginfo.rp->mutex.mutex_set_wait;
-	stats->st_region_nowait = mgr->reginfo.rp->mutex.mutex_set_nowait;
+	__mutex_set_wait_info(dbenv, region->mtx_region,
+	    &stats->st_region_wait, &stats->st_region_nowait);
 	stats->st_regsize = mgr->reginfo.rp->size;
 	if (LF_ISSET(DB_STAT_CLEAR)) {
-		mgr->reginfo.rp->mutex.mutex_set_wait = 0;
-		mgr->reginfo.rp->mutex.mutex_set_nowait = 0;
+		__mutex_clear(dbenv, region->mtx_region);
 		memset(&region->stat, 0, sizeof(region->stat));
 		region->stat.st_maxtxns = region->maxtxns;
 		region->stat.st_maxnactive =
 		    region->stat.st_nactive = stats->st_nactive;
+		region->stat.st_maxnsnapshot =
+		    region->stat.st_nsnapshot = stats->st_nsnapshot;
 	}
 
-	R_UNLOCK(dbenv, &mgr->reginfo);
+	TXN_SYSTEM_UNLOCK(dbenv);
 
 	*statp = stats;
 	return (0);
@@ -161,7 +160,8 @@ __txn_stat_print_pp(dbenv, flags)
 	DB_ENV *dbenv;
 	u_int32_t flags;
 {
-	int rep_check, ret;
+	DB_THREAD_INFO *ip;
+	int ret;
 
 	PANIC_CHECK(dbenv);
 	ENV_REQUIRES_CONFIG(dbenv,
@@ -171,12 +171,9 @@ __txn_stat_print_pp(dbenv, flags)
 	    flags, DB_STAT_ALL | DB_STAT_CLEAR)) != 0)
 		return (ret);
 
-	rep_check = IS_ENV_REPLICATED(dbenv) ? 1 : 0;
-	if (rep_check)
-		__env_rep_enter(dbenv);
-	ret = __txn_stat_print(dbenv, flags);
-	if (rep_check)
-		__env_db_rep_exit(dbenv);
+	ENV_ENTER(dbenv, ip);
+	REPLICATION_WRAP(dbenv, (__txn_stat_print(dbenv, flags)), ret);
+	ENV_LEAVE(dbenv, ip);
 	return (ret);
 }
 
@@ -220,8 +217,10 @@ __txn_print_stats(dbenv, flags)
 {
 	DB_MSGBUF mb;
 	DB_TXN_STAT *sp;
+	DB_TXN_ACTIVE *txn;
 	u_int32_t i;
 	int ret;
+	char buf[DB_THREADID_STRLEN], time_buf[CTIME_BUFLEN];
 
 	if ((ret = __txn_stat(dbenv, &sp, flags)) != 0)
 		return (ret);
@@ -236,7 +235,7 @@ __txn_print_stats(dbenv, flags)
 		__db_msg(dbenv, "0\tNo checkpoint timestamp");
 	else
 		__db_msg(dbenv, "%.24s\tCheckpoint timestamp",
-		    ctime(&sp->st_time_ckp));
+		    __db_ctime(&sp->st_time_ckp, time_buf));
 	__db_msg(dbenv, "%#lx\tLast transaction ID allocated",
 	    (u_long)sp->st_last_txnid);
 	__db_dl(dbenv, "Maximum number of active transactions configured",
@@ -250,6 +249,9 @@ __txn_print_stats(dbenv, flags)
 	    "Number of transactions aborted", (u_long)sp->st_naborts);
 	__db_dl(dbenv,
 	    "Number of transactions committed", (u_long)sp->st_ncommits);
+	__db_dl(dbenv, "Snapshot transactions", (u_long)sp->st_nsnapshot);
+	__db_dl(dbenv, "Maximum snapshot transactions",
+	    (u_long)sp->st_maxnsnapshot);
 	__db_dl(dbenv,
 	    "Number of transactions restored", (u_long)sp->st_nrestores);
 
@@ -262,19 +264,29 @@ __txn_print_stats(dbenv, flags)
 
 	qsort(sp->st_txnarray,
 	    sp->st_nactive, sizeof(sp->st_txnarray[0]), __txn_compare);
-	__db_msg(dbenv, "List of active transactions:");
+	__db_msg(dbenv, "Active transactions:");
 	DB_MSGBUF_INIT(&mb);
 	for (i = 0; i < sp->st_nactive; ++i) {
-		__db_msgadd(dbenv,
-		    &mb, "\tID: %lx; begin LSN: file/offset %lu/%lu",
-		    (u_long)sp->st_txnarray[i].txnid,
-		    (u_long)sp->st_txnarray[i].lsn.file,
-		    (u_long)sp->st_txnarray[i].lsn.offset);
-		if (sp->st_txnarray[i].parentid != 0)
-			__db_msgadd(dbenv, &mb, "; parent: %lx",
-			    (u_long)sp->st_txnarray[i].parentid);
-		if (sp->st_txnarray[i].xa_status != 0)
-			__txn_xid_stats(dbenv, &mb, &sp->st_txnarray[i]);
+		txn = &sp->st_txnarray[i];
+		__db_msgadd(dbenv, &mb,
+	    "\t%lx: %s; pid/thread %s; begin LSN: file/offset %lu/%lu",
+		    (u_long)txn->txnid, __txn_status(txn),
+		    dbenv->thread_id_string(dbenv, txn->pid, txn->tid, buf),
+		    (u_long)txn->lsn.file, (u_long)txn->lsn.offset);
+		if (txn->parentid != 0)
+			__db_msgadd(dbenv, &mb,
+			    "; parent: %lx", (u_long)txn->parentid);
+		if (!IS_MAX_LSN(txn->read_lsn))
+			__db_msgadd(dbenv, &mb, "; read LSN: %lu/%lu",
+			    (u_long)txn->read_lsn.file,
+			    (u_long)txn->read_lsn.offset);
+		if (txn->mvcc_ref != 0)
+			__db_msgadd(dbenv, &mb,
+			    "; mvcc refcount: %lu", (u_long)txn->mvcc_ref);
+		if (txn->name[0] != '\0')
+			__db_msgadd(dbenv, &mb, "; \"%s\"", txn->name);
+		if (XID_FIELD_IS_SET(txn))
+			__txn_gid(dbenv, &mb, txn);
 		DB_MSGBUF_FLUSH(dbenv, &mb);
 	}
 
@@ -298,31 +310,36 @@ __txn_print_all(dbenv, flags)
 	};
 	DB_TXNMGR *mgr;
 	DB_TXNREGION *region;
+	char time_buf[CTIME_BUFLEN];
 
 	mgr = dbenv->tx_handle;
 	region = mgr->reginfo.primary;
 
-	R_LOCK(dbenv, &mgr->reginfo);
+	TXN_SYSTEM_LOCK(dbenv);
 
 	__db_print_reginfo(dbenv, &mgr->reginfo, "Transaction");
 
 	__db_msg(dbenv, "%s", DB_GLOBAL(db_line));
 	__db_msg(dbenv, "DB_TXNMGR handle information:");
-
-	__db_print_mutex(dbenv, NULL, mgr->mutexp, "DB_TXNMGR mutex", flags);
+	__mutex_print_debug_single(dbenv, "DB_TXNMGR mutex", mgr->mutex, flags);
 	__db_dl(dbenv,
 	    "Number of transactions discarded", (u_long)mgr->n_discards);
 
 	__db_msg(dbenv, "%s", DB_GLOBAL(db_line));
 	__db_msg(dbenv, "DB_TXNREGION handle information:");
+	__mutex_print_debug_single(
+	    dbenv, "DB_TXNREGION region mutex", region->mtx_region, flags);
 	STAT_ULONG("Maximum number of active txns", region->maxtxns);
 	STAT_HEX("Last transaction ID allocated", region->last_txnid);
 	STAT_HEX("Current maximum unused ID", region->cur_maxid);
 
+	__mutex_print_debug_single(
+	    dbenv, "checkpoint mutex", region->mtx_ckp, flags);
 	STAT_LSN("Last checkpoint LSN", &region->last_ckp);
 	__db_msg(dbenv,
 	    "%.24s\tLast checkpoint timestamp",
-	    region->time_ckp == 0 ? "0" : ctime(&region->time_ckp));
+	    region->time_ckp == 0 ? "0" :
+	    __db_ctime(&region->time_ckp, time_buf));
 
 	__db_prflags(dbenv, NULL, region->flags, fn, NULL, "\tFlags");
 
@@ -334,52 +351,64 @@ __txn_print_all(dbenv, flags)
 	 * Display list of XA transactions in the DB_ENV handle.
 	 */
 
-	R_UNLOCK(dbenv, &mgr->reginfo);
+	TXN_SYSTEM_UNLOCK(dbenv);
 
 	return (0);
 }
 
+static char *
+__txn_status(txn)
+	DB_TXN_ACTIVE *txn;
+{
+	switch (txn->xa_status) {
+	case 0:
+		switch (txn->status) {
+		case TXN_ABORTED:
+			return ("aborted");
+		case TXN_COMMITTED:
+			return ("committed");
+		case TXN_PREPARED:
+			return ("prepared");
+		case TXN_RUNNING:
+			return ("running");
+		default:
+			break;
+		}
+		break;
+	case TXN_XA_ABORTED:
+		return ("XA aborted");
+	case TXN_XA_DEADLOCKED:
+		return ("XA deadlocked");
+	case TXN_XA_ENDED:
+		return ("XA ended");
+	case TXN_XA_PREPARED:
+		return ("XA prepared");
+	case TXN_XA_STARTED:
+		return ("XA started");
+	case TXN_XA_SUSPENDED:
+		return ("XA suspended");
+	default:
+		break;
+	}
+	return ("unknown state");
+}
+
 static void
-__txn_xid_stats(dbenv, mbp, txnp)
+__txn_gid(dbenv, mbp, txn)
 	DB_ENV *dbenv;
 	DB_MSGBUF *mbp;
-	DB_TXN_ACTIVE *txnp;
+	DB_TXN_ACTIVE *txn;
 {
 	u_int32_t v, *xp;
 	u_int i;
 	int cnt;
-	const char *s;
 
-	switch (txnp->xa_status) {
-	case TXN_XA_ABORTED:
-		s = "ABORTED";
-		break;
-	case TXN_XA_DEADLOCKED:
-		s = "DEADLOCKED";
-		break;
-	case TXN_XA_ENDED:
-		s = "ENDED";
-		break;
-	case TXN_XA_PREPARED:
-		s = "PREPARED";
-		break;
-	case TXN_XA_STARTED:
-		s = "STARTED";
-		break;
-	case TXN_XA_SUSPENDED:
-		s = "SUSPENDED";
-		break;
-	default:
-		s = "UNKNOWN STATE";
-		__db_err(dbenv,
-		    "XA: unknown state: %lu", (u_long)txnp->xa_status);
-		break;
-	}
-	__db_msgadd(dbenv, mbp, "\tXA: %s; XID:\n\t\t", s == NULL ? "" : s);
-	for (cnt = 0, xp = (u_int32_t *)txnp->xid,
-	    i = 0; i < DB_XIDDATASIZE; i += sizeof(u_int32_t)) {
+	__db_msgadd(dbenv, mbp, "\n\tGID/XID:");
+	for (cnt = 0, xp = (u_int32_t *)txn->xid, i = 0;;) {
 		memcpy(&v, xp++, sizeof(u_int32_t));
-		__db_msgadd(dbenv, mbp, "%#x ", v);
+		__db_msgadd(dbenv, mbp, "%#lx ", (u_long)v);
+		if ((i += sizeof(u_int32_t)) >= DB_XIDDATASIZE)
+			break;
 		if (++cnt == 4) {
 			DB_MSGBUF_FLUSH(dbenv, mbp);
 			__db_msgadd(dbenv, mbp, "\t\t");
